
namespace glow {
class AllocActivationInst final : public Instruction {
  TypeRef Ty_;

 public:
  AllocActivationInst(llvm::StringRef name, TypeRef Ty)
      : Instruction(name, Kinded::Kind::AllocActivationInstKind, Ty, {
      }), Ty_(Ty) {}

  TypeRef getTy() const { return Ty_; }

  static bool classof(const Kinded *k) {
    return k->getKind() == Kinded::Kind::AllocActivationInstKind;
  }

  bool isInplaceOp(unsigned dstIdx, unsigned srcIdx) const {
    return false;
  }

  bool isCanonical() const {
    return true;
  }

  bool isDataParallel() const {
    return false;
  }

  void setTy(TypeRef Ty);

  Instruction* clone() const;

  void dump(llvm::raw_ostream &os) const;

  llvm::StringRef getOperandName(unsigned idx) const;
  void verify() const;
};
} // namespace glow

namespace glow {
class TensorViewInst final : public Instruction {
  TypeRef Ty_;
  std::vector<dim_t> Offsets_;

 public:
  TensorViewInst(llvm::StringRef name, Value *Src, TypeRef Ty, std::vector<dim_t> Offsets)
      : Instruction(name, Kinded::Kind::TensorViewInstKind, Ty, {
          {Src, OperandKind::In},
      }), Ty_(Ty), Offsets_(Offsets) {}

  Value *getSrc() const { return getOperand(0).first; }
  TypeRef getTy() const { return Ty_; }
  llvm::ArrayRef<dim_t> getOffsets() const { return Offsets_; }

  static bool classof(const Kinded *k) {
    return k->getKind() == Kinded::Kind::TensorViewInstKind;
  }

  bool isInplaceOp(unsigned dstIdx, unsigned srcIdx) const {
    return false;
  }

  bool isCanonical() const {
    return true;
  }

  bool isDataParallel() const {
    return false;
  }

  Instruction* clone() const;

  void dump(llvm::raw_ostream &os) const;

  llvm::StringRef getOperandName(unsigned idx) const;
  void verify() const;
};
} // namespace glow

namespace glow {
class DeallocActivationInst final : public Instruction {

 public:
  DeallocActivationInst(llvm::StringRef name, Value *Src)
      : Instruction(name, Kinded::Kind::DeallocActivationInstKind, Src->getType(), {
          {Src, OperandKind::Out},
      }) {}

  Value *getSrc() const { return getOperand(0).first; }

  static bool classof(const Kinded *k) {
    return k->getKind() == Kinded::Kind::DeallocActivationInstKind;
  }

  bool isInplaceOp(unsigned dstIdx, unsigned srcIdx) const {
    return false;
  }

  bool isCanonical() const {
    return true;
  }

  bool isDataParallel() const {
    return false;
  }

  AllocActivationInst *getAlloc() const; 

  Instruction* clone() const;

  void dump(llvm::raw_ostream &os) const;

  llvm::StringRef getOperandName(unsigned idx) const;
  void verify() const;
};
} // namespace glow

namespace glow {
class CopyInst final : public Instruction {

 public:
  CopyInst(llvm::StringRef name, Value *Dest, Value *Src)
      : Instruction(name, Kinded::Kind::CopyInstKind, Src->getType(), {
          {Dest, OperandKind::Out},
          {Src, OperandKind::In},
      }) {}

  Value *getDest() const { return getOperand(0).first; }
  Value *getSrc() const { return getOperand(1).first; }

  static bool classof(const Kinded *k) {
    return k->getKind() == Kinded::Kind::CopyInstKind;
  }

  bool isInplaceOp(unsigned dstIdx, unsigned srcIdx) const {
  if (0 == dstIdx && 1 == srcIdx) { return true; }
    return false;
  }

  bool isCanonical() const {
    return true;
  }

  bool isDataParallel() const {
    return true;
  }

  Instruction* clone() const;

  void dump(llvm::raw_ostream &os) const;

  llvm::StringRef getOperandName(unsigned idx) const;
  void verify() const;
};
} // namespace glow

namespace glow {
class ConvolutionGradInst final : public Instruction {
  std::vector<unsigned_t> Kernels_;
  std::vector<unsigned_t> Strides_;
  std::vector<unsigned_t> Pads_;
  unsigned_t Group_;
  unsigned_t Dilation_;
  ConvolutionLayout Layout_;
  FusedActivation FusedActivation_;

 public:
  ConvolutionGradInst(llvm::StringRef name, Value *Src, Value *Filter, Value *DestGrad, Value *SrcGrad, Value *FilterGrad, Value *BiasGrad, std::vector<unsigned_t> Kernels, std::vector<unsigned_t> Strides, std::vector<unsigned_t> Pads, unsigned_t Group, unsigned_t Dilation, ConvolutionLayout Layout, FusedActivation FusedActivation)
      : Instruction(name, Kinded::Kind::ConvolutionGradInstKind, nullptr, {
          {Src, OperandKind::In},
          {Filter, OperandKind::In},
          {DestGrad, OperandKind::In},
          {SrcGrad, OperandKind::Out},
          {FilterGrad, OperandKind::Out},
          {BiasGrad, OperandKind::Out},
      }), Kernels_(Kernels), Strides_(Strides), Pads_(Pads), Group_(Group), Dilation_(Dilation), Layout_(Layout), FusedActivation_(FusedActivation) {}

  Value *getSrc() const { return getOperand(0).first; }
  Value *getFilter() const { return getOperand(1).first; }
  Value *getDestGrad() const { return getOperand(2).first; }
  Value *getSrcGrad() const { return getOperand(3).first; }
  Value *getFilterGrad() const { return getOperand(4).first; }
  Value *getBiasGrad() const { return getOperand(5).first; }
  llvm::ArrayRef<unsigned_t> getKernels() const { return Kernels_; }
  llvm::ArrayRef<unsigned_t> getStrides() const { return Strides_; }
  llvm::ArrayRef<unsigned_t> getPads() const { return Pads_; }
  unsigned_t getGroup() const { return Group_; }
  unsigned_t getDilation() const { return Dilation_; }
  ConvolutionLayout getLayout() const { return Layout_; }
  FusedActivation getFusedActivation() const { return FusedActivation_; }

  static bool classof(const Kinded *k) {
    return k->getKind() == Kinded::Kind::ConvolutionGradInstKind;
  }

  bool isInplaceOp(unsigned dstIdx, unsigned srcIdx) const {
    return false;
  }

  bool isCanonical() const {
    return true;
  }

  bool isDataParallel() const {
    return false;
  }

  Instruction* clone() const;

  void dump(llvm::raw_ostream &os) const;

  llvm::StringRef getOperandName(unsigned idx) const;
  void verify() const {
    assert(getDestGrad()->getElementType() == getSrcGrad()->getElementType() && "Invalid Element Type");
    assert(getDestGrad()->getElementType() == getSrc()->getElementType() && "Invalid Element Type");
    assert(getDestGrad()->getElementType() == getFilterGrad()->getElementType() && "Invalid Element Type");
    assert(getDestGrad()->getElementType() == getFilter()->getElementType() && "Invalid Element Type");
  }
};
} // namespace glow

namespace glow {
class ConvolutionInst final : public Instruction {
  std::vector<unsigned_t> Kernels_;
  std::vector<unsigned_t> Strides_;
  std::vector<unsigned_t> Pads_;
  unsigned_t Group_;
  unsigned_t Dilation_;
  ConvolutionLayout Layout_;
  FusedActivation FusedActivation_;

 public:
  ConvolutionInst(llvm::StringRef name, Value *Dest, Value *Src, Value *Filter, Value *Bias, std::vector<unsigned_t> Kernels, std::vector<unsigned_t> Strides, std::vector<unsigned_t> Pads, unsigned_t Group, unsigned_t Dilation, ConvolutionLayout Layout, FusedActivation FusedActivation)
      : Instruction(name, Kinded::Kind::ConvolutionInstKind, nullptr, {
          {Dest, OperandKind::Out},
          {Src, OperandKind::In},
          {Filter, OperandKind::In},
          {Bias, OperandKind::In},
      }), Kernels_(Kernels), Strides_(Strides), Pads_(Pads), Group_(Group), Dilation_(Dilation), Layout_(Layout), FusedActivation_(FusedActivation) {}

  Value *getDest() const { return getOperand(0).first; }
  Value *getSrc() const { return getOperand(1).first; }
  Value *getFilter() const { return getOperand(2).first; }
  Value *getBias() const { return getOperand(3).first; }
  llvm::ArrayRef<unsigned_t> getKernels() const { return Kernels_; }
  llvm::ArrayRef<unsigned_t> getStrides() const { return Strides_; }
  llvm::ArrayRef<unsigned_t> getPads() const { return Pads_; }
  unsigned_t getGroup() const { return Group_; }
  unsigned_t getDilation() const { return Dilation_; }
  ConvolutionLayout getLayout() const { return Layout_; }
  FusedActivation getFusedActivation() const { return FusedActivation_; }

  static bool classof(const Kinded *k) {
    return k->getKind() == Kinded::Kind::ConvolutionInstKind;
  }

  bool isInplaceOp(unsigned dstIdx, unsigned srcIdx) const {
    return false;
  }

  bool isCanonical() const {
    return true;
  }

  bool isDataParallel() const {
    return false;
  }

  Instruction* clone() const;

  void dump(llvm::raw_ostream &os) const;

  llvm::StringRef getOperandName(unsigned idx) const;
  void verify() const {
    assert(getDest()->getElementType() == getSrc()->getElementType() && "Invalid Element Type");
    assert(getDest()->getElementType() == getFilter()->getElementType() && "Invalid Element Type");
  }
};
} // namespace glow

namespace glow {
class ChannelwiseQuantizedConvolutionInst final : public Instruction {
  std::vector<unsigned_t> Kernels_;
  std::vector<unsigned_t> Strides_;
  std::vector<unsigned_t> Pads_;
  unsigned_t Group_;
  unsigned_t Dilation_;

 public:
  ChannelwiseQuantizedConvolutionInst(llvm::StringRef name, Value *Dest, Value *Src, Value *Filter, Value *Bias, Value *FilterScales, Value *FilterOffsets, Value *BiasScales, Value *BiasOffsets, std::vector<unsigned_t> Kernels, std::vector<unsigned_t> Strides, std::vector<unsigned_t> Pads, unsigned_t Group, unsigned_t Dilation)
      : Instruction(name, Kinded::Kind::ChannelwiseQuantizedConvolutionInstKind, nullptr, {
          {Dest, OperandKind::Out},
          {Src, OperandKind::In},
          {Filter, OperandKind::In},
          {Bias, OperandKind::In},
          {FilterScales, OperandKind::In},
          {FilterOffsets, OperandKind::In},
          {BiasScales, OperandKind::In},
          {BiasOffsets, OperandKind::In},
      }), Kernels_(Kernels), Strides_(Strides), Pads_(Pads), Group_(Group), Dilation_(Dilation) {}

  Value *getDest() const { return getOperand(0).first; }
  Value *getSrc() const { return getOperand(1).first; }
  Value *getFilter() const { return getOperand(2).first; }
  Value *getBias() const { return getOperand(3).first; }
  Value *getFilterScales() const { return getOperand(4).first; }
  Value *getFilterOffsets() const { return getOperand(5).first; }
  Value *getBiasScales() const { return getOperand(6).first; }
  Value *getBiasOffsets() const { return getOperand(7).first; }
  llvm::ArrayRef<unsigned_t> getKernels() const { return Kernels_; }
  llvm::ArrayRef<unsigned_t> getStrides() const { return Strides_; }
  llvm::ArrayRef<unsigned_t> getPads() const { return Pads_; }
  unsigned_t getGroup() const { return Group_; }
  unsigned_t getDilation() const { return Dilation_; }

  static bool classof(const Kinded *k) {
    return k->getKind() == Kinded::Kind::ChannelwiseQuantizedConvolutionInstKind;
  }

  bool isInplaceOp(unsigned dstIdx, unsigned srcIdx) const {
    return false;
  }

  bool isCanonical() const {
    return true;
  }

  bool isDataParallel() const {
    return false;
  }

  Instruction* clone() const;

  void dump(llvm::raw_ostream &os) const;

  llvm::StringRef getOperandName(unsigned idx) const;
  void verify() const {
    assert(getDest()->getElementType() == getSrc()->getElementType() && "Invalid Element Type");
    assert(getDest()->getElementType() == getFilter()->getElementType() && "Invalid Element Type");
    assert(getDest()->getElementType() == ElemKind::Int8QTy && "Invalid Element Type");
  }
};
} // namespace glow

namespace glow {
class ConvTransposeInst final : public Instruction {
  std::vector<unsigned_t> Kernels_;
  std::vector<unsigned_t> Strides_;
  std::vector<unsigned_t> Pads_;
  unsigned_t Group_;
  unsigned_t Dilation_;

 public:
  ConvTransposeInst(llvm::StringRef name, Value *Dest, Value *Src, Value *Filter, Value *Bias, std::vector<unsigned_t> Kernels, std::vector<unsigned_t> Strides, std::vector<unsigned_t> Pads, unsigned_t Group, unsigned_t Dilation)
      : Instruction(name, Kinded::Kind::ConvTransposeInstKind, nullptr, {
          {Dest, OperandKind::Out},
          {Src, OperandKind::In},
          {Filter, OperandKind::In},
          {Bias, OperandKind::In},
      }), Kernels_(Kernels), Strides_(Strides), Pads_(Pads), Group_(Group), Dilation_(Dilation) {}

  Value *getDest() const { return getOperand(0).first; }
  Value *getSrc() const { return getOperand(1).first; }
  Value *getFilter() const { return getOperand(2).first; }
  Value *getBias() const { return getOperand(3).first; }
  llvm::ArrayRef<unsigned_t> getKernels() const { return Kernels_; }
  llvm::ArrayRef<unsigned_t> getStrides() const { return Strides_; }
  llvm::ArrayRef<unsigned_t> getPads() const { return Pads_; }
  unsigned_t getGroup() const { return Group_; }
  unsigned_t getDilation() const { return Dilation_; }

  static bool classof(const Kinded *k) {
    return k->getKind() == Kinded::Kind::ConvTransposeInstKind;
  }

  bool isInplaceOp(unsigned dstIdx, unsigned srcIdx) const {
    return false;
  }

  bool isCanonical() const {
    return true;
  }

  bool isDataParallel() const {
    return false;
  }

  Instruction* clone() const;

  void dump(llvm::raw_ostream &os) const;

  llvm::StringRef getOperandName(unsigned idx) const;
  void verify() const {
    assert(getDest()->getElementType() == getSrc()->getElementType() && "Invalid Element Type");
    assert(getDest()->getElementType() == getFilter()->getElementType() && "Invalid Element Type");
  }
};
} // namespace glow

namespace glow {
class Convolution3DGradInst final : public Instruction {
  std::vector<unsigned_t> Kernels_;
  std::vector<unsigned_t> Strides_;
  std::vector<unsigned_t> Pads_;
  unsigned_t Group_;

 public:
  Convolution3DGradInst(llvm::StringRef name, Value *Src, Value *Filter, Value *DestGrad, Value *SrcGrad, Value *FilterGrad, Value *BiasGrad, std::vector<unsigned_t> Kernels, std::vector<unsigned_t> Strides, std::vector<unsigned_t> Pads, unsigned_t Group)
      : Instruction(name, Kinded::Kind::Convolution3DGradInstKind, nullptr, {
          {Src, OperandKind::In},
          {Filter, OperandKind::In},
          {DestGrad, OperandKind::In},
          {SrcGrad, OperandKind::Out},
          {FilterGrad, OperandKind::Out},
          {BiasGrad, OperandKind::Out},
      }), Kernels_(Kernels), Strides_(Strides), Pads_(Pads), Group_(Group) {}

  Value *getSrc() const { return getOperand(0).first; }
  Value *getFilter() const { return getOperand(1).first; }
  Value *getDestGrad() const { return getOperand(2).first; }
  Value *getSrcGrad() const { return getOperand(3).first; }
  Value *getFilterGrad() const { return getOperand(4).first; }
  Value *getBiasGrad() const { return getOperand(5).first; }
  llvm::ArrayRef<unsigned_t> getKernels() const { return Kernels_; }
  llvm::ArrayRef<unsigned_t> getStrides() const { return Strides_; }
  llvm::ArrayRef<unsigned_t> getPads() const { return Pads_; }
  unsigned_t getGroup() const { return Group_; }

  static bool classof(const Kinded *k) {
    return k->getKind() == Kinded::Kind::Convolution3DGradInstKind;
  }

  bool isInplaceOp(unsigned dstIdx, unsigned srcIdx) const {
    return false;
  }

  bool isCanonical() const {
    return true;
  }

  bool isDataParallel() const {
    return false;
  }

  Instruction* clone() const;

  void dump(llvm::raw_ostream &os) const;

  llvm::StringRef getOperandName(unsigned idx) const;
  void verify() const {
    assert(getDestGrad()->getElementType() == getSrcGrad()->getElementType() && "Invalid Element Type");
    assert(getDestGrad()->getElementType() == getSrc()->getElementType() && "Invalid Element Type");
    assert(getDestGrad()->getElementType() == getFilterGrad()->getElementType() && "Invalid Element Type");
    assert(getDestGrad()->getElementType() == getFilter()->getElementType() && "Invalid Element Type");
  }
};
} // namespace glow

namespace glow {
class Convolution3DInst final : public Instruction {
  std::vector<unsigned_t> Kernels_;
  std::vector<unsigned_t> Strides_;
  std::vector<unsigned_t> Pads_;
  unsigned_t Group_;

 public:
  Convolution3DInst(llvm::StringRef name, Value *Dest, Value *Src, Value *Filter, Value *Bias, std::vector<unsigned_t> Kernels, std::vector<unsigned_t> Strides, std::vector<unsigned_t> Pads, unsigned_t Group)
      : Instruction(name, Kinded::Kind::Convolution3DInstKind, nullptr, {
          {Dest, OperandKind::Out},
          {Src, OperandKind::In},
          {Filter, OperandKind::In},
          {Bias, OperandKind::In},
      }), Kernels_(Kernels), Strides_(Strides), Pads_(Pads), Group_(Group) {}

  Value *getDest() const { return getOperand(0).first; }
  Value *getSrc() const { return getOperand(1).first; }
  Value *getFilter() const { return getOperand(2).first; }
  Value *getBias() const { return getOperand(3).first; }
  llvm::ArrayRef<unsigned_t> getKernels() const { return Kernels_; }
  llvm::ArrayRef<unsigned_t> getStrides() const { return Strides_; }
  llvm::ArrayRef<unsigned_t> getPads() const { return Pads_; }
  unsigned_t getGroup() const { return Group_; }

  static bool classof(const Kinded *k) {
    return k->getKind() == Kinded::Kind::Convolution3DInstKind;
  }

  bool isInplaceOp(unsigned dstIdx, unsigned srcIdx) const {
    return false;
  }

  bool isCanonical() const {
    return true;
  }

  bool isDataParallel() const {
    return false;
  }

  Instruction* clone() const;

  void dump(llvm::raw_ostream &os) const;

  llvm::StringRef getOperandName(unsigned idx) const;
  void verify() const {
    assert(getDest()->getElementType() == getSrc()->getElementType() && "Invalid Element Type");
    assert(getDest()->getElementType() == getFilter()->getElementType() && "Invalid Element Type");
  }
};
} // namespace glow

namespace glow {
class MaxPoolWithArgmaxGradInst final : public Instruction {
  std::vector<unsigned_t> Kernels_;
  std::vector<unsigned_t> Strides_;
  std::vector<unsigned_t> Pads_;
  unsigned_t Layout_;

 public:
  MaxPoolWithArgmaxGradInst(llvm::StringRef name, Value *Dest, Value *Src, Value *Argmax, Value *DestGrad, Value *SrcGrad, std::vector<unsigned_t> Kernels, std::vector<unsigned_t> Strides, std::vector<unsigned_t> Pads, unsigned_t Layout)
      : Instruction(name, Kinded::Kind::MaxPoolWithArgmaxGradInstKind, nullptr, {
          {Dest, OperandKind::In},
          {Src, OperandKind::In},
          {Argmax, OperandKind::In},
          {DestGrad, OperandKind::In},
          {SrcGrad, OperandKind::Out},
      }), Kernels_(Kernels), Strides_(Strides), Pads_(Pads), Layout_(Layout) {}

  Value *getDest() const { return getOperand(0).first; }
  Value *getSrc() const { return getOperand(1).first; }
  Value *getArgmax() const { return getOperand(2).first; }
  Value *getDestGrad() const { return getOperand(3).first; }
  Value *getSrcGrad() const { return getOperand(4).first; }
  llvm::ArrayRef<unsigned_t> getKernels() const { return Kernels_; }
  llvm::ArrayRef<unsigned_t> getStrides() const { return Strides_; }
  llvm::ArrayRef<unsigned_t> getPads() const { return Pads_; }
  unsigned_t getLayout() const { return Layout_; }

  static bool classof(const Kinded *k) {
    return k->getKind() == Kinded::Kind::MaxPoolWithArgmaxGradInstKind;
  }

  bool isInplaceOp(unsigned dstIdx, unsigned srcIdx) const {
    return false;
  }

  bool isCanonical() const {
    return true;
  }

  bool isDataParallel() const {
    return false;
  }

  Instruction* clone() const;

  void dump(llvm::raw_ostream &os) const;

  llvm::StringRef getOperandName(unsigned idx) const;
  void verify() const {
    assert(getDestGrad()->getElementType() == getDest()->getElementType() && "Invalid Element Type");
    assert(getDestGrad()->getElementType() == getSrcGrad()->getElementType() && "Invalid Element Type");
    assert(getDestGrad()->getElementType() == getSrc()->getElementType() && "Invalid Element Type");
  }
};
} // namespace glow

namespace glow {
class MaxPoolWithArgmaxInst final : public Instruction {
  std::vector<unsigned_t> Kernels_;
  std::vector<unsigned_t> Strides_;
  std::vector<unsigned_t> Pads_;
  unsigned_t Layout_;

 public:
  MaxPoolWithArgmaxInst(llvm::StringRef name, Value *Dest, Value *Src, Value *Argmax, std::vector<unsigned_t> Kernels, std::vector<unsigned_t> Strides, std::vector<unsigned_t> Pads, unsigned_t Layout)
      : Instruction(name, Kinded::Kind::MaxPoolWithArgmaxInstKind, nullptr, {
          {Dest, OperandKind::Out},
          {Src, OperandKind::In},
          {Argmax, OperandKind::Out},
      }), Kernels_(Kernels), Strides_(Strides), Pads_(Pads), Layout_(Layout) {}

  Value *getDest() const { return getOperand(0).first; }
  Value *getSrc() const { return getOperand(1).first; }
  Value *getArgmax() const { return getOperand(2).first; }
  llvm::ArrayRef<unsigned_t> getKernels() const { return Kernels_; }
  llvm::ArrayRef<unsigned_t> getStrides() const { return Strides_; }
  llvm::ArrayRef<unsigned_t> getPads() const { return Pads_; }
  unsigned_t getLayout() const { return Layout_; }

  static bool classof(const Kinded *k) {
    return k->getKind() == Kinded::Kind::MaxPoolWithArgmaxInstKind;
  }

  bool isInplaceOp(unsigned dstIdx, unsigned srcIdx) const {
    return false;
  }

  bool isCanonical() const {
    return true;
  }

  bool isDataParallel() const {
    return false;
  }

  Instruction* clone() const;

  void dump(llvm::raw_ostream &os) const;

  llvm::StringRef getOperandName(unsigned idx) const;
  void verify() const {
    assert(getDest()->getElementType() == getSrc()->getElementType() && "Invalid Element Type");
  }
};
} // namespace glow

namespace glow {
class MaxPoolInst final : public Instruction {
  std::vector<unsigned_t> Kernels_;
  std::vector<unsigned_t> Strides_;
  std::vector<unsigned_t> Pads_;
  unsigned_t Layout_;

 public:
  MaxPoolInst(llvm::StringRef name, Value *Dest, Value *Src, std::vector<unsigned_t> Kernels, std::vector<unsigned_t> Strides, std::vector<unsigned_t> Pads, unsigned_t Layout)
      : Instruction(name, Kinded::Kind::MaxPoolInstKind, nullptr, {
          {Dest, OperandKind::Out},
          {Src, OperandKind::In},
      }), Kernels_(Kernels), Strides_(Strides), Pads_(Pads), Layout_(Layout) {}

  Value *getDest() const { return getOperand(0).first; }
  Value *getSrc() const { return getOperand(1).first; }
  llvm::ArrayRef<unsigned_t> getKernels() const { return Kernels_; }
  llvm::ArrayRef<unsigned_t> getStrides() const { return Strides_; }
  llvm::ArrayRef<unsigned_t> getPads() const { return Pads_; }
  unsigned_t getLayout() const { return Layout_; }

  static bool classof(const Kinded *k) {
    return k->getKind() == Kinded::Kind::MaxPoolInstKind;
  }

  bool isInplaceOp(unsigned dstIdx, unsigned srcIdx) const {
    return false;
  }

  bool isCanonical() const {
    return true;
  }

  bool isDataParallel() const {
    return false;
  }

  Instruction* clone() const;

  void dump(llvm::raw_ostream &os) const;

  llvm::StringRef getOperandName(unsigned idx) const;
  void verify() const {
    assert(getDest()->getElementType() == getSrc()->getElementType() && "Invalid Element Type");
  }
};
} // namespace glow

namespace glow {
class AvgPoolGradInst final : public Instruction {
  std::vector<unsigned_t> Kernels_;
  std::vector<unsigned_t> Strides_;
  std::vector<unsigned_t> Pads_;
  unsigned_t Layout_;

 public:
  AvgPoolGradInst(llvm::StringRef name, Value *Dest, Value *Src, Value *DestGrad, Value *SrcGrad, std::vector<unsigned_t> Kernels, std::vector<unsigned_t> Strides, std::vector<unsigned_t> Pads, unsigned_t Layout)
      : Instruction(name, Kinded::Kind::AvgPoolGradInstKind, nullptr, {
          {Dest, OperandKind::In},
          {Src, OperandKind::In},
          {DestGrad, OperandKind::In},
          {SrcGrad, OperandKind::Out},
      }), Kernels_(Kernels), Strides_(Strides), Pads_(Pads), Layout_(Layout) {}

  Value *getDest() const { return getOperand(0).first; }
  Value *getSrc() const { return getOperand(1).first; }
  Value *getDestGrad() const { return getOperand(2).first; }
  Value *getSrcGrad() const { return getOperand(3).first; }
  llvm::ArrayRef<unsigned_t> getKernels() const { return Kernels_; }
  llvm::ArrayRef<unsigned_t> getStrides() const { return Strides_; }
  llvm::ArrayRef<unsigned_t> getPads() const { return Pads_; }
  unsigned_t getLayout() const { return Layout_; }

  static bool classof(const Kinded *k) {
    return k->getKind() == Kinded::Kind::AvgPoolGradInstKind;
  }

  bool isInplaceOp(unsigned dstIdx, unsigned srcIdx) const {
    return false;
  }

  bool isCanonical() const {
    return true;
  }

  bool isDataParallel() const {
    return false;
  }

  Instruction* clone() const;

  void dump(llvm::raw_ostream &os) const;

  llvm::StringRef getOperandName(unsigned idx) const;
  void verify() const {
    assert(getDestGrad()->getElementType() == getDest()->getElementType() && "Invalid Element Type");
    assert(getDestGrad()->getElementType() == getSrcGrad()->getElementType() && "Invalid Element Type");
    assert(getDestGrad()->getElementType() == getSrc()->getElementType() && "Invalid Element Type");
  }
};
} // namespace glow

namespace glow {
class AvgPoolInst final : public Instruction {
  std::vector<unsigned_t> Kernels_;
  std::vector<unsigned_t> Strides_;
  std::vector<unsigned_t> Pads_;
  unsigned_t Layout_;

 public:
  AvgPoolInst(llvm::StringRef name, Value *Dest, Value *Src, std::vector<unsigned_t> Kernels, std::vector<unsigned_t> Strides, std::vector<unsigned_t> Pads, unsigned_t Layout)
      : Instruction(name, Kinded::Kind::AvgPoolInstKind, nullptr, {
          {Dest, OperandKind::Out},
          {Src, OperandKind::In},
      }), Kernels_(Kernels), Strides_(Strides), Pads_(Pads), Layout_(Layout) {}

  Value *getDest() const { return getOperand(0).first; }
  Value *getSrc() const { return getOperand(1).first; }
  llvm::ArrayRef<unsigned_t> getKernels() const { return Kernels_; }
  llvm::ArrayRef<unsigned_t> getStrides() const { return Strides_; }
  llvm::ArrayRef<unsigned_t> getPads() const { return Pads_; }
  unsigned_t getLayout() const { return Layout_; }

  static bool classof(const Kinded *k) {
    return k->getKind() == Kinded::Kind::AvgPoolInstKind;
  }

  bool isInplaceOp(unsigned dstIdx, unsigned srcIdx) const {
    return false;
  }

  bool isCanonical() const {
    return true;
  }

  bool isDataParallel() const {
    return false;
  }

  Instruction* clone() const;

  void dump(llvm::raw_ostream &os) const;

  llvm::StringRef getOperandName(unsigned idx) const;
  void verify() const {
    assert(getDest()->getElementType() == getSrc()->getElementType() && "Invalid Element Type");
  }
};
} // namespace glow

namespace glow {
class ArgMaxInst final : public Instruction {
  unsigned_t Axis_;
  bool KeepDims_;

 public:
  ArgMaxInst(llvm::StringRef name, Value *Dest, Value *Src, unsigned_t Axis, bool KeepDims)
      : Instruction(name, Kinded::Kind::ArgMaxInstKind, nullptr, {
          {Dest, OperandKind::Out},
          {Src, OperandKind::In},
      }), Axis_(Axis), KeepDims_(KeepDims) {}

  Value *getDest() const { return getOperand(0).first; }
  Value *getSrc() const { return getOperand(1).first; }
  unsigned_t getAxis() const { return Axis_; }
  bool getKeepDims() const { return KeepDims_; }

  static bool classof(const Kinded *k) {
    return k->getKind() == Kinded::Kind::ArgMaxInstKind;
  }

  bool isInplaceOp(unsigned dstIdx, unsigned srcIdx) const {
    return false;
  }

  bool isCanonical() const {
    return true;
  }

  bool isDataParallel() const {
    return false;
  }

  Instruction* clone() const;

  void dump(llvm::raw_ostream &os) const;

  llvm::StringRef getOperandName(unsigned idx) const;
  void verify() const {
    // Nothing to verify.
  }
};
} // namespace glow

namespace glow {
class ArgMinInst final : public Instruction {
  unsigned_t Axis_;
  bool KeepDims_;

 public:
  ArgMinInst(llvm::StringRef name, Value *Dest, Value *Src, unsigned_t Axis, bool KeepDims)
      : Instruction(name, Kinded::Kind::ArgMinInstKind, nullptr, {
          {Dest, OperandKind::Out},
          {Src, OperandKind::In},
      }), Axis_(Axis), KeepDims_(KeepDims) {}

  Value *getDest() const { return getOperand(0).first; }
  Value *getSrc() const { return getOperand(1).first; }
  unsigned_t getAxis() const { return Axis_; }
  bool getKeepDims() const { return KeepDims_; }

  static bool classof(const Kinded *k) {
    return k->getKind() == Kinded::Kind::ArgMinInstKind;
  }

  bool isInplaceOp(unsigned dstIdx, unsigned srcIdx) const {
    return false;
  }

  bool isCanonical() const {
    return true;
  }

  bool isDataParallel() const {
    return false;
  }

  Instruction* clone() const;

  void dump(llvm::raw_ostream &os) const;

  llvm::StringRef getOperandName(unsigned idx) const;
  void verify() const {
    // Nothing to verify.
  }
};
} // namespace glow

namespace glow {
class AdaptiveAvgPoolGradInst final : public Instruction {

 public:
  AdaptiveAvgPoolGradInst(llvm::StringRef name, Value *Dest, Value *DestGrad, Value *SrcGrad)
      : Instruction(name, Kinded::Kind::AdaptiveAvgPoolGradInstKind, nullptr, {
          {Dest, OperandKind::In},
          {DestGrad, OperandKind::In},
          {SrcGrad, OperandKind::Out},
      }) {}

  Value *getDest() const { return getOperand(0).first; }
  Value *getDestGrad() const { return getOperand(1).first; }
  Value *getSrcGrad() const { return getOperand(2).first; }

  static bool classof(const Kinded *k) {
    return k->getKind() == Kinded::Kind::AdaptiveAvgPoolGradInstKind;
  }

  bool isInplaceOp(unsigned dstIdx, unsigned srcIdx) const {
    return false;
  }

  bool isCanonical() const {
    return true;
  }

  bool isDataParallel() const {
    return false;
  }

  Instruction* clone() const;

  void dump(llvm::raw_ostream &os) const;

  llvm::StringRef getOperandName(unsigned idx) const;
  void verify() const {
    assert(getDestGrad()->getElementType() == getDest()->getElementType() && "Invalid Element Type");
    assert(getDestGrad()->getElementType() == getSrcGrad()->getElementType() && "Invalid Element Type");
  }
};
} // namespace glow

namespace glow {
class AdaptiveAvgPoolInst final : public Instruction {

 public:
  AdaptiveAvgPoolInst(llvm::StringRef name, Value *Dest, Value *Src)
      : Instruction(name, Kinded::Kind::AdaptiveAvgPoolInstKind, nullptr, {
          {Dest, OperandKind::Out},
          {Src, OperandKind::In},
      }) {}

  Value *getDest() const { return getOperand(0).first; }
  Value *getSrc() const { return getOperand(1).first; }

  static bool classof(const Kinded *k) {
    return k->getKind() == Kinded::Kind::AdaptiveAvgPoolInstKind;
  }

  bool isInplaceOp(unsigned dstIdx, unsigned srcIdx) const {
    return false;
  }

  bool isCanonical() const {
    return true;
  }

  bool isDataParallel() const {
    return false;
  }

  Instruction* clone() const;

  void dump(llvm::raw_ostream &os) const;

  llvm::StringRef getOperandName(unsigned idx) const;
  void verify() const {
    assert(getDest()->getElementType() == getSrc()->getElementType() && "Invalid Element Type");
  }
};
} // namespace glow

namespace glow {
class FullyConnectedInst final : public Instruction {

 public:
  FullyConnectedInst(llvm::StringRef name, Value *Dest, Value *Src, Value *Weights, Value *Bias)
      : Instruction(name, Kinded::Kind::FullyConnectedInstKind, nullptr, {
          {Dest, OperandKind::Out},
          {Src, OperandKind::In},
          {Weights, OperandKind::In},
          {Bias, OperandKind::In},
      }) {}

  Value *getDest() const { return getOperand(0).first; }
  Value *getSrc() const { return getOperand(1).first; }
  Value *getWeights() const { return getOperand(2).first; }
  Value *getBias() const { return getOperand(3).first; }

  static bool classof(const Kinded *k) {
    return k->getKind() == Kinded::Kind::FullyConnectedInstKind;
  }

  bool isInplaceOp(unsigned dstIdx, unsigned srcIdx) const {
    return false;
  }

  bool isCanonical() const {
    return true;
  }

  bool isDataParallel() const {
    return false;
  }

  Instruction* clone() const;

  void dump(llvm::raw_ostream &os) const;

  llvm::StringRef getOperandName(unsigned idx) const;
  void verify() const {
    assert(getDest()->getElementType() == getSrc()->getElementType() && "Invalid Element Type");
  }
};
} // namespace glow

namespace glow {
class RowwiseQuantizedFullyConnectedInst final : public Instruction {

 public:
  RowwiseQuantizedFullyConnectedInst(llvm::StringRef name, Value *Dest, Value *Src, Value *Weights, Value *Bias, Value *Scales, Value *Offsets)
      : Instruction(name, Kinded::Kind::RowwiseQuantizedFullyConnectedInstKind, nullptr, {
          {Dest, OperandKind::Out},
          {Src, OperandKind::In},
          {Weights, OperandKind::In},
          {Bias, OperandKind::In},
          {Scales, OperandKind::In},
          {Offsets, OperandKind::In},
      }) {}

  Value *getDest() const { return getOperand(0).first; }
  Value *getSrc() const { return getOperand(1).first; }
  Value *getWeights() const { return getOperand(2).first; }
  Value *getBias() const { return getOperand(3).first; }
  Value *getScales() const { return getOperand(4).first; }
  Value *getOffsets() const { return getOperand(5).first; }

  static bool classof(const Kinded *k) {
    return k->getKind() == Kinded::Kind::RowwiseQuantizedFullyConnectedInstKind;
  }

  bool isInplaceOp(unsigned dstIdx, unsigned srcIdx) const {
    return false;
  }

  bool isCanonical() const {
    return true;
  }

  bool isDataParallel() const {
    return false;
  }

  Instruction* clone() const;

  void dump(llvm::raw_ostream &os) const;

  llvm::StringRef getOperandName(unsigned idx) const;
  void verify() const {
    assert(getDest()->getElementType() == getSrc()->getElementType() && "Invalid Element Type");
    assert(getDest()->getElementType() == ElemKind::Int8QTy && "Invalid Element Type");
  }
};
} // namespace glow

namespace glow {
class LocalResponseNormalizationGradInst final : public Instruction {
  unsigned_t HalfWindowSize_;
  float Alpha_;
  float Beta_;
  float K_;

 public:
  LocalResponseNormalizationGradInst(llvm::StringRef name, Value *Dest, Value *Src, Value *Scale, Value *DestGrad, Value *SrcGrad, unsigned_t HalfWindowSize, float Alpha, float Beta, float K)
      : Instruction(name, Kinded::Kind::LocalResponseNormalizationGradInstKind, Src->getType(), {
          {Dest, OperandKind::In},
          {Src, OperandKind::In},
          {Scale, OperandKind::In},
          {DestGrad, OperandKind::In},
          {SrcGrad, OperandKind::Out},
      }), HalfWindowSize_(HalfWindowSize), Alpha_(Alpha), Beta_(Beta), K_(K) {}

  Value *getDest() const { return getOperand(0).first; }
  Value *getSrc() const { return getOperand(1).first; }
  Value *getScale() const { return getOperand(2).first; }
  Value *getDestGrad() const { return getOperand(3).first; }
  Value *getSrcGrad() const { return getOperand(4).first; }
  unsigned_t getHalfWindowSize() const { return HalfWindowSize_; }
  float getAlpha() const { return Alpha_; }
  float getBeta() const { return Beta_; }
  float getK() const { return K_; }

  static bool classof(const Kinded *k) {
    return k->getKind() == Kinded::Kind::LocalResponseNormalizationGradInstKind;
  }

  bool isInplaceOp(unsigned dstIdx, unsigned srcIdx) const {
    return false;
  }

  bool isCanonical() const {
    return true;
  }

  bool isDataParallel() const {
    return false;
  }

  Instruction* clone() const;

  void dump(llvm::raw_ostream &os) const;

  llvm::StringRef getOperandName(unsigned idx) const;
  void verify() const {
    assert(getDestGrad()->getType() == getDest()->getType() && "Invalid Type");
    assert(getDestGrad()->getType() == getSrcGrad()->getType() && "Invalid Type");
    assert(getDestGrad()->getType() == getSrc()->getType() && "Invalid Type");
    assert(getDestGrad()->getType() == getScale()->getType() && "Invalid Type");
  }
};
} // namespace glow

namespace glow {
class LocalResponseNormalizationInst final : public Instruction {
  unsigned_t HalfWindowSize_;
  float Alpha_;
  float Beta_;
  float K_;

 public:
  LocalResponseNormalizationInst(llvm::StringRef name, Value *Dest, Value *Src, Value *Scale, unsigned_t HalfWindowSize, float Alpha, float Beta, float K)
      : Instruction(name, Kinded::Kind::LocalResponseNormalizationInstKind, Src->getType(), {
          {Dest, OperandKind::Out},
          {Src, OperandKind::In},
          {Scale, OperandKind::Out},
      }), HalfWindowSize_(HalfWindowSize), Alpha_(Alpha), Beta_(Beta), K_(K) {}

  Value *getDest() const { return getOperand(0).first; }
  Value *getSrc() const { return getOperand(1).first; }
  Value *getScale() const { return getOperand(2).first; }
  unsigned_t getHalfWindowSize() const { return HalfWindowSize_; }
  float getAlpha() const { return Alpha_; }
  float getBeta() const { return Beta_; }
  float getK() const { return K_; }

  static bool classof(const Kinded *k) {
    return k->getKind() == Kinded::Kind::LocalResponseNormalizationInstKind;
  }

  bool isInplaceOp(unsigned dstIdx, unsigned srcIdx) const {
    return false;
  }

  bool isCanonical() const {
    return true;
  }

  bool isDataParallel() const {
    return false;
  }

  Instruction* clone() const;

  void dump(llvm::raw_ostream &os) const;

  llvm::StringRef getOperandName(unsigned idx) const;
  void verify() const {
    assert(getDest()->getType() == getSrc()->getType() && "Invalid Type");
    assert(getDest()->getType() == getScale()->getType() && "Invalid Type");
  }
};
} // namespace glow

namespace glow {
class SoftMaxInst final : public Instruction {

 public:
  SoftMaxInst(llvm::StringRef name, Value *Dest, Value *Src)
      : Instruction(name, Kinded::Kind::SoftMaxInstKind, nullptr, {
          {Dest, OperandKind::Out},
          {Src, OperandKind::In},
      }) {}

  Value *getDest() const { return getOperand(0).first; }
  Value *getSrc() const { return getOperand(1).first; }

  static bool classof(const Kinded *k) {
    return k->getKind() == Kinded::Kind::SoftMaxInstKind;
  }

  bool isInplaceOp(unsigned dstIdx, unsigned srcIdx) const {
    return false;
  }

  bool isCanonical() const {
    return true;
  }

  bool isDataParallel() const {
    return false;
  }

  Instruction* clone() const;

  void dump(llvm::raw_ostream &os) const;

  llvm::StringRef getOperandName(unsigned idx) const;
  void verify() const {
    assert(getDest()->dims().equals(getSrc()->dims()) && "Invalid Shape");
  }
};
} // namespace glow

namespace glow {
class SoftMaxGradInst final : public Instruction {

 public:
  SoftMaxGradInst(llvm::StringRef name, Value *OrigDest, Value *OrigSrc, Value *Selected, Value *SrcGrad)
      : Instruction(name, Kinded::Kind::SoftMaxGradInstKind, nullptr, {
          {OrigDest, OperandKind::In},
          {OrigSrc, OperandKind::In},
          {Selected, OperandKind::In},
          {SrcGrad, OperandKind::Out},
      }) {}

  Value *getOrigDest() const { return getOperand(0).first; }
  Value *getOrigSrc() const { return getOperand(1).first; }
  Value *getSelected() const { return getOperand(2).first; }
  Value *getSrcGrad() const { return getOperand(3).first; }

  static bool classof(const Kinded *k) {
    return k->getKind() == Kinded::Kind::SoftMaxGradInstKind;
  }

  bool isInplaceOp(unsigned dstIdx, unsigned srcIdx) const {
    return false;
  }

  bool isCanonical() const {
    return true;
  }

  bool isDataParallel() const {
    return false;
  }

  Instruction* clone() const;

  void dump(llvm::raw_ostream &os) const;

  llvm::StringRef getOperandName(unsigned idx) const;
  void verify() const {
    assert(getOrigDest()->getType() == getOrigSrc()->getType() && "Invalid Type");
    assert(getOrigDest()->getType() == getSrcGrad()->getType() && "Invalid Type");
  }
};
} // namespace glow

namespace glow {
class CrossEntropyLossInst final : public Instruction {

 public:
  CrossEntropyLossInst(llvm::StringRef name, Value *P, Value *Labels, Value *CE)
      : Instruction(name, Kinded::Kind::CrossEntropyLossInstKind, nullptr, {
          {P, OperandKind::In},
          {Labels, OperandKind::In},
          {CE, OperandKind::Out},
      }) {}

  Value *getP() const { return getOperand(0).first; }
  Value *getLabels() const { return getOperand(1).first; }
  Value *getCE() const { return getOperand(2).first; }

  static bool classof(const Kinded *k) {
    return k->getKind() == Kinded::Kind::CrossEntropyLossInstKind;
  }

  bool isInplaceOp(unsigned dstIdx, unsigned srcIdx) const {
    return false;
  }

  bool isCanonical() const {
    return true;
  }

  bool isDataParallel() const {
    return false;
  }

  Instruction* clone() const;

  void dump(llvm::raw_ostream &os) const;

  llvm::StringRef getOperandName(unsigned idx) const;
  void verify() const {
    // Nothing to verify.
  }
};
} // namespace glow

namespace glow {
class CrossEntropyLossGradInst final : public Instruction {

 public:
  CrossEntropyLossGradInst(llvm::StringRef name, Value *CEGrad, Value *P, Value *Labels, Value *Pgrad, Value *Labelsgrad)
      : Instruction(name, Kinded::Kind::CrossEntropyLossGradInstKind, nullptr, {
          {CEGrad, OperandKind::In},
          {P, OperandKind::In},
          {Labels, OperandKind::In},
          {Pgrad, OperandKind::Out},
          {Labelsgrad, OperandKind::Out},
      }) {}

  Value *getCEGrad() const { return getOperand(0).first; }
  Value *getP() const { return getOperand(1).first; }
  Value *getLabels() const { return getOperand(2).first; }
  Value *getPgrad() const { return getOperand(3).first; }
  Value *getLabelsgrad() const { return getOperand(4).first; }

  static bool classof(const Kinded *k) {
    return k->getKind() == Kinded::Kind::CrossEntropyLossGradInstKind;
  }

  bool isInplaceOp(unsigned dstIdx, unsigned srcIdx) const {
    return false;
  }

  bool isCanonical() const {
    return true;
  }

  bool isDataParallel() const {
    return false;
  }

  Instruction* clone() const;

  void dump(llvm::raw_ostream &os) const;

  llvm::StringRef getOperandName(unsigned idx) const;
  void verify() const {
    // Nothing to verify.
  }
};
} // namespace glow

namespace glow {
class MatMulInst final : public Instruction {

 public:
  MatMulInst(llvm::StringRef name, Value *Dest, Value *LHS, Value *RHS)
      : Instruction(name, Kinded::Kind::MatMulInstKind, nullptr, {
          {Dest, OperandKind::Out},
          {LHS, OperandKind::In},
          {RHS, OperandKind::In},
      }) {}

  Value *getDest() const { return getOperand(0).first; }
  Value *getLHS() const { return getOperand(1).first; }
  Value *getRHS() const { return getOperand(2).first; }

  static bool classof(const Kinded *k) {
    return k->getKind() == Kinded::Kind::MatMulInstKind;
  }

  bool isInplaceOp(unsigned dstIdx, unsigned srcIdx) const {
    return false;
  }

  bool isCanonical() const {
    return true;
  }

  bool isDataParallel() const {
    return false;
  }

  Instruction* clone() const;

  void dump(llvm::raw_ostream &os) const;

  llvm::StringRef getOperandName(unsigned idx) const;
  void verify() const {
    assert(getDest()->getElementType() == getLHS()->getElementType() && "Invalid Element Type");
    assert(getDest()->getElementType() == getRHS()->getElementType() && "Invalid Element Type");
  }
};
} // namespace glow

namespace glow {
class BatchMatMulInst final : public Instruction {

 public:
  BatchMatMulInst(llvm::StringRef name, Value *Dest, Value *LHS, Value *RHS)
      : Instruction(name, Kinded::Kind::BatchMatMulInstKind, nullptr, {
          {Dest, OperandKind::Out},
          {LHS, OperandKind::In},
          {RHS, OperandKind::In},
      }) {}

  Value *getDest() const { return getOperand(0).first; }
  Value *getLHS() const { return getOperand(1).first; }
  Value *getRHS() const { return getOperand(2).first; }

  static bool classof(const Kinded *k) {
    return k->getKind() == Kinded::Kind::BatchMatMulInstKind;
  }

  bool isInplaceOp(unsigned dstIdx, unsigned srcIdx) const {
    return false;
  }

  bool isCanonical() const {
    return true;
  }

  bool isDataParallel() const {
    return false;
  }

  Instruction* clone() const;

  void dump(llvm::raw_ostream &os) const;

  llvm::StringRef getOperandName(unsigned idx) const;
  void verify() const {
    assert(getDest()->getElementType() == getLHS()->getElementType() && "Invalid Element Type");
    assert(getDest()->getElementType() == getRHS()->getElementType() && "Invalid Element Type");
  }
};
} // namespace glow

namespace glow {
class BatchedReduceAddInst final : public Instruction {
  unsigned_t Axis_;

 public:
  BatchedReduceAddInst(llvm::StringRef name, Value *Dest, Value *Batch, unsigned_t Axis)
      : Instruction(name, Kinded::Kind::BatchedReduceAddInstKind, nullptr, {
          {Dest, OperandKind::Out},
          {Batch, OperandKind::In},
      }), Axis_(Axis) {}

  Value *getDest() const { return getOperand(0).first; }
  Value *getBatch() const { return getOperand(1).first; }
  unsigned_t getAxis() const { return Axis_; }

  static bool classof(const Kinded *k) {
    return k->getKind() == Kinded::Kind::BatchedReduceAddInstKind;
  }

  bool isInplaceOp(unsigned dstIdx, unsigned srcIdx) const {
    return false;
  }

  bool isCanonical() const {
    return true;
  }

  bool isDataParallel() const {
    return false;
  }

  Instruction* clone() const;

  void dump(llvm::raw_ostream &os) const;

  llvm::StringRef getOperandName(unsigned idx) const;
  void verify() const {
    assert(getDest()->getElementType() == getBatch()->getElementType() && "Invalid Element Type");
  }
};
} // namespace glow

namespace glow {
class BatchedReduceMinInst final : public Instruction {
  std::vector<unsigned_t> Axes_;

 public:
  BatchedReduceMinInst(llvm::StringRef name, Value *Dest, Value *Batch, std::vector<unsigned_t> Axes)
      : Instruction(name, Kinded::Kind::BatchedReduceMinInstKind, nullptr, {
          {Dest, OperandKind::Out},
          {Batch, OperandKind::In},
      }), Axes_(Axes) {}

  Value *getDest() const { return getOperand(0).first; }
  Value *getBatch() const { return getOperand(1).first; }
  llvm::ArrayRef<unsigned_t> getAxes() const { return Axes_; }

  static bool classof(const Kinded *k) {
    return k->getKind() == Kinded::Kind::BatchedReduceMinInstKind;
  }

  bool isInplaceOp(unsigned dstIdx, unsigned srcIdx) const {
    return false;
  }

  bool isCanonical() const {
    return true;
  }

  bool isDataParallel() const {
    return false;
  }

  Instruction* clone() const;

  void dump(llvm::raw_ostream &os) const;

  llvm::StringRef getOperandName(unsigned idx) const;
  void verify() const {
    assert(getDest()->getElementType() == getBatch()->getElementType() && "Invalid Element Type");
  }
};
} // namespace glow

namespace glow {
class CumSumInst final : public Instruction {
  unsigned_t Exclusive_;
  unsigned_t Reverse_;

 public:
  CumSumInst(llvm::StringRef name, Value *Dest, Value *Input, unsigned_t Exclusive, unsigned_t Reverse)
      : Instruction(name, Kinded::Kind::CumSumInstKind, nullptr, {
          {Dest, OperandKind::Out},
          {Input, OperandKind::In},
      }), Exclusive_(Exclusive), Reverse_(Reverse) {}

  Value *getDest() const { return getOperand(0).first; }
  Value *getInput() const { return getOperand(1).first; }
  unsigned_t getExclusive() const { return Exclusive_; }
  unsigned_t getReverse() const { return Reverse_; }

  static bool classof(const Kinded *k) {
    return k->getKind() == Kinded::Kind::CumSumInstKind;
  }

  bool isInplaceOp(unsigned dstIdx, unsigned srcIdx) const {
  if (0 == dstIdx && 1 == srcIdx) { return true; }
    return false;
  }

  bool isCanonical() const {
    return true;
  }

  bool isDataParallel() const {
    return false;
  }

  Instruction* clone() const;

  void dump(llvm::raw_ostream &os) const;

  llvm::StringRef getOperandName(unsigned idx) const;
  void verify() const {
    assert(getDest()->getType() == getInput()->getType() && "Invalid Type");
  }
};
} // namespace glow

namespace glow {
class LengthsSumInst final : public Instruction {

 public:
  LengthsSumInst(llvm::StringRef name, Value *Dest, Value *Data, Value *Lengths)
      : Instruction(name, Kinded::Kind::LengthsSumInstKind, nullptr, {
          {Dest, OperandKind::Out},
          {Data, OperandKind::In},
          {Lengths, OperandKind::In},
      }) {}

  Value *getDest() const { return getOperand(0).first; }
  Value *getData() const { return getOperand(1).first; }
  Value *getLengths() const { return getOperand(2).first; }

  static bool classof(const Kinded *k) {
    return k->getKind() == Kinded::Kind::LengthsSumInstKind;
  }

  bool isInplaceOp(unsigned dstIdx, unsigned srcIdx) const {
    return false;
  }

  bool isCanonical() const {
    return true;
  }

  bool isDataParallel() const {
    return false;
  }

  Instruction* clone() const;

  void dump(llvm::raw_ostream &os) const;

  llvm::StringRef getOperandName(unsigned idx) const;
  void verify() const {
    assert(getDest()->getElementType() == getData()->getElementType() && "Invalid Element Type");
    assert(getLengths()->getElementType() == ElemKind::Int32ITy && "Invalid Element Type");
  }
};
} // namespace glow

namespace glow {
class SparseLengthsSumGradInst final : public Instruction {
  glow::LengthsMode LengthsMode_;
  float AvgLength_;

 public:
  SparseLengthsSumGradInst(llvm::StringRef name, Value *Data, Value *Indices, Value *Lengths, Value *DestGrad, Value *DataGrad, glow::LengthsMode LengthsMode, float AvgLength)
      : Instruction(name, Kinded::Kind::SparseLengthsSumGradInstKind, nullptr, {
          {Data, OperandKind::In},
          {Indices, OperandKind::In},
          {Lengths, OperandKind::In},
          {DestGrad, OperandKind::In},
          {DataGrad, OperandKind::Out},
      }), LengthsMode_(LengthsMode), AvgLength_(AvgLength) {}

  Value *getData() const { return getOperand(0).first; }
  Value *getIndices() const { return getOperand(1).first; }
  Value *getLengths() const { return getOperand(2).first; }
  Value *getDestGrad() const { return getOperand(3).first; }
  Value *getDataGrad() const { return getOperand(4).first; }
  glow::LengthsMode getLengthsMode() const { return LengthsMode_; }
  float getAvgLength() const { return AvgLength_; }

  static bool classof(const Kinded *k) {
    return k->getKind() == Kinded::Kind::SparseLengthsSumGradInstKind;
  }

  bool isInplaceOp(unsigned dstIdx, unsigned srcIdx) const {
    return false;
  }

  bool isCanonical() const {
    return true;
  }

  bool isDataParallel() const {
    return false;
  }

  Instruction* clone() const;

  void dump(llvm::raw_ostream &os) const;

  llvm::StringRef getOperandName(unsigned idx) const;
  void verify() const {
    assert(getDestGrad()->getElementType() == getDataGrad()->getElementType() && "Invalid Element Type");
    assert(getDestGrad()->getElementType() == getData()->getElementType() && "Invalid Element Type");
  }
};
} // namespace glow

namespace glow {
class SparseLengthsSumInst final : public Instruction {
  glow::LengthsMode LengthsMode_;
  float AvgLength_;

 public:
  SparseLengthsSumInst(llvm::StringRef name, Value *Dest, Value *Data, Value *Indices, Value *Lengths, glow::LengthsMode LengthsMode, float AvgLength)
      : Instruction(name, Kinded::Kind::SparseLengthsSumInstKind, nullptr, {
          {Dest, OperandKind::Out},
          {Data, OperandKind::In},
          {Indices, OperandKind::In},
          {Lengths, OperandKind::In},
      }), LengthsMode_(LengthsMode), AvgLength_(AvgLength) {}

  Value *getDest() const { return getOperand(0).first; }
  Value *getData() const { return getOperand(1).first; }
  Value *getIndices() const { return getOperand(2).first; }
  Value *getLengths() const { return getOperand(3).first; }
  glow::LengthsMode getLengthsMode() const { return LengthsMode_; }
  float getAvgLength() const { return AvgLength_; }

  static bool classof(const Kinded *k) {
    return k->getKind() == Kinded::Kind::SparseLengthsSumInstKind;
  }

  bool isInplaceOp(unsigned dstIdx, unsigned srcIdx) const {
    return false;
  }

  bool isCanonical() const {
    return true;
  }

  bool isDataParallel() const {
    return false;
  }

  Instruction* clone() const;

  void dump(llvm::raw_ostream &os) const;

  llvm::StringRef getOperandName(unsigned idx) const;
  void verify() const {
    assert(getDest()->getElementType() == getData()->getElementType() && "Invalid Element Type");
    assert(getLengths()->getElementType() == ElemKind::Int32ITy && "Invalid Element Type");
  }
};
} // namespace glow

namespace glow {
class SparseLengthsWeightedSumGradInst final : public Instruction {
  glow::LengthsMode LengthsMode_;
  float AvgLength_;

 public:
  SparseLengthsWeightedSumGradInst(llvm::StringRef name, Value *Data, Value *Weights, Value *Indices, Value *Lengths, Value *DestGrad, Value *DataGrad, Value *WeightsGrad, glow::LengthsMode LengthsMode, float AvgLength)
      : Instruction(name, Kinded::Kind::SparseLengthsWeightedSumGradInstKind, nullptr, {
          {Data, OperandKind::In},
          {Weights, OperandKind::In},
          {Indices, OperandKind::In},
          {Lengths, OperandKind::In},
          {DestGrad, OperandKind::In},
          {DataGrad, OperandKind::Out},
          {WeightsGrad, OperandKind::Out},
      }), LengthsMode_(LengthsMode), AvgLength_(AvgLength) {}

  Value *getData() const { return getOperand(0).first; }
  Value *getWeights() const { return getOperand(1).first; }
  Value *getIndices() const { return getOperand(2).first; }
  Value *getLengths() const { return getOperand(3).first; }
  Value *getDestGrad() const { return getOperand(4).first; }
  Value *getDataGrad() const { return getOperand(5).first; }
  Value *getWeightsGrad() const { return getOperand(6).first; }
  glow::LengthsMode getLengthsMode() const { return LengthsMode_; }
  float getAvgLength() const { return AvgLength_; }

  static bool classof(const Kinded *k) {
    return k->getKind() == Kinded::Kind::SparseLengthsWeightedSumGradInstKind;
  }

  bool isInplaceOp(unsigned dstIdx, unsigned srcIdx) const {
    return false;
  }

  bool isCanonical() const {
    return true;
  }

  bool isDataParallel() const {
    return false;
  }

  Instruction* clone() const;

  void dump(llvm::raw_ostream &os) const;

  llvm::StringRef getOperandName(unsigned idx) const;
  void verify() const {
    assert(getDestGrad()->getElementType() == getDataGrad()->getElementType() && "Invalid Element Type");
    assert(getDestGrad()->getElementType() == getData()->getElementType() && "Invalid Element Type");
    assert(getDestGrad()->getElementType() == getWeightsGrad()->getElementType() && "Invalid Element Type");
    assert(getDestGrad()->getElementType() == getWeights()->getElementType() && "Invalid Element Type");
    assert(getWeightsGrad()->dims().equals(getWeights()->dims()) && "Invalid Shape");
    assert(getWeightsGrad()->dims().equals(getIndices()->dims()) && "Invalid Shape");
  }
};
} // namespace glow

namespace glow {
class SparseLengthsWeightedSumInst final : public Instruction {
  glow::LengthsMode LengthsMode_;
  float AvgLength_;

 public:
  SparseLengthsWeightedSumInst(llvm::StringRef name, Value *Dest, Value *Data, Value *Weights, Value *Indices, Value *Lengths, glow::LengthsMode LengthsMode, float AvgLength)
      : Instruction(name, Kinded::Kind::SparseLengthsWeightedSumInstKind, nullptr, {
          {Dest, OperandKind::Out},
          {Data, OperandKind::In},
          {Weights, OperandKind::In},
          {Indices, OperandKind::In},
          {Lengths, OperandKind::In},
      }), LengthsMode_(LengthsMode), AvgLength_(AvgLength) {}

  Value *getDest() const { return getOperand(0).first; }
  Value *getData() const { return getOperand(1).first; }
  Value *getWeights() const { return getOperand(2).first; }
  Value *getIndices() const { return getOperand(3).first; }
  Value *getLengths() const { return getOperand(4).first; }
  glow::LengthsMode getLengthsMode() const { return LengthsMode_; }
  float getAvgLength() const { return AvgLength_; }

  static bool classof(const Kinded *k) {
    return k->getKind() == Kinded::Kind::SparseLengthsWeightedSumInstKind;
  }

  bool isInplaceOp(unsigned dstIdx, unsigned srcIdx) const {
    return false;
  }

  bool isCanonical() const {
    return true;
  }

  bool isDataParallel() const {
    return false;
  }

  Instruction* clone() const;

  void dump(llvm::raw_ostream &os) const;

  llvm::StringRef getOperandName(unsigned idx) const;
  void verify() const {
    assert(getDest()->getElementType() == getData()->getElementType() && "Invalid Element Type");
    assert(getDest()->getElementType() == getWeights()->getElementType() && "Invalid Element Type");
    assert(getLengths()->getElementType() == ElemKind::Int32ITy && "Invalid Element Type");
    assert(getWeights()->dims().equals(getIndices()->dims()) && "Invalid Shape");
  }
};
} // namespace glow

namespace glow {
class EmbeddingBagInst final : public Instruction {
  bool HasEndOffset_;
  glow::LengthsMode LengthsMode_;
  float AvgLength_;

 public:
  EmbeddingBagInst(llvm::StringRef name, Value *Dest, Value *Data, Value *Weights, Value *Indices, Value *Offsets, bool HasEndOffset, glow::LengthsMode LengthsMode, float AvgLength)
      : Instruction(name, Kinded::Kind::EmbeddingBagInstKind, nullptr, {
          {Dest, OperandKind::Out},
          {Data, OperandKind::In},
          {Weights, OperandKind::In},
          {Indices, OperandKind::In},
          {Offsets, OperandKind::In},
      }), HasEndOffset_(HasEndOffset), LengthsMode_(LengthsMode), AvgLength_(AvgLength) {}

  Value *getDest() const { return getOperand(0).first; }
  Value *getData() const { return getOperand(1).first; }
  Value *getWeights() const { return getOperand(2).first; }
  Value *getIndices() const { return getOperand(3).first; }
  Value *getOffsets() const { return getOperand(4).first; }
  bool getHasEndOffset() const { return HasEndOffset_; }
  glow::LengthsMode getLengthsMode() const { return LengthsMode_; }
  float getAvgLength() const { return AvgLength_; }

  static bool classof(const Kinded *k) {
    return k->getKind() == Kinded::Kind::EmbeddingBagInstKind;
  }

  bool isInplaceOp(unsigned dstIdx, unsigned srcIdx) const {
    return false;
  }

  bool isCanonical() const {
    return true;
  }

  bool isDataParallel() const {
    return false;
  }

  Instruction* clone() const;

  void dump(llvm::raw_ostream &os) const;

  llvm::StringRef getOperandName(unsigned idx) const;
  void verify() const {
    assert(getDest()->getElementType() == getData()->getElementType() && "Invalid Element Type");
    assert(getDest()->getElementType() == getWeights()->getElementType() && "Invalid Element Type");
    assert(getIndices()->getElementType() == ElemKind::Int64ITy && "Invalid Element Type");
    assert(getOffsets()->getElementType() == ElemKind::Int64ITy && "Invalid Element Type");
    assert(getWeights()->dims().equals(getIndices()->dims()) && "Invalid Shape");
  }
};
} // namespace glow

namespace glow {
class RowwiseQuantizedSparseLengthsWeightedSumInst final : public Instruction {
  bool UseFP16Accumulation_;
  glow::LengthsMode LengthsMode_;
  float AvgLength_;

 public:
  RowwiseQuantizedSparseLengthsWeightedSumInst(llvm::StringRef name, Value *Dest, Value *Data, Value *Scales, Value *Offsets, Value *Weights, Value *Indices, Value *Lengths, bool UseFP16Accumulation, glow::LengthsMode LengthsMode, float AvgLength)
      : Instruction(name, Kinded::Kind::RowwiseQuantizedSparseLengthsWeightedSumInstKind, nullptr, {
          {Dest, OperandKind::Out},
          {Data, OperandKind::In},
          {Scales, OperandKind::In},
          {Offsets, OperandKind::In},
          {Weights, OperandKind::In},
          {Indices, OperandKind::In},
          {Lengths, OperandKind::In},
      }), UseFP16Accumulation_(UseFP16Accumulation), LengthsMode_(LengthsMode), AvgLength_(AvgLength) {}

  Value *getDest() const { return getOperand(0).first; }
  Value *getData() const { return getOperand(1).first; }
  Value *getScales() const { return getOperand(2).first; }
  Value *getOffsets() const { return getOperand(3).first; }
  Value *getWeights() const { return getOperand(4).first; }
  Value *getIndices() const { return getOperand(5).first; }
  Value *getLengths() const { return getOperand(6).first; }
  bool getUseFP16Accumulation() const { return UseFP16Accumulation_; }
  glow::LengthsMode getLengthsMode() const { return LengthsMode_; }
  float getAvgLength() const { return AvgLength_; }

  static bool classof(const Kinded *k) {
    return k->getKind() == Kinded::Kind::RowwiseQuantizedSparseLengthsWeightedSumInstKind;
  }

  bool isInplaceOp(unsigned dstIdx, unsigned srcIdx) const {
    return false;
  }

  bool isCanonical() const {
    return true;
  }

  bool isDataParallel() const {
    return false;
  }

  Instruction* clone() const;

  void dump(llvm::raw_ostream &os) const;

  llvm::StringRef getOperandName(unsigned idx) const;
  void verify() const {
    assert(getData()->getElementType() == ElemKind::UInt8QTy && "Invalid Element Type");
    assert(getLengths()->getElementType() == ElemKind::Int32ITy && "Invalid Element Type");
    assert(getWeights()->dims().equals(getIndices()->dims()) && "Invalid Shape");
  }
};
} // namespace glow

namespace glow {
class FusedRowwiseQuantizedSparseLengthsWeightedSumInst final : public Instruction {
  bool UseFP16Accumulation_;
  glow::LengthsMode LengthsMode_;
  float AvgLength_;

 public:
  FusedRowwiseQuantizedSparseLengthsWeightedSumInst(llvm::StringRef name, Value *Dest, Value *Data, Value *Weights, Value *Indices, Value *Lengths, bool UseFP16Accumulation, glow::LengthsMode LengthsMode, float AvgLength)
      : Instruction(name, Kinded::Kind::FusedRowwiseQuantizedSparseLengthsWeightedSumInstKind, nullptr, {
          {Dest, OperandKind::Out},
          {Data, OperandKind::In},
          {Weights, OperandKind::In},
          {Indices, OperandKind::In},
          {Lengths, OperandKind::In},
      }), UseFP16Accumulation_(UseFP16Accumulation), LengthsMode_(LengthsMode), AvgLength_(AvgLength) {}

  Value *getDest() const { return getOperand(0).first; }
  Value *getData() const { return getOperand(1).first; }
  Value *getWeights() const { return getOperand(2).first; }
  Value *getIndices() const { return getOperand(3).first; }
  Value *getLengths() const { return getOperand(4).first; }
  bool getUseFP16Accumulation() const { return UseFP16Accumulation_; }
  glow::LengthsMode getLengthsMode() const { return LengthsMode_; }
  float getAvgLength() const { return AvgLength_; }

  static bool classof(const Kinded *k) {
    return k->getKind() == Kinded::Kind::FusedRowwiseQuantizedSparseLengthsWeightedSumInstKind;
  }

  bool isInplaceOp(unsigned dstIdx, unsigned srcIdx) const {
    return false;
  }

  bool isCanonical() const {
    return true;
  }

  bool isDataParallel() const {
    return false;
  }

  Instruction* clone() const;

  void dump(llvm::raw_ostream &os) const;

  llvm::StringRef getOperandName(unsigned idx) const;
  void verify() const {
    assert(getLengths()->getElementType() == ElemKind::Int32ITy && "Invalid Element Type");
    assert(getWeights()->dims().equals(getIndices()->dims()) && "Invalid Shape");
  }
};
} // namespace glow

namespace glow {
class EmbeddingBagByteRowwiseOffsetsInst final : public Instruction {
  bool UseFP16Accumulation_;
  bool HasEndOffset_;
  glow::LengthsMode LengthsMode_;
  float AvgLength_;

 public:
  EmbeddingBagByteRowwiseOffsetsInst(llvm::StringRef name, Value *Dest, Value *Data, Value *Weights, Value *Indices, Value *Offsets, bool UseFP16Accumulation, bool HasEndOffset, glow::LengthsMode LengthsMode, float AvgLength)
      : Instruction(name, Kinded::Kind::EmbeddingBagByteRowwiseOffsetsInstKind, nullptr, {
          {Dest, OperandKind::Out},
          {Data, OperandKind::In},
          {Weights, OperandKind::In},
          {Indices, OperandKind::In},
          {Offsets, OperandKind::In},
      }), UseFP16Accumulation_(UseFP16Accumulation), HasEndOffset_(HasEndOffset), LengthsMode_(LengthsMode), AvgLength_(AvgLength) {}

  Value *getDest() const { return getOperand(0).first; }
  Value *getData() const { return getOperand(1).first; }
  Value *getWeights() const { return getOperand(2).first; }
  Value *getIndices() const { return getOperand(3).first; }
  Value *getOffsets() const { return getOperand(4).first; }
  bool getUseFP16Accumulation() const { return UseFP16Accumulation_; }
  bool getHasEndOffset() const { return HasEndOffset_; }
  glow::LengthsMode getLengthsMode() const { return LengthsMode_; }
  float getAvgLength() const { return AvgLength_; }

  static bool classof(const Kinded *k) {
    return k->getKind() == Kinded::Kind::EmbeddingBagByteRowwiseOffsetsInstKind;
  }

  bool isInplaceOp(unsigned dstIdx, unsigned srcIdx) const {
    return false;
  }

  bool isCanonical() const {
    return true;
  }

  bool isDataParallel() const {
    return false;
  }

  Instruction* clone() const;

  void dump(llvm::raw_ostream &os) const;

  llvm::StringRef getOperandName(unsigned idx) const;
  void verify() const {
    assert(getIndices()->getElementType() == ElemKind::Int64ITy && "Invalid Element Type");
    assert(getOffsets()->getElementType() == ElemKind::Int64ITy && "Invalid Element Type");
    assert(getWeights()->dims().equals(getIndices()->dims()) && "Invalid Shape");
  }
};
} // namespace glow

namespace glow {
class LengthsToRangesInst final : public Instruction {

 public:
  LengthsToRangesInst(llvm::StringRef name, Value *Dest, Value *Lengths)
      : Instruction(name, Kinded::Kind::LengthsToRangesInstKind, nullptr, {
          {Dest, OperandKind::Out},
          {Lengths, OperandKind::In},
      }) {}

  Value *getDest() const { return getOperand(0).first; }
  Value *getLengths() const { return getOperand(1).first; }

  static bool classof(const Kinded *k) {
    return k->getKind() == Kinded::Kind::LengthsToRangesInstKind;
  }

  bool isInplaceOp(unsigned dstIdx, unsigned srcIdx) const {
    return false;
  }

  bool isCanonical() const {
    return true;
  }

  bool isDataParallel() const {
    return false;
  }

  Instruction* clone() const;

  void dump(llvm::raw_ostream &os) const;

  llvm::StringRef getOperandName(unsigned idx) const;
  void verify() const {
    assert(getDest()->getElementType() == ElemKind::Int32ITy && "Invalid Element Type");
    assert(getLengths()->getElementType() == ElemKind::Int32ITy && "Invalid Element Type");
  }
};
} // namespace glow

namespace glow {
class LengthsRangeFillInst final : public Instruction {

 public:
  LengthsRangeFillInst(llvm::StringRef name, Value *Dest, Value *Lengths)
      : Instruction(name, Kinded::Kind::LengthsRangeFillInstKind, nullptr, {
          {Dest, OperandKind::Out},
          {Lengths, OperandKind::In},
      }) {}

  Value *getDest() const { return getOperand(0).first; }
  Value *getLengths() const { return getOperand(1).first; }

  static bool classof(const Kinded *k) {
    return k->getKind() == Kinded::Kind::LengthsRangeFillInstKind;
  }

  bool isInplaceOp(unsigned dstIdx, unsigned srcIdx) const {
    return false;
  }

  bool isCanonical() const {
    return true;
  }

  bool isDataParallel() const {
    return false;
  }

  Instruction* clone() const;

  void dump(llvm::raw_ostream &os) const;

  llvm::StringRef getOperandName(unsigned idx) const;
  void verify() const {
    assert(getDest()->getElementType() == ElemKind::Int32ITy && "Invalid Element Type");
    assert(getLengths()->getElementType() == ElemKind::Int32ITy && "Invalid Element Type");
  }
};
} // namespace glow

namespace glow {
class SparseToDenseInst final : public Instruction {

 public:
  SparseToDenseInst(llvm::StringRef name, Value *Dest, Value *Indices, Value *Values)
      : Instruction(name, Kinded::Kind::SparseToDenseInstKind, nullptr, {
          {Dest, OperandKind::Out},
          {Indices, OperandKind::In},
          {Values, OperandKind::In},
      }) {}

  Value *getDest() const { return getOperand(0).first; }
  Value *getIndices() const { return getOperand(1).first; }
  Value *getValues() const { return getOperand(2).first; }

  static bool classof(const Kinded *k) {
    return k->getKind() == Kinded::Kind::SparseToDenseInstKind;
  }

  bool isInplaceOp(unsigned dstIdx, unsigned srcIdx) const {
    return false;
  }

  bool isCanonical() const {
    return true;
  }

  bool isDataParallel() const {
    return false;
  }

  Instruction* clone() const;

  void dump(llvm::raw_ostream &os) const;

  llvm::StringRef getOperandName(unsigned idx) const;
  void verify() const {
    assert(getDest()->getElementType() == getValues()->getElementType() && "Invalid Element Type");
  }
};
} // namespace glow

namespace glow {
class SparseToDenseMaskInst final : public Instruction {
  std::vector<dim_t> Mask_;

 public:
  SparseToDenseMaskInst(llvm::StringRef name, Value *Dest, Value *Indices, Value *Values, Value *DefaultValue, Value *Lengths, std::vector<dim_t> Mask)
      : Instruction(name, Kinded::Kind::SparseToDenseMaskInstKind, nullptr, {
          {Dest, OperandKind::Out},
          {Indices, OperandKind::In},
          {Values, OperandKind::In},
          {DefaultValue, OperandKind::In},
          {Lengths, OperandKind::In},
      }), Mask_(Mask) {}

  Value *getDest() const { return getOperand(0).first; }
  Value *getIndices() const { return getOperand(1).first; }
  Value *getValues() const { return getOperand(2).first; }
  Value *getDefaultValue() const { return getOperand(3).first; }
  Value *getLengths() const { return getOperand(4).first; }
  llvm::ArrayRef<dim_t> getMask() const { return Mask_; }

  static bool classof(const Kinded *k) {
    return k->getKind() == Kinded::Kind::SparseToDenseMaskInstKind;
  }

  bool isInplaceOp(unsigned dstIdx, unsigned srcIdx) const {
    return false;
  }

  bool isCanonical() const {
    return true;
  }

  bool isDataParallel() const {
    return false;
  }

  Instruction* clone() const;

  void dump(llvm::raw_ostream &os) const;

  llvm::StringRef getOperandName(unsigned idx) const;
  void verify() const {
    assert(getDest()->getElementType() == getValues()->getElementType() && "Invalid Element Type");
    assert(getDest()->getElementType() == getDefaultValue()->getElementType() && "Invalid Element Type");
    assert(getIndices()->getElementType() == ElemKind::Int64ITy && "Invalid Element Type");
    assert(getLengths()->getElementType() == ElemKind::Int32ITy && "Invalid Element Type");
  }
};
} // namespace glow

namespace glow {
class BatchedAddInst final : public Instruction {

 public:
  BatchedAddInst(llvm::StringRef name, Value *Dest, Value *Batch, Value *Slice)
      : Instruction(name, Kinded::Kind::BatchedAddInstKind, nullptr, {
          {Dest, OperandKind::Out},
          {Batch, OperandKind::In},
          {Slice, OperandKind::In},
      }) {}

  Value *getDest() const { return getOperand(0).first; }
  Value *getBatch() const { return getOperand(1).first; }
  Value *getSlice() const { return getOperand(2).first; }

  static bool classof(const Kinded *k) {
    return k->getKind() == Kinded::Kind::BatchedAddInstKind;
  }

  bool isInplaceOp(unsigned dstIdx, unsigned srcIdx) const {
  if (0 == dstIdx && 1 == srcIdx) { return true; }
    return false;
  }

  bool isCanonical() const {
    return true;
  }

  bool isDataParallel() const {
    return false;
  }

  Instruction* clone() const;

  void dump(llvm::raw_ostream &os) const;

  llvm::StringRef getOperandName(unsigned idx) const;
  void verify() const {
    assert(getBatch()->dims().equals(getDest()->dims()) && "Invalid Shape");
  }
};
} // namespace glow

namespace glow {
class ElementAddInst final : public Instruction {

 public:
  ElementAddInst(llvm::StringRef name, Value *Dest, Value *LHS, Value *RHS)
      : Instruction(name, Kinded::Kind::ElementAddInstKind, nullptr, {
          {Dest, OperandKind::Out},
          {LHS, OperandKind::In},
          {RHS, OperandKind::In},
      }) {}

  Value *getDest() const { return getOperand(0).first; }
  Value *getLHS() const { return getOperand(1).first; }
  Value *getRHS() const { return getOperand(2).first; }

  static bool classof(const Kinded *k) {
    return k->getKind() == Kinded::Kind::ElementAddInstKind;
  }

  bool isInplaceOp(unsigned dstIdx, unsigned srcIdx) const {
  if (0 == dstIdx && 1 == srcIdx) { return true; }
  if (0 == dstIdx && 2 == srcIdx) { return true; }
    return false;
  }

  bool isCanonical() const {
    return true;
  }

  bool isDataParallel() const {
    return true;
  }

  Instruction* clone() const;

  void dump(llvm::raw_ostream &os) const;

  llvm::StringRef getOperandName(unsigned idx) const;
  void verify() const {
    assert(getDest()->dims().equals(getLHS()->dims()) && "Invalid Shape");
    assert(getDest()->dims().equals(getRHS()->dims()) && "Invalid Shape");
  }
};
} // namespace glow

namespace glow {
class ElementSubInst final : public Instruction {

 public:
  ElementSubInst(llvm::StringRef name, Value *Dest, Value *LHS, Value *RHS)
      : Instruction(name, Kinded::Kind::ElementSubInstKind, nullptr, {
          {Dest, OperandKind::Out},
          {LHS, OperandKind::In},
          {RHS, OperandKind::In},
      }) {}

  Value *getDest() const { return getOperand(0).first; }
  Value *getLHS() const { return getOperand(1).first; }
  Value *getRHS() const { return getOperand(2).first; }

  static bool classof(const Kinded *k) {
    return k->getKind() == Kinded::Kind::ElementSubInstKind;
  }

  bool isInplaceOp(unsigned dstIdx, unsigned srcIdx) const {
  if (0 == dstIdx && 1 == srcIdx) { return true; }
  if (0 == dstIdx && 2 == srcIdx) { return true; }
    return false;
  }

  bool isCanonical() const {
    return true;
  }

  bool isDataParallel() const {
    return true;
  }

  Instruction* clone() const;

  void dump(llvm::raw_ostream &os) const;

  llvm::StringRef getOperandName(unsigned idx) const;
  void verify() const {
    assert(getDest()->dims().equals(getLHS()->dims()) && "Invalid Shape");
    assert(getDest()->dims().equals(getRHS()->dims()) && "Invalid Shape");
  }
};
} // namespace glow

namespace glow {
class ElementMulInst final : public Instruction {

 public:
  ElementMulInst(llvm::StringRef name, Value *Dest, Value *LHS, Value *RHS)
      : Instruction(name, Kinded::Kind::ElementMulInstKind, nullptr, {
          {Dest, OperandKind::Out},
          {LHS, OperandKind::In},
          {RHS, OperandKind::In},
      }) {}

  Value *getDest() const { return getOperand(0).first; }
  Value *getLHS() const { return getOperand(1).first; }
  Value *getRHS() const { return getOperand(2).first; }

  static bool classof(const Kinded *k) {
    return k->getKind() == Kinded::Kind::ElementMulInstKind;
  }

  bool isInplaceOp(unsigned dstIdx, unsigned srcIdx) const {
  if (0 == dstIdx && 1 == srcIdx) { return true; }
  if (0 == dstIdx && 2 == srcIdx) { return true; }
    return false;
  }

  bool isCanonical() const {
    return true;
  }

  bool isDataParallel() const {
    return true;
  }

  Instruction* clone() const;

  void dump(llvm::raw_ostream &os) const;

  llvm::StringRef getOperandName(unsigned idx) const;
  void verify() const {
    assert(getDest()->dims().equals(getLHS()->dims()) && "Invalid Shape");
    assert(getDest()->dims().equals(getRHS()->dims()) && "Invalid Shape");
  }
};
} // namespace glow

namespace glow {
class ElementDivInst final : public Instruction {

 public:
  ElementDivInst(llvm::StringRef name, Value *Dest, Value *LHS, Value *RHS)
      : Instruction(name, Kinded::Kind::ElementDivInstKind, nullptr, {
          {Dest, OperandKind::Out},
          {LHS, OperandKind::In},
          {RHS, OperandKind::In},
      }) {}

  Value *getDest() const { return getOperand(0).first; }
  Value *getLHS() const { return getOperand(1).first; }
  Value *getRHS() const { return getOperand(2).first; }

  static bool classof(const Kinded *k) {
    return k->getKind() == Kinded::Kind::ElementDivInstKind;
  }

  bool isInplaceOp(unsigned dstIdx, unsigned srcIdx) const {
  if (0 == dstIdx && 1 == srcIdx) { return true; }
  if (0 == dstIdx && 2 == srcIdx) { return true; }
    return false;
  }

  bool isCanonical() const {
    return true;
  }

  bool isDataParallel() const {
    return true;
  }

  Instruction* clone() const;

  void dump(llvm::raw_ostream &os) const;

  llvm::StringRef getOperandName(unsigned idx) const;
  void verify() const {
    assert(getDest()->dims().equals(getLHS()->dims()) && "Invalid Shape");
    assert(getDest()->dims().equals(getRHS()->dims()) && "Invalid Shape");
  }
};
} // namespace glow

namespace glow {
class ElementMaxInst final : public Instruction {

 public:
  ElementMaxInst(llvm::StringRef name, Value *Dest, Value *LHS, Value *RHS)
      : Instruction(name, Kinded::Kind::ElementMaxInstKind, nullptr, {
          {Dest, OperandKind::Out},
          {LHS, OperandKind::In},
          {RHS, OperandKind::In},
      }) {}

  Value *getDest() const { return getOperand(0).first; }
  Value *getLHS() const { return getOperand(1).first; }
  Value *getRHS() const { return getOperand(2).first; }

  static bool classof(const Kinded *k) {
    return k->getKind() == Kinded::Kind::ElementMaxInstKind;
  }

  bool isInplaceOp(unsigned dstIdx, unsigned srcIdx) const {
  if (0 == dstIdx && 1 == srcIdx) { return true; }
  if (0 == dstIdx && 2 == srcIdx) { return true; }
    return false;
  }

  bool isCanonical() const {
    return true;
  }

  bool isDataParallel() const {
    return true;
  }

  Instruction* clone() const;

  void dump(llvm::raw_ostream &os) const;

  llvm::StringRef getOperandName(unsigned idx) const;
  void verify() const {
    assert(getDest()->dims().equals(getLHS()->dims()) && "Invalid Shape");
    assert(getDest()->dims().equals(getRHS()->dims()) && "Invalid Shape");
  }
};
} // namespace glow

namespace glow {
class ElementMinInst final : public Instruction {

 public:
  ElementMinInst(llvm::StringRef name, Value *Dest, Value *LHS, Value *RHS)
      : Instruction(name, Kinded::Kind::ElementMinInstKind, nullptr, {
          {Dest, OperandKind::Out},
          {LHS, OperandKind::In},
          {RHS, OperandKind::In},
      }) {}

  Value *getDest() const { return getOperand(0).first; }
  Value *getLHS() const { return getOperand(1).first; }
  Value *getRHS() const { return getOperand(2).first; }

  static bool classof(const Kinded *k) {
    return k->getKind() == Kinded::Kind::ElementMinInstKind;
  }

  bool isInplaceOp(unsigned dstIdx, unsigned srcIdx) const {
  if (0 == dstIdx && 1 == srcIdx) { return true; }
  if (0 == dstIdx && 2 == srcIdx) { return true; }
    return false;
  }

  bool isCanonical() const {
    return true;
  }

  bool isDataParallel() const {
    return true;
  }

  Instruction* clone() const;

  void dump(llvm::raw_ostream &os) const;

  llvm::StringRef getOperandName(unsigned idx) const;
  void verify() const {
    assert(getDest()->dims().equals(getLHS()->dims()) && "Invalid Shape");
    assert(getDest()->dims().equals(getRHS()->dims()) && "Invalid Shape");
  }
};
} // namespace glow

namespace glow {
class ElementCmpEQInst final : public Instruction {

 public:
  ElementCmpEQInst(llvm::StringRef name, Value *Dest, Value *LHS, Value *RHS)
      : Instruction(name, Kinded::Kind::ElementCmpEQInstKind, nullptr, {
          {Dest, OperandKind::Out},
          {LHS, OperandKind::In},
          {RHS, OperandKind::In},
      }) {}

  Value *getDest() const { return getOperand(0).first; }
  Value *getLHS() const { return getOperand(1).first; }
  Value *getRHS() const { return getOperand(2).first; }

  static bool classof(const Kinded *k) {
    return k->getKind() == Kinded::Kind::ElementCmpEQInstKind;
  }

  bool isInplaceOp(unsigned dstIdx, unsigned srcIdx) const {
    return false;
  }

  bool isCanonical() const {
    return true;
  }

  bool isDataParallel() const {
    return true;
  }

  Instruction* clone() const;

  void dump(llvm::raw_ostream &os) const;

  llvm::StringRef getOperandName(unsigned idx) const;
  void verify() const {
    assert(getDest()->dims().equals(getLHS()->dims()) && "Invalid Shape");
    assert(getDest()->dims().equals(getRHS()->dims()) && "Invalid Shape");
    assert(getLHS()->getElementType() == getRHS()->getElementType() && "Invalid Element Type");
    assert(getDest()->getElementType() == ElemKind::BoolTy && "Invalid Element Type");
  }
};
} // namespace glow

namespace glow {
class ElementCmpNEQInst final : public Instruction {

 public:
  ElementCmpNEQInst(llvm::StringRef name, Value *Dest, Value *LHS, Value *RHS)
      : Instruction(name, Kinded::Kind::ElementCmpNEQInstKind, nullptr, {
          {Dest, OperandKind::Out},
          {LHS, OperandKind::In},
          {RHS, OperandKind::In},
      }) {}

  Value *getDest() const { return getOperand(0).first; }
  Value *getLHS() const { return getOperand(1).first; }
  Value *getRHS() const { return getOperand(2).first; }

  static bool classof(const Kinded *k) {
    return k->getKind() == Kinded::Kind::ElementCmpNEQInstKind;
  }

  bool isInplaceOp(unsigned dstIdx, unsigned srcIdx) const {
    return false;
  }

  bool isCanonical() const {
    return true;
  }

  bool isDataParallel() const {
    return true;
  }

  Instruction* clone() const;

  void dump(llvm::raw_ostream &os) const;

  llvm::StringRef getOperandName(unsigned idx) const;
  void verify() const {
    assert(getDest()->dims().equals(getLHS()->dims()) && "Invalid Shape");
    assert(getDest()->dims().equals(getRHS()->dims()) && "Invalid Shape");
    assert(getLHS()->getElementType() == getRHS()->getElementType() && "Invalid Element Type");
    assert(getDest()->getElementType() == ElemKind::BoolTy && "Invalid Element Type");
  }
};
} // namespace glow

namespace glow {
class ElementCmpLTInst final : public Instruction {

 public:
  ElementCmpLTInst(llvm::StringRef name, Value *Dest, Value *LHS, Value *RHS)
      : Instruction(name, Kinded::Kind::ElementCmpLTInstKind, nullptr, {
          {Dest, OperandKind::Out},
          {LHS, OperandKind::In},
          {RHS, OperandKind::In},
      }) {}

  Value *getDest() const { return getOperand(0).first; }
  Value *getLHS() const { return getOperand(1).first; }
  Value *getRHS() const { return getOperand(2).first; }

  static bool classof(const Kinded *k) {
    return k->getKind() == Kinded::Kind::ElementCmpLTInstKind;
  }

  bool isInplaceOp(unsigned dstIdx, unsigned srcIdx) const {
    return false;
  }

  bool isCanonical() const {
    return true;
  }

  bool isDataParallel() const {
    return true;
  }

  Instruction* clone() const;

  void dump(llvm::raw_ostream &os) const;

  llvm::StringRef getOperandName(unsigned idx) const;
  void verify() const {
    assert(getDest()->dims().equals(getLHS()->dims()) && "Invalid Shape");
    assert(getDest()->dims().equals(getRHS()->dims()) && "Invalid Shape");
    assert(getLHS()->getElementType() == getRHS()->getElementType() && "Invalid Element Type");
    assert(getDest()->getElementType() == ElemKind::BoolTy && "Invalid Element Type");
  }
};
} // namespace glow

namespace glow {
class ElementCmpLTEInst final : public Instruction {

 public:
  ElementCmpLTEInst(llvm::StringRef name, Value *Dest, Value *LHS, Value *RHS)
      : Instruction(name, Kinded::Kind::ElementCmpLTEInstKind, nullptr, {
          {Dest, OperandKind::Out},
          {LHS, OperandKind::In},
          {RHS, OperandKind::In},
      }) {}

  Value *getDest() const { return getOperand(0).first; }
  Value *getLHS() const { return getOperand(1).first; }
  Value *getRHS() const { return getOperand(2).first; }

  static bool classof(const Kinded *k) {
    return k->getKind() == Kinded::Kind::ElementCmpLTEInstKind;
  }

  bool isInplaceOp(unsigned dstIdx, unsigned srcIdx) const {
    return false;
  }

  bool isCanonical() const {
    return true;
  }

  bool isDataParallel() const {
    return true;
  }

  Instruction* clone() const;

  void dump(llvm::raw_ostream &os) const;

  llvm::StringRef getOperandName(unsigned idx) const;
  void verify() const {
    assert(getDest()->dims().equals(getLHS()->dims()) && "Invalid Shape");
    assert(getDest()->dims().equals(getRHS()->dims()) && "Invalid Shape");
    assert(getLHS()->getElementType() == getRHS()->getElementType() && "Invalid Element Type");
    assert(getDest()->getElementType() == ElemKind::BoolTy && "Invalid Element Type");
  }
};
} // namespace glow

namespace glow {
class ElementIsNaNInst final : public Instruction {

 public:
  ElementIsNaNInst(llvm::StringRef name, Value *Dest, Value *Src)
      : Instruction(name, Kinded::Kind::ElementIsNaNInstKind, nullptr, {
          {Dest, OperandKind::Out},
          {Src, OperandKind::In},
      }) {}

  Value *getDest() const { return getOperand(0).first; }
  Value *getSrc() const { return getOperand(1).first; }

  static bool classof(const Kinded *k) {
    return k->getKind() == Kinded::Kind::ElementIsNaNInstKind;
  }

  bool isInplaceOp(unsigned dstIdx, unsigned srcIdx) const {
    return false;
  }

  bool isCanonical() const {
    return true;
  }

  bool isDataParallel() const {
    return true;
  }

  Instruction* clone() const;

  void dump(llvm::raw_ostream &os) const;

  llvm::StringRef getOperandName(unsigned idx) const;
  void verify() const {
    assert(getDest()->dims().equals(getSrc()->dims()) && "Invalid Shape");
    assert(getDest()->getElementType() == ElemKind::BoolTy && "Invalid Element Type");
  }
};
} // namespace glow

namespace glow {
class ElementPowInst final : public Instruction {

 public:
  ElementPowInst(llvm::StringRef name, Value *Dest, Value *LHS, Value *RHS)
      : Instruction(name, Kinded::Kind::ElementPowInstKind, nullptr, {
          {Dest, OperandKind::Out},
          {LHS, OperandKind::In},
          {RHS, OperandKind::In},
      }) {}

  Value *getDest() const { return getOperand(0).first; }
  Value *getLHS() const { return getOperand(1).first; }
  Value *getRHS() const { return getOperand(2).first; }

  static bool classof(const Kinded *k) {
    return k->getKind() == Kinded::Kind::ElementPowInstKind;
  }

  bool isInplaceOp(unsigned dstIdx, unsigned srcIdx) const {
  if (0 == dstIdx && 1 == srcIdx) { return true; }
  if (0 == dstIdx && 2 == srcIdx) { return true; }
    return false;
  }

  bool isCanonical() const {
    return true;
  }

  bool isDataParallel() const {
    return true;
  }

  Instruction* clone() const;

  void dump(llvm::raw_ostream &os) const;

  llvm::StringRef getOperandName(unsigned idx) const;
  void verify() const {
    assert(getDest()->dims().equals(getLHS()->dims()) && "Invalid Shape");
    assert(getDest()->dims().equals(getRHS()->dims()) && "Invalid Shape");
  }
};
} // namespace glow

namespace glow {
class ElementAndInst final : public Instruction {

 public:
  ElementAndInst(llvm::StringRef name, Value *Dest, Value *LHS, Value *RHS)
      : Instruction(name, Kinded::Kind::ElementAndInstKind, nullptr, {
          {Dest, OperandKind::Out},
          {LHS, OperandKind::In},
          {RHS, OperandKind::In},
      }) {}

  Value *getDest() const { return getOperand(0).first; }
  Value *getLHS() const { return getOperand(1).first; }
  Value *getRHS() const { return getOperand(2).first; }

  static bool classof(const Kinded *k) {
    return k->getKind() == Kinded::Kind::ElementAndInstKind;
  }

  bool isInplaceOp(unsigned dstIdx, unsigned srcIdx) const {
  if (0 == dstIdx && 1 == srcIdx) { return true; }
  if (0 == dstIdx && 2 == srcIdx) { return true; }
    return false;
  }

  bool isCanonical() const {
    return true;
  }

  bool isDataParallel() const {
    return true;
  }

  Instruction* clone() const;

  void dump(llvm::raw_ostream &os) const;

  llvm::StringRef getOperandName(unsigned idx) const;
  void verify() const {
    assert(getDest()->dims().equals(getLHS()->dims()) && "Invalid Shape");
    assert(getDest()->dims().equals(getRHS()->dims()) && "Invalid Shape");
    assert(getLHS()->getElementType() == ElemKind::BoolTy && "Invalid Element Type");
    assert(getRHS()->getElementType() == ElemKind::BoolTy && "Invalid Element Type");
    assert(getDest()->getElementType() == ElemKind::BoolTy && "Invalid Element Type");
  }
};
} // namespace glow

namespace glow {
class ElementOrInst final : public Instruction {

 public:
  ElementOrInst(llvm::StringRef name, Value *Dest, Value *LHS, Value *RHS)
      : Instruction(name, Kinded::Kind::ElementOrInstKind, nullptr, {
          {Dest, OperandKind::Out},
          {LHS, OperandKind::In},
          {RHS, OperandKind::In},
      }) {}

  Value *getDest() const { return getOperand(0).first; }
  Value *getLHS() const { return getOperand(1).first; }
  Value *getRHS() const { return getOperand(2).first; }

  static bool classof(const Kinded *k) {
    return k->getKind() == Kinded::Kind::ElementOrInstKind;
  }

  bool isInplaceOp(unsigned dstIdx, unsigned srcIdx) const {
  if (0 == dstIdx && 1 == srcIdx) { return true; }
  if (0 == dstIdx && 2 == srcIdx) { return true; }
    return false;
  }

  bool isCanonical() const {
    return true;
  }

  bool isDataParallel() const {
    return true;
  }

  Instruction* clone() const;

  void dump(llvm::raw_ostream &os) const;

  llvm::StringRef getOperandName(unsigned idx) const;
  void verify() const {
    assert(getDest()->dims().equals(getLHS()->dims()) && "Invalid Shape");
    assert(getDest()->dims().equals(getRHS()->dims()) && "Invalid Shape");
    assert(getLHS()->getElementType() == ElemKind::BoolTy && "Invalid Element Type");
    assert(getRHS()->getElementType() == ElemKind::BoolTy && "Invalid Element Type");
    assert(getDest()->getElementType() == ElemKind::BoolTy && "Invalid Element Type");
  }
};
} // namespace glow

namespace glow {
class ElementXorInst final : public Instruction {

 public:
  ElementXorInst(llvm::StringRef name, Value *Dest, Value *LHS, Value *RHS)
      : Instruction(name, Kinded::Kind::ElementXorInstKind, nullptr, {
          {Dest, OperandKind::Out},
          {LHS, OperandKind::In},
          {RHS, OperandKind::In},
      }) {}

  Value *getDest() const { return getOperand(0).first; }
  Value *getLHS() const { return getOperand(1).first; }
  Value *getRHS() const { return getOperand(2).first; }

  static bool classof(const Kinded *k) {
    return k->getKind() == Kinded::Kind::ElementXorInstKind;
  }

  bool isInplaceOp(unsigned dstIdx, unsigned srcIdx) const {
  if (0 == dstIdx && 1 == srcIdx) { return true; }
  if (0 == dstIdx && 2 == srcIdx) { return true; }
    return false;
  }

  bool isCanonical() const {
    return true;
  }

  bool isDataParallel() const {
    return true;
  }

  Instruction* clone() const;

  void dump(llvm::raw_ostream &os) const;

  llvm::StringRef getOperandName(unsigned idx) const;
  void verify() const {
    assert(getDest()->dims().equals(getLHS()->dims()) && "Invalid Shape");
    assert(getDest()->dims().equals(getRHS()->dims()) && "Invalid Shape");
    assert(getLHS()->getElementType() == ElemKind::BoolTy && "Invalid Element Type");
    assert(getRHS()->getElementType() == ElemKind::BoolTy && "Invalid Element Type");
    assert(getDest()->getElementType() == ElemKind::BoolTy && "Invalid Element Type");
  }
};
} // namespace glow

namespace glow {
class ElementNotInst final : public Instruction {

 public:
  ElementNotInst(llvm::StringRef name, Value *Dest, Value *Src)
      : Instruction(name, Kinded::Kind::ElementNotInstKind, nullptr, {
          {Dest, OperandKind::Out},
          {Src, OperandKind::In},
      }) {}

  Value *getDest() const { return getOperand(0).first; }
  Value *getSrc() const { return getOperand(1).first; }

  static bool classof(const Kinded *k) {
    return k->getKind() == Kinded::Kind::ElementNotInstKind;
  }

  bool isInplaceOp(unsigned dstIdx, unsigned srcIdx) const {
  if (0 == dstIdx && 1 == srcIdx) { return true; }
    return false;
  }

  bool isCanonical() const {
    return true;
  }

  bool isDataParallel() const {
    return true;
  }

  Instruction* clone() const;

  void dump(llvm::raw_ostream &os) const;

  llvm::StringRef getOperandName(unsigned idx) const;
  void verify() const {
    assert(getDest()->dims().equals(getSrc()->dims()) && "Invalid Shape");
    assert(getSrc()->getElementType() == ElemKind::BoolTy && "Invalid Element Type");
    assert(getDest()->getElementType() == ElemKind::BoolTy && "Invalid Element Type");
  }
};
} // namespace glow

namespace glow {
class ElementNegInst final : public Instruction {

 public:
  ElementNegInst(llvm::StringRef name, Value *Dest, Value *Src)
      : Instruction(name, Kinded::Kind::ElementNegInstKind, nullptr, {
          {Dest, OperandKind::Out},
          {Src, OperandKind::In},
      }) {}

  Value *getDest() const { return getOperand(0).first; }
  Value *getSrc() const { return getOperand(1).first; }

  static bool classof(const Kinded *k) {
    return k->getKind() == Kinded::Kind::ElementNegInstKind;
  }

  bool isInplaceOp(unsigned dstIdx, unsigned srcIdx) const {
  if (0 == dstIdx && 1 == srcIdx) { return true; }
    return false;
  }

  bool isCanonical() const {
    return true;
  }

  bool isDataParallel() const {
    return true;
  }

  Instruction* clone() const;

  void dump(llvm::raw_ostream &os) const;

  llvm::StringRef getOperandName(unsigned idx) const;
  void verify() const {
    assert(getDest()->dims().equals(getSrc()->dims()) && "Invalid Shape");
    assert(getDest()->getElementType() == getSrc()->getElementType() && "Invalid Element Type");
  }
};
} // namespace glow

namespace glow {
class ElementAbsInst final : public Instruction {

 public:
  ElementAbsInst(llvm::StringRef name, Value *Dest, Value *Src)
      : Instruction(name, Kinded::Kind::ElementAbsInstKind, nullptr, {
          {Dest, OperandKind::Out},
          {Src, OperandKind::In},
      }) {}

  Value *getDest() const { return getOperand(0).first; }
  Value *getSrc() const { return getOperand(1).first; }

  static bool classof(const Kinded *k) {
    return k->getKind() == Kinded::Kind::ElementAbsInstKind;
  }

  bool isInplaceOp(unsigned dstIdx, unsigned srcIdx) const {
  if (0 == dstIdx && 1 == srcIdx) { return true; }
    return false;
  }

  bool isCanonical() const {
    return true;
  }

  bool isDataParallel() const {
    return true;
  }

  Instruction* clone() const;

  void dump(llvm::raw_ostream &os) const;

  llvm::StringRef getOperandName(unsigned idx) const;
  void verify() const {
    assert(getDest()->dims().equals(getSrc()->dims()) && "Invalid Shape");
    assert(getDest()->getElementType() == getSrc()->getElementType() && "Invalid Element Type");
  }
};
} // namespace glow

namespace glow {
class ElementFloorInst final : public Instruction {

 public:
  ElementFloorInst(llvm::StringRef name, Value *Dest, Value *Src)
      : Instruction(name, Kinded::Kind::ElementFloorInstKind, nullptr, {
          {Dest, OperandKind::Out},
          {Src, OperandKind::In},
      }) {}

  Value *getDest() const { return getOperand(0).first; }
  Value *getSrc() const { return getOperand(1).first; }

  static bool classof(const Kinded *k) {
    return k->getKind() == Kinded::Kind::ElementFloorInstKind;
  }

  bool isInplaceOp(unsigned dstIdx, unsigned srcIdx) const {
  if (0 == dstIdx && 1 == srcIdx) { return true; }
    return false;
  }

  bool isCanonical() const {
    return true;
  }

  bool isDataParallel() const {
    return true;
  }

  Instruction* clone() const;

  void dump(llvm::raw_ostream &os) const;

  llvm::StringRef getOperandName(unsigned idx) const;
  void verify() const {
    assert(getDest()->dims().equals(getSrc()->dims()) && "Invalid Shape");
    assert(getDest()->getElementType() == getSrc()->getElementType() && "Invalid Element Type");
  }
};
} // namespace glow

namespace glow {
class ElementCeilInst final : public Instruction {

 public:
  ElementCeilInst(llvm::StringRef name, Value *Dest, Value *Src)
      : Instruction(name, Kinded::Kind::ElementCeilInstKind, nullptr, {
          {Dest, OperandKind::Out},
          {Src, OperandKind::In},
      }) {}

  Value *getDest() const { return getOperand(0).first; }
  Value *getSrc() const { return getOperand(1).first; }

  static bool classof(const Kinded *k) {
    return k->getKind() == Kinded::Kind::ElementCeilInstKind;
  }

  bool isInplaceOp(unsigned dstIdx, unsigned srcIdx) const {
  if (0 == dstIdx && 1 == srcIdx) { return true; }
    return false;
  }

  bool isCanonical() const {
    return true;
  }

  bool isDataParallel() const {
    return true;
  }

  Instruction* clone() const;

  void dump(llvm::raw_ostream &os) const;

  llvm::StringRef getOperandName(unsigned idx) const;
  void verify() const {
    assert(getDest()->dims().equals(getSrc()->dims()) && "Invalid Shape");
    assert(getDest()->getElementType() == getSrc()->getElementType() && "Invalid Element Type");
  }
};
} // namespace glow

namespace glow {
class ElementRoundInst final : public Instruction {

 public:
  ElementRoundInst(llvm::StringRef name, Value *Dest, Value *Src)
      : Instruction(name, Kinded::Kind::ElementRoundInstKind, nullptr, {
          {Dest, OperandKind::Out},
          {Src, OperandKind::In},
      }) {}

  Value *getDest() const { return getOperand(0).first; }
  Value *getSrc() const { return getOperand(1).first; }

  static bool classof(const Kinded *k) {
    return k->getKind() == Kinded::Kind::ElementRoundInstKind;
  }

  bool isInplaceOp(unsigned dstIdx, unsigned srcIdx) const {
  if (0 == dstIdx && 1 == srcIdx) { return true; }
    return false;
  }

  bool isCanonical() const {
    return true;
  }

  bool isDataParallel() const {
    return true;
  }

  Instruction* clone() const;

  void dump(llvm::raw_ostream &os) const;

  llvm::StringRef getOperandName(unsigned idx) const;
  void verify() const {
    assert(getDest()->dims().equals(getSrc()->dims()) && "Invalid Shape");
    assert(getDest()->getElementType() == getSrc()->getElementType() && "Invalid Element Type");
  }
};
} // namespace glow

namespace glow {
class ElementSqrtInst final : public Instruction {

 public:
  ElementSqrtInst(llvm::StringRef name, Value *Dest, Value *Src)
      : Instruction(name, Kinded::Kind::ElementSqrtInstKind, nullptr, {
          {Dest, OperandKind::Out},
          {Src, OperandKind::In},
      }) {}

  Value *getDest() const { return getOperand(0).first; }
  Value *getSrc() const { return getOperand(1).first; }

  static bool classof(const Kinded *k) {
    return k->getKind() == Kinded::Kind::ElementSqrtInstKind;
  }

  bool isInplaceOp(unsigned dstIdx, unsigned srcIdx) const {
  if (0 == dstIdx && 1 == srcIdx) { return true; }
    return false;
  }

  bool isCanonical() const {
    return true;
  }

  bool isDataParallel() const {
    return true;
  }

  Instruction* clone() const;

  void dump(llvm::raw_ostream &os) const;

  llvm::StringRef getOperandName(unsigned idx) const;
  void verify() const {
    assert(getDest()->dims().equals(getSrc()->dims()) && "Invalid Shape");
    assert(getDest()->getElementType() == getSrc()->getElementType() && "Invalid Element Type");
  }
};
} // namespace glow

namespace glow {
class ElementRsqrtInst final : public Instruction {

 public:
  ElementRsqrtInst(llvm::StringRef name, Value *Dest, Value *Src)
      : Instruction(name, Kinded::Kind::ElementRsqrtInstKind, nullptr, {
          {Dest, OperandKind::Out},
          {Src, OperandKind::In},
      }) {}

  Value *getDest() const { return getOperand(0).first; }
  Value *getSrc() const { return getOperand(1).first; }

  static bool classof(const Kinded *k) {
    return k->getKind() == Kinded::Kind::ElementRsqrtInstKind;
  }

  bool isInplaceOp(unsigned dstIdx, unsigned srcIdx) const {
  if (0 == dstIdx && 1 == srcIdx) { return true; }
    return false;
  }

  bool isCanonical() const {
    return true;
  }

  bool isDataParallel() const {
    return true;
  }

  Instruction* clone() const;

  void dump(llvm::raw_ostream &os) const;

  llvm::StringRef getOperandName(unsigned idx) const;
  void verify() const {
    assert(getDest()->dims().equals(getSrc()->dims()) && "Invalid Shape");
    assert(getDest()->getElementType() == getSrc()->getElementType() && "Invalid Element Type");
  }
};
} // namespace glow

namespace glow {
class ElementReciprocalInst final : public Instruction {

 public:
  ElementReciprocalInst(llvm::StringRef name, Value *Dest, Value *Src)
      : Instruction(name, Kinded::Kind::ElementReciprocalInstKind, nullptr, {
          {Dest, OperandKind::Out},
          {Src, OperandKind::In},
      }) {}

  Value *getDest() const { return getOperand(0).first; }
  Value *getSrc() const { return getOperand(1).first; }

  static bool classof(const Kinded *k) {
    return k->getKind() == Kinded::Kind::ElementReciprocalInstKind;
  }

  bool isInplaceOp(unsigned dstIdx, unsigned srcIdx) const {
  if (0 == dstIdx && 1 == srcIdx) { return true; }
    return false;
  }

  bool isCanonical() const {
    return true;
  }

  bool isDataParallel() const {
    return true;
  }

  Instruction* clone() const;

  void dump(llvm::raw_ostream &os) const;

  llvm::StringRef getOperandName(unsigned idx) const;
  void verify() const {
    assert(getDest()->dims().equals(getSrc()->dims()) && "Invalid Shape");
    assert(getDest()->getElementType() == getSrc()->getElementType() && "Invalid Element Type");
  }
};
} // namespace glow

namespace glow {
class ElementSinInst final : public Instruction {

 public:
  ElementSinInst(llvm::StringRef name, Value *Dest, Value *Src)
      : Instruction(name, Kinded::Kind::ElementSinInstKind, nullptr, {
          {Dest, OperandKind::Out},
          {Src, OperandKind::In},
      }) {}

  Value *getDest() const { return getOperand(0).first; }
  Value *getSrc() const { return getOperand(1).first; }

  static bool classof(const Kinded *k) {
    return k->getKind() == Kinded::Kind::ElementSinInstKind;
  }

  bool isInplaceOp(unsigned dstIdx, unsigned srcIdx) const {
  if (0 == dstIdx && 1 == srcIdx) { return true; }
    return false;
  }

  bool isCanonical() const {
    return true;
  }

  bool isDataParallel() const {
    return true;
  }

  Instruction* clone() const;

  void dump(llvm::raw_ostream &os) const;

  llvm::StringRef getOperandName(unsigned idx) const;
  void verify() const {
    assert(getDest()->dims().equals(getSrc()->dims()) && "Invalid Shape");
    assert(getDest()->getElementType() == getSrc()->getElementType() && "Invalid Element Type");
  }
};
} // namespace glow

namespace glow {
class ElementCosInst final : public Instruction {

 public:
  ElementCosInst(llvm::StringRef name, Value *Dest, Value *Src)
      : Instruction(name, Kinded::Kind::ElementCosInstKind, nullptr, {
          {Dest, OperandKind::Out},
          {Src, OperandKind::In},
      }) {}

  Value *getDest() const { return getOperand(0).first; }
  Value *getSrc() const { return getOperand(1).first; }

  static bool classof(const Kinded *k) {
    return k->getKind() == Kinded::Kind::ElementCosInstKind;
  }

  bool isInplaceOp(unsigned dstIdx, unsigned srcIdx) const {
  if (0 == dstIdx && 1 == srcIdx) { return true; }
    return false;
  }

  bool isCanonical() const {
    return true;
  }

  bool isDataParallel() const {
    return true;
  }

  Instruction* clone() const;

  void dump(llvm::raw_ostream &os) const;

  llvm::StringRef getOperandName(unsigned idx) const;
  void verify() const {
    assert(getDest()->dims().equals(getSrc()->dims()) && "Invalid Shape");
    assert(getDest()->getElementType() == getSrc()->getElementType() && "Invalid Element Type");
  }
};
} // namespace glow

namespace glow {
class ElementLogInst final : public Instruction {

 public:
  ElementLogInst(llvm::StringRef name, Value *Dest, Value *Src)
      : Instruction(name, Kinded::Kind::ElementLogInstKind, nullptr, {
          {Dest, OperandKind::Out},
          {Src, OperandKind::In},
      }) {}

  Value *getDest() const { return getOperand(0).first; }
  Value *getSrc() const { return getOperand(1).first; }

  static bool classof(const Kinded *k) {
    return k->getKind() == Kinded::Kind::ElementLogInstKind;
  }

  bool isInplaceOp(unsigned dstIdx, unsigned srcIdx) const {
  if (0 == dstIdx && 1 == srcIdx) { return true; }
    return false;
  }

  bool isCanonical() const {
    return true;
  }

  bool isDataParallel() const {
    return true;
  }

  Instruction* clone() const;

  void dump(llvm::raw_ostream &os) const;

  llvm::StringRef getOperandName(unsigned idx) const;
  void verify() const {
    assert(getDest()->getType() == getSrc()->getType() && "Invalid Type");
  }
};
} // namespace glow

namespace glow {
class ElementExpInst final : public Instruction {

 public:
  ElementExpInst(llvm::StringRef name, Value *Dest, Value *Src)
      : Instruction(name, Kinded::Kind::ElementExpInstKind, nullptr, {
          {Dest, OperandKind::Out},
          {Src, OperandKind::In},
      }) {}

  Value *getDest() const { return getOperand(0).first; }
  Value *getSrc() const { return getOperand(1).first; }

  static bool classof(const Kinded *k) {
    return k->getKind() == Kinded::Kind::ElementExpInstKind;
  }

  bool isInplaceOp(unsigned dstIdx, unsigned srcIdx) const {
  if (0 == dstIdx && 1 == srcIdx) { return true; }
    return false;
  }

  bool isCanonical() const {
    return true;
  }

  bool isDataParallel() const {
    return true;
  }

  Instruction* clone() const;

  void dump(llvm::raw_ostream &os) const;

  llvm::StringRef getOperandName(unsigned idx) const;
  void verify() const {
    assert(getDest()->getType() == getSrc()->getType() && "Invalid Type");
  }
};
} // namespace glow

namespace glow {
class ElementSelectInst final : public Instruction {

 public:
  ElementSelectInst(llvm::StringRef name, Value *Dest, Value *Cond, Value *LHS, Value *RHS)
      : Instruction(name, Kinded::Kind::ElementSelectInstKind, nullptr, {
          {Dest, OperandKind::Out},
          {Cond, OperandKind::In},
          {LHS, OperandKind::In},
          {RHS, OperandKind::In},
      }) {}

  Value *getDest() const { return getOperand(0).first; }
  Value *getCond() const { return getOperand(1).first; }
  Value *getLHS() const { return getOperand(2).first; }
  Value *getRHS() const { return getOperand(3).first; }

  static bool classof(const Kinded *k) {
    return k->getKind() == Kinded::Kind::ElementSelectInstKind;
  }

  bool isInplaceOp(unsigned dstIdx, unsigned srcIdx) const {
  if (0 == dstIdx && 2 == srcIdx) { return true; }
  if (0 == dstIdx && 3 == srcIdx) { return true; }
  if (0 == dstIdx && 1 == srcIdx) { return true; }
    return false;
  }

  bool isCanonical() const {
    return true;
  }

  bool isDataParallel() const {
    return true;
  }

  Instruction* clone() const;

  void dump(llvm::raw_ostream &os) const;

  llvm::StringRef getOperandName(unsigned idx) const;
  void verify() const {
    assert(getDest()->dims().equals(getLHS()->dims()) && "Invalid Shape");
    assert(getDest()->dims().equals(getRHS()->dims()) && "Invalid Shape");
    assert(getDest()->dims().equals(getCond()->dims()) && "Invalid Shape");
    assert(getDest()->getElementType() == getLHS()->getElementType() && "Invalid Element Type");
    assert(getDest()->getElementType() == getRHS()->getElementType() && "Invalid Element Type");
    assert(getCond()->getElementType() == ElemKind::BoolTy && "Invalid Element Type");
  }
};
} // namespace glow

namespace glow {
class ModuloInst final : public Instruction {
  int64_t Divisor_;
  bool SignFollowDivisor_;

 public:
  ModuloInst(llvm::StringRef name, Value *Dest, Value *Src, int64_t Divisor, bool SignFollowDivisor)
      : Instruction(name, Kinded::Kind::ModuloInstKind, nullptr, {
          {Dest, OperandKind::Out},
          {Src, OperandKind::In},
      }), Divisor_(Divisor), SignFollowDivisor_(SignFollowDivisor) {}

  Value *getDest() const { return getOperand(0).first; }
  Value *getSrc() const { return getOperand(1).first; }
  int64_t getDivisor() const { return Divisor_; }
  bool getSignFollowDivisor() const { return SignFollowDivisor_; }

  static bool classof(const Kinded *k) {
    return k->getKind() == Kinded::Kind::ModuloInstKind;
  }

  bool isInplaceOp(unsigned dstIdx, unsigned srcIdx) const {
  if (0 == dstIdx && 1 == srcIdx) { return true; }
    return false;
  }

  bool isCanonical() const {
    return true;
  }

  bool isDataParallel() const {
    return true;
  }

  Instruction* clone() const;

  void dump(llvm::raw_ostream &os) const;

  llvm::StringRef getOperandName(unsigned idx) const;
  void verify() const {
    assert(getDest()->getType() == getSrc()->getType() && "Invalid Type");
  }
};
} // namespace glow

namespace glow {
class BatchedPairwiseDotProductInst final : public Instruction {
  unsigned_t NumInputs_;
  unsigned_t VectorSize_;

 public:
  BatchedPairwiseDotProductInst(llvm::StringRef name, Value *Dest, unsigned_t NumInputs, unsigned_t VectorSize)
      : Instruction(name, Kinded::Kind::BatchedPairwiseDotProductInstKind, nullptr, {
          {Dest, OperandKind::Out},
      }), NumInputs_(NumInputs), VectorSize_(VectorSize) {}

  Value *getDest() const { return getOperand(0).first; }
  unsigned_t getNumInputs() const { return NumInputs_; }
  unsigned_t getVectorSize() const { return VectorSize_; }

  static bool classof(const Kinded *k) {
    return k->getKind() == Kinded::Kind::BatchedPairwiseDotProductInstKind;
  }

  bool isInplaceOp(unsigned dstIdx, unsigned srcIdx) const {
    return false;
  }

  bool isCanonical() const {
    return true;
  }

  bool isDataParallel() const {
    return false;
  }

  Instruction* clone() const;

  void dump(llvm::raw_ostream &os) const;

  llvm::StringRef getOperandName(unsigned idx) const;
  void verify() const {
    // Nothing to verify.
  }
};
} // namespace glow

namespace glow {
class BatchedPairwiseDotProductGradInst final : public Instruction {
  unsigned_t NumInputs_;
  unsigned_t VectorSize_;

 public:
  BatchedPairwiseDotProductGradInst(llvm::StringRef name, Value *DestGrad, unsigned_t NumInputs, unsigned_t VectorSize)
      : Instruction(name, Kinded::Kind::BatchedPairwiseDotProductGradInstKind, nullptr, {
          {DestGrad, OperandKind::In},
      }), NumInputs_(NumInputs), VectorSize_(VectorSize) {}

  Value *getDestGrad() const { return getOperand(0).first; }
  unsigned_t getNumInputs() const { return NumInputs_; }
  unsigned_t getVectorSize() const { return VectorSize_; }

  static bool classof(const Kinded *k) {
    return k->getKind() == Kinded::Kind::BatchedPairwiseDotProductGradInstKind;
  }

  bool isInplaceOp(unsigned dstIdx, unsigned srcIdx) const {
    return false;
  }

  bool isCanonical() const {
    return true;
  }

  bool isDataParallel() const {
    return false;
  }

  Instruction* clone() const;

  void dump(llvm::raw_ostream &os) const;

  llvm::StringRef getOperandName(unsigned idx) const;
  void verify() const {
    // Nothing to verify.
  }
};
} // namespace glow

namespace glow {
class ReluGradInst final : public Instruction {

 public:
  ReluGradInst(llvm::StringRef name, Value *Dest, Value *DestGrad, Value *SrcGrad)
      : Instruction(name, Kinded::Kind::ReluGradInstKind, nullptr, {
          {Dest, OperandKind::In},
          {DestGrad, OperandKind::In},
          {SrcGrad, OperandKind::Out},
      }) {}

  Value *getDest() const { return getOperand(0).first; }
  Value *getDestGrad() const { return getOperand(1).first; }
  Value *getSrcGrad() const { return getOperand(2).first; }

  static bool classof(const Kinded *k) {
    return k->getKind() == Kinded::Kind::ReluGradInstKind;
  }

  bool isInplaceOp(unsigned dstIdx, unsigned srcIdx) const {
    return false;
  }

  bool isCanonical() const {
    return true;
  }

  bool isDataParallel() const {
    return false;
  }

  Instruction* clone() const;

  void dump(llvm::raw_ostream &os) const;

  llvm::StringRef getOperandName(unsigned idx) const;
  void verify() const;
};
} // namespace glow

namespace glow {
class ReluInst final : public Instruction {

 public:
  ReluInst(llvm::StringRef name, Value *Dest, Value *Src)
      : Instruction(name, Kinded::Kind::ReluInstKind, nullptr, {
          {Dest, OperandKind::Out},
          {Src, OperandKind::In},
      }) {}

  Value *getDest() const { return getOperand(0).first; }
  Value *getSrc() const { return getOperand(1).first; }

  static bool classof(const Kinded *k) {
    return k->getKind() == Kinded::Kind::ReluInstKind;
  }

  bool isInplaceOp(unsigned dstIdx, unsigned srcIdx) const {
  if (0 == dstIdx && 1 == srcIdx) { return true; }
    return false;
  }

  bool isCanonical() const {
    return true;
  }

  bool isDataParallel() const {
    return true;
  }

  Instruction* clone() const;

  void dump(llvm::raw_ostream &os) const;

  llvm::StringRef getOperandName(unsigned idx) const;
  void verify() const;
};
} // namespace glow

namespace glow {
class ClipInst final : public Instruction {
  float Min_;
  float Max_;

 public:
  ClipInst(llvm::StringRef name, Value *Dest, Value *Src, float Min, float Max)
      : Instruction(name, Kinded::Kind::ClipInstKind, nullptr, {
          {Dest, OperandKind::Out},
          {Src, OperandKind::In},
      }), Min_(Min), Max_(Max) {}

  Value *getDest() const { return getOperand(0).first; }
  Value *getSrc() const { return getOperand(1).first; }
  float getMin() const { return Min_; }
  float getMax() const { return Max_; }

  static bool classof(const Kinded *k) {
    return k->getKind() == Kinded::Kind::ClipInstKind;
  }

  bool isInplaceOp(unsigned dstIdx, unsigned srcIdx) const {
  if (0 == dstIdx && 1 == srcIdx) { return true; }
    return false;
  }

  bool isCanonical() const {
    return true;
  }

  bool isDataParallel() const {
    return true;
  }

  Instruction* clone() const;

  void dump(llvm::raw_ostream &os) const;

  llvm::StringRef getOperandName(unsigned idx) const;
  void verify() const {
    assert(getDest()->dims().equals(getSrc()->dims()) && "Invalid Shape");
    assert(getDest()->getElementType() == getSrc()->getElementType() && "Invalid Element Type");
  }
};
} // namespace glow

namespace glow {
class SigmoidInst final : public Instruction {

 public:
  SigmoidInst(llvm::StringRef name, Value *Dest, Value *Src)
      : Instruction(name, Kinded::Kind::SigmoidInstKind, nullptr, {
          {Dest, OperandKind::Out},
          {Src, OperandKind::In},
      }) {}

  Value *getDest() const { return getOperand(0).first; }
  Value *getSrc() const { return getOperand(1).first; }

  static bool classof(const Kinded *k) {
    return k->getKind() == Kinded::Kind::SigmoidInstKind;
  }

  bool isInplaceOp(unsigned dstIdx, unsigned srcIdx) const {
  if (0 == dstIdx && 1 == srcIdx) { return true; }
    return false;
  }

  bool isCanonical() const {
    return true;
  }

  bool isDataParallel() const {
    return true;
  }

  Instruction* clone() const;

  void dump(llvm::raw_ostream &os) const;

  llvm::StringRef getOperandName(unsigned idx) const;
  void verify() const {
    assert(getDest()->getType() == getSrc()->getType() && "Invalid Type");
  }
};
} // namespace glow

namespace glow {
class TanhInst final : public Instruction {

 public:
  TanhInst(llvm::StringRef name, Value *Dest, Value *Src)
      : Instruction(name, Kinded::Kind::TanhInstKind, nullptr, {
          {Dest, OperandKind::Out},
          {Src, OperandKind::In},
      }) {}

  Value *getDest() const { return getOperand(0).first; }
  Value *getSrc() const { return getOperand(1).first; }

  static bool classof(const Kinded *k) {
    return k->getKind() == Kinded::Kind::TanhInstKind;
  }

  bool isInplaceOp(unsigned dstIdx, unsigned srcIdx) const {
  if (0 == dstIdx && 1 == srcIdx) { return true; }
    return false;
  }

  bool isCanonical() const {
    return true;
  }

  bool isDataParallel() const {
    return true;
  }

  Instruction* clone() const;

  void dump(llvm::raw_ostream &os) const;

  llvm::StringRef getOperandName(unsigned idx) const;
  void verify() const {
    assert(getDest()->getType() == getSrc()->getType() && "Invalid Type");
  }
};
} // namespace glow

namespace glow {
class TransposeInst final : public Instruction {
  std::vector<unsigned_t> Shuffle_;

 public:
  TransposeInst(llvm::StringRef name, Value *Dest, Value *Src, std::vector<unsigned_t> Shuffle)
      : Instruction(name, Kinded::Kind::TransposeInstKind, nullptr, {
          {Dest, OperandKind::Out},
          {Src, OperandKind::In},
      }), Shuffle_(Shuffle) {}

  Value *getDest() const { return getOperand(0).first; }
  Value *getSrc() const { return getOperand(1).first; }
  llvm::ArrayRef<unsigned_t> getShuffle() const { return Shuffle_; }

  static bool classof(const Kinded *k) {
    return k->getKind() == Kinded::Kind::TransposeInstKind;
  }

  bool isInplaceOp(unsigned dstIdx, unsigned srcIdx) const {
    return false;
  }

  bool isCanonical() const {
    return true;
  }

  bool isDataParallel() const {
    return false;
  }

  Instruction* clone() const;

  void dump(llvm::raw_ostream &os) const;

  llvm::StringRef getOperandName(unsigned idx) const;
  void verify() const {
    assert(getDest()->getElementType() == getSrc()->getElementType() && "Invalid Element Type");
  }
};
} // namespace glow

namespace glow {
class SplatInst final : public Instruction {
  float Value_;

 public:
  SplatInst(llvm::StringRef name, Value *Dest, float Value)
      : Instruction(name, Kinded::Kind::SplatInstKind, nullptr, {
          {Dest, OperandKind::Out},
      }), Value_(Value) {}

  Value *getDest() const { return getOperand(0).first; }
  float getValue() const { return Value_; }

  static bool classof(const Kinded *k) {
    return k->getKind() == Kinded::Kind::SplatInstKind;
  }

  bool isInplaceOp(unsigned dstIdx, unsigned srcIdx) const {
    return false;
  }

  bool isCanonical() const {
    return true;
  }

  bool isDataParallel() const {
    return true;
  }

  Instruction* clone() const;

  void dump(llvm::raw_ostream &os) const;

  llvm::StringRef getOperandName(unsigned idx) const;
  void verify() const {
    // Nothing to verify.
  }
};
} // namespace glow

namespace glow {
class TouchInst final : public Instruction {

 public:
  TouchInst(llvm::StringRef name, Value *Dest)
      : Instruction(name, Kinded::Kind::TouchInstKind, nullptr, {
          {Dest, OperandKind::Out},
      }) {}

  Value *getDest() const { return getOperand(0).first; }

  static bool classof(const Kinded *k) {
    return k->getKind() == Kinded::Kind::TouchInstKind;
  }

  bool isInplaceOp(unsigned dstIdx, unsigned srcIdx) const {
    return false;
  }

  bool isCanonical() const {
    return true;
  }

  bool isDataParallel() const {
    return true;
  }

  Instruction* clone() const;

  void dump(llvm::raw_ostream &os) const;

  llvm::StringRef getOperandName(unsigned idx) const;
  void verify() const {
    // Nothing to verify.
  }
};
} // namespace glow

namespace glow {
class InsertTensorInst final : public Instruction {
  std::vector<dim_t> Offsets_;
  unsigned_t Count_;
  unsigned_t Axis_;

 public:
  InsertTensorInst(llvm::StringRef name, Value *Dest, Value *Src, std::vector<dim_t> Offsets, unsigned_t Count, unsigned_t Axis)
      : Instruction(name, Kinded::Kind::InsertTensorInstKind, nullptr, {
          {Dest, OperandKind::InOut},
          {Src, OperandKind::In},
      }), Offsets_(Offsets), Count_(Count), Axis_(Axis) {}

  Value *getDest() const { return getOperand(0).first; }
  Value *getSrc() const { return getOperand(1).first; }
  llvm::ArrayRef<dim_t> getOffsets() const { return Offsets_; }
  unsigned_t getCount() const { return Count_; }
  unsigned_t getAxis() const { return Axis_; }

  static bool classof(const Kinded *k) {
    return k->getKind() == Kinded::Kind::InsertTensorInstKind;
  }

  bool isInplaceOp(unsigned dstIdx, unsigned srcIdx) const {
    return false;
  }

  bool isCanonical() const {
    return true;
  }

  bool isDataParallel() const {
    return false;
  }

  Instruction* clone() const;

  void dump(llvm::raw_ostream &os) const;

  llvm::StringRef getOperandName(unsigned idx) const;
  void verify() const;
};
} // namespace glow

namespace glow {
class ExtractTensorInst final : public Instruction {
  std::vector<dim_t> Offsets_;

 public:
  ExtractTensorInst(llvm::StringRef name, Value *Dest, Value *Src, std::vector<dim_t> Offsets)
      : Instruction(name, Kinded::Kind::ExtractTensorInstKind, nullptr, {
          {Dest, OperandKind::Out},
          {Src, OperandKind::In},
      }), Offsets_(Offsets) {}

  Value *getDest() const { return getOperand(0).first; }
  Value *getSrc() const { return getOperand(1).first; }
  llvm::ArrayRef<dim_t> getOffsets() const { return Offsets_; }

  static bool classof(const Kinded *k) {
    return k->getKind() == Kinded::Kind::ExtractTensorInstKind;
  }

  bool isInplaceOp(unsigned dstIdx, unsigned srcIdx) const {
    return false;
  }

  bool isCanonical() const {
    return true;
  }

  bool isDataParallel() const {
    return false;
  }

  Instruction* clone() const;

  void dump(llvm::raw_ostream &os) const;

  llvm::StringRef getOperandName(unsigned idx) const;
  void verify() const;
};
} // namespace glow

namespace glow {
class GatherInst final : public Instruction {
  unsigned_t BatchDims_;

 public:
  GatherInst(llvm::StringRef name, Value *Dest, Value *Data, Value *Indices, unsigned_t BatchDims)
      : Instruction(name, Kinded::Kind::GatherInstKind, nullptr, {
          {Dest, OperandKind::Out},
          {Data, OperandKind::In},
          {Indices, OperandKind::In},
      }), BatchDims_(BatchDims) {}

  Value *getDest() const { return getOperand(0).first; }
  Value *getData() const { return getOperand(1).first; }
  Value *getIndices() const { return getOperand(2).first; }
  unsigned_t getBatchDims() const { return BatchDims_; }

  static bool classof(const Kinded *k) {
    return k->getKind() == Kinded::Kind::GatherInstKind;
  }

  bool isInplaceOp(unsigned dstIdx, unsigned srcIdx) const {
    return false;
  }

  bool isCanonical() const {
    return true;
  }

  bool isDataParallel() const {
    return false;
  }

  Instruction* clone() const;

  void dump(llvm::raw_ostream &os) const;

  llvm::StringRef getOperandName(unsigned idx) const;
  void verify() const {
    assert(getDest()->getElementType() == getData()->getElementType() && "Invalid Element Type");
  }
};
} // namespace glow

namespace glow {
class GatherRangesInst final : public Instruction {

 public:
  GatherRangesInst(llvm::StringRef name, Value *Output, Value *Lengths, Value *Data, Value *Ranges)
      : Instruction(name, Kinded::Kind::GatherRangesInstKind, nullptr, {
          {Output, OperandKind::Out},
          {Lengths, OperandKind::Out},
          {Data, OperandKind::In},
          {Ranges, OperandKind::In},
      }) {}

  Value *getOutput() const { return getOperand(0).first; }
  Value *getLengths() const { return getOperand(1).first; }
  Value *getData() const { return getOperand(2).first; }
  Value *getRanges() const { return getOperand(3).first; }

  static bool classof(const Kinded *k) {
    return k->getKind() == Kinded::Kind::GatherRangesInstKind;
  }

  bool isInplaceOp(unsigned dstIdx, unsigned srcIdx) const {
    return false;
  }

  bool isCanonical() const {
    return true;
  }

  bool isDataParallel() const {
    return false;
  }

  Instruction* clone() const;

  void dump(llvm::raw_ostream &os) const;

  llvm::StringRef getOperandName(unsigned idx) const;
  void verify() const {
    assert(getData()->getElementType() == getOutput()->getElementType() && "Invalid Element Type");
    assert(getRanges()->getElementType() == getLengths()->getElementType() && "Invalid Element Type");
  }
};
} // namespace glow

namespace glow {
class ScatterDataInst final : public Instruction {
  bool Cumulative_;

 public:
  ScatterDataInst(llvm::StringRef name, Value *Data, Value *Indices, Value *Slices, bool Cumulative)
      : Instruction(name, Kinded::Kind::ScatterDataInstKind, nullptr, {
          {Data, OperandKind::InOut},
          {Indices, OperandKind::In},
          {Slices, OperandKind::In},
      }), Cumulative_(Cumulative) {}

  Value *getData() const { return getOperand(0).first; }
  Value *getIndices() const { return getOperand(1).first; }
  Value *getSlices() const { return getOperand(2).first; }
  bool getCumulative() const { return Cumulative_; }

  static bool classof(const Kinded *k) {
    return k->getKind() == Kinded::Kind::ScatterDataInstKind;
  }

  bool isInplaceOp(unsigned dstIdx, unsigned srcIdx) const {
    return false;
  }

  bool isCanonical() const {
    return true;
  }

  bool isDataParallel() const {
    return false;
  }

  Instruction* clone() const;

  void dump(llvm::raw_ostream &os) const;

  llvm::StringRef getOperandName(unsigned idx) const;
  void verify() const {
    // Nothing to verify.
  }
};
} // namespace glow

namespace glow {
class BatchOneHotInst final : public Instruction {

 public:
  BatchOneHotInst(llvm::StringRef name, Value *Dest, Value *Data, Value *Lengths, Value *Values)
      : Instruction(name, Kinded::Kind::BatchOneHotInstKind, nullptr, {
          {Dest, OperandKind::Out},
          {Data, OperandKind::In},
          {Lengths, OperandKind::In},
          {Values, OperandKind::In},
      }) {}

  Value *getDest() const { return getOperand(0).first; }
  Value *getData() const { return getOperand(1).first; }
  Value *getLengths() const { return getOperand(2).first; }
  Value *getValues() const { return getOperand(3).first; }

  static bool classof(const Kinded *k) {
    return k->getKind() == Kinded::Kind::BatchOneHotInstKind;
  }

  bool isInplaceOp(unsigned dstIdx, unsigned srcIdx) const {
    return false;
  }

  bool isCanonical() const {
    return true;
  }

  bool isDataParallel() const {
    return false;
  }

  Instruction* clone() const;

  void dump(llvm::raw_ostream &os) const;

  llvm::StringRef getOperandName(unsigned idx) const;
  void verify() const {
    assert(getValues()->getElementType() == getData()->getElementType() && "Invalid Element Type");
    assert(getValues()->getElementType() == getDest()->getElementType() && "Invalid Element Type");
    assert(getLengths()->getElementType() == ElemKind::Int32ITy && "Invalid Element Type");
  }
};
} // namespace glow

namespace glow {
class SpaceToDepthInst final : public Instruction {
  unsigned_t BlockSize_;

 public:
  SpaceToDepthInst(llvm::StringRef name, Value *Dest, Value *Src, unsigned_t BlockSize)
      : Instruction(name, Kinded::Kind::SpaceToDepthInstKind, nullptr, {
          {Dest, OperandKind::Out},
          {Src, OperandKind::In},
      }), BlockSize_(BlockSize) {}

  Value *getDest() const { return getOperand(0).first; }
  Value *getSrc() const { return getOperand(1).first; }
  unsigned_t getBlockSize() const { return BlockSize_; }

  static bool classof(const Kinded *k) {
    return k->getKind() == Kinded::Kind::SpaceToDepthInstKind;
  }

  bool isInplaceOp(unsigned dstIdx, unsigned srcIdx) const {
    return false;
  }

  bool isCanonical() const {
    return true;
  }

  bool isDataParallel() const {
    return false;
  }

  Instruction* clone() const;

  void dump(llvm::raw_ostream &os) const;

  llvm::StringRef getOperandName(unsigned idx) const;
  void verify() const {
    assert(getDest()->getElementType() == getSrc()->getElementType() && "Invalid Element Type");
  }
};
} // namespace glow

namespace glow {
class ResizeNearestInst final : public Instruction {
  std::vector<float> Scale_;

 public:
  ResizeNearestInst(llvm::StringRef name, Value *Dest, Value *Src, std::vector<float> Scale)
      : Instruction(name, Kinded::Kind::ResizeNearestInstKind, nullptr, {
          {Dest, OperandKind::Out},
          {Src, OperandKind::In},
      }), Scale_(Scale) {}

  Value *getDest() const { return getOperand(0).first; }
  Value *getSrc() const { return getOperand(1).first; }
  llvm::ArrayRef<float> getScale() const { return Scale_; }

  static bool classof(const Kinded *k) {
    return k->getKind() == Kinded::Kind::ResizeNearestInstKind;
  }

  bool isInplaceOp(unsigned dstIdx, unsigned srcIdx) const {
    return false;
  }

  bool isCanonical() const {
    return true;
  }

  bool isDataParallel() const {
    return false;
  }

  Instruction* clone() const;

  void dump(llvm::raw_ostream &os) const;

  llvm::StringRef getOperandName(unsigned idx) const;
  void verify() const {
    assert(getDest()->getElementType() == getSrc()->getElementType() && "Invalid Element Type");
  }
};
} // namespace glow

namespace glow {
class ResizeBilinearInst final : public Instruction {
  std::vector<float> Scale_;

 public:
  ResizeBilinearInst(llvm::StringRef name, Value *Dest, Value *Src, std::vector<float> Scale)
      : Instruction(name, Kinded::Kind::ResizeBilinearInstKind, nullptr, {
          {Dest, OperandKind::Out},
          {Src, OperandKind::In},
      }), Scale_(Scale) {}

  Value *getDest() const { return getOperand(0).first; }
  Value *getSrc() const { return getOperand(1).first; }
  llvm::ArrayRef<float> getScale() const { return Scale_; }

  static bool classof(const Kinded *k) {
    return k->getKind() == Kinded::Kind::ResizeBilinearInstKind;
  }

  bool isInplaceOp(unsigned dstIdx, unsigned srcIdx) const {
    return false;
  }

  bool isCanonical() const {
    return true;
  }

  bool isDataParallel() const {
    return false;
  }

  Instruction* clone() const;

  void dump(llvm::raw_ostream &os) const;

  llvm::StringRef getOperandName(unsigned idx) const;
  void verify() const {
    assert(getDest()->getElementType() == getSrc()->getElementType() && "Invalid Element Type");
  }
};
} // namespace glow

namespace glow {
class FlipInst final : public Instruction {
  unsigned_t Axis_;

 public:
  FlipInst(llvm::StringRef name, Value *Dest, Value *Src, unsigned_t Axis)
      : Instruction(name, Kinded::Kind::FlipInstKind, nullptr, {
          {Dest, OperandKind::Out},
          {Src, OperandKind::In},
      }), Axis_(Axis) {}

  Value *getDest() const { return getOperand(0).first; }
  Value *getSrc() const { return getOperand(1).first; }
  unsigned_t getAxis() const { return Axis_; }

  static bool classof(const Kinded *k) {
    return k->getKind() == Kinded::Kind::FlipInstKind;
  }

  bool isInplaceOp(unsigned dstIdx, unsigned srcIdx) const {
    return false;
  }

  bool isCanonical() const {
    return true;
  }

  bool isDataParallel() const {
    return false;
  }

  Instruction* clone() const;

  void dump(llvm::raw_ostream &os) const;

  llvm::StringRef getOperandName(unsigned idx) const;
  void verify() const {
    assert(getDest()->getElementType() == getSrc()->getElementType() && "Invalid Element Type");
    assert(getDest()->dims().equals(getSrc()->dims()) && "Invalid Shape");
  }
};
} // namespace glow

namespace glow {
class DebugPrintInst final : public Instruction {
  std::string Format_;
  std::string FileName_;

 public:
  DebugPrintInst(llvm::StringRef name, Value *Src, std::string Format, std::string FileName)
      : Instruction(name, Kinded::Kind::DebugPrintInstKind, nullptr, {
          {Src, OperandKind::In},
      }), Format_(Format), FileName_(FileName) {}

  Value *getSrc() const { return getOperand(0).first; }
  std::string getFormat() const { return Format_; }
  std::string getFileName() const { return FileName_; }

  static bool classof(const Kinded *k) {
    return k->getKind() == Kinded::Kind::DebugPrintInstKind;
  }

  bool isInplaceOp(unsigned dstIdx, unsigned srcIdx) const {
    return false;
  }

  bool isCanonical() const {
    return true;
  }

  bool isDataParallel() const {
    return false;
  }

  Instruction* clone() const;

  void dump(llvm::raw_ostream &os) const;

  llvm::StringRef getOperandName(unsigned idx) const;
  void verify() const {
    // Nothing to verify.
  }
};
} // namespace glow

namespace glow {
class TraceEventInst final : public Instruction {
  unsigned_t Index_;

 public:
  TraceEventInst(llvm::StringRef name, Value *Data, unsigned_t Index)
      : Instruction(name, Kinded::Kind::TraceEventInstKind, nullptr, {
          {Data, OperandKind::In},
      }), Index_(Index) {}

  Value *getData() const { return getOperand(0).first; }
  unsigned_t getIndex() const { return Index_; }

  static bool classof(const Kinded *k) {
    return k->getKind() == Kinded::Kind::TraceEventInstKind;
  }

  bool isInplaceOp(unsigned dstIdx, unsigned srcIdx) const {
    return false;
  }

  bool isCanonical() const {
    return true;
  }

  bool isDataParallel() const {
    return false;
  }

  Instruction* clone() const;

  void dump(llvm::raw_ostream &os) const;

  llvm::StringRef getOperandName(unsigned idx) const;
  void verify() const {
    // Nothing to verify.
  }
};
} // namespace glow

namespace glow {
class QuantizationProfileInst final : public Instruction {

 public:
  QuantizationProfileInst(llvm::StringRef name, Value *InputTensor, Value *Histogram, Value *ComputationInfo)
      : Instruction(name, Kinded::Kind::QuantizationProfileInstKind, nullptr, {
          {InputTensor, OperandKind::In},
          {Histogram, OperandKind::InOut},
          {ComputationInfo, OperandKind::InOut},
      }) {}

  Value *getInputTensor() const { return getOperand(0).first; }
  Value *getHistogram() const { return getOperand(1).first; }
  Value *getComputationInfo() const { return getOperand(2).first; }

  static bool classof(const Kinded *k) {
    return k->getKind() == Kinded::Kind::QuantizationProfileInstKind;
  }

  bool isInplaceOp(unsigned dstIdx, unsigned srcIdx) const {
    return false;
  }

  bool isCanonical() const {
    return true;
  }

  bool isDataParallel() const {
    return false;
  }

  Instruction* clone() const;

  void dump(llvm::raw_ostream &os) const;

  llvm::StringRef getOperandName(unsigned idx) const;
  void verify() const {
    assert(getInputTensor()->getElementType() == ElemKind::FloatTy && "Invalid Element Type");
  }
};
} // namespace glow

namespace glow {
class IntLookupTableInst final : public Instruction {

 public:
  IntLookupTableInst(llvm::StringRef name, Value *Dest, Value *Src, Value *Mapping)
      : Instruction(name, Kinded::Kind::IntLookupTableInstKind, nullptr, {
          {Dest, OperandKind::Out},
          {Src, OperandKind::In},
          {Mapping, OperandKind::In},
      }) {}

  Value *getDest() const { return getOperand(0).first; }
  Value *getSrc() const { return getOperand(1).first; }
  Value *getMapping() const { return getOperand(2).first; }

  static bool classof(const Kinded *k) {
    return k->getKind() == Kinded::Kind::IntLookupTableInstKind;
  }

  bool isInplaceOp(unsigned dstIdx, unsigned srcIdx) const {
    return false;
  }

  bool isCanonical() const {
    return true;
  }

  bool isDataParallel() const {
    return true;
  }

  Instruction* clone() const;

  void dump(llvm::raw_ostream &os) const;

  llvm::StringRef getOperandName(unsigned idx) const;
  void verify() const {
    assert(getDest()->getElementType() == getSrc()->getElementType() && "Invalid Element Type");
    assert(getDest()->getType()->isQuantizedType() && "Invalid Type");
  }
};
} // namespace glow

namespace glow {
class QuantizeInst final : public Instruction {

 public:
  QuantizeInst(llvm::StringRef name, Value *Dest, Value *Src)
      : Instruction(name, Kinded::Kind::QuantizeInstKind, nullptr, {
          {Dest, OperandKind::Out},
          {Src, OperandKind::In},
      }) {}

  Value *getDest() const { return getOperand(0).first; }
  Value *getSrc() const { return getOperand(1).first; }

  static bool classof(const Kinded *k) {
    return k->getKind() == Kinded::Kind::QuantizeInstKind;
  }

  bool isInplaceOp(unsigned dstIdx, unsigned srcIdx) const {
    return false;
  }

  bool isCanonical() const {
    return true;
  }

  bool isDataParallel() const {
    return true;
  }

  Instruction* clone() const;

  void dump(llvm::raw_ostream &os) const;

  llvm::StringRef getOperandName(unsigned idx) const;
  void verify() const {
    assert(getSrc()->getType()->isFPType() && "Invalid Type");
    assert(getDest()->getType()->isQuantizedType() && "Invalid Type");
    assert(getDest()->dims().equals(getSrc()->dims()) && "Invalid Shape");
  }
};
} // namespace glow

namespace glow {
class DequantizeInst final : public Instruction {

 public:
  DequantizeInst(llvm::StringRef name, Value *Dest, Value *Src)
      : Instruction(name, Kinded::Kind::DequantizeInstKind, nullptr, {
          {Dest, OperandKind::Out},
          {Src, OperandKind::In},
      }) {}

  Value *getDest() const { return getOperand(0).first; }
  Value *getSrc() const { return getOperand(1).first; }

  static bool classof(const Kinded *k) {
    return k->getKind() == Kinded::Kind::DequantizeInstKind;
  }

  bool isInplaceOp(unsigned dstIdx, unsigned srcIdx) const {
    return false;
  }

  bool isCanonical() const {
    return true;
  }

  bool isDataParallel() const {
    return true;
  }

  Instruction* clone() const;

  void dump(llvm::raw_ostream &os) const;

  llvm::StringRef getOperandName(unsigned idx) const;
  void verify() const {
    assert(getDest()->getType()->isFPType() && "Invalid Type");
    assert(getSrc()->getType()->isQuantizedType() && "Invalid Type");
  }
};
} // namespace glow

namespace glow {
class RescaleQuantizedInst final : public Instruction {

 public:
  RescaleQuantizedInst(llvm::StringRef name, Value *Dest, Value *Src)
      : Instruction(name, Kinded::Kind::RescaleQuantizedInstKind, nullptr, {
          {Dest, OperandKind::Out},
          {Src, OperandKind::In},
      }) {}

  Value *getDest() const { return getOperand(0).first; }
  Value *getSrc() const { return getOperand(1).first; }

  static bool classof(const Kinded *k) {
    return k->getKind() == Kinded::Kind::RescaleQuantizedInstKind;
  }

  bool isInplaceOp(unsigned dstIdx, unsigned srcIdx) const {
    return false;
  }

  bool isCanonical() const {
    return true;
  }

  bool isDataParallel() const {
    return true;
  }

  Instruction* clone() const;

  void dump(llvm::raw_ostream &os) const;

  llvm::StringRef getOperandName(unsigned idx) const;
  void verify() const {
    assert(getDest()->getElementType() == getSrc()->getElementType() && "Invalid Element Type");
    assert(getDest()->getType()->isQuantizedType() && "Invalid Type");
    assert(getDest()->dims().equals(getSrc()->dims()) && "Invalid Shape");
  }
};
} // namespace glow

namespace glow {
class TopKInst final : public Instruction {
  unsigned_t K_;

 public:
  TopKInst(llvm::StringRef name, Value *Values, Value *Indices, Value *Input, Value *Scratch, unsigned_t K)
      : Instruction(name, Kinded::Kind::TopKInstKind, nullptr, {
          {Values, OperandKind::Out},
          {Indices, OperandKind::Out},
          {Input, OperandKind::In},
          {Scratch, OperandKind::Out},
      }), K_(K) {}

  Value *getValues() const { return getOperand(0).first; }
  Value *getIndices() const { return getOperand(1).first; }
  Value *getInput() const { return getOperand(2).first; }
  Value *getScratch() const { return getOperand(3).first; }
  unsigned_t getK() const { return K_; }
  dim_t getScratchSize() const;

  static bool classof(const Kinded *k) {
    return k->getKind() == Kinded::Kind::TopKInstKind;
  }

  bool isInplaceOp(unsigned dstIdx, unsigned srcIdx) const {
    return false;
  }

  bool isCanonical() const {
    return true;
  }

  bool isDataParallel() const {
    return false;
  }

  Instruction* clone() const;

  void dump(llvm::raw_ostream &os) const;

  llvm::StringRef getOperandName(unsigned idx) const;
  void verify() const {
    assert(getValues()->getElementType() == getInput()->getElementType() && "Invalid Element Type");
    assert(getValues()->dims().equals(getIndices()->dims()) && "Invalid Shape");
  }
};
} // namespace glow

namespace glow {
class ConvertToInst final : public Instruction {

 public:
  ConvertToInst(llvm::StringRef name, Value *Result, Value *Input)
      : Instruction(name, Kinded::Kind::ConvertToInstKind, nullptr, {
          {Result, OperandKind::Out},
          {Input, OperandKind::In},
      }) {}

  Value *getResult() const { return getOperand(0).first; }
  Value *getInput() const { return getOperand(1).first; }

  static bool classof(const Kinded *k) {
    return k->getKind() == Kinded::Kind::ConvertToInstKind;
  }

  bool isInplaceOp(unsigned dstIdx, unsigned srcIdx) const {
    return false;
  }

  bool isCanonical() const {
    return true;
  }

  bool isDataParallel() const {
    return false;
  }

  Instruction* clone() const;

  void dump(llvm::raw_ostream &os) const;

  llvm::StringRef getOperandName(unsigned idx) const;
  void verify() const {
    assert(getResult()->dims().equals(getInput()->dims()) && "Invalid Shape");
  }
};
} // namespace glow

namespace glow {
class AudioSpectrogramInst final : public Instruction {
  int64_t WindowSize_;
  int64_t WindowStride_;
  bool MagnitudeSquared_;

 public:
  AudioSpectrogramInst(llvm::StringRef name, Value *Spectrogram, Value *Input, Value *Window, Value *TwiddleFactors, Value *BitReverseIndices, Value *ComplexToRealWeights, Value *WinOutScratch, Value *FftOutScratch, int64_t WindowSize, int64_t WindowStride, bool MagnitudeSquared)
      : Instruction(name, Kinded::Kind::AudioSpectrogramInstKind, nullptr, {
          {Spectrogram, OperandKind::Out},
          {Input, OperandKind::In},
          {Window, OperandKind::In},
          {TwiddleFactors, OperandKind::In},
          {BitReverseIndices, OperandKind::In},
          {ComplexToRealWeights, OperandKind::In},
          {WinOutScratch, OperandKind::Out},
          {FftOutScratch, OperandKind::Out},
      }), WindowSize_(WindowSize), WindowStride_(WindowStride), MagnitudeSquared_(MagnitudeSquared) {}

  Value *getSpectrogram() const { return getOperand(0).first; }
  Value *getInput() const { return getOperand(1).first; }
  Value *getWindow() const { return getOperand(2).first; }
  Value *getTwiddleFactors() const { return getOperand(3).first; }
  Value *getBitReverseIndices() const { return getOperand(4).first; }
  Value *getComplexToRealWeights() const { return getOperand(5).first; }
  Value *getWinOutScratch() const { return getOperand(6).first; }
  Value *getFftOutScratch() const { return getOperand(7).first; }
  int64_t getWindowSize() const { return WindowSize_; }
  int64_t getWindowStride() const { return WindowStride_; }
  bool getMagnitudeSquared() const { return MagnitudeSquared_; }
  dim_t getWinOutScratchSize() const;
  dim_t getFftOutScratchSize() const;

  static bool classof(const Kinded *k) {
    return k->getKind() == Kinded::Kind::AudioSpectrogramInstKind;
  }

  bool isInplaceOp(unsigned dstIdx, unsigned srcIdx) const {
    return false;
  }

  bool isCanonical() const {
    return true;
  }

  bool isDataParallel() const {
    return false;
  }

  Instruction* clone() const;

  void dump(llvm::raw_ostream &os) const;

  llvm::StringRef getOperandName(unsigned idx) const;
  void verify() const {
    assert(getSpectrogram()->getElementType() == getInput()->getElementType() && "Invalid Element Type");
    assert(getSpectrogram()->getElementType() == getWindow()->getElementType() && "Invalid Element Type");
    assert(getSpectrogram()->getElementType() == getTwiddleFactors()->getElementType() && "Invalid Element Type");
    assert(getSpectrogram()->getElementType() == getComplexToRealWeights()->getElementType() && "Invalid Element Type");
    assert(getSpectrogram()->getElementType() == ElemKind::FloatTy && "Invalid Element Type");
    assert(getBitReverseIndices()->getElementType() == ElemKind::Int32ITy && "Invalid Element Type");
  }
};
} // namespace glow

namespace glow {
class MFCCInst final : public Instruction {
  float SampleRate_;
  float LowerFrequency_;
  float UpperFrequency_;
  int64_t FilterBankCount_;
  int64_t NumCoefficients_;

 public:
  MFCCInst(llvm::StringRef name, Value *Coefficients, Value *Spectrogram, Value *MelWeights, Value *MelRanges, Value *DctMat, Value *Scratch, float SampleRate, float LowerFrequency, float UpperFrequency, int64_t FilterBankCount, int64_t NumCoefficients)
      : Instruction(name, Kinded::Kind::MFCCInstKind, nullptr, {
          {Coefficients, OperandKind::Out},
          {Spectrogram, OperandKind::In},
          {MelWeights, OperandKind::In},
          {MelRanges, OperandKind::In},
          {DctMat, OperandKind::In},
          {Scratch, OperandKind::Out},
      }), SampleRate_(SampleRate), LowerFrequency_(LowerFrequency), UpperFrequency_(UpperFrequency), FilterBankCount_(FilterBankCount), NumCoefficients_(NumCoefficients) {}

  Value *getCoefficients() const { return getOperand(0).first; }
  Value *getSpectrogram() const { return getOperand(1).first; }
  Value *getMelWeights() const { return getOperand(2).first; }
  Value *getMelRanges() const { return getOperand(3).first; }
  Value *getDctMat() const { return getOperand(4).first; }
  Value *getScratch() const { return getOperand(5).first; }
  float getSampleRate() const { return SampleRate_; }
  float getLowerFrequency() const { return LowerFrequency_; }
  float getUpperFrequency() const { return UpperFrequency_; }
  int64_t getFilterBankCount() const { return FilterBankCount_; }
  int64_t getNumCoefficients() const { return NumCoefficients_; }
  dim_t getScratchSize() const;

  static bool classof(const Kinded *k) {
    return k->getKind() == Kinded::Kind::MFCCInstKind;
  }

  bool isInplaceOp(unsigned dstIdx, unsigned srcIdx) const {
    return false;
  }

  bool isCanonical() const {
    return true;
  }

  bool isDataParallel() const {
    return false;
  }

  Instruction* clone() const;

  void dump(llvm::raw_ostream &os) const;

  llvm::StringRef getOperandName(unsigned idx) const;
  void verify() const {
    assert(getCoefficients()->getElementType() == getSpectrogram()->getElementType() && "Invalid Element Type");
    assert(getCoefficients()->getElementType() == getMelWeights()->getElementType() && "Invalid Element Type");
    assert(getCoefficients()->getElementType() == getDctMat()->getElementType() && "Invalid Element Type");
    assert(getCoefficients()->getElementType() == ElemKind::FloatTy && "Invalid Element Type");
    assert(getMelRanges()->getElementType() == ElemKind::Int32ITy && "Invalid Element Type");
  }
};
} // namespace glow

namespace glow {
class NonMaxSuppressionInst final : public Instruction {
  int64_t CenterPointBox_;
  int64_t MaxOutputBoxesPerClass_;
  float IouThreshold_;
  float ScoreThreshold_;
  bool IsTFVersion4_;

 public:
  NonMaxSuppressionInst(llvm::StringRef name, Value *Indices, Value *NumberOfSelectedIndices, Value *Boxes, Value *Scores, int64_t CenterPointBox, int64_t MaxOutputBoxesPerClass, float IouThreshold, float ScoreThreshold, bool IsTFVersion4)
      : Instruction(name, Kinded::Kind::NonMaxSuppressionInstKind, nullptr, {
          {Indices, OperandKind::Out},
          {NumberOfSelectedIndices, OperandKind::Out},
          {Boxes, OperandKind::In},
          {Scores, OperandKind::In},
      }), CenterPointBox_(CenterPointBox), MaxOutputBoxesPerClass_(MaxOutputBoxesPerClass), IouThreshold_(IouThreshold), ScoreThreshold_(ScoreThreshold), IsTFVersion4_(IsTFVersion4) {}

  Value *getIndices() const { return getOperand(0).first; }
  Value *getNumberOfSelectedIndices() const { return getOperand(1).first; }
  Value *getBoxes() const { return getOperand(2).first; }
  Value *getScores() const { return getOperand(3).first; }
  int64_t getCenterPointBox() const { return CenterPointBox_; }
  int64_t getMaxOutputBoxesPerClass() const { return MaxOutputBoxesPerClass_; }
  float getIouThreshold() const { return IouThreshold_; }
  float getScoreThreshold() const { return ScoreThreshold_; }
  bool getIsTFVersion4() const { return IsTFVersion4_; }

  static bool classof(const Kinded *k) {
    return k->getKind() == Kinded::Kind::NonMaxSuppressionInstKind;
  }

  bool isInplaceOp(unsigned dstIdx, unsigned srcIdx) const {
    return false;
  }

  bool isCanonical() const {
    return true;
  }

  bool isDataParallel() const {
    return false;
  }

  Instruction* clone() const;

  void dump(llvm::raw_ostream &os) const;

  llvm::StringRef getOperandName(unsigned idx) const;
  void verify() const {
    assert(getBoxes()->getElementType() == getScores()->getElementType() && "Invalid Element Type");
    assert(getIndices()->getElementType() == getNumberOfSelectedIndices()->getElementType() && "Invalid Element Type");
  }
};
} // namespace glow

namespace glow {
class CPUMaxSplatInst final : public Instruction {
  float SplatValue_;

 public:
  CPUMaxSplatInst(llvm::StringRef name, Value *Dest, Value *Src, float SplatValue)
      : Instruction(name, Kinded::Kind::CPUMaxSplatInstKind, nullptr, {
          {Dest, OperandKind::Out},
          {Src, OperandKind::In},
      }), SplatValue_(SplatValue) {}

  Value *getDest() const { return getOperand(0).first; }
  Value *getSrc() const { return getOperand(1).first; }
  float getSplatValue() const { return SplatValue_; }

  static bool classof(const Kinded *k) {
    return k->getKind() == Kinded::Kind::CPUMaxSplatInstKind;
  }

  bool isInplaceOp(unsigned dstIdx, unsigned srcIdx) const {
  if (0 == dstIdx && 1 == srcIdx) { return true; }
    return false;
  }

  bool isCanonical() const {
    return false;
  }

  bool isDataParallel() const {
    return true;
  }

  Instruction* clone() const;

  void dump(llvm::raw_ostream &os) const;

  llvm::StringRef getOperandName(unsigned idx) const;
  void verify() const;
};
} // namespace glow

namespace glow {
class CPUConvDKKC8Inst final : public Instruction {
  std::vector<unsigned_t> Kernels_;
  std::vector<unsigned_t> Strides_;
  std::vector<unsigned_t> Pads_;
  unsigned_t Group_;

 public:
  CPUConvDKKC8Inst(llvm::StringRef name, Value *Dest, Value *Src, Value *Filter, Value *Bias, std::vector<unsigned_t> Kernels, std::vector<unsigned_t> Strides, std::vector<unsigned_t> Pads, unsigned_t Group)
      : Instruction(name, Kinded::Kind::CPUConvDKKC8InstKind, nullptr, {
          {Dest, OperandKind::Out},
          {Src, OperandKind::In},
          {Filter, OperandKind::In},
          {Bias, OperandKind::In},
      }), Kernels_(Kernels), Strides_(Strides), Pads_(Pads), Group_(Group) {}

  Value *getDest() const { return getOperand(0).first; }
  Value *getSrc() const { return getOperand(1).first; }
  Value *getFilter() const { return getOperand(2).first; }
  Value *getBias() const { return getOperand(3).first; }
  llvm::ArrayRef<unsigned_t> getKernels() const { return Kernels_; }
  llvm::ArrayRef<unsigned_t> getStrides() const { return Strides_; }
  llvm::ArrayRef<unsigned_t> getPads() const { return Pads_; }
  unsigned_t getGroup() const { return Group_; }

  static bool classof(const Kinded *k) {
    return k->getKind() == Kinded::Kind::CPUConvDKKC8InstKind;
  }

  bool isInplaceOp(unsigned dstIdx, unsigned srcIdx) const {
    return false;
  }

  bool isCanonical() const {
    return false;
  }

  bool isDataParallel() const {
    return false;
  }

  Instruction* clone() const;

  void dump(llvm::raw_ostream &os) const;

  llvm::StringRef getOperandName(unsigned idx) const;
  void verify() const;
};
} // namespace glow
