#include "glow/Graph/Nodes.h"
#include "glow/Base/Type.h"
#include "glow/Support/Support.h"
using namespace glow;

unsigned SaveNode::getNumInputs() const {
  return 2;
}

std::string SaveNode::getInputName(unsigned idx) const {
  if (idx == 0) { return "Input"; }
  if (idx == 1) { return "Output"; }
  idx -= 2;
  llvm_unreachable("Invalid index");
}

NodeValue SaveNode::getNthInput(unsigned idx) {
  if (idx == 0) { return Input_; }
  if (idx == 1) { return Output_; }
  idx -= 2;
  llvm_unreachable("Invalid index");
}

void SaveNode::setNthInput(unsigned idx, NodeValue val) {
  if (idx == 0) { Input_ = val; return; }
  if (idx == 1) { Output_ = val; return; }
  idx -= 2;
  llvm_unreachable("Invalid index");
}

llvm::StringRef SaveNode::getOutputName(unsigned idx) const {
  llvm_unreachable("Invalid index");
}

std::string SaveNode::getDebugDesc() const {
  DescriptionBuilder db(getKindName());
  db.addParam("name", getName());
  if (hasPredicate()) db.addParam("Predicate", "Yes");
  db
    .addParam("Input", *(getInput().getType()))
    .addParam("Output", *(getOutput().getType()))
    .addParam("users", getNumUsers());
  return db;
}

void SaveNode::visit(Node *parent, NodeWalker *visitor) {
  if (!visitor->shouldVisit(parent, this)) { return; }
  visitor->pre(parent, this);
if (hasPredicate())
 getPredicate().getNode()->visit(this, visitor);
  getInput().getNode()->visit(this, visitor);
  getOutput().getNode()->visit(this, visitor);
  visitor->post(parent, this);
}

bool SaveNode::isEqual(const SaveNode &other) const {
  return true &&
      Input_ == other.Input_ &&
      Output_ == other.Output_ &&
      predicate_ == other.predicate_;
}

Node* SaveNode::clone() const {
  return new SaveNode(getName(), getInput(), getOutput());
}

llvm::hash_code SaveNode::getHash() const {
  return llvm::hash_combine(
      Input_,
      Output_);
}
Placeholder *SaveNode::getPlaceholder() const { return llvm::cast<Placeholder>(Output_.getNode()); };
unsigned PadNode::getNumInputs() const {
  return 1;
}

std::string PadNode::getInputName(unsigned idx) const {
  if (idx == 0) { return "Input"; }
  idx -= 1;
  llvm_unreachable("Invalid index");
}

NodeValue PadNode::getNthInput(unsigned idx) {
  if (idx == 0) { return Input_; }
  idx -= 1;
  llvm_unreachable("Invalid index");
}

void PadNode::setNthInput(unsigned idx, NodeValue val) {
  if (idx == 0) { Input_ = val; return; }
  idx -= 1;
  llvm_unreachable("Invalid index");
}

llvm::StringRef PadNode::getOutputName(unsigned idx) const {
  if (idx == 0) { return "Result"; }
  llvm_unreachable("Invalid index");
}

std::string PadNode::getDebugDesc() const {
  DescriptionBuilder db(getKindName());
  db.addParam("name", getName());
  if (hasPredicate()) db.addParam("Predicate", "Yes");
  db
    .addParam("Input", *(getInput().getType()))
    .addParam("Mode", getMode())
    .addParam("Pads", getPads())
    .addParam("Value", getValue())
    .addParam("users", getNumUsers());
  db.addParam("Result", *(getResult().getType()));
  return db;
}

void PadNode::visit(Node *parent, NodeWalker *visitor) {
  if (!visitor->shouldVisit(parent, this)) { return; }
  visitor->pre(parent, this);
if (hasPredicate())
 getPredicate().getNode()->visit(this, visitor);
  getInput().getNode()->visit(this, visitor);
  visitor->post(parent, this);
}

bool PadNode::isEqual(const PadNode &other) const {
  return true &&
      Input_ == other.Input_ &&
      predicate_ == other.predicate_ &&
      Mode_ == other.Mode_ &&
      Pads_ == other.Pads_ &&
      Value_ == other.Value_ &&
      getType(0) == other.getType(0);
}

Node* PadNode::clone() const {
  return new PadNode(getName(), getResult().getType(), getInput(), getMode(), getPads(), getValue());
}

llvm::hash_code PadNode::getHash() const {
  return llvm::hash_combine(
      Mode_,
      llvm::hash_combine_range(Pads_.begin(), Pads_.end()),
      toBinary(Value_),
      Input_);
}

unsigned ConvolutionGradNode::getNumInputs() const {
  return 5;
}

std::string ConvolutionGradNode::getInputName(unsigned idx) const {
  if (idx == 0) { return "Input"; }
  if (idx == 1) { return "Filter"; }
  if (idx == 2) { return "Bias"; }
  if (idx == 3) { return "OriginalOutputForResult"; }
  if (idx == 4) { return "GradOfOriginalOutputNamedResult"; }
  idx -= 5;
  llvm_unreachable("Invalid index");
}

NodeValue ConvolutionGradNode::getNthInput(unsigned idx) {
  if (idx == 0) { return Input_; }
  if (idx == 1) { return Filter_; }
  if (idx == 2) { return Bias_; }
  if (idx == 3) { return OriginalOutputForResult_; }
  if (idx == 4) { return GradOfOriginalOutputNamedResult_; }
  idx -= 5;
  llvm_unreachable("Invalid index");
}

void ConvolutionGradNode::setNthInput(unsigned idx, NodeValue val) {
  if (idx == 0) { Input_ = val; return; }
  if (idx == 1) { Filter_ = val; return; }
  if (idx == 2) { Bias_ = val; return; }
  if (idx == 3) { OriginalOutputForResult_ = val; return; }
  if (idx == 4) { GradOfOriginalOutputNamedResult_ = val; return; }
  idx -= 5;
  llvm_unreachable("Invalid index");
}

llvm::StringRef ConvolutionGradNode::getOutputName(unsigned idx) const {
  if (idx == 0) { return "GradOfInputNamedInput"; }
  if (idx == 1) { return "GradOfInputNamedFilter"; }
  if (idx == 2) { return "GradOfInputNamedBias"; }
  llvm_unreachable("Invalid index");
}

std::string ConvolutionGradNode::getDebugDesc() const {
  DescriptionBuilder db(getKindName());
  db.addParam("name", getName());
  if (hasPredicate()) db.addParam("Predicate", "Yes");
  db
    .addParam("Input", *(getInput().getType()))
    .addParam("Filter", *(getFilter().getType()))
    .addParam("Bias", *(getBias().getType()))
    .addParam("OriginalOutputForResult", *(getOriginalOutputForResult().getType()))
    .addParam("GradOfOriginalOutputNamedResult", *(getGradOfOriginalOutputNamedResult().getType()))
    .addParam("Kernels", getKernels())
    .addParam("Strides", getStrides())
    .addParam("Pads", getPads())
    .addParam("Group", getGroup())
    .addParam("Dilation", getDilation())
    .addParam("Layout", getLayout())
    .addParam("FusedActivation", getFusedActivation())
    .addParam("users", getNumUsers());
  db.addParam("GradOfInputNamedInput", *(getGradOfInputNamedInput().getType()));
  db.addParam("GradOfInputNamedFilter", *(getGradOfInputNamedFilter().getType()));
  db.addParam("GradOfInputNamedBias", *(getGradOfInputNamedBias().getType()));
  return db;
}

void ConvolutionGradNode::visit(Node *parent, NodeWalker *visitor) {
  if (!visitor->shouldVisit(parent, this)) { return; }
  visitor->pre(parent, this);
if (hasPredicate())
 getPredicate().getNode()->visit(this, visitor);
  getInput().getNode()->visit(this, visitor);
  getFilter().getNode()->visit(this, visitor);
  getBias().getNode()->visit(this, visitor);
  getOriginalOutputForResult().getNode()->visit(this, visitor);
  getGradOfOriginalOutputNamedResult().getNode()->visit(this, visitor);
  visitor->post(parent, this);
}

bool ConvolutionGradNode::isEqual(const ConvolutionGradNode &other) const {
  return true &&
      Input_ == other.Input_ &&
      Filter_ == other.Filter_ &&
      Bias_ == other.Bias_ &&
      OriginalOutputForResult_ == other.OriginalOutputForResult_ &&
      GradOfOriginalOutputNamedResult_ == other.GradOfOriginalOutputNamedResult_ &&
      predicate_ == other.predicate_ &&
      Kernels_ == other.Kernels_ &&
      Strides_ == other.Strides_ &&
      Pads_ == other.Pads_ &&
      Group_ == other.Group_ &&
      Dilation_ == other.Dilation_ &&
      Layout_ == other.Layout_ &&
      FusedActivation_ == other.FusedActivation_ &&
      getType(0) == other.getType(0) &&
      getType(1) == other.getType(1) &&
      getType(2) == other.getType(2);
}

Node* ConvolutionGradNode::clone() const {
  return new ConvolutionGradNode(getName(), getInput(), getFilter(), getBias(), getOriginalOutputForResult(), getGradOfOriginalOutputNamedResult(), getKernels(), getStrides(), getPads(), getGroup(), getDilation(), getLayout(), getFusedActivation());
}

llvm::hash_code ConvolutionGradNode::getHash() const {
  return llvm::hash_combine(
      llvm::hash_combine_range(Kernels_.begin(), Kernels_.end()),
      llvm::hash_combine_range(Strides_.begin(), Strides_.end()),
      llvm::hash_combine_range(Pads_.begin(), Pads_.end()),
      Group_,
      Dilation_,
      Layout_,
      FusedActivation_,
      Input_,
      Filter_,
      Bias_,
      OriginalOutputForResult_,
      GradOfOriginalOutputNamedResult_);
}

unsigned ConvolutionNode::getNumInputs() const {
  return 3;
}

std::string ConvolutionNode::getInputName(unsigned idx) const {
  if (idx == 0) { return "Input"; }
  if (idx == 1) { return "Filter"; }
  if (idx == 2) { return "Bias"; }
  idx -= 3;
  llvm_unreachable("Invalid index");
}

NodeValue ConvolutionNode::getNthInput(unsigned idx) {
  if (idx == 0) { return Input_; }
  if (idx == 1) { return Filter_; }
  if (idx == 2) { return Bias_; }
  idx -= 3;
  llvm_unreachable("Invalid index");
}

void ConvolutionNode::setNthInput(unsigned idx, NodeValue val) {
  if (idx == 0) { Input_ = val; return; }
  if (idx == 1) { Filter_ = val; return; }
  if (idx == 2) { Bias_ = val; return; }
  idx -= 3;
  llvm_unreachable("Invalid index");
}

llvm::StringRef ConvolutionNode::getOutputName(unsigned idx) const {
  if (idx == 0) { return "Result"; }
  llvm_unreachable("Invalid index");
}

std::string ConvolutionNode::getDebugDesc() const {
  DescriptionBuilder db(getKindName());
  db.addParam("name", getName());
  if (hasPredicate()) db.addParam("Predicate", "Yes");
  db
    .addParam("Input", *(getInput().getType()))
    .addParam("Filter", *(getFilter().getType()))
    .addParam("Bias", *(getBias().getType()))
    .addParam("Kernels", getKernels())
    .addParam("Strides", getStrides())
    .addParam("Pads", getPads())
    .addParam("Group", getGroup())
    .addParam("Dilation", getDilation())
    .addParam("Layout", getLayout())
    .addParam("FusedActivation", getFusedActivation())
    .addParam("users", getNumUsers());
  db.addParam("Result", *(getResult().getType()));
  return db;
}

void ConvolutionNode::visit(Node *parent, NodeWalker *visitor) {
  if (!visitor->shouldVisit(parent, this)) { return; }
  visitor->pre(parent, this);
if (hasPredicate())
 getPredicate().getNode()->visit(this, visitor);
  getInput().getNode()->visit(this, visitor);
  getFilter().getNode()->visit(this, visitor);
  getBias().getNode()->visit(this, visitor);
  visitor->post(parent, this);
}

bool ConvolutionNode::isEqual(const ConvolutionNode &other) const {
  return true &&
      Input_ == other.Input_ &&
      Filter_ == other.Filter_ &&
      Bias_ == other.Bias_ &&
      predicate_ == other.predicate_ &&
      Kernels_ == other.Kernels_ &&
      Strides_ == other.Strides_ &&
      Pads_ == other.Pads_ &&
      Group_ == other.Group_ &&
      Dilation_ == other.Dilation_ &&
      Layout_ == other.Layout_ &&
      FusedActivation_ == other.FusedActivation_ &&
      getType(0) == other.getType(0);
}

Node* ConvolutionNode::clone() const {
  return new ConvolutionNode(getName(), getResult().getType(), getInput(), getFilter(), getBias(), getKernels(), getStrides(), getPads(), getGroup(), getDilation(), getLayout(), getFusedActivation());
}

llvm::hash_code ConvolutionNode::getHash() const {
  return llvm::hash_combine(
      llvm::hash_combine_range(Kernels_.begin(), Kernels_.end()),
      llvm::hash_combine_range(Strides_.begin(), Strides_.end()),
      llvm::hash_combine_range(Pads_.begin(), Pads_.end()),
      Group_,
      Dilation_,
      Layout_,
      FusedActivation_,
      Input_,
      Filter_,
      Bias_);
}
bool ConvolutionNode::hasFusedActivation() const { return getFusedActivation() != FusedActivation::NONE; }
ConvolutionGradNode *ConvolutionNode::getGrad(GraphGradMapper &builder) {
  auto *x = new ConvolutionGradNode(getName().str() + "_grad", getInput(), getFilter(), getBias(), getResult(), builder.getGradient(getResult()), getKernels(), getStrides(), getPads(), getGroup(), getDilation(), getLayout(), getFusedActivation());
  builder.addGradient(getInput(), x->getGradOfInputNamedInput());
  builder.addGradient(getFilter(), x->getGradOfInputNamedFilter());
  builder.addGradient(getBias(), x->getGradOfInputNamedBias());
  return x;
}

unsigned ChannelwiseQuantizedConvolutionNode::getNumInputs() const {
  return 7;
}

std::string ChannelwiseQuantizedConvolutionNode::getInputName(unsigned idx) const {
  if (idx == 0) { return "Input"; }
  if (idx == 1) { return "Filter"; }
  if (idx == 2) { return "Bias"; }
  if (idx == 3) { return "FilterScales"; }
  if (idx == 4) { return "FilterOffsets"; }
  if (idx == 5) { return "BiasScales"; }
  if (idx == 6) { return "BiasOffsets"; }
  idx -= 7;
  llvm_unreachable("Invalid index");
}

NodeValue ChannelwiseQuantizedConvolutionNode::getNthInput(unsigned idx) {
  if (idx == 0) { return Input_; }
  if (idx == 1) { return Filter_; }
  if (idx == 2) { return Bias_; }
  if (idx == 3) { return FilterScales_; }
  if (idx == 4) { return FilterOffsets_; }
  if (idx == 5) { return BiasScales_; }
  if (idx == 6) { return BiasOffsets_; }
  idx -= 7;
  llvm_unreachable("Invalid index");
}

void ChannelwiseQuantizedConvolutionNode::setNthInput(unsigned idx, NodeValue val) {
  if (idx == 0) { Input_ = val; return; }
  if (idx == 1) { Filter_ = val; return; }
  if (idx == 2) { Bias_ = val; return; }
  if (idx == 3) { FilterScales_ = val; return; }
  if (idx == 4) { FilterOffsets_ = val; return; }
  if (idx == 5) { BiasScales_ = val; return; }
  if (idx == 6) { BiasOffsets_ = val; return; }
  idx -= 7;
  llvm_unreachable("Invalid index");
}

llvm::StringRef ChannelwiseQuantizedConvolutionNode::getOutputName(unsigned idx) const {
  if (idx == 0) { return "Result"; }
  llvm_unreachable("Invalid index");
}

std::string ChannelwiseQuantizedConvolutionNode::getDebugDesc() const {
  DescriptionBuilder db(getKindName());
  db.addParam("name", getName());
  if (hasPredicate()) db.addParam("Predicate", "Yes");
  db
    .addParam("Input", *(getInput().getType()))
    .addParam("Filter", *(getFilter().getType()))
    .addParam("Bias", *(getBias().getType()))
    .addParam("FilterScales", *(getFilterScales().getType()))
    .addParam("FilterOffsets", *(getFilterOffsets().getType()))
    .addParam("BiasScales", *(getBiasScales().getType()))
    .addParam("BiasOffsets", *(getBiasOffsets().getType()))
    .addParam("Kernels", getKernels())
    .addParam("Strides", getStrides())
    .addParam("Pads", getPads())
    .addParam("Group", getGroup())
    .addParam("Dilation", getDilation())
    .addParam("users", getNumUsers());
  db.addParam("Result", *(getResult().getType()));
  return db;
}

void ChannelwiseQuantizedConvolutionNode::visit(Node *parent, NodeWalker *visitor) {
  if (!visitor->shouldVisit(parent, this)) { return; }
  visitor->pre(parent, this);
if (hasPredicate())
 getPredicate().getNode()->visit(this, visitor);
  getInput().getNode()->visit(this, visitor);
  getFilter().getNode()->visit(this, visitor);
  getBias().getNode()->visit(this, visitor);
  getFilterScales().getNode()->visit(this, visitor);
  getFilterOffsets().getNode()->visit(this, visitor);
  getBiasScales().getNode()->visit(this, visitor);
  getBiasOffsets().getNode()->visit(this, visitor);
  visitor->post(parent, this);
}

bool ChannelwiseQuantizedConvolutionNode::isEqual(const ChannelwiseQuantizedConvolutionNode &other) const {
  return true &&
      Input_ == other.Input_ &&
      Filter_ == other.Filter_ &&
      Bias_ == other.Bias_ &&
      FilterScales_ == other.FilterScales_ &&
      FilterOffsets_ == other.FilterOffsets_ &&
      BiasScales_ == other.BiasScales_ &&
      BiasOffsets_ == other.BiasOffsets_ &&
      predicate_ == other.predicate_ &&
      Kernels_ == other.Kernels_ &&
      Strides_ == other.Strides_ &&
      Pads_ == other.Pads_ &&
      Group_ == other.Group_ &&
      Dilation_ == other.Dilation_ &&
      getType(0) == other.getType(0);
}

Node* ChannelwiseQuantizedConvolutionNode::clone() const {
  return new ChannelwiseQuantizedConvolutionNode(getName(), getResult().getType(), getInput(), getFilter(), getBias(), getFilterScales(), getFilterOffsets(), getBiasScales(), getBiasOffsets(), getKernels(), getStrides(), getPads(), getGroup(), getDilation());
}

llvm::hash_code ChannelwiseQuantizedConvolutionNode::getHash() const {
  return llvm::hash_combine(
      llvm::hash_combine_range(Kernels_.begin(), Kernels_.end()),
      llvm::hash_combine_range(Strides_.begin(), Strides_.end()),
      llvm::hash_combine_range(Pads_.begin(), Pads_.end()),
      Group_,
      Dilation_,
      Input_,
      Filter_,
      Bias_,
      FilterScales_,
      FilterOffsets_,
      BiasScales_,
      BiasOffsets_);
}

unsigned ConvTransposeNode::getNumInputs() const {
  return 3;
}

std::string ConvTransposeNode::getInputName(unsigned idx) const {
  if (idx == 0) { return "Input"; }
  if (idx == 1) { return "Filter"; }
  if (idx == 2) { return "Bias"; }
  idx -= 3;
  llvm_unreachable("Invalid index");
}

NodeValue ConvTransposeNode::getNthInput(unsigned idx) {
  if (idx == 0) { return Input_; }
  if (idx == 1) { return Filter_; }
  if (idx == 2) { return Bias_; }
  idx -= 3;
  llvm_unreachable("Invalid index");
}

void ConvTransposeNode::setNthInput(unsigned idx, NodeValue val) {
  if (idx == 0) { Input_ = val; return; }
  if (idx == 1) { Filter_ = val; return; }
  if (idx == 2) { Bias_ = val; return; }
  idx -= 3;
  llvm_unreachable("Invalid index");
}

llvm::StringRef ConvTransposeNode::getOutputName(unsigned idx) const {
  if (idx == 0) { return "Result"; }
  llvm_unreachable("Invalid index");
}

std::string ConvTransposeNode::getDebugDesc() const {
  DescriptionBuilder db(getKindName());
  db.addParam("name", getName());
  if (hasPredicate()) db.addParam("Predicate", "Yes");
  db
    .addParam("Input", *(getInput().getType()))
    .addParam("Filter", *(getFilter().getType()))
    .addParam("Bias", *(getBias().getType()))
    .addParam("Kernels", getKernels())
    .addParam("Strides", getStrides())
    .addParam("Pads", getPads())
    .addParam("Group", getGroup())
    .addParam("Dilation", getDilation())
    .addParam("users", getNumUsers());
  db.addParam("Result", *(getResult().getType()));
  return db;
}

void ConvTransposeNode::visit(Node *parent, NodeWalker *visitor) {
  if (!visitor->shouldVisit(parent, this)) { return; }
  visitor->pre(parent, this);
if (hasPredicate())
 getPredicate().getNode()->visit(this, visitor);
  getInput().getNode()->visit(this, visitor);
  getFilter().getNode()->visit(this, visitor);
  getBias().getNode()->visit(this, visitor);
  visitor->post(parent, this);
}

bool ConvTransposeNode::isEqual(const ConvTransposeNode &other) const {
  return true &&
      Input_ == other.Input_ &&
      Filter_ == other.Filter_ &&
      Bias_ == other.Bias_ &&
      predicate_ == other.predicate_ &&
      Kernels_ == other.Kernels_ &&
      Strides_ == other.Strides_ &&
      Pads_ == other.Pads_ &&
      Group_ == other.Group_ &&
      Dilation_ == other.Dilation_ &&
      getType(0) == other.getType(0);
}

Node* ConvTransposeNode::clone() const {
  return new ConvTransposeNode(getName(), getResult().getType(), getInput(), getFilter(), getBias(), getKernels(), getStrides(), getPads(), getGroup(), getDilation());
}

llvm::hash_code ConvTransposeNode::getHash() const {
  return llvm::hash_combine(
      llvm::hash_combine_range(Kernels_.begin(), Kernels_.end()),
      llvm::hash_combine_range(Strides_.begin(), Strides_.end()),
      llvm::hash_combine_range(Pads_.begin(), Pads_.end()),
      Group_,
      Dilation_,
      Input_,
      Filter_,
      Bias_);
}

unsigned Convolution3DGradNode::getNumInputs() const {
  return 5;
}

std::string Convolution3DGradNode::getInputName(unsigned idx) const {
  if (idx == 0) { return "Input"; }
  if (idx == 1) { return "Filter"; }
  if (idx == 2) { return "Bias"; }
  if (idx == 3) { return "OriginalOutputForResult"; }
  if (idx == 4) { return "GradOfOriginalOutputNamedResult"; }
  idx -= 5;
  llvm_unreachable("Invalid index");
}

NodeValue Convolution3DGradNode::getNthInput(unsigned idx) {
  if (idx == 0) { return Input_; }
  if (idx == 1) { return Filter_; }
  if (idx == 2) { return Bias_; }
  if (idx == 3) { return OriginalOutputForResult_; }
  if (idx == 4) { return GradOfOriginalOutputNamedResult_; }
  idx -= 5;
  llvm_unreachable("Invalid index");
}

void Convolution3DGradNode::setNthInput(unsigned idx, NodeValue val) {
  if (idx == 0) { Input_ = val; return; }
  if (idx == 1) { Filter_ = val; return; }
  if (idx == 2) { Bias_ = val; return; }
  if (idx == 3) { OriginalOutputForResult_ = val; return; }
  if (idx == 4) { GradOfOriginalOutputNamedResult_ = val; return; }
  idx -= 5;
  llvm_unreachable("Invalid index");
}

llvm::StringRef Convolution3DGradNode::getOutputName(unsigned idx) const {
  if (idx == 0) { return "GradOfInputNamedInput"; }
  if (idx == 1) { return "GradOfInputNamedFilter"; }
  if (idx == 2) { return "GradOfInputNamedBias"; }
  llvm_unreachable("Invalid index");
}

std::string Convolution3DGradNode::getDebugDesc() const {
  DescriptionBuilder db(getKindName());
  db.addParam("name", getName());
  if (hasPredicate()) db.addParam("Predicate", "Yes");
  db
    .addParam("Input", *(getInput().getType()))
    .addParam("Filter", *(getFilter().getType()))
    .addParam("Bias", *(getBias().getType()))
    .addParam("OriginalOutputForResult", *(getOriginalOutputForResult().getType()))
    .addParam("GradOfOriginalOutputNamedResult", *(getGradOfOriginalOutputNamedResult().getType()))
    .addParam("Kernels", getKernels())
    .addParam("Strides", getStrides())
    .addParam("Pads", getPads())
    .addParam("Group", getGroup())
    .addParam("users", getNumUsers());
  db.addParam("GradOfInputNamedInput", *(getGradOfInputNamedInput().getType()));
  db.addParam("GradOfInputNamedFilter", *(getGradOfInputNamedFilter().getType()));
  db.addParam("GradOfInputNamedBias", *(getGradOfInputNamedBias().getType()));
  return db;
}

void Convolution3DGradNode::visit(Node *parent, NodeWalker *visitor) {
  if (!visitor->shouldVisit(parent, this)) { return; }
  visitor->pre(parent, this);
if (hasPredicate())
 getPredicate().getNode()->visit(this, visitor);
  getInput().getNode()->visit(this, visitor);
  getFilter().getNode()->visit(this, visitor);
  getBias().getNode()->visit(this, visitor);
  getOriginalOutputForResult().getNode()->visit(this, visitor);
  getGradOfOriginalOutputNamedResult().getNode()->visit(this, visitor);
  visitor->post(parent, this);
}

bool Convolution3DGradNode::isEqual(const Convolution3DGradNode &other) const {
  return true &&
      Input_ == other.Input_ &&
      Filter_ == other.Filter_ &&
      Bias_ == other.Bias_ &&
      OriginalOutputForResult_ == other.OriginalOutputForResult_ &&
      GradOfOriginalOutputNamedResult_ == other.GradOfOriginalOutputNamedResult_ &&
      predicate_ == other.predicate_ &&
      Kernels_ == other.Kernels_ &&
      Strides_ == other.Strides_ &&
      Pads_ == other.Pads_ &&
      Group_ == other.Group_ &&
      getType(0) == other.getType(0) &&
      getType(1) == other.getType(1) &&
      getType(2) == other.getType(2);
}

Node* Convolution3DGradNode::clone() const {
  return new Convolution3DGradNode(getName(), getInput(), getFilter(), getBias(), getOriginalOutputForResult(), getGradOfOriginalOutputNamedResult(), getKernels(), getStrides(), getPads(), getGroup());
}

llvm::hash_code Convolution3DGradNode::getHash() const {
  return llvm::hash_combine(
      llvm::hash_combine_range(Kernels_.begin(), Kernels_.end()),
      llvm::hash_combine_range(Strides_.begin(), Strides_.end()),
      llvm::hash_combine_range(Pads_.begin(), Pads_.end()),
      Group_,
      Input_,
      Filter_,
      Bias_,
      OriginalOutputForResult_,
      GradOfOriginalOutputNamedResult_);
}

unsigned Convolution3DNode::getNumInputs() const {
  return 3;
}

std::string Convolution3DNode::getInputName(unsigned idx) const {
  if (idx == 0) { return "Input"; }
  if (idx == 1) { return "Filter"; }
  if (idx == 2) { return "Bias"; }
  idx -= 3;
  llvm_unreachable("Invalid index");
}

NodeValue Convolution3DNode::getNthInput(unsigned idx) {
  if (idx == 0) { return Input_; }
  if (idx == 1) { return Filter_; }
  if (idx == 2) { return Bias_; }
  idx -= 3;
  llvm_unreachable("Invalid index");
}

void Convolution3DNode::setNthInput(unsigned idx, NodeValue val) {
  if (idx == 0) { Input_ = val; return; }
  if (idx == 1) { Filter_ = val; return; }
  if (idx == 2) { Bias_ = val; return; }
  idx -= 3;
  llvm_unreachable("Invalid index");
}

llvm::StringRef Convolution3DNode::getOutputName(unsigned idx) const {
  if (idx == 0) { return "Result"; }
  llvm_unreachable("Invalid index");
}

std::string Convolution3DNode::getDebugDesc() const {
  DescriptionBuilder db(getKindName());
  db.addParam("name", getName());
  if (hasPredicate()) db.addParam("Predicate", "Yes");
  db
    .addParam("Input", *(getInput().getType()))
    .addParam("Filter", *(getFilter().getType()))
    .addParam("Bias", *(getBias().getType()))
    .addParam("Kernels", getKernels())
    .addParam("Strides", getStrides())
    .addParam("Pads", getPads())
    .addParam("Group", getGroup())
    .addParam("users", getNumUsers());
  db.addParam("Result", *(getResult().getType()));
  return db;
}

void Convolution3DNode::visit(Node *parent, NodeWalker *visitor) {
  if (!visitor->shouldVisit(parent, this)) { return; }
  visitor->pre(parent, this);
if (hasPredicate())
 getPredicate().getNode()->visit(this, visitor);
  getInput().getNode()->visit(this, visitor);
  getFilter().getNode()->visit(this, visitor);
  getBias().getNode()->visit(this, visitor);
  visitor->post(parent, this);
}

bool Convolution3DNode::isEqual(const Convolution3DNode &other) const {
  return true &&
      Input_ == other.Input_ &&
      Filter_ == other.Filter_ &&
      Bias_ == other.Bias_ &&
      predicate_ == other.predicate_ &&
      Kernels_ == other.Kernels_ &&
      Strides_ == other.Strides_ &&
      Pads_ == other.Pads_ &&
      Group_ == other.Group_ &&
      getType(0) == other.getType(0);
}

Node* Convolution3DNode::clone() const {
  return new Convolution3DNode(getName(), getResult().getType(), getInput(), getFilter(), getBias(), getKernels(), getStrides(), getPads(), getGroup());
}

llvm::hash_code Convolution3DNode::getHash() const {
  return llvm::hash_combine(
      llvm::hash_combine_range(Kernels_.begin(), Kernels_.end()),
      llvm::hash_combine_range(Strides_.begin(), Strides_.end()),
      llvm::hash_combine_range(Pads_.begin(), Pads_.end()),
      Group_,
      Input_,
      Filter_,
      Bias_);
}

Convolution3DGradNode *Convolution3DNode::getGrad(GraphGradMapper &builder) {
  auto *x = new Convolution3DGradNode(getName().str() + "_grad", getInput(), getFilter(), getBias(), getResult(), builder.getGradient(getResult()), getKernels(), getStrides(), getPads(), getGroup());
  builder.addGradient(getInput(), x->getGradOfInputNamedInput());
  builder.addGradient(getFilter(), x->getGradOfInputNamedFilter());
  builder.addGradient(getBias(), x->getGradOfInputNamedBias());
  return x;
}

unsigned MaxPoolGradNode::getNumInputs() const {
  return 5;
}

std::string MaxPoolGradNode::getInputName(unsigned idx) const {
  if (idx == 0) { return "Input"; }
  if (idx == 1) { return "OriginalOutputForResult"; }
  if (idx == 2) { return "GradOfOriginalOutputNamedResult"; }
  if (idx == 3) { return "OriginalOutputForArgmax"; }
  if (idx == 4) { return "GradOfOriginalOutputNamedArgmax"; }
  idx -= 5;
  llvm_unreachable("Invalid index");
}

NodeValue MaxPoolGradNode::getNthInput(unsigned idx) {
  if (idx == 0) { return Input_; }
  if (idx == 1) { return OriginalOutputForResult_; }
  if (idx == 2) { return GradOfOriginalOutputNamedResult_; }
  if (idx == 3) { return OriginalOutputForArgmax_; }
  if (idx == 4) { return GradOfOriginalOutputNamedArgmax_; }
  idx -= 5;
  llvm_unreachable("Invalid index");
}

void MaxPoolGradNode::setNthInput(unsigned idx, NodeValue val) {
  if (idx == 0) { Input_ = val; return; }
  if (idx == 1) { OriginalOutputForResult_ = val; return; }
  if (idx == 2) { GradOfOriginalOutputNamedResult_ = val; return; }
  if (idx == 3) { OriginalOutputForArgmax_ = val; return; }
  if (idx == 4) { GradOfOriginalOutputNamedArgmax_ = val; return; }
  idx -= 5;
  llvm_unreachable("Invalid index");
}

llvm::StringRef MaxPoolGradNode::getOutputName(unsigned idx) const {
  if (idx == 0) { return "GradOfInputNamedInput"; }
  llvm_unreachable("Invalid index");
}

std::string MaxPoolGradNode::getDebugDesc() const {
  DescriptionBuilder db(getKindName());
  db.addParam("name", getName());
  if (hasPredicate()) db.addParam("Predicate", "Yes");
  db
    .addParam("Input", *(getInput().getType()))
    .addParam("OriginalOutputForResult", *(getOriginalOutputForResult().getType()))
    .addParam("GradOfOriginalOutputNamedResult", *(getGradOfOriginalOutputNamedResult().getType()))
    .addParam("OriginalOutputForArgmax", *(getOriginalOutputForArgmax().getType()))
    .addParam("GradOfOriginalOutputNamedArgmax", *(getGradOfOriginalOutputNamedArgmax().getType()))
    .addParam("Kernels", getKernels())
    .addParam("Strides", getStrides())
    .addParam("Pads", getPads())
    .addParam("Layout", getLayout())
    .addParam("users", getNumUsers());
  db.addParam("GradOfInputNamedInput", *(getGradOfInputNamedInput().getType()));
  return db;
}

void MaxPoolGradNode::visit(Node *parent, NodeWalker *visitor) {
  if (!visitor->shouldVisit(parent, this)) { return; }
  visitor->pre(parent, this);
if (hasPredicate())
 getPredicate().getNode()->visit(this, visitor);
  getInput().getNode()->visit(this, visitor);
  getOriginalOutputForResult().getNode()->visit(this, visitor);
  getGradOfOriginalOutputNamedResult().getNode()->visit(this, visitor);
  getOriginalOutputForArgmax().getNode()->visit(this, visitor);
  getGradOfOriginalOutputNamedArgmax().getNode()->visit(this, visitor);
  visitor->post(parent, this);
}

bool MaxPoolGradNode::isEqual(const MaxPoolGradNode &other) const {
  return true &&
      Input_ == other.Input_ &&
      OriginalOutputForResult_ == other.OriginalOutputForResult_ &&
      GradOfOriginalOutputNamedResult_ == other.GradOfOriginalOutputNamedResult_ &&
      OriginalOutputForArgmax_ == other.OriginalOutputForArgmax_ &&
      GradOfOriginalOutputNamedArgmax_ == other.GradOfOriginalOutputNamedArgmax_ &&
      predicate_ == other.predicate_ &&
      Kernels_ == other.Kernels_ &&
      Strides_ == other.Strides_ &&
      Pads_ == other.Pads_ &&
      Layout_ == other.Layout_ &&
      getType(0) == other.getType(0);
}

Node* MaxPoolGradNode::clone() const {
  return new MaxPoolGradNode(getName(), getInput(), getOriginalOutputForResult(), getGradOfOriginalOutputNamedResult(), getOriginalOutputForArgmax(), getGradOfOriginalOutputNamedArgmax(), getKernels(), getStrides(), getPads(), getLayout());
}

llvm::hash_code MaxPoolGradNode::getHash() const {
  return llvm::hash_combine(
      llvm::hash_combine_range(Kernels_.begin(), Kernels_.end()),
      llvm::hash_combine_range(Strides_.begin(), Strides_.end()),
      llvm::hash_combine_range(Pads_.begin(), Pads_.end()),
      Layout_,
      Input_,
      OriginalOutputForResult_,
      GradOfOriginalOutputNamedResult_,
      OriginalOutputForArgmax_,
      GradOfOriginalOutputNamedArgmax_);
}

unsigned MaxPoolNode::getNumInputs() const {
  return 1;
}

std::string MaxPoolNode::getInputName(unsigned idx) const {
  if (idx == 0) { return "Input"; }
  idx -= 1;
  llvm_unreachable("Invalid index");
}

NodeValue MaxPoolNode::getNthInput(unsigned idx) {
  if (idx == 0) { return Input_; }
  idx -= 1;
  llvm_unreachable("Invalid index");
}

void MaxPoolNode::setNthInput(unsigned idx, NodeValue val) {
  if (idx == 0) { Input_ = val; return; }
  idx -= 1;
  llvm_unreachable("Invalid index");
}

llvm::StringRef MaxPoolNode::getOutputName(unsigned idx) const {
  if (idx == 0) { return "Result"; }
  if (idx == 1) { return "Argmax"; }
  llvm_unreachable("Invalid index");
}

std::string MaxPoolNode::getDebugDesc() const {
  DescriptionBuilder db(getKindName());
  db.addParam("name", getName());
  if (hasPredicate()) db.addParam("Predicate", "Yes");
  db
    .addParam("Input", *(getInput().getType()))
    .addParam("Kernels", getKernels())
    .addParam("Strides", getStrides())
    .addParam("Pads", getPads())
    .addParam("Layout", getLayout())
    .addParam("users", getNumUsers());
  db.addParam("Result", *(getResult().getType()));
  db.addParam("Argmax", *(getArgmax().getType()));
  return db;
}

void MaxPoolNode::visit(Node *parent, NodeWalker *visitor) {
  if (!visitor->shouldVisit(parent, this)) { return; }
  visitor->pre(parent, this);
if (hasPredicate())
 getPredicate().getNode()->visit(this, visitor);
  getInput().getNode()->visit(this, visitor);
  visitor->post(parent, this);
}

bool MaxPoolNode::isEqual(const MaxPoolNode &other) const {
  return true &&
      Input_ == other.Input_ &&
      predicate_ == other.predicate_ &&
      Kernels_ == other.Kernels_ &&
      Strides_ == other.Strides_ &&
      Pads_ == other.Pads_ &&
      Layout_ == other.Layout_ &&
      getType(0) == other.getType(0) &&
      getType(1) == other.getType(1);
}

Node* MaxPoolNode::clone() const {
  return new MaxPoolNode(getName(), getResult().getType(), getArgmax().getType(), getInput(), getKernels(), getStrides(), getPads(), getLayout());
}

llvm::hash_code MaxPoolNode::getHash() const {
  return llvm::hash_combine(
      llvm::hash_combine_range(Kernels_.begin(), Kernels_.end()),
      llvm::hash_combine_range(Strides_.begin(), Strides_.end()),
      llvm::hash_combine_range(Pads_.begin(), Pads_.end()),
      Layout_,
      Input_);
}

MaxPoolGradNode *MaxPoolNode::getGrad(GraphGradMapper &builder) {
  auto *x = new MaxPoolGradNode(getName().str() + "_grad", getInput(), getResult(), builder.getGradient(getResult()), getArgmax(), builder.getGradient(getArgmax()), getKernels(), getStrides(), getPads(), getLayout());
  builder.addGradient(getInput(), x->getGradOfInputNamedInput());
  return x;
}

unsigned ArgMaxNode::getNumInputs() const {
  return 1;
}

std::string ArgMaxNode::getInputName(unsigned idx) const {
  if (idx == 0) { return "Input"; }
  idx -= 1;
  llvm_unreachable("Invalid index");
}

NodeValue ArgMaxNode::getNthInput(unsigned idx) {
  if (idx == 0) { return Input_; }
  idx -= 1;
  llvm_unreachable("Invalid index");
}

void ArgMaxNode::setNthInput(unsigned idx, NodeValue val) {
  if (idx == 0) { Input_ = val; return; }
  idx -= 1;
  llvm_unreachable("Invalid index");
}

llvm::StringRef ArgMaxNode::getOutputName(unsigned idx) const {
  if (idx == 0) { return "Result"; }
  llvm_unreachable("Invalid index");
}

std::string ArgMaxNode::getDebugDesc() const {
  DescriptionBuilder db(getKindName());
  db.addParam("name", getName());
  if (hasPredicate()) db.addParam("Predicate", "Yes");
  db
    .addParam("Input", *(getInput().getType()))
    .addParam("Axis", getAxis())
    .addParam("KeepDims", getKeepDims())
    .addParam("users", getNumUsers());
  db.addParam("Result", *(getResult().getType()));
  return db;
}

void ArgMaxNode::visit(Node *parent, NodeWalker *visitor) {
  if (!visitor->shouldVisit(parent, this)) { return; }
  visitor->pre(parent, this);
if (hasPredicate())
 getPredicate().getNode()->visit(this, visitor);
  getInput().getNode()->visit(this, visitor);
  visitor->post(parent, this);
}

bool ArgMaxNode::isEqual(const ArgMaxNode &other) const {
  return true &&
      Input_ == other.Input_ &&
      predicate_ == other.predicate_ &&
      Axis_ == other.Axis_ &&
      KeepDims_ == other.KeepDims_ &&
      getType(0) == other.getType(0);
}

Node* ArgMaxNode::clone() const {
  return new ArgMaxNode(getName(), getResult().getType(), getInput(), getAxis(), getKeepDims());
}

llvm::hash_code ArgMaxNode::getHash() const {
  return llvm::hash_combine(
      Axis_,
      KeepDims_,
      Input_);
}

unsigned ArgMinNode::getNumInputs() const {
  return 1;
}

std::string ArgMinNode::getInputName(unsigned idx) const {
  if (idx == 0) { return "Input"; }
  idx -= 1;
  llvm_unreachable("Invalid index");
}

NodeValue ArgMinNode::getNthInput(unsigned idx) {
  if (idx == 0) { return Input_; }
  idx -= 1;
  llvm_unreachable("Invalid index");
}

void ArgMinNode::setNthInput(unsigned idx, NodeValue val) {
  if (idx == 0) { Input_ = val; return; }
  idx -= 1;
  llvm_unreachable("Invalid index");
}

llvm::StringRef ArgMinNode::getOutputName(unsigned idx) const {
  if (idx == 0) { return "Result"; }
  llvm_unreachable("Invalid index");
}

std::string ArgMinNode::getDebugDesc() const {
  DescriptionBuilder db(getKindName());
  db.addParam("name", getName());
  if (hasPredicate()) db.addParam("Predicate", "Yes");
  db
    .addParam("Input", *(getInput().getType()))
    .addParam("Axis", getAxis())
    .addParam("KeepDims", getKeepDims())
    .addParam("users", getNumUsers());
  db.addParam("Result", *(getResult().getType()));
  return db;
}

void ArgMinNode::visit(Node *parent, NodeWalker *visitor) {
  if (!visitor->shouldVisit(parent, this)) { return; }
  visitor->pre(parent, this);
if (hasPredicate())
 getPredicate().getNode()->visit(this, visitor);
  getInput().getNode()->visit(this, visitor);
  visitor->post(parent, this);
}

bool ArgMinNode::isEqual(const ArgMinNode &other) const {
  return true &&
      Input_ == other.Input_ &&
      predicate_ == other.predicate_ &&
      Axis_ == other.Axis_ &&
      KeepDims_ == other.KeepDims_ &&
      getType(0) == other.getType(0);
}

Node* ArgMinNode::clone() const {
  return new ArgMinNode(getName(), getResult().getType(), getInput(), getAxis(), getKeepDims());
}

llvm::hash_code ArgMinNode::getHash() const {
  return llvm::hash_combine(
      Axis_,
      KeepDims_,
      Input_);
}

unsigned AvgPoolGradNode::getNumInputs() const {
  return 3;
}

std::string AvgPoolGradNode::getInputName(unsigned idx) const {
  if (idx == 0) { return "Input"; }
  if (idx == 1) { return "OriginalOutputForResult"; }
  if (idx == 2) { return "GradOfOriginalOutputNamedResult"; }
  idx -= 3;
  llvm_unreachable("Invalid index");
}

NodeValue AvgPoolGradNode::getNthInput(unsigned idx) {
  if (idx == 0) { return Input_; }
  if (idx == 1) { return OriginalOutputForResult_; }
  if (idx == 2) { return GradOfOriginalOutputNamedResult_; }
  idx -= 3;
  llvm_unreachable("Invalid index");
}

void AvgPoolGradNode::setNthInput(unsigned idx, NodeValue val) {
  if (idx == 0) { Input_ = val; return; }
  if (idx == 1) { OriginalOutputForResult_ = val; return; }
  if (idx == 2) { GradOfOriginalOutputNamedResult_ = val; return; }
  idx -= 3;
  llvm_unreachable("Invalid index");
}

llvm::StringRef AvgPoolGradNode::getOutputName(unsigned idx) const {
  if (idx == 0) { return "GradOfInputNamedInput"; }
  llvm_unreachable("Invalid index");
}

std::string AvgPoolGradNode::getDebugDesc() const {
  DescriptionBuilder db(getKindName());
  db.addParam("name", getName());
  if (hasPredicate()) db.addParam("Predicate", "Yes");
  db
    .addParam("Input", *(getInput().getType()))
    .addParam("OriginalOutputForResult", *(getOriginalOutputForResult().getType()))
    .addParam("GradOfOriginalOutputNamedResult", *(getGradOfOriginalOutputNamedResult().getType()))
    .addParam("Kernels", getKernels())
    .addParam("Strides", getStrides())
    .addParam("Pads", getPads())
    .addParam("Layout", getLayout())
    .addParam("users", getNumUsers());
  db.addParam("GradOfInputNamedInput", *(getGradOfInputNamedInput().getType()));
  return db;
}

void AvgPoolGradNode::visit(Node *parent, NodeWalker *visitor) {
  if (!visitor->shouldVisit(parent, this)) { return; }
  visitor->pre(parent, this);
if (hasPredicate())
 getPredicate().getNode()->visit(this, visitor);
  getInput().getNode()->visit(this, visitor);
  getOriginalOutputForResult().getNode()->visit(this, visitor);
  getGradOfOriginalOutputNamedResult().getNode()->visit(this, visitor);
  visitor->post(parent, this);
}

bool AvgPoolGradNode::isEqual(const AvgPoolGradNode &other) const {
  return true &&
      Input_ == other.Input_ &&
      OriginalOutputForResult_ == other.OriginalOutputForResult_ &&
      GradOfOriginalOutputNamedResult_ == other.GradOfOriginalOutputNamedResult_ &&
      predicate_ == other.predicate_ &&
      Kernels_ == other.Kernels_ &&
      Strides_ == other.Strides_ &&
      Pads_ == other.Pads_ &&
      Layout_ == other.Layout_ &&
      getType(0) == other.getType(0);
}

Node* AvgPoolGradNode::clone() const {
  return new AvgPoolGradNode(getName(), getInput(), getOriginalOutputForResult(), getGradOfOriginalOutputNamedResult(), getKernels(), getStrides(), getPads(), getLayout());
}

llvm::hash_code AvgPoolGradNode::getHash() const {
  return llvm::hash_combine(
      llvm::hash_combine_range(Kernels_.begin(), Kernels_.end()),
      llvm::hash_combine_range(Strides_.begin(), Strides_.end()),
      llvm::hash_combine_range(Pads_.begin(), Pads_.end()),
      Layout_,
      Input_,
      OriginalOutputForResult_,
      GradOfOriginalOutputNamedResult_);
}

unsigned AvgPoolNode::getNumInputs() const {
  return 1;
}

std::string AvgPoolNode::getInputName(unsigned idx) const {
  if (idx == 0) { return "Input"; }
  idx -= 1;
  llvm_unreachable("Invalid index");
}

NodeValue AvgPoolNode::getNthInput(unsigned idx) {
  if (idx == 0) { return Input_; }
  idx -= 1;
  llvm_unreachable("Invalid index");
}

void AvgPoolNode::setNthInput(unsigned idx, NodeValue val) {
  if (idx == 0) { Input_ = val; return; }
  idx -= 1;
  llvm_unreachable("Invalid index");
}

llvm::StringRef AvgPoolNode::getOutputName(unsigned idx) const {
  if (idx == 0) { return "Result"; }
  llvm_unreachable("Invalid index");
}

std::string AvgPoolNode::getDebugDesc() const {
  DescriptionBuilder db(getKindName());
  db.addParam("name", getName());
  if (hasPredicate()) db.addParam("Predicate", "Yes");
  db
    .addParam("Input", *(getInput().getType()))
    .addParam("Kernels", getKernels())
    .addParam("Strides", getStrides())
    .addParam("Pads", getPads())
    .addParam("Layout", getLayout())
    .addParam("users", getNumUsers());
  db.addParam("Result", *(getResult().getType()));
  return db;
}

void AvgPoolNode::visit(Node *parent, NodeWalker *visitor) {
  if (!visitor->shouldVisit(parent, this)) { return; }
  visitor->pre(parent, this);
if (hasPredicate())
 getPredicate().getNode()->visit(this, visitor);
  getInput().getNode()->visit(this, visitor);
  visitor->post(parent, this);
}

bool AvgPoolNode::isEqual(const AvgPoolNode &other) const {
  return true &&
      Input_ == other.Input_ &&
      predicate_ == other.predicate_ &&
      Kernels_ == other.Kernels_ &&
      Strides_ == other.Strides_ &&
      Pads_ == other.Pads_ &&
      Layout_ == other.Layout_ &&
      getType(0) == other.getType(0);
}

Node* AvgPoolNode::clone() const {
  return new AvgPoolNode(getName(), getResult().getType(), getInput(), getKernels(), getStrides(), getPads(), getLayout());
}

llvm::hash_code AvgPoolNode::getHash() const {
  return llvm::hash_combine(
      llvm::hash_combine_range(Kernels_.begin(), Kernels_.end()),
      llvm::hash_combine_range(Strides_.begin(), Strides_.end()),
      llvm::hash_combine_range(Pads_.begin(), Pads_.end()),
      Layout_,
      Input_);
}

AvgPoolGradNode *AvgPoolNode::getGrad(GraphGradMapper &builder) {
  auto *x = new AvgPoolGradNode(getName().str() + "_grad", getInput(), getResult(), builder.getGradient(getResult()), getKernels(), getStrides(), getPads(), getLayout());
  builder.addGradient(getInput(), x->getGradOfInputNamedInput());
  return x;
}

unsigned AdaptiveAvgPoolGradNode::getNumInputs() const {
  return 3;
}

std::string AdaptiveAvgPoolGradNode::getInputName(unsigned idx) const {
  if (idx == 0) { return "Input"; }
  if (idx == 1) { return "OriginalOutputForResult"; }
  if (idx == 2) { return "GradOfOriginalOutputNamedResult"; }
  idx -= 3;
  llvm_unreachable("Invalid index");
}

NodeValue AdaptiveAvgPoolGradNode::getNthInput(unsigned idx) {
  if (idx == 0) { return Input_; }
  if (idx == 1) { return OriginalOutputForResult_; }
  if (idx == 2) { return GradOfOriginalOutputNamedResult_; }
  idx -= 3;
  llvm_unreachable("Invalid index");
}

void AdaptiveAvgPoolGradNode::setNthInput(unsigned idx, NodeValue val) {
  if (idx == 0) { Input_ = val; return; }
  if (idx == 1) { OriginalOutputForResult_ = val; return; }
  if (idx == 2) { GradOfOriginalOutputNamedResult_ = val; return; }
  idx -= 3;
  llvm_unreachable("Invalid index");
}

llvm::StringRef AdaptiveAvgPoolGradNode::getOutputName(unsigned idx) const {
  if (idx == 0) { return "GradOfInputNamedInput"; }
  llvm_unreachable("Invalid index");
}

std::string AdaptiveAvgPoolGradNode::getDebugDesc() const {
  DescriptionBuilder db(getKindName());
  db.addParam("name", getName());
  if (hasPredicate()) db.addParam("Predicate", "Yes");
  db
    .addParam("Input", *(getInput().getType()))
    .addParam("OriginalOutputForResult", *(getOriginalOutputForResult().getType()))
    .addParam("GradOfOriginalOutputNamedResult", *(getGradOfOriginalOutputNamedResult().getType()))
    .addParam("users", getNumUsers());
  db.addParam("GradOfInputNamedInput", *(getGradOfInputNamedInput().getType()));
  return db;
}

void AdaptiveAvgPoolGradNode::visit(Node *parent, NodeWalker *visitor) {
  if (!visitor->shouldVisit(parent, this)) { return; }
  visitor->pre(parent, this);
if (hasPredicate())
 getPredicate().getNode()->visit(this, visitor);
  getInput().getNode()->visit(this, visitor);
  getOriginalOutputForResult().getNode()->visit(this, visitor);
  getGradOfOriginalOutputNamedResult().getNode()->visit(this, visitor);
  visitor->post(parent, this);
}

bool AdaptiveAvgPoolGradNode::isEqual(const AdaptiveAvgPoolGradNode &other) const {
  return true &&
      Input_ == other.Input_ &&
      OriginalOutputForResult_ == other.OriginalOutputForResult_ &&
      GradOfOriginalOutputNamedResult_ == other.GradOfOriginalOutputNamedResult_ &&
      predicate_ == other.predicate_ &&
      getType(0) == other.getType(0);
}

Node* AdaptiveAvgPoolGradNode::clone() const {
  return new AdaptiveAvgPoolGradNode(getName(), getInput(), getOriginalOutputForResult(), getGradOfOriginalOutputNamedResult());
}

llvm::hash_code AdaptiveAvgPoolGradNode::getHash() const {
  return llvm::hash_combine(
      Input_,
      OriginalOutputForResult_,
      GradOfOriginalOutputNamedResult_);
}

unsigned AdaptiveAvgPoolNode::getNumInputs() const {
  return 1;
}

std::string AdaptiveAvgPoolNode::getInputName(unsigned idx) const {
  if (idx == 0) { return "Input"; }
  idx -= 1;
  llvm_unreachable("Invalid index");
}

NodeValue AdaptiveAvgPoolNode::getNthInput(unsigned idx) {
  if (idx == 0) { return Input_; }
  idx -= 1;
  llvm_unreachable("Invalid index");
}

void AdaptiveAvgPoolNode::setNthInput(unsigned idx, NodeValue val) {
  if (idx == 0) { Input_ = val; return; }
  idx -= 1;
  llvm_unreachable("Invalid index");
}

llvm::StringRef AdaptiveAvgPoolNode::getOutputName(unsigned idx) const {
  if (idx == 0) { return "Result"; }
  llvm_unreachable("Invalid index");
}

std::string AdaptiveAvgPoolNode::getDebugDesc() const {
  DescriptionBuilder db(getKindName());
  db.addParam("name", getName());
  if (hasPredicate()) db.addParam("Predicate", "Yes");
  db
    .addParam("Input", *(getInput().getType()))
    .addParam("users", getNumUsers());
  db.addParam("Result", *(getResult().getType()));
  return db;
}

void AdaptiveAvgPoolNode::visit(Node *parent, NodeWalker *visitor) {
  if (!visitor->shouldVisit(parent, this)) { return; }
  visitor->pre(parent, this);
if (hasPredicate())
 getPredicate().getNode()->visit(this, visitor);
  getInput().getNode()->visit(this, visitor);
  visitor->post(parent, this);
}

bool AdaptiveAvgPoolNode::isEqual(const AdaptiveAvgPoolNode &other) const {
  return true &&
      Input_ == other.Input_ &&
      predicate_ == other.predicate_ &&
      getType(0) == other.getType(0);
}

Node* AdaptiveAvgPoolNode::clone() const {
  return new AdaptiveAvgPoolNode(getName(), getResult().getType(), getInput());
}

llvm::hash_code AdaptiveAvgPoolNode::getHash() const {
  return llvm::hash_combine(
      Input_);
}

AdaptiveAvgPoolGradNode *AdaptiveAvgPoolNode::getGrad(GraphGradMapper &builder) {
  auto *x = new AdaptiveAvgPoolGradNode(getName().str() + "_grad", getInput(), getResult(), builder.getGradient(getResult()));
  builder.addGradient(getInput(), x->getGradOfInputNamedInput());
  return x;
}

unsigned GemmNode::getNumInputs() const {
  return 3;
}

std::string GemmNode::getInputName(unsigned idx) const {
  if (idx == 0) { return "A"; }
  if (idx == 1) { return "B"; }
  if (idx == 2) { return "C"; }
  idx -= 3;
  llvm_unreachable("Invalid index");
}

NodeValue GemmNode::getNthInput(unsigned idx) {
  if (idx == 0) { return A_; }
  if (idx == 1) { return B_; }
  if (idx == 2) { return C_; }
  idx -= 3;
  llvm_unreachable("Invalid index");
}

void GemmNode::setNthInput(unsigned idx, NodeValue val) {
  if (idx == 0) { A_ = val; return; }
  if (idx == 1) { B_ = val; return; }
  if (idx == 2) { C_ = val; return; }
  idx -= 3;
  llvm_unreachable("Invalid index");
}

llvm::StringRef GemmNode::getOutputName(unsigned idx) const {
  if (idx == 0) { return "Result"; }
  llvm_unreachable("Invalid index");
}

std::string GemmNode::getDebugDesc() const {
  DescriptionBuilder db(getKindName());
  db.addParam("name", getName());
  if (hasPredicate()) db.addParam("Predicate", "Yes");
  db
    .addParam("A", *(getA().getType()))
    .addParam("B", *(getB().getType()))
    .addParam("C", *(getC().getType()))
    .addParam("Alpha", getAlpha())
    .addParam("Beta", getBeta())
    .addParam("TransposeA", getTransposeA())
    .addParam("TransposeB", getTransposeB())
    .addParam("users", getNumUsers());
  db.addParam("Result", *(getResult().getType()));
  return db;
}

void GemmNode::visit(Node *parent, NodeWalker *visitor) {
  if (!visitor->shouldVisit(parent, this)) { return; }
  visitor->pre(parent, this);
if (hasPredicate())
 getPredicate().getNode()->visit(this, visitor);
  getA().getNode()->visit(this, visitor);
  getB().getNode()->visit(this, visitor);
  getC().getNode()->visit(this, visitor);
  visitor->post(parent, this);
}

bool GemmNode::isEqual(const GemmNode &other) const {
  return true &&
      A_ == other.A_ &&
      B_ == other.B_ &&
      C_ == other.C_ &&
      predicate_ == other.predicate_ &&
      Alpha_ == other.Alpha_ &&
      Beta_ == other.Beta_ &&
      TransposeA_ == other.TransposeA_ &&
      TransposeB_ == other.TransposeB_ &&
      getType(0) == other.getType(0);
}

Node* GemmNode::clone() const {
  return new GemmNode(getName(), getResult().getType(), getA(), getB(), getC(), getAlpha(), getBeta(), getTransposeA(), getTransposeB());
}

llvm::hash_code GemmNode::getHash() const {
  return llvm::hash_combine(
      toBinary(Alpha_),
      toBinary(Beta_),
      TransposeA_,
      TransposeB_,
      A_,
      B_,
      C_);
}

unsigned FullyConnectedGradNode::getNumInputs() const {
  return 5;
}

std::string FullyConnectedGradNode::getInputName(unsigned idx) const {
  if (idx == 0) { return "Input"; }
  if (idx == 1) { return "Weights"; }
  if (idx == 2) { return "Bias"; }
  if (idx == 3) { return "OriginalOutputForResult"; }
  if (idx == 4) { return "GradOfOriginalOutputNamedResult"; }
  idx -= 5;
  llvm_unreachable("Invalid index");
}

NodeValue FullyConnectedGradNode::getNthInput(unsigned idx) {
  if (idx == 0) { return Input_; }
  if (idx == 1) { return Weights_; }
  if (idx == 2) { return Bias_; }
  if (idx == 3) { return OriginalOutputForResult_; }
  if (idx == 4) { return GradOfOriginalOutputNamedResult_; }
  idx -= 5;
  llvm_unreachable("Invalid index");
}

void FullyConnectedGradNode::setNthInput(unsigned idx, NodeValue val) {
  if (idx == 0) { Input_ = val; return; }
  if (idx == 1) { Weights_ = val; return; }
  if (idx == 2) { Bias_ = val; return; }
  if (idx == 3) { OriginalOutputForResult_ = val; return; }
  if (idx == 4) { GradOfOriginalOutputNamedResult_ = val; return; }
  idx -= 5;
  llvm_unreachable("Invalid index");
}

llvm::StringRef FullyConnectedGradNode::getOutputName(unsigned idx) const {
  if (idx == 0) { return "GradOfInputNamedInput"; }
  if (idx == 1) { return "GradOfInputNamedWeights"; }
  if (idx == 2) { return "GradOfInputNamedBias"; }
  llvm_unreachable("Invalid index");
}

std::string FullyConnectedGradNode::getDebugDesc() const {
  DescriptionBuilder db(getKindName());
  db.addParam("name", getName());
  if (hasPredicate()) db.addParam("Predicate", "Yes");
  db
    .addParam("Input", *(getInput().getType()))
    .addParam("Weights", *(getWeights().getType()))
    .addParam("Bias", *(getBias().getType()))
    .addParam("OriginalOutputForResult", *(getOriginalOutputForResult().getType()))
    .addParam("GradOfOriginalOutputNamedResult", *(getGradOfOriginalOutputNamedResult().getType()))
    .addParam("users", getNumUsers());
  db.addParam("GradOfInputNamedInput", *(getGradOfInputNamedInput().getType()));
  db.addParam("GradOfInputNamedWeights", *(getGradOfInputNamedWeights().getType()));
  db.addParam("GradOfInputNamedBias", *(getGradOfInputNamedBias().getType()));
  return db;
}

void FullyConnectedGradNode::visit(Node *parent, NodeWalker *visitor) {
  if (!visitor->shouldVisit(parent, this)) { return; }
  visitor->pre(parent, this);
if (hasPredicate())
 getPredicate().getNode()->visit(this, visitor);
  getInput().getNode()->visit(this, visitor);
  getWeights().getNode()->visit(this, visitor);
  getBias().getNode()->visit(this, visitor);
  getOriginalOutputForResult().getNode()->visit(this, visitor);
  getGradOfOriginalOutputNamedResult().getNode()->visit(this, visitor);
  visitor->post(parent, this);
}

bool FullyConnectedGradNode::isEqual(const FullyConnectedGradNode &other) const {
  return true &&
      Input_ == other.Input_ &&
      Weights_ == other.Weights_ &&
      Bias_ == other.Bias_ &&
      OriginalOutputForResult_ == other.OriginalOutputForResult_ &&
      GradOfOriginalOutputNamedResult_ == other.GradOfOriginalOutputNamedResult_ &&
      predicate_ == other.predicate_ &&
      getType(0) == other.getType(0) &&
      getType(1) == other.getType(1) &&
      getType(2) == other.getType(2);
}

Node* FullyConnectedGradNode::clone() const {
  return new FullyConnectedGradNode(getName(), getInput(), getWeights(), getBias(), getOriginalOutputForResult(), getGradOfOriginalOutputNamedResult());
}

llvm::hash_code FullyConnectedGradNode::getHash() const {
  return llvm::hash_combine(
      Input_,
      Weights_,
      Bias_,
      OriginalOutputForResult_,
      GradOfOriginalOutputNamedResult_);
}

unsigned FullyConnectedNode::getNumInputs() const {
  return 3;
}

std::string FullyConnectedNode::getInputName(unsigned idx) const {
  if (idx == 0) { return "Input"; }
  if (idx == 1) { return "Weights"; }
  if (idx == 2) { return "Bias"; }
  idx -= 3;
  llvm_unreachable("Invalid index");
}

NodeValue FullyConnectedNode::getNthInput(unsigned idx) {
  if (idx == 0) { return Input_; }
  if (idx == 1) { return Weights_; }
  if (idx == 2) { return Bias_; }
  idx -= 3;
  llvm_unreachable("Invalid index");
}

void FullyConnectedNode::setNthInput(unsigned idx, NodeValue val) {
  if (idx == 0) { Input_ = val; return; }
  if (idx == 1) { Weights_ = val; return; }
  if (idx == 2) { Bias_ = val; return; }
  idx -= 3;
  llvm_unreachable("Invalid index");
}

llvm::StringRef FullyConnectedNode::getOutputName(unsigned idx) const {
  if (idx == 0) { return "Result"; }
  llvm_unreachable("Invalid index");
}

std::string FullyConnectedNode::getDebugDesc() const {
  DescriptionBuilder db(getKindName());
  db.addParam("name", getName());
  if (hasPredicate()) db.addParam("Predicate", "Yes");
  db
    .addParam("Input", *(getInput().getType()))
    .addParam("Weights", *(getWeights().getType()))
    .addParam("Bias", *(getBias().getType()))
    .addParam("users", getNumUsers());
  db.addParam("Result", *(getResult().getType()));
  return db;
}

void FullyConnectedNode::visit(Node *parent, NodeWalker *visitor) {
  if (!visitor->shouldVisit(parent, this)) { return; }
  visitor->pre(parent, this);
if (hasPredicate())
 getPredicate().getNode()->visit(this, visitor);
  getInput().getNode()->visit(this, visitor);
  getWeights().getNode()->visit(this, visitor);
  getBias().getNode()->visit(this, visitor);
  visitor->post(parent, this);
}

bool FullyConnectedNode::isEqual(const FullyConnectedNode &other) const {
  return true &&
      Input_ == other.Input_ &&
      Weights_ == other.Weights_ &&
      Bias_ == other.Bias_ &&
      predicate_ == other.predicate_ &&
      getType(0) == other.getType(0);
}

Node* FullyConnectedNode::clone() const {
  return new FullyConnectedNode(getName(), getResult().getType(), getInput(), getWeights(), getBias());
}

llvm::hash_code FullyConnectedNode::getHash() const {
  return llvm::hash_combine(
      Input_,
      Weights_,
      Bias_);
}

FullyConnectedGradNode *FullyConnectedNode::getGrad(GraphGradMapper &builder) {
  auto *x = new FullyConnectedGradNode(getName().str() + "_grad", getInput(), getWeights(), getBias(), getResult(), builder.getGradient(getResult()));
  builder.addGradient(getInput(), x->getGradOfInputNamedInput());
  builder.addGradient(getWeights(), x->getGradOfInputNamedWeights());
  builder.addGradient(getBias(), x->getGradOfInputNamedBias());
  return x;
}

unsigned RowwiseQuantizedFullyConnectedNode::getNumInputs() const {
  return 5;
}

std::string RowwiseQuantizedFullyConnectedNode::getInputName(unsigned idx) const {
  if (idx == 0) { return "Input"; }
  if (idx == 1) { return "Weights"; }
  if (idx == 2) { return "Scales"; }
  if (idx == 3) { return "Offsets"; }
  if (idx == 4) { return "Bias"; }
  idx -= 5;
  llvm_unreachable("Invalid index");
}

NodeValue RowwiseQuantizedFullyConnectedNode::getNthInput(unsigned idx) {
  if (idx == 0) { return Input_; }
  if (idx == 1) { return Weights_; }
  if (idx == 2) { return Scales_; }
  if (idx == 3) { return Offsets_; }
  if (idx == 4) { return Bias_; }
  idx -= 5;
  llvm_unreachable("Invalid index");
}

void RowwiseQuantizedFullyConnectedNode::setNthInput(unsigned idx, NodeValue val) {
  if (idx == 0) { Input_ = val; return; }
  if (idx == 1) { Weights_ = val; return; }
  if (idx == 2) { Scales_ = val; return; }
  if (idx == 3) { Offsets_ = val; return; }
  if (idx == 4) { Bias_ = val; return; }
  idx -= 5;
  llvm_unreachable("Invalid index");
}

llvm::StringRef RowwiseQuantizedFullyConnectedNode::getOutputName(unsigned idx) const {
  if (idx == 0) { return "Result"; }
  llvm_unreachable("Invalid index");
}

std::string RowwiseQuantizedFullyConnectedNode::getDebugDesc() const {
  DescriptionBuilder db(getKindName());
  db.addParam("name", getName());
  if (hasPredicate()) db.addParam("Predicate", "Yes");
  db
    .addParam("Input", *(getInput().getType()))
    .addParam("Weights", *(getWeights().getType()))
    .addParam("Scales", *(getScales().getType()))
    .addParam("Offsets", *(getOffsets().getType()))
    .addParam("Bias", *(getBias().getType()))
    .addParam("users", getNumUsers());
  db.addParam("Result", *(getResult().getType()));
  return db;
}

void RowwiseQuantizedFullyConnectedNode::visit(Node *parent, NodeWalker *visitor) {
  if (!visitor->shouldVisit(parent, this)) { return; }
  visitor->pre(parent, this);
if (hasPredicate())
 getPredicate().getNode()->visit(this, visitor);
  getInput().getNode()->visit(this, visitor);
  getWeights().getNode()->visit(this, visitor);
  getScales().getNode()->visit(this, visitor);
  getOffsets().getNode()->visit(this, visitor);
  getBias().getNode()->visit(this, visitor);
  visitor->post(parent, this);
}

bool RowwiseQuantizedFullyConnectedNode::isEqual(const RowwiseQuantizedFullyConnectedNode &other) const {
  return true &&
      Input_ == other.Input_ &&
      Weights_ == other.Weights_ &&
      Scales_ == other.Scales_ &&
      Offsets_ == other.Offsets_ &&
      Bias_ == other.Bias_ &&
      predicate_ == other.predicate_ &&
      getType(0) == other.getType(0);
}

Node* RowwiseQuantizedFullyConnectedNode::clone() const {
  return new RowwiseQuantizedFullyConnectedNode(getName(), getResult().getType(), getInput(), getWeights(), getScales(), getOffsets(), getBias());
}

llvm::hash_code RowwiseQuantizedFullyConnectedNode::getHash() const {
  return llvm::hash_combine(
      Input_,
      Weights_,
      Scales_,
      Offsets_,
      Bias_);
}

unsigned BatchNormalizationGradNode::getNumInputs() const {
  return 7;
}

std::string BatchNormalizationGradNode::getInputName(unsigned idx) const {
  if (idx == 0) { return "Input"; }
  if (idx == 1) { return "Scale"; }
  if (idx == 2) { return "Bias"; }
  if (idx == 3) { return "Mean"; }
  if (idx == 4) { return "Var"; }
  if (idx == 5) { return "OriginalOutputForResult"; }
  if (idx == 6) { return "GradOfOriginalOutputNamedResult"; }
  idx -= 7;
  llvm_unreachable("Invalid index");
}

NodeValue BatchNormalizationGradNode::getNthInput(unsigned idx) {
  if (idx == 0) { return Input_; }
  if (idx == 1) { return Scale_; }
  if (idx == 2) { return Bias_; }
  if (idx == 3) { return Mean_; }
  if (idx == 4) { return Var_; }
  if (idx == 5) { return OriginalOutputForResult_; }
  if (idx == 6) { return GradOfOriginalOutputNamedResult_; }
  idx -= 7;
  llvm_unreachable("Invalid index");
}

void BatchNormalizationGradNode::setNthInput(unsigned idx, NodeValue val) {
  if (idx == 0) { Input_ = val; return; }
  if (idx == 1) { Scale_ = val; return; }
  if (idx == 2) { Bias_ = val; return; }
  if (idx == 3) { Mean_ = val; return; }
  if (idx == 4) { Var_ = val; return; }
  if (idx == 5) { OriginalOutputForResult_ = val; return; }
  if (idx == 6) { GradOfOriginalOutputNamedResult_ = val; return; }
  idx -= 7;
  llvm_unreachable("Invalid index");
}

llvm::StringRef BatchNormalizationGradNode::getOutputName(unsigned idx) const {
  if (idx == 0) { return "GradOfInputNamedInput"; }
  if (idx == 1) { return "GradOfInputNamedScale"; }
  if (idx == 2) { return "GradOfInputNamedBias"; }
  if (idx == 3) { return "GradOfInputNamedMean"; }
  if (idx == 4) { return "GradOfInputNamedVar"; }
  llvm_unreachable("Invalid index");
}

std::string BatchNormalizationGradNode::getDebugDesc() const {
  DescriptionBuilder db(getKindName());
  db.addParam("name", getName());
  if (hasPredicate()) db.addParam("Predicate", "Yes");
  db
    .addParam("Input", *(getInput().getType()))
    .addParam("Scale", *(getScale().getType()))
    .addParam("Bias", *(getBias().getType()))
    .addParam("Mean", *(getMean().getType()))
    .addParam("Var", *(getVar().getType()))
    .addParam("OriginalOutputForResult", *(getOriginalOutputForResult().getType()))
    .addParam("GradOfOriginalOutputNamedResult", *(getGradOfOriginalOutputNamedResult().getType()))
    .addParam("ChannelIdx", getChannelIdx())
    .addParam("Epsilon", getEpsilon())
    .addParam("Momentum", getMomentum())
    .addParam("users", getNumUsers());
  db.addParam("GradOfInputNamedInput", *(getGradOfInputNamedInput().getType()));
  db.addParam("GradOfInputNamedScale", *(getGradOfInputNamedScale().getType()));
  db.addParam("GradOfInputNamedBias", *(getGradOfInputNamedBias().getType()));
  db.addParam("GradOfInputNamedMean", *(getGradOfInputNamedMean().getType()));
  db.addParam("GradOfInputNamedVar", *(getGradOfInputNamedVar().getType()));
  return db;
}

void BatchNormalizationGradNode::visit(Node *parent, NodeWalker *visitor) {
  if (!visitor->shouldVisit(parent, this)) { return; }
  visitor->pre(parent, this);
if (hasPredicate())
 getPredicate().getNode()->visit(this, visitor);
  getInput().getNode()->visit(this, visitor);
  getScale().getNode()->visit(this, visitor);
  getBias().getNode()->visit(this, visitor);
  getMean().getNode()->visit(this, visitor);
  getVar().getNode()->visit(this, visitor);
  getOriginalOutputForResult().getNode()->visit(this, visitor);
  getGradOfOriginalOutputNamedResult().getNode()->visit(this, visitor);
  visitor->post(parent, this);
}

bool BatchNormalizationGradNode::isEqual(const BatchNormalizationGradNode &other) const {
  return true &&
      Input_ == other.Input_ &&
      Scale_ == other.Scale_ &&
      Bias_ == other.Bias_ &&
      Mean_ == other.Mean_ &&
      Var_ == other.Var_ &&
      OriginalOutputForResult_ == other.OriginalOutputForResult_ &&
      GradOfOriginalOutputNamedResult_ == other.GradOfOriginalOutputNamedResult_ &&
      predicate_ == other.predicate_ &&
      ChannelIdx_ == other.ChannelIdx_ &&
      Epsilon_ == other.Epsilon_ &&
      Momentum_ == other.Momentum_ &&
      getType(0) == other.getType(0) &&
      getType(1) == other.getType(1) &&
      getType(2) == other.getType(2) &&
      getType(3) == other.getType(3) &&
      getType(4) == other.getType(4);
}

Node* BatchNormalizationGradNode::clone() const {
  return new BatchNormalizationGradNode(getName(), getInput(), getScale(), getBias(), getMean(), getVar(), getOriginalOutputForResult(), getGradOfOriginalOutputNamedResult(), getChannelIdx(), getEpsilon(), getMomentum());
}

llvm::hash_code BatchNormalizationGradNode::getHash() const {
  return llvm::hash_combine(
      ChannelIdx_,
      toBinary(Epsilon_),
      toBinary(Momentum_),
      Input_,
      Scale_,
      Bias_,
      Mean_,
      Var_,
      OriginalOutputForResult_,
      GradOfOriginalOutputNamedResult_);
}

unsigned BatchNormalizationNode::getNumInputs() const {
  return 5;
}

std::string BatchNormalizationNode::getInputName(unsigned idx) const {
  if (idx == 0) { return "Input"; }
  if (idx == 1) { return "Scale"; }
  if (idx == 2) { return "Bias"; }
  if (idx == 3) { return "Mean"; }
  if (idx == 4) { return "Var"; }
  idx -= 5;
  llvm_unreachable("Invalid index");
}

NodeValue BatchNormalizationNode::getNthInput(unsigned idx) {
  if (idx == 0) { return Input_; }
  if (idx == 1) { return Scale_; }
  if (idx == 2) { return Bias_; }
  if (idx == 3) { return Mean_; }
  if (idx == 4) { return Var_; }
  idx -= 5;
  llvm_unreachable("Invalid index");
}

void BatchNormalizationNode::setNthInput(unsigned idx, NodeValue val) {
  if (idx == 0) { Input_ = val; return; }
  if (idx == 1) { Scale_ = val; return; }
  if (idx == 2) { Bias_ = val; return; }
  if (idx == 3) { Mean_ = val; return; }
  if (idx == 4) { Var_ = val; return; }
  idx -= 5;
  llvm_unreachable("Invalid index");
}

llvm::StringRef BatchNormalizationNode::getOutputName(unsigned idx) const {
  if (idx == 0) { return "Result"; }
  llvm_unreachable("Invalid index");
}

std::string BatchNormalizationNode::getDebugDesc() const {
  DescriptionBuilder db(getKindName());
  db.addParam("name", getName());
  if (hasPredicate()) db.addParam("Predicate", "Yes");
  db
    .addParam("Input", *(getInput().getType()))
    .addParam("Scale", *(getScale().getType()))
    .addParam("Bias", *(getBias().getType()))
    .addParam("Mean", *(getMean().getType()))
    .addParam("Var", *(getVar().getType()))
    .addParam("ChannelIdx", getChannelIdx())
    .addParam("Epsilon", getEpsilon())
    .addParam("Momentum", getMomentum())
    .addParam("users", getNumUsers());
  db.addParam("Result", *(getResult().getType()));
  return db;
}

void BatchNormalizationNode::visit(Node *parent, NodeWalker *visitor) {
  if (!visitor->shouldVisit(parent, this)) { return; }
  visitor->pre(parent, this);
if (hasPredicate())
 getPredicate().getNode()->visit(this, visitor);
  getInput().getNode()->visit(this, visitor);
  getScale().getNode()->visit(this, visitor);
  getBias().getNode()->visit(this, visitor);
  getMean().getNode()->visit(this, visitor);
  getVar().getNode()->visit(this, visitor);
  visitor->post(parent, this);
}

bool BatchNormalizationNode::isEqual(const BatchNormalizationNode &other) const {
  return true &&
      Input_ == other.Input_ &&
      Scale_ == other.Scale_ &&
      Bias_ == other.Bias_ &&
      Mean_ == other.Mean_ &&
      Var_ == other.Var_ &&
      predicate_ == other.predicate_ &&
      ChannelIdx_ == other.ChannelIdx_ &&
      Epsilon_ == other.Epsilon_ &&
      Momentum_ == other.Momentum_ &&
      getType(0) == other.getType(0);
}

Node* BatchNormalizationNode::clone() const {
  return new BatchNormalizationNode(getName(), getInput(), getScale(), getBias(), getMean(), getVar(), getChannelIdx(), getEpsilon(), getMomentum());
}

llvm::hash_code BatchNormalizationNode::getHash() const {
  return llvm::hash_combine(
      ChannelIdx_,
      toBinary(Epsilon_),
      toBinary(Momentum_),
      Input_,
      Scale_,
      Bias_,
      Mean_,
      Var_);
}

BatchNormalizationGradNode *BatchNormalizationNode::getGrad(GraphGradMapper &builder) {
  auto *x = new BatchNormalizationGradNode(getName().str() + "_grad", getInput(), getScale(), getBias(), getMean(), getVar(), getResult(), builder.getGradient(getResult()), getChannelIdx(), getEpsilon(), getMomentum());
  builder.addGradient(getInput(), x->getGradOfInputNamedInput());
  builder.addGradient(getScale(), x->getGradOfInputNamedScale());
  builder.addGradient(getBias(), x->getGradOfInputNamedBias());
  builder.addGradient(getMean(), x->getGradOfInputNamedMean());
  builder.addGradient(getVar(), x->getGradOfInputNamedVar());
  return x;
}

unsigned MeanVarNormalizationNode::getNumInputs() const {
  return 3;
}

std::string MeanVarNormalizationNode::getInputName(unsigned idx) const {
  if (idx == 0) { return "Input"; }
  if (idx == 1) { return "Mean"; }
  if (idx == 2) { return "Var"; }
  idx -= 3;
  llvm_unreachable("Invalid index");
}

NodeValue MeanVarNormalizationNode::getNthInput(unsigned idx) {
  if (idx == 0) { return Input_; }
  if (idx == 1) { return Mean_; }
  if (idx == 2) { return Var_; }
  idx -= 3;
  llvm_unreachable("Invalid index");
}

void MeanVarNormalizationNode::setNthInput(unsigned idx, NodeValue val) {
  if (idx == 0) { Input_ = val; return; }
  if (idx == 1) { Mean_ = val; return; }
  if (idx == 2) { Var_ = val; return; }
  idx -= 3;
  llvm_unreachable("Invalid index");
}

llvm::StringRef MeanVarNormalizationNode::getOutputName(unsigned idx) const {
  if (idx == 0) { return "NewMean"; }
  if (idx == 1) { return "NewVar"; }
  llvm_unreachable("Invalid index");
}

std::string MeanVarNormalizationNode::getDebugDesc() const {
  DescriptionBuilder db(getKindName());
  db.addParam("name", getName());
  if (hasPredicate()) db.addParam("Predicate", "Yes");
  db
    .addParam("Input", *(getInput().getType()))
    .addParam("Mean", *(getMean().getType()))
    .addParam("Var", *(getVar().getType()))
    .addParam("ChannelIdx", getChannelIdx())
    .addParam("Momentum", getMomentum())
    .addParam("users", getNumUsers());
  db.addParam("NewMean", *(getNewMean().getType()));
  db.addParam("NewVar", *(getNewVar().getType()));
  return db;
}

void MeanVarNormalizationNode::visit(Node *parent, NodeWalker *visitor) {
  if (!visitor->shouldVisit(parent, this)) { return; }
  visitor->pre(parent, this);
if (hasPredicate())
 getPredicate().getNode()->visit(this, visitor);
  getInput().getNode()->visit(this, visitor);
  getMean().getNode()->visit(this, visitor);
  getVar().getNode()->visit(this, visitor);
  visitor->post(parent, this);
}

bool MeanVarNormalizationNode::isEqual(const MeanVarNormalizationNode &other) const {
  return true &&
      Input_ == other.Input_ &&
      Mean_ == other.Mean_ &&
      Var_ == other.Var_ &&
      predicate_ == other.predicate_ &&
      ChannelIdx_ == other.ChannelIdx_ &&
      Momentum_ == other.Momentum_ &&
      getType(0) == other.getType(0) &&
      getType(1) == other.getType(1);
}

Node* MeanVarNormalizationNode::clone() const {
  return new MeanVarNormalizationNode(getName(), getInput(), getMean(), getVar(), getChannelIdx(), getMomentum());
}

llvm::hash_code MeanVarNormalizationNode::getHash() const {
  return llvm::hash_combine(
      ChannelIdx_,
      toBinary(Momentum_),
      Input_,
      Mean_,
      Var_);
}

unsigned LocalResponseNormalizationGradNode::getNumInputs() const {
  return 3;
}

std::string LocalResponseNormalizationGradNode::getInputName(unsigned idx) const {
  if (idx == 0) { return "Input"; }
  if (idx == 1) { return "OriginalOutputForResult"; }
  if (idx == 2) { return "GradOfOriginalOutputNamedResult"; }
  idx -= 3;
  llvm_unreachable("Invalid index");
}

NodeValue LocalResponseNormalizationGradNode::getNthInput(unsigned idx) {
  if (idx == 0) { return Input_; }
  if (idx == 1) { return OriginalOutputForResult_; }
  if (idx == 2) { return GradOfOriginalOutputNamedResult_; }
  idx -= 3;
  llvm_unreachable("Invalid index");
}

void LocalResponseNormalizationGradNode::setNthInput(unsigned idx, NodeValue val) {
  if (idx == 0) { Input_ = val; return; }
  if (idx == 1) { OriginalOutputForResult_ = val; return; }
  if (idx == 2) { GradOfOriginalOutputNamedResult_ = val; return; }
  idx -= 3;
  llvm_unreachable("Invalid index");
}

llvm::StringRef LocalResponseNormalizationGradNode::getOutputName(unsigned idx) const {
  if (idx == 0) { return "GradOfInputNamedInput"; }
  llvm_unreachable("Invalid index");
}

std::string LocalResponseNormalizationGradNode::getDebugDesc() const {
  DescriptionBuilder db(getKindName());
  db.addParam("name", getName());
  if (hasPredicate()) db.addParam("Predicate", "Yes");
  db
    .addParam("Input", *(getInput().getType()))
    .addParam("OriginalOutputForResult", *(getOriginalOutputForResult().getType()))
    .addParam("GradOfOriginalOutputNamedResult", *(getGradOfOriginalOutputNamedResult().getType()))
    .addParam("HalfWindowSize", getHalfWindowSize())
    .addParam("Alpha", getAlpha())
    .addParam("Beta", getBeta())
    .addParam("K", getK())
    .addParam("users", getNumUsers());
  db.addParam("GradOfInputNamedInput", *(getGradOfInputNamedInput().getType()));
  return db;
}

void LocalResponseNormalizationGradNode::visit(Node *parent, NodeWalker *visitor) {
  if (!visitor->shouldVisit(parent, this)) { return; }
  visitor->pre(parent, this);
if (hasPredicate())
 getPredicate().getNode()->visit(this, visitor);
  getInput().getNode()->visit(this, visitor);
  getOriginalOutputForResult().getNode()->visit(this, visitor);
  getGradOfOriginalOutputNamedResult().getNode()->visit(this, visitor);
  visitor->post(parent, this);
}

bool LocalResponseNormalizationGradNode::isEqual(const LocalResponseNormalizationGradNode &other) const {
  return true &&
      Input_ == other.Input_ &&
      OriginalOutputForResult_ == other.OriginalOutputForResult_ &&
      GradOfOriginalOutputNamedResult_ == other.GradOfOriginalOutputNamedResult_ &&
      predicate_ == other.predicate_ &&
      HalfWindowSize_ == other.HalfWindowSize_ &&
      Alpha_ == other.Alpha_ &&
      Beta_ == other.Beta_ &&
      K_ == other.K_ &&
      getType(0) == other.getType(0);
}

Node* LocalResponseNormalizationGradNode::clone() const {
  return new LocalResponseNormalizationGradNode(getName(), getInput(), getOriginalOutputForResult(), getGradOfOriginalOutputNamedResult(), getHalfWindowSize(), getAlpha(), getBeta(), getK());
}

llvm::hash_code LocalResponseNormalizationGradNode::getHash() const {
  return llvm::hash_combine(
      HalfWindowSize_,
      toBinary(Alpha_),
      toBinary(Beta_),
      toBinary(K_),
      Input_,
      OriginalOutputForResult_,
      GradOfOriginalOutputNamedResult_);
}

unsigned LocalResponseNormalizationNode::getNumInputs() const {
  return 1;
}

std::string LocalResponseNormalizationNode::getInputName(unsigned idx) const {
  if (idx == 0) { return "Input"; }
  idx -= 1;
  llvm_unreachable("Invalid index");
}

NodeValue LocalResponseNormalizationNode::getNthInput(unsigned idx) {
  if (idx == 0) { return Input_; }
  idx -= 1;
  llvm_unreachable("Invalid index");
}

void LocalResponseNormalizationNode::setNthInput(unsigned idx, NodeValue val) {
  if (idx == 0) { Input_ = val; return; }
  idx -= 1;
  llvm_unreachable("Invalid index");
}

llvm::StringRef LocalResponseNormalizationNode::getOutputName(unsigned idx) const {
  if (idx == 0) { return "Result"; }
  llvm_unreachable("Invalid index");
}

std::string LocalResponseNormalizationNode::getDebugDesc() const {
  DescriptionBuilder db(getKindName());
  db.addParam("name", getName());
  if (hasPredicate()) db.addParam("Predicate", "Yes");
  db
    .addParam("Input", *(getInput().getType()))
    .addParam("HalfWindowSize", getHalfWindowSize())
    .addParam("Alpha", getAlpha())
    .addParam("Beta", getBeta())
    .addParam("K", getK())
    .addParam("users", getNumUsers());
  db.addParam("Result", *(getResult().getType()));
  return db;
}

void LocalResponseNormalizationNode::visit(Node *parent, NodeWalker *visitor) {
  if (!visitor->shouldVisit(parent, this)) { return; }
  visitor->pre(parent, this);
if (hasPredicate())
 getPredicate().getNode()->visit(this, visitor);
  getInput().getNode()->visit(this, visitor);
  visitor->post(parent, this);
}

bool LocalResponseNormalizationNode::isEqual(const LocalResponseNormalizationNode &other) const {
  return true &&
      Input_ == other.Input_ &&
      predicate_ == other.predicate_ &&
      HalfWindowSize_ == other.HalfWindowSize_ &&
      Alpha_ == other.Alpha_ &&
      Beta_ == other.Beta_ &&
      K_ == other.K_ &&
      getType(0) == other.getType(0);
}

Node* LocalResponseNormalizationNode::clone() const {
  return new LocalResponseNormalizationNode(getName(), getInput(), getHalfWindowSize(), getAlpha(), getBeta(), getK());
}

llvm::hash_code LocalResponseNormalizationNode::getHash() const {
  return llvm::hash_combine(
      HalfWindowSize_,
      toBinary(Alpha_),
      toBinary(Beta_),
      toBinary(K_),
      Input_);
}

LocalResponseNormalizationGradNode *LocalResponseNormalizationNode::getGrad(GraphGradMapper &builder) {
  auto *x = new LocalResponseNormalizationGradNode(getName().str() + "_grad", getInput(), getResult(), builder.getGradient(getResult()), getHalfWindowSize(), getAlpha(), getBeta(), getK());
  builder.addGradient(getInput(), x->getGradOfInputNamedInput());
  return x;
}

unsigned LayerNormalizationNode::getNumInputs() const {
  return 3;
}

std::string LayerNormalizationNode::getInputName(unsigned idx) const {
  if (idx == 0) { return "Input"; }
  if (idx == 1) { return "Scale"; }
  if (idx == 2) { return "Bias"; }
  idx -= 3;
  llvm_unreachable("Invalid index");
}

NodeValue LayerNormalizationNode::getNthInput(unsigned idx) {
  if (idx == 0) { return Input_; }
  if (idx == 1) { return Scale_; }
  if (idx == 2) { return Bias_; }
  idx -= 3;
  llvm_unreachable("Invalid index");
}

void LayerNormalizationNode::setNthInput(unsigned idx, NodeValue val) {
  if (idx == 0) { Input_ = val; return; }
  if (idx == 1) { Scale_ = val; return; }
  if (idx == 2) { Bias_ = val; return; }
  idx -= 3;
  llvm_unreachable("Invalid index");
}

llvm::StringRef LayerNormalizationNode::getOutputName(unsigned idx) const {
  if (idx == 0) { return "Result"; }
  llvm_unreachable("Invalid index");
}

std::string LayerNormalizationNode::getDebugDesc() const {
  DescriptionBuilder db(getKindName());
  db.addParam("name", getName());
  if (hasPredicate()) db.addParam("Predicate", "Yes");
  db
    .addParam("Input", *(getInput().getType()))
    .addParam("Scale", *(getScale().getType()))
    .addParam("Bias", *(getBias().getType()))
    .addParam("Epsilon", getEpsilon())
    .addParam("users", getNumUsers());
  db.addParam("Result", *(getResult().getType()));
  return db;
}

void LayerNormalizationNode::visit(Node *parent, NodeWalker *visitor) {
  if (!visitor->shouldVisit(parent, this)) { return; }
  visitor->pre(parent, this);
if (hasPredicate())
 getPredicate().getNode()->visit(this, visitor);
  getInput().getNode()->visit(this, visitor);
  getScale().getNode()->visit(this, visitor);
  getBias().getNode()->visit(this, visitor);
  visitor->post(parent, this);
}

bool LayerNormalizationNode::isEqual(const LayerNormalizationNode &other) const {
  return true &&
      Input_ == other.Input_ &&
      Scale_ == other.Scale_ &&
      Bias_ == other.Bias_ &&
      predicate_ == other.predicate_ &&
      Epsilon_ == other.Epsilon_ &&
      getType(0) == other.getType(0);
}

Node* LayerNormalizationNode::clone() const {
  return new LayerNormalizationNode(getName(), getInput(), getScale(), getBias(), getEpsilon());
}

llvm::hash_code LayerNormalizationNode::getHash() const {
  return llvm::hash_combine(
      toBinary(Epsilon_),
      Input_,
      Scale_,
      Bias_);
}

unsigned BatchBoxCoxNode::getNumInputs() const {
  return 3;
}

std::string BatchBoxCoxNode::getInputName(unsigned idx) const {
  if (idx == 0) { return "Input"; }
  if (idx == 1) { return "Lambda1"; }
  if (idx == 2) { return "Lambda2"; }
  idx -= 3;
  llvm_unreachable("Invalid index");
}

NodeValue BatchBoxCoxNode::getNthInput(unsigned idx) {
  if (idx == 0) { return Input_; }
  if (idx == 1) { return Lambda1_; }
  if (idx == 2) { return Lambda2_; }
  idx -= 3;
  llvm_unreachable("Invalid index");
}

void BatchBoxCoxNode::setNthInput(unsigned idx, NodeValue val) {
  if (idx == 0) { Input_ = val; return; }
  if (idx == 1) { Lambda1_ = val; return; }
  if (idx == 2) { Lambda2_ = val; return; }
  idx -= 3;
  llvm_unreachable("Invalid index");
}

llvm::StringRef BatchBoxCoxNode::getOutputName(unsigned idx) const {
  if (idx == 0) { return "Result"; }
  llvm_unreachable("Invalid index");
}

std::string BatchBoxCoxNode::getDebugDesc() const {
  DescriptionBuilder db(getKindName());
  db.addParam("name", getName());
  if (hasPredicate()) db.addParam("Predicate", "Yes");
  db
    .addParam("Input", *(getInput().getType()))
    .addParam("Lambda1", *(getLambda1().getType()))
    .addParam("Lambda2", *(getLambda2().getType()))
    .addParam("Epsilon", getEpsilon())
    .addParam("users", getNumUsers());
  db.addParam("Result", *(getResult().getType()));
  return db;
}

void BatchBoxCoxNode::visit(Node *parent, NodeWalker *visitor) {
  if (!visitor->shouldVisit(parent, this)) { return; }
  visitor->pre(parent, this);
if (hasPredicate())
 getPredicate().getNode()->visit(this, visitor);
  getInput().getNode()->visit(this, visitor);
  getLambda1().getNode()->visit(this, visitor);
  getLambda2().getNode()->visit(this, visitor);
  visitor->post(parent, this);
}

bool BatchBoxCoxNode::isEqual(const BatchBoxCoxNode &other) const {
  return true &&
      Input_ == other.Input_ &&
      Lambda1_ == other.Lambda1_ &&
      Lambda2_ == other.Lambda2_ &&
      predicate_ == other.predicate_ &&
      Epsilon_ == other.Epsilon_ &&
      getType(0) == other.getType(0);
}

Node* BatchBoxCoxNode::clone() const {
  return new BatchBoxCoxNode(getName(), getInput(), getLambda1(), getLambda2(), getEpsilon());
}

llvm::hash_code BatchBoxCoxNode::getHash() const {
  return llvm::hash_combine(
      toBinary(Epsilon_),
      Input_,
      Lambda1_,
      Lambda2_);
}

unsigned BucketizeNode::getNumInputs() const {
  return 1;
}

std::string BucketizeNode::getInputName(unsigned idx) const {
  if (idx == 0) { return "Input"; }
  idx -= 1;
  llvm_unreachable("Invalid index");
}

NodeValue BucketizeNode::getNthInput(unsigned idx) {
  if (idx == 0) { return Input_; }
  idx -= 1;
  llvm_unreachable("Invalid index");
}

void BucketizeNode::setNthInput(unsigned idx, NodeValue val) {
  if (idx == 0) { Input_ = val; return; }
  idx -= 1;
  llvm_unreachable("Invalid index");
}

llvm::StringRef BucketizeNode::getOutputName(unsigned idx) const {
  if (idx == 0) { return "Result"; }
  llvm_unreachable("Invalid index");
}

std::string BucketizeNode::getDebugDesc() const {
  DescriptionBuilder db(getKindName());
  db.addParam("name", getName());
  if (hasPredicate()) db.addParam("Predicate", "Yes");
  db
    .addParam("Input", *(getInput().getType()))
    .addParam("Boundaries", getBoundaries())
    .addParam("users", getNumUsers());
  db.addParam("Result", *(getResult().getType()));
  return db;
}

void BucketizeNode::visit(Node *parent, NodeWalker *visitor) {
  if (!visitor->shouldVisit(parent, this)) { return; }
  visitor->pre(parent, this);
if (hasPredicate())
 getPredicate().getNode()->visit(this, visitor);
  getInput().getNode()->visit(this, visitor);
  visitor->post(parent, this);
}

bool BucketizeNode::isEqual(const BucketizeNode &other) const {
  return true &&
      Input_ == other.Input_ &&
      predicate_ == other.predicate_ &&
      Boundaries_ == other.Boundaries_ &&
      getType(0) == other.getType(0);
}

Node* BucketizeNode::clone() const {
  return new BucketizeNode(getName(), getResult().getType(), getInput(), getBoundaries());
}

llvm::hash_code BucketizeNode::getHash() const {
  return llvm::hash_combine(
      [](const std::vector<float>& floatVec) -> llvm::hash_code {
        std::vector<size_t> sizeVec = toBinary(floatVec);
        return llvm::hash_combine_range(sizeVec.begin(), sizeVec.end());
      }(Boundaries_),
      Input_);
}

unsigned SoftMaxGradNode::getNumInputs() const {
  return 4;
}

std::string SoftMaxGradNode::getInputName(unsigned idx) const {
  if (idx == 0) { return "Input"; }
  if (idx == 1) { return "Selected"; }
  if (idx == 2) { return "OriginalOutputForResult"; }
  if (idx == 3) { return "GradOfOriginalOutputNamedResult"; }
  idx -= 4;
  llvm_unreachable("Invalid index");
}

NodeValue SoftMaxGradNode::getNthInput(unsigned idx) {
  if (idx == 0) { return Input_; }
  if (idx == 1) { return Selected_; }
  if (idx == 2) { return OriginalOutputForResult_; }
  if (idx == 3) { return GradOfOriginalOutputNamedResult_; }
  idx -= 4;
  llvm_unreachable("Invalid index");
}

void SoftMaxGradNode::setNthInput(unsigned idx, NodeValue val) {
  if (idx == 0) { Input_ = val; return; }
  if (idx == 1) { Selected_ = val; return; }
  if (idx == 2) { OriginalOutputForResult_ = val; return; }
  if (idx == 3) { GradOfOriginalOutputNamedResult_ = val; return; }
  idx -= 4;
  llvm_unreachable("Invalid index");
}

llvm::StringRef SoftMaxGradNode::getOutputName(unsigned idx) const {
  if (idx == 0) { return "GradOfInputNamedInput"; }
  if (idx == 1) { return "GradOfInputNamedSelected"; }
  llvm_unreachable("Invalid index");
}

std::string SoftMaxGradNode::getDebugDesc() const {
  DescriptionBuilder db(getKindName());
  db.addParam("name", getName());
  if (hasPredicate()) db.addParam("Predicate", "Yes");
  db
    .addParam("Input", *(getInput().getType()))
    .addParam("Selected", *(getSelected().getType()))
    .addParam("OriginalOutputForResult", *(getOriginalOutputForResult().getType()))
    .addParam("GradOfOriginalOutputNamedResult", *(getGradOfOriginalOutputNamedResult().getType()))
    .addParam("users", getNumUsers());
  db.addParam("GradOfInputNamedInput", *(getGradOfInputNamedInput().getType()));
  db.addParam("GradOfInputNamedSelected", *(getGradOfInputNamedSelected().getType()));
  return db;
}

void SoftMaxGradNode::visit(Node *parent, NodeWalker *visitor) {
  if (!visitor->shouldVisit(parent, this)) { return; }
  visitor->pre(parent, this);
if (hasPredicate())
 getPredicate().getNode()->visit(this, visitor);
  getInput().getNode()->visit(this, visitor);
  getSelected().getNode()->visit(this, visitor);
  getOriginalOutputForResult().getNode()->visit(this, visitor);
  getGradOfOriginalOutputNamedResult().getNode()->visit(this, visitor);
  visitor->post(parent, this);
}

bool SoftMaxGradNode::isEqual(const SoftMaxGradNode &other) const {
  return true &&
      Input_ == other.Input_ &&
      Selected_ == other.Selected_ &&
      OriginalOutputForResult_ == other.OriginalOutputForResult_ &&
      GradOfOriginalOutputNamedResult_ == other.GradOfOriginalOutputNamedResult_ &&
      predicate_ == other.predicate_ &&
      getType(0) == other.getType(0) &&
      getType(1) == other.getType(1);
}

Node* SoftMaxGradNode::clone() const {
  return new SoftMaxGradNode(getName(), getInput(), getSelected(), getOriginalOutputForResult(), getGradOfOriginalOutputNamedResult());
}

llvm::hash_code SoftMaxGradNode::getHash() const {
  return llvm::hash_combine(
      Input_,
      Selected_,
      OriginalOutputForResult_,
      GradOfOriginalOutputNamedResult_);
}

unsigned SoftMaxNode::getNumInputs() const {
  return 2;
}

std::string SoftMaxNode::getInputName(unsigned idx) const {
  if (idx == 0) { return "Input"; }
  if (idx == 1) { return "Selected"; }
  idx -= 2;
  llvm_unreachable("Invalid index");
}

NodeValue SoftMaxNode::getNthInput(unsigned idx) {
  if (idx == 0) { return Input_; }
  if (idx == 1) { return Selected_; }
  idx -= 2;
  llvm_unreachable("Invalid index");
}

void SoftMaxNode::setNthInput(unsigned idx, NodeValue val) {
  if (idx == 0) { Input_ = val; return; }
  if (idx == 1) { Selected_ = val; return; }
  idx -= 2;
  llvm_unreachable("Invalid index");
}

llvm::StringRef SoftMaxNode::getOutputName(unsigned idx) const {
  if (idx == 0) { return "Result"; }
  llvm_unreachable("Invalid index");
}

std::string SoftMaxNode::getDebugDesc() const {
  DescriptionBuilder db(getKindName());
  db.addParam("name", getName());
  if (hasPredicate()) db.addParam("Predicate", "Yes");
  db
    .addParam("Input", *(getInput().getType()))
    .addParam("Selected", *(getSelected().getType()))
    .addParam("users", getNumUsers());
  db.addParam("Result", *(getResult().getType()));
  return db;
}

void SoftMaxNode::visit(Node *parent, NodeWalker *visitor) {
  if (!visitor->shouldVisit(parent, this)) { return; }
  visitor->pre(parent, this);
if (hasPredicate())
 getPredicate().getNode()->visit(this, visitor);
  getInput().getNode()->visit(this, visitor);
  getSelected().getNode()->visit(this, visitor);
  visitor->post(parent, this);
}

bool SoftMaxNode::isEqual(const SoftMaxNode &other) const {
  return true &&
      Input_ == other.Input_ &&
      Selected_ == other.Selected_ &&
      predicate_ == other.predicate_ &&
      getType(0) == other.getType(0);
}

Node* SoftMaxNode::clone() const {
  return new SoftMaxNode(getName(), getResult().getType(), getInput(), getSelected());
}

llvm::hash_code SoftMaxNode::getHash() const {
  return llvm::hash_combine(
      Input_,
      Selected_);
}

SoftMaxGradNode *SoftMaxNode::getGrad(GraphGradMapper &builder) {
  auto *x = new SoftMaxGradNode(getName().str() + "_grad", getInput(), getSelected(), getResult(), builder.getGradient(getResult()));
  builder.addGradient(getInput(), x->getGradOfInputNamedInput());
  builder.addGradient(getSelected(), x->getGradOfInputNamedSelected());
  return x;
}

unsigned CrossEntropyLossGradNode::getNumInputs() const {
  return 4;
}

std::string CrossEntropyLossGradNode::getInputName(unsigned idx) const {
  if (idx == 0) { return "P"; }
  if (idx == 1) { return "Labels"; }
  if (idx == 2) { return "OriginalOutputForCE"; }
  if (idx == 3) { return "GradOfOriginalOutputNamedCE"; }
  idx -= 4;
  llvm_unreachable("Invalid index");
}

NodeValue CrossEntropyLossGradNode::getNthInput(unsigned idx) {
  if (idx == 0) { return P_; }
  if (idx == 1) { return Labels_; }
  if (idx == 2) { return OriginalOutputForCE_; }
  if (idx == 3) { return GradOfOriginalOutputNamedCE_; }
  idx -= 4;
  llvm_unreachable("Invalid index");
}

void CrossEntropyLossGradNode::setNthInput(unsigned idx, NodeValue val) {
  if (idx == 0) { P_ = val; return; }
  if (idx == 1) { Labels_ = val; return; }
  if (idx == 2) { OriginalOutputForCE_ = val; return; }
  if (idx == 3) { GradOfOriginalOutputNamedCE_ = val; return; }
  idx -= 4;
  llvm_unreachable("Invalid index");
}

llvm::StringRef CrossEntropyLossGradNode::getOutputName(unsigned idx) const {
  if (idx == 0) { return "GradOfInputNamedP"; }
  if (idx == 1) { return "GradOfInputNamedLabels"; }
  llvm_unreachable("Invalid index");
}

std::string CrossEntropyLossGradNode::getDebugDesc() const {
  DescriptionBuilder db(getKindName());
  db.addParam("name", getName());
  if (hasPredicate()) db.addParam("Predicate", "Yes");
  db
    .addParam("P", *(getP().getType()))
    .addParam("Labels", *(getLabels().getType()))
    .addParam("OriginalOutputForCE", *(getOriginalOutputForCE().getType()))
    .addParam("GradOfOriginalOutputNamedCE", *(getGradOfOriginalOutputNamedCE().getType()))
    .addParam("users", getNumUsers());
  db.addParam("GradOfInputNamedP", *(getGradOfInputNamedP().getType()));
  db.addParam("GradOfInputNamedLabels", *(getGradOfInputNamedLabels().getType()));
  return db;
}

void CrossEntropyLossGradNode::visit(Node *parent, NodeWalker *visitor) {
  if (!visitor->shouldVisit(parent, this)) { return; }
  visitor->pre(parent, this);
if (hasPredicate())
 getPredicate().getNode()->visit(this, visitor);
  getP().getNode()->visit(this, visitor);
  getLabels().getNode()->visit(this, visitor);
  getOriginalOutputForCE().getNode()->visit(this, visitor);
  getGradOfOriginalOutputNamedCE().getNode()->visit(this, visitor);
  visitor->post(parent, this);
}

bool CrossEntropyLossGradNode::isEqual(const CrossEntropyLossGradNode &other) const {
  return true &&
      P_ == other.P_ &&
      Labels_ == other.Labels_ &&
      OriginalOutputForCE_ == other.OriginalOutputForCE_ &&
      GradOfOriginalOutputNamedCE_ == other.GradOfOriginalOutputNamedCE_ &&
      predicate_ == other.predicate_ &&
      getType(0) == other.getType(0) &&
      getType(1) == other.getType(1);
}

Node* CrossEntropyLossGradNode::clone() const {
  return new CrossEntropyLossGradNode(getName(), getP(), getLabels(), getOriginalOutputForCE(), getGradOfOriginalOutputNamedCE());
}

llvm::hash_code CrossEntropyLossGradNode::getHash() const {
  return llvm::hash_combine(
      P_,
      Labels_,
      OriginalOutputForCE_,
      GradOfOriginalOutputNamedCE_);
}

unsigned CrossEntropyLossNode::getNumInputs() const {
  return 2;
}

std::string CrossEntropyLossNode::getInputName(unsigned idx) const {
  if (idx == 0) { return "P"; }
  if (idx == 1) { return "Labels"; }
  idx -= 2;
  llvm_unreachable("Invalid index");
}

NodeValue CrossEntropyLossNode::getNthInput(unsigned idx) {
  if (idx == 0) { return P_; }
  if (idx == 1) { return Labels_; }
  idx -= 2;
  llvm_unreachable("Invalid index");
}

void CrossEntropyLossNode::setNthInput(unsigned idx, NodeValue val) {
  if (idx == 0) { P_ = val; return; }
  if (idx == 1) { Labels_ = val; return; }
  idx -= 2;
  llvm_unreachable("Invalid index");
}

llvm::StringRef CrossEntropyLossNode::getOutputName(unsigned idx) const {
  if (idx == 0) { return "CE"; }
  llvm_unreachable("Invalid index");
}

std::string CrossEntropyLossNode::getDebugDesc() const {
  DescriptionBuilder db(getKindName());
  db.addParam("name", getName());
  if (hasPredicate()) db.addParam("Predicate", "Yes");
  db
    .addParam("P", *(getP().getType()))
    .addParam("Labels", *(getLabels().getType()))
    .addParam("users", getNumUsers());
  db.addParam("CE", *(getCE().getType()));
  return db;
}

void CrossEntropyLossNode::visit(Node *parent, NodeWalker *visitor) {
  if (!visitor->shouldVisit(parent, this)) { return; }
  visitor->pre(parent, this);
if (hasPredicate())
 getPredicate().getNode()->visit(this, visitor);
  getP().getNode()->visit(this, visitor);
  getLabels().getNode()->visit(this, visitor);
  visitor->post(parent, this);
}

bool CrossEntropyLossNode::isEqual(const CrossEntropyLossNode &other) const {
  return true &&
      P_ == other.P_ &&
      Labels_ == other.Labels_ &&
      predicate_ == other.predicate_ &&
      getType(0) == other.getType(0);
}

Node* CrossEntropyLossNode::clone() const {
  return new CrossEntropyLossNode(getName(), getCE().getType(), getP(), getLabels());
}

llvm::hash_code CrossEntropyLossNode::getHash() const {
  return llvm::hash_combine(
      P_,
      Labels_);
}

CrossEntropyLossGradNode *CrossEntropyLossNode::getGrad(GraphGradMapper &builder) {
  auto *x = new CrossEntropyLossGradNode(getName().str() + "_grad", getP(), getLabels(), getCE(), builder.getGradient(getCE()));
  builder.addGradient(getP(), x->getGradOfInputNamedP());
  builder.addGradient(getLabels(), x->getGradOfInputNamedLabels());
  return x;
}

unsigned RegressionGradNode::getNumInputs() const {
  return 4;
}

std::string RegressionGradNode::getInputName(unsigned idx) const {
  if (idx == 0) { return "Input"; }
  if (idx == 1) { return "Expected"; }
  if (idx == 2) { return "OriginalOutputForResult"; }
  if (idx == 3) { return "GradOfOriginalOutputNamedResult"; }
  idx -= 4;
  llvm_unreachable("Invalid index");
}

NodeValue RegressionGradNode::getNthInput(unsigned idx) {
  if (idx == 0) { return Input_; }
  if (idx == 1) { return Expected_; }
  if (idx == 2) { return OriginalOutputForResult_; }
  if (idx == 3) { return GradOfOriginalOutputNamedResult_; }
  idx -= 4;
  llvm_unreachable("Invalid index");
}

void RegressionGradNode::setNthInput(unsigned idx, NodeValue val) {
  if (idx == 0) { Input_ = val; return; }
  if (idx == 1) { Expected_ = val; return; }
  if (idx == 2) { OriginalOutputForResult_ = val; return; }
  if (idx == 3) { GradOfOriginalOutputNamedResult_ = val; return; }
  idx -= 4;
  llvm_unreachable("Invalid index");
}

llvm::StringRef RegressionGradNode::getOutputName(unsigned idx) const {
  if (idx == 0) { return "GradOfInputNamedInput"; }
  if (idx == 1) { return "GradOfInputNamedExpected"; }
  llvm_unreachable("Invalid index");
}

std::string RegressionGradNode::getDebugDesc() const {
  DescriptionBuilder db(getKindName());
  db.addParam("name", getName());
  if (hasPredicate()) db.addParam("Predicate", "Yes");
  db
    .addParam("Input", *(getInput().getType()))
    .addParam("Expected", *(getExpected().getType()))
    .addParam("OriginalOutputForResult", *(getOriginalOutputForResult().getType()))
    .addParam("GradOfOriginalOutputNamedResult", *(getGradOfOriginalOutputNamedResult().getType()))
    .addParam("users", getNumUsers());
  db.addParam("GradOfInputNamedInput", *(getGradOfInputNamedInput().getType()));
  db.addParam("GradOfInputNamedExpected", *(getGradOfInputNamedExpected().getType()));
  return db;
}

void RegressionGradNode::visit(Node *parent, NodeWalker *visitor) {
  if (!visitor->shouldVisit(parent, this)) { return; }
  visitor->pre(parent, this);
if (hasPredicate())
 getPredicate().getNode()->visit(this, visitor);
  getInput().getNode()->visit(this, visitor);
  getExpected().getNode()->visit(this, visitor);
  getOriginalOutputForResult().getNode()->visit(this, visitor);
  getGradOfOriginalOutputNamedResult().getNode()->visit(this, visitor);
  visitor->post(parent, this);
}

bool RegressionGradNode::isEqual(const RegressionGradNode &other) const {
  return true &&
      Input_ == other.Input_ &&
      Expected_ == other.Expected_ &&
      OriginalOutputForResult_ == other.OriginalOutputForResult_ &&
      GradOfOriginalOutputNamedResult_ == other.GradOfOriginalOutputNamedResult_ &&
      predicate_ == other.predicate_ &&
      getType(0) == other.getType(0) &&
      getType(1) == other.getType(1);
}

Node* RegressionGradNode::clone() const {
  return new RegressionGradNode(getName(), getInput(), getExpected(), getOriginalOutputForResult(), getGradOfOriginalOutputNamedResult());
}

llvm::hash_code RegressionGradNode::getHash() const {
  return llvm::hash_combine(
      Input_,
      Expected_,
      OriginalOutputForResult_,
      GradOfOriginalOutputNamedResult_);
}

unsigned RegressionNode::getNumInputs() const {
  return 2;
}

std::string RegressionNode::getInputName(unsigned idx) const {
  if (idx == 0) { return "Input"; }
  if (idx == 1) { return "Expected"; }
  idx -= 2;
  llvm_unreachable("Invalid index");
}

NodeValue RegressionNode::getNthInput(unsigned idx) {
  if (idx == 0) { return Input_; }
  if (idx == 1) { return Expected_; }
  idx -= 2;
  llvm_unreachable("Invalid index");
}

void RegressionNode::setNthInput(unsigned idx, NodeValue val) {
  if (idx == 0) { Input_ = val; return; }
  if (idx == 1) { Expected_ = val; return; }
  idx -= 2;
  llvm_unreachable("Invalid index");
}

llvm::StringRef RegressionNode::getOutputName(unsigned idx) const {
  if (idx == 0) { return "Result"; }
  llvm_unreachable("Invalid index");
}

std::string RegressionNode::getDebugDesc() const {
  DescriptionBuilder db(getKindName());
  db.addParam("name", getName());
  if (hasPredicate()) db.addParam("Predicate", "Yes");
  db
    .addParam("Input", *(getInput().getType()))
    .addParam("Expected", *(getExpected().getType()))
    .addParam("users", getNumUsers());
  db.addParam("Result", *(getResult().getType()));
  return db;
}

void RegressionNode::visit(Node *parent, NodeWalker *visitor) {
  if (!visitor->shouldVisit(parent, this)) { return; }
  visitor->pre(parent, this);
if (hasPredicate())
 getPredicate().getNode()->visit(this, visitor);
  getInput().getNode()->visit(this, visitor);
  getExpected().getNode()->visit(this, visitor);
  visitor->post(parent, this);
}

bool RegressionNode::isEqual(const RegressionNode &other) const {
  return true &&
      Input_ == other.Input_ &&
      Expected_ == other.Expected_ &&
      predicate_ == other.predicate_ &&
      getType(0) == other.getType(0);
}

Node* RegressionNode::clone() const {
  return new RegressionNode(getName(), getInput(), getExpected());
}

llvm::hash_code RegressionNode::getHash() const {
  return llvm::hash_combine(
      Input_,
      Expected_);
}

RegressionGradNode *RegressionNode::getGrad(GraphGradMapper &builder) {
  auto *x = new RegressionGradNode(getName().str() + "_grad", getInput(), getExpected(), getResult(), builder.getGradient(getResult()));
  builder.addGradient(getInput(), x->getGradOfInputNamedInput());
  builder.addGradient(getExpected(), x->getGradOfInputNamedExpected());
  return x;
}

unsigned SigmoidCrossEntropyWithLogitsNode::getNumInputs() const {
  return 2;
}

std::string SigmoidCrossEntropyWithLogitsNode::getInputName(unsigned idx) const {
  if (idx == 0) { return "Logits"; }
  if (idx == 1) { return "Targets"; }
  idx -= 2;
  llvm_unreachable("Invalid index");
}

NodeValue SigmoidCrossEntropyWithLogitsNode::getNthInput(unsigned idx) {
  if (idx == 0) { return Logits_; }
  if (idx == 1) { return Targets_; }
  idx -= 2;
  llvm_unreachable("Invalid index");
}

void SigmoidCrossEntropyWithLogitsNode::setNthInput(unsigned idx, NodeValue val) {
  if (idx == 0) { Logits_ = val; return; }
  if (idx == 1) { Targets_ = val; return; }
  idx -= 2;
  llvm_unreachable("Invalid index");
}

llvm::StringRef SigmoidCrossEntropyWithLogitsNode::getOutputName(unsigned idx) const {
  if (idx == 0) { return "Result"; }
  llvm_unreachable("Invalid index");
}

std::string SigmoidCrossEntropyWithLogitsNode::getDebugDesc() const {
  DescriptionBuilder db(getKindName());
  db.addParam("name", getName());
  if (hasPredicate()) db.addParam("Predicate", "Yes");
  db
    .addParam("Logits", *(getLogits().getType()))
    .addParam("Targets", *(getTargets().getType()))
    .addParam("users", getNumUsers());
  db.addParam("Result", *(getResult().getType()));
  return db;
}

void SigmoidCrossEntropyWithLogitsNode::visit(Node *parent, NodeWalker *visitor) {
  if (!visitor->shouldVisit(parent, this)) { return; }
  visitor->pre(parent, this);
if (hasPredicate())
 getPredicate().getNode()->visit(this, visitor);
  getLogits().getNode()->visit(this, visitor);
  getTargets().getNode()->visit(this, visitor);
  visitor->post(parent, this);
}

bool SigmoidCrossEntropyWithLogitsNode::isEqual(const SigmoidCrossEntropyWithLogitsNode &other) const {
  return true &&
      Logits_ == other.Logits_ &&
      Targets_ == other.Targets_ &&
      predicate_ == other.predicate_ &&
      getType(0) == other.getType(0);
}

Node* SigmoidCrossEntropyWithLogitsNode::clone() const {
  return new SigmoidCrossEntropyWithLogitsNode(getName(), getResult().getType(), getLogits(), getTargets());
}

llvm::hash_code SigmoidCrossEntropyWithLogitsNode::getHash() const {
  return llvm::hash_combine(
      Logits_,
      Targets_);
}

unsigned AddGradNode::getNumInputs() const {
  return 4;
}

std::string AddGradNode::getInputName(unsigned idx) const {
  if (idx == 0) { return "LHS"; }
  if (idx == 1) { return "RHS"; }
  if (idx == 2) { return "OriginalOutputForResult"; }
  if (idx == 3) { return "GradOfOriginalOutputNamedResult"; }
  idx -= 4;
  llvm_unreachable("Invalid index");
}

NodeValue AddGradNode::getNthInput(unsigned idx) {
  if (idx == 0) { return LHS_; }
  if (idx == 1) { return RHS_; }
  if (idx == 2) { return OriginalOutputForResult_; }
  if (idx == 3) { return GradOfOriginalOutputNamedResult_; }
  idx -= 4;
  llvm_unreachable("Invalid index");
}

void AddGradNode::setNthInput(unsigned idx, NodeValue val) {
  if (idx == 0) { LHS_ = val; return; }
  if (idx == 1) { RHS_ = val; return; }
  if (idx == 2) { OriginalOutputForResult_ = val; return; }
  if (idx == 3) { GradOfOriginalOutputNamedResult_ = val; return; }
  idx -= 4;
  llvm_unreachable("Invalid index");
}

llvm::StringRef AddGradNode::getOutputName(unsigned idx) const {
  if (idx == 0) { return "GradOfInputNamedLHS"; }
  if (idx == 1) { return "GradOfInputNamedRHS"; }
  llvm_unreachable("Invalid index");
}

std::string AddGradNode::getDebugDesc() const {
  DescriptionBuilder db(getKindName());
  db.addParam("name", getName());
  if (hasPredicate()) db.addParam("Predicate", "Yes");
  db
    .addParam("LHS", *(getLHS().getType()))
    .addParam("RHS", *(getRHS().getType()))
    .addParam("OriginalOutputForResult", *(getOriginalOutputForResult().getType()))
    .addParam("GradOfOriginalOutputNamedResult", *(getGradOfOriginalOutputNamedResult().getType()))
    .addParam("users", getNumUsers());
  db.addParam("GradOfInputNamedLHS", *(getGradOfInputNamedLHS().getType()));
  db.addParam("GradOfInputNamedRHS", *(getGradOfInputNamedRHS().getType()));
  return db;
}

void AddGradNode::visit(Node *parent, NodeWalker *visitor) {
  if (!visitor->shouldVisit(parent, this)) { return; }
  visitor->pre(parent, this);
if (hasPredicate())
 getPredicate().getNode()->visit(this, visitor);
  getLHS().getNode()->visit(this, visitor);
  getRHS().getNode()->visit(this, visitor);
  getOriginalOutputForResult().getNode()->visit(this, visitor);
  getGradOfOriginalOutputNamedResult().getNode()->visit(this, visitor);
  visitor->post(parent, this);
}

bool AddGradNode::isEqual(const AddGradNode &other) const {
  return true &&
      LHS_ == other.LHS_ &&
      RHS_ == other.RHS_ &&
      OriginalOutputForResult_ == other.OriginalOutputForResult_ &&
      GradOfOriginalOutputNamedResult_ == other.GradOfOriginalOutputNamedResult_ &&
      predicate_ == other.predicate_ &&
      getType(0) == other.getType(0) &&
      getType(1) == other.getType(1);
}

Node* AddGradNode::clone() const {
  return new AddGradNode(getName(), getLHS(), getRHS(), getOriginalOutputForResult(), getGradOfOriginalOutputNamedResult());
}

llvm::hash_code AddGradNode::getHash() const {
  return llvm::hash_combine(
      LHS_,
      RHS_,
      OriginalOutputForResult_,
      GradOfOriginalOutputNamedResult_);
}

unsigned AddNode::getNumInputs() const {
  return 2;
}

std::string AddNode::getInputName(unsigned idx) const {
  if (idx == 0) { return "LHS"; }
  if (idx == 1) { return "RHS"; }
  idx -= 2;
  llvm_unreachable("Invalid index");
}

NodeValue AddNode::getNthInput(unsigned idx) {
  if (idx == 0) { return LHS_; }
  if (idx == 1) { return RHS_; }
  idx -= 2;
  llvm_unreachable("Invalid index");
}

void AddNode::setNthInput(unsigned idx, NodeValue val) {
  if (idx == 0) { LHS_ = val; return; }
  if (idx == 1) { RHS_ = val; return; }
  idx -= 2;
  llvm_unreachable("Invalid index");
}

llvm::StringRef AddNode::getOutputName(unsigned idx) const {
  if (idx == 0) { return "Result"; }
  llvm_unreachable("Invalid index");
}

std::string AddNode::getDebugDesc() const {
  DescriptionBuilder db(getKindName());
  db.addParam("name", getName());
  if (hasPredicate()) db.addParam("Predicate", "Yes");
  db
    .addParam("LHS", *(getLHS().getType()))
    .addParam("RHS", *(getRHS().getType()))
    .addParam("users", getNumUsers());
  db.addParam("Result", *(getResult().getType()));
  return db;
}

void AddNode::visit(Node *parent, NodeWalker *visitor) {
  if (!visitor->shouldVisit(parent, this)) { return; }
  visitor->pre(parent, this);
if (hasPredicate())
 getPredicate().getNode()->visit(this, visitor);
  getLHS().getNode()->visit(this, visitor);
  getRHS().getNode()->visit(this, visitor);
  visitor->post(parent, this);
}

bool AddNode::isEqual(const AddNode &other) const {
  return true &&
      LHS_ == other.LHS_ &&
      RHS_ == other.RHS_ &&
      predicate_ == other.predicate_ &&
      getType(0) == other.getType(0);
}

Node* AddNode::clone() const {
  return new AddNode(getName(), getResult().getType(), getLHS(), getRHS());
}

llvm::hash_code AddNode::getHash() const {
  return llvm::hash_combine(
      LHS_,
      RHS_);
}

AddGradNode *AddNode::getGrad(GraphGradMapper &builder) {
  auto *x = new AddGradNode(getName().str() + "_grad", getLHS(), getRHS(), getResult(), builder.getGradient(getResult()));
  builder.addGradient(getLHS(), x->getGradOfInputNamedLHS());
  builder.addGradient(getRHS(), x->getGradOfInputNamedRHS());
  return x;
}

unsigned MulGradNode::getNumInputs() const {
  return 4;
}

std::string MulGradNode::getInputName(unsigned idx) const {
  if (idx == 0) { return "LHS"; }
  if (idx == 1) { return "RHS"; }
  if (idx == 2) { return "OriginalOutputForResult"; }
  if (idx == 3) { return "GradOfOriginalOutputNamedResult"; }
  idx -= 4;
  llvm_unreachable("Invalid index");
}

NodeValue MulGradNode::getNthInput(unsigned idx) {
  if (idx == 0) { return LHS_; }
  if (idx == 1) { return RHS_; }
  if (idx == 2) { return OriginalOutputForResult_; }
  if (idx == 3) { return GradOfOriginalOutputNamedResult_; }
  idx -= 4;
  llvm_unreachable("Invalid index");
}

void MulGradNode::setNthInput(unsigned idx, NodeValue val) {
  if (idx == 0) { LHS_ = val; return; }
  if (idx == 1) { RHS_ = val; return; }
  if (idx == 2) { OriginalOutputForResult_ = val; return; }
  if (idx == 3) { GradOfOriginalOutputNamedResult_ = val; return; }
  idx -= 4;
  llvm_unreachable("Invalid index");
}

llvm::StringRef MulGradNode::getOutputName(unsigned idx) const {
  if (idx == 0) { return "GradOfInputNamedLHS"; }
  if (idx == 1) { return "GradOfInputNamedRHS"; }
  llvm_unreachable("Invalid index");
}

std::string MulGradNode::getDebugDesc() const {
  DescriptionBuilder db(getKindName());
  db.addParam("name", getName());
  if (hasPredicate()) db.addParam("Predicate", "Yes");
  db
    .addParam("LHS", *(getLHS().getType()))
    .addParam("RHS", *(getRHS().getType()))
    .addParam("OriginalOutputForResult", *(getOriginalOutputForResult().getType()))
    .addParam("GradOfOriginalOutputNamedResult", *(getGradOfOriginalOutputNamedResult().getType()))
    .addParam("users", getNumUsers());
  db.addParam("GradOfInputNamedLHS", *(getGradOfInputNamedLHS().getType()));
  db.addParam("GradOfInputNamedRHS", *(getGradOfInputNamedRHS().getType()));
  return db;
}

void MulGradNode::visit(Node *parent, NodeWalker *visitor) {
  if (!visitor->shouldVisit(parent, this)) { return; }
  visitor->pre(parent, this);
if (hasPredicate())
 getPredicate().getNode()->visit(this, visitor);
  getLHS().getNode()->visit(this, visitor);
  getRHS().getNode()->visit(this, visitor);
  getOriginalOutputForResult().getNode()->visit(this, visitor);
  getGradOfOriginalOutputNamedResult().getNode()->visit(this, visitor);
  visitor->post(parent, this);
}

bool MulGradNode::isEqual(const MulGradNode &other) const {
  return true &&
      LHS_ == other.LHS_ &&
      RHS_ == other.RHS_ &&
      OriginalOutputForResult_ == other.OriginalOutputForResult_ &&
      GradOfOriginalOutputNamedResult_ == other.GradOfOriginalOutputNamedResult_ &&
      predicate_ == other.predicate_ &&
      getType(0) == other.getType(0) &&
      getType(1) == other.getType(1);
}

Node* MulGradNode::clone() const {
  return new MulGradNode(getName(), getLHS(), getRHS(), getOriginalOutputForResult(), getGradOfOriginalOutputNamedResult());
}

llvm::hash_code MulGradNode::getHash() const {
  return llvm::hash_combine(
      LHS_,
      RHS_,
      OriginalOutputForResult_,
      GradOfOriginalOutputNamedResult_);
}

unsigned MulNode::getNumInputs() const {
  return 2;
}

std::string MulNode::getInputName(unsigned idx) const {
  if (idx == 0) { return "LHS"; }
  if (idx == 1) { return "RHS"; }
  idx -= 2;
  llvm_unreachable("Invalid index");
}

NodeValue MulNode::getNthInput(unsigned idx) {
  if (idx == 0) { return LHS_; }
  if (idx == 1) { return RHS_; }
  idx -= 2;
  llvm_unreachable("Invalid index");
}

void MulNode::setNthInput(unsigned idx, NodeValue val) {
  if (idx == 0) { LHS_ = val; return; }
  if (idx == 1) { RHS_ = val; return; }
  idx -= 2;
  llvm_unreachable("Invalid index");
}

llvm::StringRef MulNode::getOutputName(unsigned idx) const {
  if (idx == 0) { return "Result"; }
  llvm_unreachable("Invalid index");
}

std::string MulNode::getDebugDesc() const {
  DescriptionBuilder db(getKindName());
  db.addParam("name", getName());
  if (hasPredicate()) db.addParam("Predicate", "Yes");
  db
    .addParam("LHS", *(getLHS().getType()))
    .addParam("RHS", *(getRHS().getType()))
    .addParam("users", getNumUsers());
  db.addParam("Result", *(getResult().getType()));
  return db;
}

void MulNode::visit(Node *parent, NodeWalker *visitor) {
  if (!visitor->shouldVisit(parent, this)) { return; }
  visitor->pre(parent, this);
if (hasPredicate())
 getPredicate().getNode()->visit(this, visitor);
  getLHS().getNode()->visit(this, visitor);
  getRHS().getNode()->visit(this, visitor);
  visitor->post(parent, this);
}

bool MulNode::isEqual(const MulNode &other) const {
  return true &&
      LHS_ == other.LHS_ &&
      RHS_ == other.RHS_ &&
      predicate_ == other.predicate_ &&
      getType(0) == other.getType(0);
}

Node* MulNode::clone() const {
  return new MulNode(getName(), getResult().getType(), getLHS(), getRHS());
}

llvm::hash_code MulNode::getHash() const {
  return llvm::hash_combine(
      LHS_,
      RHS_);
}

MulGradNode *MulNode::getGrad(GraphGradMapper &builder) {
  auto *x = new MulGradNode(getName().str() + "_grad", getLHS(), getRHS(), getResult(), builder.getGradient(getResult()));
  builder.addGradient(getLHS(), x->getGradOfInputNamedLHS());
  builder.addGradient(getRHS(), x->getGradOfInputNamedRHS());
  return x;
}

unsigned SubGradNode::getNumInputs() const {
  return 4;
}

std::string SubGradNode::getInputName(unsigned idx) const {
  if (idx == 0) { return "LHS"; }
  if (idx == 1) { return "RHS"; }
  if (idx == 2) { return "OriginalOutputForResult"; }
  if (idx == 3) { return "GradOfOriginalOutputNamedResult"; }
  idx -= 4;
  llvm_unreachable("Invalid index");
}

NodeValue SubGradNode::getNthInput(unsigned idx) {
  if (idx == 0) { return LHS_; }
  if (idx == 1) { return RHS_; }
  if (idx == 2) { return OriginalOutputForResult_; }
  if (idx == 3) { return GradOfOriginalOutputNamedResult_; }
  idx -= 4;
  llvm_unreachable("Invalid index");
}

void SubGradNode::setNthInput(unsigned idx, NodeValue val) {
  if (idx == 0) { LHS_ = val; return; }
  if (idx == 1) { RHS_ = val; return; }
  if (idx == 2) { OriginalOutputForResult_ = val; return; }
  if (idx == 3) { GradOfOriginalOutputNamedResult_ = val; return; }
  idx -= 4;
  llvm_unreachable("Invalid index");
}

llvm::StringRef SubGradNode::getOutputName(unsigned idx) const {
  if (idx == 0) { return "GradOfInputNamedLHS"; }
  if (idx == 1) { return "GradOfInputNamedRHS"; }
  llvm_unreachable("Invalid index");
}

std::string SubGradNode::getDebugDesc() const {
  DescriptionBuilder db(getKindName());
  db.addParam("name", getName());
  if (hasPredicate()) db.addParam("Predicate", "Yes");
  db
    .addParam("LHS", *(getLHS().getType()))
    .addParam("RHS", *(getRHS().getType()))
    .addParam("OriginalOutputForResult", *(getOriginalOutputForResult().getType()))
    .addParam("GradOfOriginalOutputNamedResult", *(getGradOfOriginalOutputNamedResult().getType()))
    .addParam("users", getNumUsers());
  db.addParam("GradOfInputNamedLHS", *(getGradOfInputNamedLHS().getType()));
  db.addParam("GradOfInputNamedRHS", *(getGradOfInputNamedRHS().getType()));
  return db;
}

void SubGradNode::visit(Node *parent, NodeWalker *visitor) {
  if (!visitor->shouldVisit(parent, this)) { return; }
  visitor->pre(parent, this);
if (hasPredicate())
 getPredicate().getNode()->visit(this, visitor);
  getLHS().getNode()->visit(this, visitor);
  getRHS().getNode()->visit(this, visitor);
  getOriginalOutputForResult().getNode()->visit(this, visitor);
  getGradOfOriginalOutputNamedResult().getNode()->visit(this, visitor);
  visitor->post(parent, this);
}

bool SubGradNode::isEqual(const SubGradNode &other) const {
  return true &&
      LHS_ == other.LHS_ &&
      RHS_ == other.RHS_ &&
      OriginalOutputForResult_ == other.OriginalOutputForResult_ &&
      GradOfOriginalOutputNamedResult_ == other.GradOfOriginalOutputNamedResult_ &&
      predicate_ == other.predicate_ &&
      getType(0) == other.getType(0) &&
      getType(1) == other.getType(1);
}

Node* SubGradNode::clone() const {
  return new SubGradNode(getName(), getLHS(), getRHS(), getOriginalOutputForResult(), getGradOfOriginalOutputNamedResult());
}

llvm::hash_code SubGradNode::getHash() const {
  return llvm::hash_combine(
      LHS_,
      RHS_,
      OriginalOutputForResult_,
      GradOfOriginalOutputNamedResult_);
}

unsigned SubNode::getNumInputs() const {
  return 2;
}

std::string SubNode::getInputName(unsigned idx) const {
  if (idx == 0) { return "LHS"; }
  if (idx == 1) { return "RHS"; }
  idx -= 2;
  llvm_unreachable("Invalid index");
}

NodeValue SubNode::getNthInput(unsigned idx) {
  if (idx == 0) { return LHS_; }
  if (idx == 1) { return RHS_; }
  idx -= 2;
  llvm_unreachable("Invalid index");
}

void SubNode::setNthInput(unsigned idx, NodeValue val) {
  if (idx == 0) { LHS_ = val; return; }
  if (idx == 1) { RHS_ = val; return; }
  idx -= 2;
  llvm_unreachable("Invalid index");
}

llvm::StringRef SubNode::getOutputName(unsigned idx) const {
  if (idx == 0) { return "Result"; }
  llvm_unreachable("Invalid index");
}

std::string SubNode::getDebugDesc() const {
  DescriptionBuilder db(getKindName());
  db.addParam("name", getName());
  if (hasPredicate()) db.addParam("Predicate", "Yes");
  db
    .addParam("LHS", *(getLHS().getType()))
    .addParam("RHS", *(getRHS().getType()))
    .addParam("users", getNumUsers());
  db.addParam("Result", *(getResult().getType()));
  return db;
}

void SubNode::visit(Node *parent, NodeWalker *visitor) {
  if (!visitor->shouldVisit(parent, this)) { return; }
  visitor->pre(parent, this);
if (hasPredicate())
 getPredicate().getNode()->visit(this, visitor);
  getLHS().getNode()->visit(this, visitor);
  getRHS().getNode()->visit(this, visitor);
  visitor->post(parent, this);
}

bool SubNode::isEqual(const SubNode &other) const {
  return true &&
      LHS_ == other.LHS_ &&
      RHS_ == other.RHS_ &&
      predicate_ == other.predicate_ &&
      getType(0) == other.getType(0);
}

Node* SubNode::clone() const {
  return new SubNode(getName(), getResult().getType(), getLHS(), getRHS());
}

llvm::hash_code SubNode::getHash() const {
  return llvm::hash_combine(
      LHS_,
      RHS_);
}

SubGradNode *SubNode::getGrad(GraphGradMapper &builder) {
  auto *x = new SubGradNode(getName().str() + "_grad", getLHS(), getRHS(), getResult(), builder.getGradient(getResult()));
  builder.addGradient(getLHS(), x->getGradOfInputNamedLHS());
  builder.addGradient(getRHS(), x->getGradOfInputNamedRHS());
  return x;
}

unsigned DivGradNode::getNumInputs() const {
  return 4;
}

std::string DivGradNode::getInputName(unsigned idx) const {
  if (idx == 0) { return "LHS"; }
  if (idx == 1) { return "RHS"; }
  if (idx == 2) { return "OriginalOutputForResult"; }
  if (idx == 3) { return "GradOfOriginalOutputNamedResult"; }
  idx -= 4;
  llvm_unreachable("Invalid index");
}

NodeValue DivGradNode::getNthInput(unsigned idx) {
  if (idx == 0) { return LHS_; }
  if (idx == 1) { return RHS_; }
  if (idx == 2) { return OriginalOutputForResult_; }
  if (idx == 3) { return GradOfOriginalOutputNamedResult_; }
  idx -= 4;
  llvm_unreachable("Invalid index");
}

void DivGradNode::setNthInput(unsigned idx, NodeValue val) {
  if (idx == 0) { LHS_ = val; return; }
  if (idx == 1) { RHS_ = val; return; }
  if (idx == 2) { OriginalOutputForResult_ = val; return; }
  if (idx == 3) { GradOfOriginalOutputNamedResult_ = val; return; }
  idx -= 4;
  llvm_unreachable("Invalid index");
}

llvm::StringRef DivGradNode::getOutputName(unsigned idx) const {
  if (idx == 0) { return "GradOfInputNamedLHS"; }
  if (idx == 1) { return "GradOfInputNamedRHS"; }
  llvm_unreachable("Invalid index");
}

std::string DivGradNode::getDebugDesc() const {
  DescriptionBuilder db(getKindName());
  db.addParam("name", getName());
  if (hasPredicate()) db.addParam("Predicate", "Yes");
  db
    .addParam("LHS", *(getLHS().getType()))
    .addParam("RHS", *(getRHS().getType()))
    .addParam("OriginalOutputForResult", *(getOriginalOutputForResult().getType()))
    .addParam("GradOfOriginalOutputNamedResult", *(getGradOfOriginalOutputNamedResult().getType()))
    .addParam("users", getNumUsers());
  db.addParam("GradOfInputNamedLHS", *(getGradOfInputNamedLHS().getType()));
  db.addParam("GradOfInputNamedRHS", *(getGradOfInputNamedRHS().getType()));
  return db;
}

void DivGradNode::visit(Node *parent, NodeWalker *visitor) {
  if (!visitor->shouldVisit(parent, this)) { return; }
  visitor->pre(parent, this);
if (hasPredicate())
 getPredicate().getNode()->visit(this, visitor);
  getLHS().getNode()->visit(this, visitor);
  getRHS().getNode()->visit(this, visitor);
  getOriginalOutputForResult().getNode()->visit(this, visitor);
  getGradOfOriginalOutputNamedResult().getNode()->visit(this, visitor);
  visitor->post(parent, this);
}

bool DivGradNode::isEqual(const DivGradNode &other) const {
  return true &&
      LHS_ == other.LHS_ &&
      RHS_ == other.RHS_ &&
      OriginalOutputForResult_ == other.OriginalOutputForResult_ &&
      GradOfOriginalOutputNamedResult_ == other.GradOfOriginalOutputNamedResult_ &&
      predicate_ == other.predicate_ &&
      getType(0) == other.getType(0) &&
      getType(1) == other.getType(1);
}

Node* DivGradNode::clone() const {
  return new DivGradNode(getName(), getLHS(), getRHS(), getOriginalOutputForResult(), getGradOfOriginalOutputNamedResult());
}

llvm::hash_code DivGradNode::getHash() const {
  return llvm::hash_combine(
      LHS_,
      RHS_,
      OriginalOutputForResult_,
      GradOfOriginalOutputNamedResult_);
}

unsigned DivNode::getNumInputs() const {
  return 2;
}

std::string DivNode::getInputName(unsigned idx) const {
  if (idx == 0) { return "LHS"; }
  if (idx == 1) { return "RHS"; }
  idx -= 2;
  llvm_unreachable("Invalid index");
}

NodeValue DivNode::getNthInput(unsigned idx) {
  if (idx == 0) { return LHS_; }
  if (idx == 1) { return RHS_; }
  idx -= 2;
  llvm_unreachable("Invalid index");
}

void DivNode::setNthInput(unsigned idx, NodeValue val) {
  if (idx == 0) { LHS_ = val; return; }
  if (idx == 1) { RHS_ = val; return; }
  idx -= 2;
  llvm_unreachable("Invalid index");
}

llvm::StringRef DivNode::getOutputName(unsigned idx) const {
  if (idx == 0) { return "Result"; }
  llvm_unreachable("Invalid index");
}

std::string DivNode::getDebugDesc() const {
  DescriptionBuilder db(getKindName());
  db.addParam("name", getName());
  if (hasPredicate()) db.addParam("Predicate", "Yes");
  db
    .addParam("LHS", *(getLHS().getType()))
    .addParam("RHS", *(getRHS().getType()))
    .addParam("users", getNumUsers());
  db.addParam("Result", *(getResult().getType()));
  return db;
}

void DivNode::visit(Node *parent, NodeWalker *visitor) {
  if (!visitor->shouldVisit(parent, this)) { return; }
  visitor->pre(parent, this);
if (hasPredicate())
 getPredicate().getNode()->visit(this, visitor);
  getLHS().getNode()->visit(this, visitor);
  getRHS().getNode()->visit(this, visitor);
  visitor->post(parent, this);
}

bool DivNode::isEqual(const DivNode &other) const {
  return true &&
      LHS_ == other.LHS_ &&
      RHS_ == other.RHS_ &&
      predicate_ == other.predicate_ &&
      getType(0) == other.getType(0);
}

Node* DivNode::clone() const {
  return new DivNode(getName(), getResult().getType(), getLHS(), getRHS());
}

llvm::hash_code DivNode::getHash() const {
  return llvm::hash_combine(
      LHS_,
      RHS_);
}

DivGradNode *DivNode::getGrad(GraphGradMapper &builder) {
  auto *x = new DivGradNode(getName().str() + "_grad", getLHS(), getRHS(), getResult(), builder.getGradient(getResult()));
  builder.addGradient(getLHS(), x->getGradOfInputNamedLHS());
  builder.addGradient(getRHS(), x->getGradOfInputNamedRHS());
  return x;
}

unsigned MaxNode::getNumInputs() const {
  return 2;
}

std::string MaxNode::getInputName(unsigned idx) const {
  if (idx == 0) { return "LHS"; }
  if (idx == 1) { return "RHS"; }
  idx -= 2;
  llvm_unreachable("Invalid index");
}

NodeValue MaxNode::getNthInput(unsigned idx) {
  if (idx == 0) { return LHS_; }
  if (idx == 1) { return RHS_; }
  idx -= 2;
  llvm_unreachable("Invalid index");
}

void MaxNode::setNthInput(unsigned idx, NodeValue val) {
  if (idx == 0) { LHS_ = val; return; }
  if (idx == 1) { RHS_ = val; return; }
  idx -= 2;
  llvm_unreachable("Invalid index");
}

llvm::StringRef MaxNode::getOutputName(unsigned idx) const {
  if (idx == 0) { return "Result"; }
  llvm_unreachable("Invalid index");
}

std::string MaxNode::getDebugDesc() const {
  DescriptionBuilder db(getKindName());
  db.addParam("name", getName());
  if (hasPredicate()) db.addParam("Predicate", "Yes");
  db
    .addParam("LHS", *(getLHS().getType()))
    .addParam("RHS", *(getRHS().getType()))
    .addParam("users", getNumUsers());
  db.addParam("Result", *(getResult().getType()));
  return db;
}

void MaxNode::visit(Node *parent, NodeWalker *visitor) {
  if (!visitor->shouldVisit(parent, this)) { return; }
  visitor->pre(parent, this);
if (hasPredicate())
 getPredicate().getNode()->visit(this, visitor);
  getLHS().getNode()->visit(this, visitor);
  getRHS().getNode()->visit(this, visitor);
  visitor->post(parent, this);
}

bool MaxNode::isEqual(const MaxNode &other) const {
  return true &&
      LHS_ == other.LHS_ &&
      RHS_ == other.RHS_ &&
      predicate_ == other.predicate_ &&
      getType(0) == other.getType(0);
}

Node* MaxNode::clone() const {
  return new MaxNode(getName(), getResult().getType(), getLHS(), getRHS());
}

llvm::hash_code MaxNode::getHash() const {
  return llvm::hash_combine(
      LHS_,
      RHS_);
}

unsigned MinNode::getNumInputs() const {
  return 2;
}

std::string MinNode::getInputName(unsigned idx) const {
  if (idx == 0) { return "LHS"; }
  if (idx == 1) { return "RHS"; }
  idx -= 2;
  llvm_unreachable("Invalid index");
}

NodeValue MinNode::getNthInput(unsigned idx) {
  if (idx == 0) { return LHS_; }
  if (idx == 1) { return RHS_; }
  idx -= 2;
  llvm_unreachable("Invalid index");
}

void MinNode::setNthInput(unsigned idx, NodeValue val) {
  if (idx == 0) { LHS_ = val; return; }
  if (idx == 1) { RHS_ = val; return; }
  idx -= 2;
  llvm_unreachable("Invalid index");
}

llvm::StringRef MinNode::getOutputName(unsigned idx) const {
  if (idx == 0) { return "Result"; }
  llvm_unreachable("Invalid index");
}

std::string MinNode::getDebugDesc() const {
  DescriptionBuilder db(getKindName());
  db.addParam("name", getName());
  if (hasPredicate()) db.addParam("Predicate", "Yes");
  db
    .addParam("LHS", *(getLHS().getType()))
    .addParam("RHS", *(getRHS().getType()))
    .addParam("users", getNumUsers());
  db.addParam("Result", *(getResult().getType()));
  return db;
}

void MinNode::visit(Node *parent, NodeWalker *visitor) {
  if (!visitor->shouldVisit(parent, this)) { return; }
  visitor->pre(parent, this);
if (hasPredicate())
 getPredicate().getNode()->visit(this, visitor);
  getLHS().getNode()->visit(this, visitor);
  getRHS().getNode()->visit(this, visitor);
  visitor->post(parent, this);
}

bool MinNode::isEqual(const MinNode &other) const {
  return true &&
      LHS_ == other.LHS_ &&
      RHS_ == other.RHS_ &&
      predicate_ == other.predicate_ &&
      getType(0) == other.getType(0);
}

Node* MinNode::clone() const {
  return new MinNode(getName(), getResult().getType(), getLHS(), getRHS());
}

llvm::hash_code MinNode::getHash() const {
  return llvm::hash_combine(
      LHS_,
      RHS_);
}

unsigned CmpEQNode::getNumInputs() const {
  return 2;
}

std::string CmpEQNode::getInputName(unsigned idx) const {
  if (idx == 0) { return "LHS"; }
  if (idx == 1) { return "RHS"; }
  idx -= 2;
  llvm_unreachable("Invalid index");
}

NodeValue CmpEQNode::getNthInput(unsigned idx) {
  if (idx == 0) { return LHS_; }
  if (idx == 1) { return RHS_; }
  idx -= 2;
  llvm_unreachable("Invalid index");
}

void CmpEQNode::setNthInput(unsigned idx, NodeValue val) {
  if (idx == 0) { LHS_ = val; return; }
  if (idx == 1) { RHS_ = val; return; }
  idx -= 2;
  llvm_unreachable("Invalid index");
}

llvm::StringRef CmpEQNode::getOutputName(unsigned idx) const {
  if (idx == 0) { return "Result"; }
  llvm_unreachable("Invalid index");
}

std::string CmpEQNode::getDebugDesc() const {
  DescriptionBuilder db(getKindName());
  db.addParam("name", getName());
  if (hasPredicate()) db.addParam("Predicate", "Yes");
  db
    .addParam("LHS", *(getLHS().getType()))
    .addParam("RHS", *(getRHS().getType()))
    .addParam("users", getNumUsers());
  db.addParam("Result", *(getResult().getType()));
  return db;
}

void CmpEQNode::visit(Node *parent, NodeWalker *visitor) {
  if (!visitor->shouldVisit(parent, this)) { return; }
  visitor->pre(parent, this);
if (hasPredicate())
 getPredicate().getNode()->visit(this, visitor);
  getLHS().getNode()->visit(this, visitor);
  getRHS().getNode()->visit(this, visitor);
  visitor->post(parent, this);
}

bool CmpEQNode::isEqual(const CmpEQNode &other) const {
  return true &&
      LHS_ == other.LHS_ &&
      RHS_ == other.RHS_ &&
      predicate_ == other.predicate_ &&
      getType(0) == other.getType(0);
}

Node* CmpEQNode::clone() const {
  return new CmpEQNode(getName(), getResult().getType(), getLHS(), getRHS());
}

llvm::hash_code CmpEQNode::getHash() const {
  return llvm::hash_combine(
      LHS_,
      RHS_);
}

unsigned CmpNEQNode::getNumInputs() const {
  return 2;
}

std::string CmpNEQNode::getInputName(unsigned idx) const {
  if (idx == 0) { return "LHS"; }
  if (idx == 1) { return "RHS"; }
  idx -= 2;
  llvm_unreachable("Invalid index");
}

NodeValue CmpNEQNode::getNthInput(unsigned idx) {
  if (idx == 0) { return LHS_; }
  if (idx == 1) { return RHS_; }
  idx -= 2;
  llvm_unreachable("Invalid index");
}

void CmpNEQNode::setNthInput(unsigned idx, NodeValue val) {
  if (idx == 0) { LHS_ = val; return; }
  if (idx == 1) { RHS_ = val; return; }
  idx -= 2;
  llvm_unreachable("Invalid index");
}

llvm::StringRef CmpNEQNode::getOutputName(unsigned idx) const {
  if (idx == 0) { return "Result"; }
  llvm_unreachable("Invalid index");
}

std::string CmpNEQNode::getDebugDesc() const {
  DescriptionBuilder db(getKindName());
  db.addParam("name", getName());
  if (hasPredicate()) db.addParam("Predicate", "Yes");
  db
    .addParam("LHS", *(getLHS().getType()))
    .addParam("RHS", *(getRHS().getType()))
    .addParam("users", getNumUsers());
  db.addParam("Result", *(getResult().getType()));
  return db;
}

void CmpNEQNode::visit(Node *parent, NodeWalker *visitor) {
  if (!visitor->shouldVisit(parent, this)) { return; }
  visitor->pre(parent, this);
if (hasPredicate())
 getPredicate().getNode()->visit(this, visitor);
  getLHS().getNode()->visit(this, visitor);
  getRHS().getNode()->visit(this, visitor);
  visitor->post(parent, this);
}

bool CmpNEQNode::isEqual(const CmpNEQNode &other) const {
  return true &&
      LHS_ == other.LHS_ &&
      RHS_ == other.RHS_ &&
      predicate_ == other.predicate_ &&
      getType(0) == other.getType(0);
}

Node* CmpNEQNode::clone() const {
  return new CmpNEQNode(getName(), getResult().getType(), getLHS(), getRHS());
}

llvm::hash_code CmpNEQNode::getHash() const {
  return llvm::hash_combine(
      LHS_,
      RHS_);
}

unsigned CmpLTNode::getNumInputs() const {
  return 2;
}

std::string CmpLTNode::getInputName(unsigned idx) const {
  if (idx == 0) { return "LHS"; }
  if (idx == 1) { return "RHS"; }
  idx -= 2;
  llvm_unreachable("Invalid index");
}

NodeValue CmpLTNode::getNthInput(unsigned idx) {
  if (idx == 0) { return LHS_; }
  if (idx == 1) { return RHS_; }
  idx -= 2;
  llvm_unreachable("Invalid index");
}

void CmpLTNode::setNthInput(unsigned idx, NodeValue val) {
  if (idx == 0) { LHS_ = val; return; }
  if (idx == 1) { RHS_ = val; return; }
  idx -= 2;
  llvm_unreachable("Invalid index");
}

llvm::StringRef CmpLTNode::getOutputName(unsigned idx) const {
  if (idx == 0) { return "Result"; }
  llvm_unreachable("Invalid index");
}

std::string CmpLTNode::getDebugDesc() const {
  DescriptionBuilder db(getKindName());
  db.addParam("name", getName());
  if (hasPredicate()) db.addParam("Predicate", "Yes");
  db
    .addParam("LHS", *(getLHS().getType()))
    .addParam("RHS", *(getRHS().getType()))
    .addParam("users", getNumUsers());
  db.addParam("Result", *(getResult().getType()));
  return db;
}

void CmpLTNode::visit(Node *parent, NodeWalker *visitor) {
  if (!visitor->shouldVisit(parent, this)) { return; }
  visitor->pre(parent, this);
if (hasPredicate())
 getPredicate().getNode()->visit(this, visitor);
  getLHS().getNode()->visit(this, visitor);
  getRHS().getNode()->visit(this, visitor);
  visitor->post(parent, this);
}

bool CmpLTNode::isEqual(const CmpLTNode &other) const {
  return true &&
      LHS_ == other.LHS_ &&
      RHS_ == other.RHS_ &&
      predicate_ == other.predicate_ &&
      getType(0) == other.getType(0);
}

Node* CmpLTNode::clone() const {
  return new CmpLTNode(getName(), getResult().getType(), getLHS(), getRHS());
}

llvm::hash_code CmpLTNode::getHash() const {
  return llvm::hash_combine(
      LHS_,
      RHS_);
}

unsigned CmpLTENode::getNumInputs() const {
  return 2;
}

std::string CmpLTENode::getInputName(unsigned idx) const {
  if (idx == 0) { return "LHS"; }
  if (idx == 1) { return "RHS"; }
  idx -= 2;
  llvm_unreachable("Invalid index");
}

NodeValue CmpLTENode::getNthInput(unsigned idx) {
  if (idx == 0) { return LHS_; }
  if (idx == 1) { return RHS_; }
  idx -= 2;
  llvm_unreachable("Invalid index");
}

void CmpLTENode::setNthInput(unsigned idx, NodeValue val) {
  if (idx == 0) { LHS_ = val; return; }
  if (idx == 1) { RHS_ = val; return; }
  idx -= 2;
  llvm_unreachable("Invalid index");
}

llvm::StringRef CmpLTENode::getOutputName(unsigned idx) const {
  if (idx == 0) { return "Result"; }
  llvm_unreachable("Invalid index");
}

std::string CmpLTENode::getDebugDesc() const {
  DescriptionBuilder db(getKindName());
  db.addParam("name", getName());
  if (hasPredicate()) db.addParam("Predicate", "Yes");
  db
    .addParam("LHS", *(getLHS().getType()))
    .addParam("RHS", *(getRHS().getType()))
    .addParam("users", getNumUsers());
  db.addParam("Result", *(getResult().getType()));
  return db;
}

void CmpLTENode::visit(Node *parent, NodeWalker *visitor) {
  if (!visitor->shouldVisit(parent, this)) { return; }
  visitor->pre(parent, this);
if (hasPredicate())
 getPredicate().getNode()->visit(this, visitor);
  getLHS().getNode()->visit(this, visitor);
  getRHS().getNode()->visit(this, visitor);
  visitor->post(parent, this);
}

bool CmpLTENode::isEqual(const CmpLTENode &other) const {
  return true &&
      LHS_ == other.LHS_ &&
      RHS_ == other.RHS_ &&
      predicate_ == other.predicate_ &&
      getType(0) == other.getType(0);
}

Node* CmpLTENode::clone() const {
  return new CmpLTENode(getName(), getResult().getType(), getLHS(), getRHS());
}

llvm::hash_code CmpLTENode::getHash() const {
  return llvm::hash_combine(
      LHS_,
      RHS_);
}

unsigned PowNode::getNumInputs() const {
  return 2;
}

std::string PowNode::getInputName(unsigned idx) const {
  if (idx == 0) { return "LHS"; }
  if (idx == 1) { return "RHS"; }
  idx -= 2;
  llvm_unreachable("Invalid index");
}

NodeValue PowNode::getNthInput(unsigned idx) {
  if (idx == 0) { return LHS_; }
  if (idx == 1) { return RHS_; }
  idx -= 2;
  llvm_unreachable("Invalid index");
}

void PowNode::setNthInput(unsigned idx, NodeValue val) {
  if (idx == 0) { LHS_ = val; return; }
  if (idx == 1) { RHS_ = val; return; }
  idx -= 2;
  llvm_unreachable("Invalid index");
}

llvm::StringRef PowNode::getOutputName(unsigned idx) const {
  if (idx == 0) { return "Result"; }
  llvm_unreachable("Invalid index");
}

std::string PowNode::getDebugDesc() const {
  DescriptionBuilder db(getKindName());
  db.addParam("name", getName());
  if (hasPredicate()) db.addParam("Predicate", "Yes");
  db
    .addParam("LHS", *(getLHS().getType()))
    .addParam("RHS", *(getRHS().getType()))
    .addParam("users", getNumUsers());
  db.addParam("Result", *(getResult().getType()));
  return db;
}

void PowNode::visit(Node *parent, NodeWalker *visitor) {
  if (!visitor->shouldVisit(parent, this)) { return; }
  visitor->pre(parent, this);
if (hasPredicate())
 getPredicate().getNode()->visit(this, visitor);
  getLHS().getNode()->visit(this, visitor);
  getRHS().getNode()->visit(this, visitor);
  visitor->post(parent, this);
}

bool PowNode::isEqual(const PowNode &other) const {
  return true &&
      LHS_ == other.LHS_ &&
      RHS_ == other.RHS_ &&
      predicate_ == other.predicate_ &&
      getType(0) == other.getType(0);
}

Node* PowNode::clone() const {
  return new PowNode(getName(), getResult().getType(), getLHS(), getRHS());
}

llvm::hash_code PowNode::getHash() const {
  return llvm::hash_combine(
      LHS_,
      RHS_);
}

unsigned AndNode::getNumInputs() const {
  return 2;
}

std::string AndNode::getInputName(unsigned idx) const {
  if (idx == 0) { return "LHS"; }
  if (idx == 1) { return "RHS"; }
  idx -= 2;
  llvm_unreachable("Invalid index");
}

NodeValue AndNode::getNthInput(unsigned idx) {
  if (idx == 0) { return LHS_; }
  if (idx == 1) { return RHS_; }
  idx -= 2;
  llvm_unreachable("Invalid index");
}

void AndNode::setNthInput(unsigned idx, NodeValue val) {
  if (idx == 0) { LHS_ = val; return; }
  if (idx == 1) { RHS_ = val; return; }
  idx -= 2;
  llvm_unreachable("Invalid index");
}

llvm::StringRef AndNode::getOutputName(unsigned idx) const {
  if (idx == 0) { return "Result"; }
  llvm_unreachable("Invalid index");
}

std::string AndNode::getDebugDesc() const {
  DescriptionBuilder db(getKindName());
  db.addParam("name", getName());
  if (hasPredicate()) db.addParam("Predicate", "Yes");
  db
    .addParam("LHS", *(getLHS().getType()))
    .addParam("RHS", *(getRHS().getType()))
    .addParam("users", getNumUsers());
  db.addParam("Result", *(getResult().getType()));
  return db;
}

void AndNode::visit(Node *parent, NodeWalker *visitor) {
  if (!visitor->shouldVisit(parent, this)) { return; }
  visitor->pre(parent, this);
if (hasPredicate())
 getPredicate().getNode()->visit(this, visitor);
  getLHS().getNode()->visit(this, visitor);
  getRHS().getNode()->visit(this, visitor);
  visitor->post(parent, this);
}

bool AndNode::isEqual(const AndNode &other) const {
  return true &&
      LHS_ == other.LHS_ &&
      RHS_ == other.RHS_ &&
      predicate_ == other.predicate_ &&
      getType(0) == other.getType(0);
}

Node* AndNode::clone() const {
  return new AndNode(getName(), getResult().getType(), getLHS(), getRHS());
}

llvm::hash_code AndNode::getHash() const {
  return llvm::hash_combine(
      LHS_,
      RHS_);
}

unsigned OrNode::getNumInputs() const {
  return 2;
}

std::string OrNode::getInputName(unsigned idx) const {
  if (idx == 0) { return "LHS"; }
  if (idx == 1) { return "RHS"; }
  idx -= 2;
  llvm_unreachable("Invalid index");
}

NodeValue OrNode::getNthInput(unsigned idx) {
  if (idx == 0) { return LHS_; }
  if (idx == 1) { return RHS_; }
  idx -= 2;
  llvm_unreachable("Invalid index");
}

void OrNode::setNthInput(unsigned idx, NodeValue val) {
  if (idx == 0) { LHS_ = val; return; }
  if (idx == 1) { RHS_ = val; return; }
  idx -= 2;
  llvm_unreachable("Invalid index");
}

llvm::StringRef OrNode::getOutputName(unsigned idx) const {
  if (idx == 0) { return "Result"; }
  llvm_unreachable("Invalid index");
}

std::string OrNode::getDebugDesc() const {
  DescriptionBuilder db(getKindName());
  db.addParam("name", getName());
  if (hasPredicate()) db.addParam("Predicate", "Yes");
  db
    .addParam("LHS", *(getLHS().getType()))
    .addParam("RHS", *(getRHS().getType()))
    .addParam("users", getNumUsers());
  db.addParam("Result", *(getResult().getType()));
  return db;
}

void OrNode::visit(Node *parent, NodeWalker *visitor) {
  if (!visitor->shouldVisit(parent, this)) { return; }
  visitor->pre(parent, this);
if (hasPredicate())
 getPredicate().getNode()->visit(this, visitor);
  getLHS().getNode()->visit(this, visitor);
  getRHS().getNode()->visit(this, visitor);
  visitor->post(parent, this);
}

bool OrNode::isEqual(const OrNode &other) const {
  return true &&
      LHS_ == other.LHS_ &&
      RHS_ == other.RHS_ &&
      predicate_ == other.predicate_ &&
      getType(0) == other.getType(0);
}

Node* OrNode::clone() const {
  return new OrNode(getName(), getResult().getType(), getLHS(), getRHS());
}

llvm::hash_code OrNode::getHash() const {
  return llvm::hash_combine(
      LHS_,
      RHS_);
}

unsigned XorNode::getNumInputs() const {
  return 2;
}

std::string XorNode::getInputName(unsigned idx) const {
  if (idx == 0) { return "LHS"; }
  if (idx == 1) { return "RHS"; }
  idx -= 2;
  llvm_unreachable("Invalid index");
}

NodeValue XorNode::getNthInput(unsigned idx) {
  if (idx == 0) { return LHS_; }
  if (idx == 1) { return RHS_; }
  idx -= 2;
  llvm_unreachable("Invalid index");
}

void XorNode::setNthInput(unsigned idx, NodeValue val) {
  if (idx == 0) { LHS_ = val; return; }
  if (idx == 1) { RHS_ = val; return; }
  idx -= 2;
  llvm_unreachable("Invalid index");
}

llvm::StringRef XorNode::getOutputName(unsigned idx) const {
  if (idx == 0) { return "Result"; }
  llvm_unreachable("Invalid index");
}

std::string XorNode::getDebugDesc() const {
  DescriptionBuilder db(getKindName());
  db.addParam("name", getName());
  if (hasPredicate()) db.addParam("Predicate", "Yes");
  db
    .addParam("LHS", *(getLHS().getType()))
    .addParam("RHS", *(getRHS().getType()))
    .addParam("users", getNumUsers());
  db.addParam("Result", *(getResult().getType()));
  return db;
}

void XorNode::visit(Node *parent, NodeWalker *visitor) {
  if (!visitor->shouldVisit(parent, this)) { return; }
  visitor->pre(parent, this);
if (hasPredicate())
 getPredicate().getNode()->visit(this, visitor);
  getLHS().getNode()->visit(this, visitor);
  getRHS().getNode()->visit(this, visitor);
  visitor->post(parent, this);
}

bool XorNode::isEqual(const XorNode &other) const {
  return true &&
      LHS_ == other.LHS_ &&
      RHS_ == other.RHS_ &&
      predicate_ == other.predicate_ &&
      getType(0) == other.getType(0);
}

Node* XorNode::clone() const {
  return new XorNode(getName(), getResult().getType(), getLHS(), getRHS());
}

llvm::hash_code XorNode::getHash() const {
  return llvm::hash_combine(
      LHS_,
      RHS_);
}

unsigned NotNode::getNumInputs() const {
  return 1;
}

std::string NotNode::getInputName(unsigned idx) const {
  if (idx == 0) { return "Input"; }
  idx -= 1;
  llvm_unreachable("Invalid index");
}

NodeValue NotNode::getNthInput(unsigned idx) {
  if (idx == 0) { return Input_; }
  idx -= 1;
  llvm_unreachable("Invalid index");
}

void NotNode::setNthInput(unsigned idx, NodeValue val) {
  if (idx == 0) { Input_ = val; return; }
  idx -= 1;
  llvm_unreachable("Invalid index");
}

llvm::StringRef NotNode::getOutputName(unsigned idx) const {
  if (idx == 0) { return "Result"; }
  llvm_unreachable("Invalid index");
}

std::string NotNode::getDebugDesc() const {
  DescriptionBuilder db(getKindName());
  db.addParam("name", getName());
  if (hasPredicate()) db.addParam("Predicate", "Yes");
  db
    .addParam("Input", *(getInput().getType()))
    .addParam("users", getNumUsers());
  db.addParam("Result", *(getResult().getType()));
  return db;
}

void NotNode::visit(Node *parent, NodeWalker *visitor) {
  if (!visitor->shouldVisit(parent, this)) { return; }
  visitor->pre(parent, this);
if (hasPredicate())
 getPredicate().getNode()->visit(this, visitor);
  getInput().getNode()->visit(this, visitor);
  visitor->post(parent, this);
}

bool NotNode::isEqual(const NotNode &other) const {
  return true &&
      Input_ == other.Input_ &&
      predicate_ == other.predicate_ &&
      getType(0) == other.getType(0);
}

Node* NotNode::clone() const {
  return new NotNode(getName(), getResult().getType(), getInput());
}

llvm::hash_code NotNode::getHash() const {
  return llvm::hash_combine(
      Input_);
}

unsigned NegNode::getNumInputs() const {
  return 1;
}

std::string NegNode::getInputName(unsigned idx) const {
  if (idx == 0) { return "Input"; }
  idx -= 1;
  llvm_unreachable("Invalid index");
}

NodeValue NegNode::getNthInput(unsigned idx) {
  if (idx == 0) { return Input_; }
  idx -= 1;
  llvm_unreachable("Invalid index");
}

void NegNode::setNthInput(unsigned idx, NodeValue val) {
  if (idx == 0) { Input_ = val; return; }
  idx -= 1;
  llvm_unreachable("Invalid index");
}

llvm::StringRef NegNode::getOutputName(unsigned idx) const {
  if (idx == 0) { return "Result"; }
  llvm_unreachable("Invalid index");
}

std::string NegNode::getDebugDesc() const {
  DescriptionBuilder db(getKindName());
  db.addParam("name", getName());
  if (hasPredicate()) db.addParam("Predicate", "Yes");
  db
    .addParam("Input", *(getInput().getType()))
    .addParam("users", getNumUsers());
  db.addParam("Result", *(getResult().getType()));
  return db;
}

void NegNode::visit(Node *parent, NodeWalker *visitor) {
  if (!visitor->shouldVisit(parent, this)) { return; }
  visitor->pre(parent, this);
if (hasPredicate())
 getPredicate().getNode()->visit(this, visitor);
  getInput().getNode()->visit(this, visitor);
  visitor->post(parent, this);
}

bool NegNode::isEqual(const NegNode &other) const {
  return true &&
      Input_ == other.Input_ &&
      predicate_ == other.predicate_ &&
      getType(0) == other.getType(0);
}

Node* NegNode::clone() const {
  return new NegNode(getName(), getResult().getType(), getInput());
}

llvm::hash_code NegNode::getHash() const {
  return llvm::hash_combine(
      Input_);
}

unsigned AbsNode::getNumInputs() const {
  return 1;
}

std::string AbsNode::getInputName(unsigned idx) const {
  if (idx == 0) { return "Input"; }
  idx -= 1;
  llvm_unreachable("Invalid index");
}

NodeValue AbsNode::getNthInput(unsigned idx) {
  if (idx == 0) { return Input_; }
  idx -= 1;
  llvm_unreachable("Invalid index");
}

void AbsNode::setNthInput(unsigned idx, NodeValue val) {
  if (idx == 0) { Input_ = val; return; }
  idx -= 1;
  llvm_unreachable("Invalid index");
}

llvm::StringRef AbsNode::getOutputName(unsigned idx) const {
  if (idx == 0) { return "Result"; }
  llvm_unreachable("Invalid index");
}

std::string AbsNode::getDebugDesc() const {
  DescriptionBuilder db(getKindName());
  db.addParam("name", getName());
  if (hasPredicate()) db.addParam("Predicate", "Yes");
  db
    .addParam("Input", *(getInput().getType()))
    .addParam("users", getNumUsers());
  db.addParam("Result", *(getResult().getType()));
  return db;
}

void AbsNode::visit(Node *parent, NodeWalker *visitor) {
  if (!visitor->shouldVisit(parent, this)) { return; }
  visitor->pre(parent, this);
if (hasPredicate())
 getPredicate().getNode()->visit(this, visitor);
  getInput().getNode()->visit(this, visitor);
  visitor->post(parent, this);
}

bool AbsNode::isEqual(const AbsNode &other) const {
  return true &&
      Input_ == other.Input_ &&
      predicate_ == other.predicate_ &&
      getType(0) == other.getType(0);
}

Node* AbsNode::clone() const {
  return new AbsNode(getName(), getResult().getType(), getInput());
}

llvm::hash_code AbsNode::getHash() const {
  return llvm::hash_combine(
      Input_);
}

unsigned FloorNode::getNumInputs() const {
  return 1;
}

std::string FloorNode::getInputName(unsigned idx) const {
  if (idx == 0) { return "Input"; }
  idx -= 1;
  llvm_unreachable("Invalid index");
}

NodeValue FloorNode::getNthInput(unsigned idx) {
  if (idx == 0) { return Input_; }
  idx -= 1;
  llvm_unreachable("Invalid index");
}

void FloorNode::setNthInput(unsigned idx, NodeValue val) {
  if (idx == 0) { Input_ = val; return; }
  idx -= 1;
  llvm_unreachable("Invalid index");
}

llvm::StringRef FloorNode::getOutputName(unsigned idx) const {
  if (idx == 0) { return "Result"; }
  llvm_unreachable("Invalid index");
}

std::string FloorNode::getDebugDesc() const {
  DescriptionBuilder db(getKindName());
  db.addParam("name", getName());
  if (hasPredicate()) db.addParam("Predicate", "Yes");
  db
    .addParam("Input", *(getInput().getType()))
    .addParam("users", getNumUsers());
  db.addParam("Result", *(getResult().getType()));
  return db;
}

void FloorNode::visit(Node *parent, NodeWalker *visitor) {
  if (!visitor->shouldVisit(parent, this)) { return; }
  visitor->pre(parent, this);
if (hasPredicate())
 getPredicate().getNode()->visit(this, visitor);
  getInput().getNode()->visit(this, visitor);
  visitor->post(parent, this);
}

bool FloorNode::isEqual(const FloorNode &other) const {
  return true &&
      Input_ == other.Input_ &&
      predicate_ == other.predicate_ &&
      getType(0) == other.getType(0);
}

Node* FloorNode::clone() const {
  return new FloorNode(getName(), getResult().getType(), getInput());
}

llvm::hash_code FloorNode::getHash() const {
  return llvm::hash_combine(
      Input_);
}

unsigned CeilNode::getNumInputs() const {
  return 1;
}

std::string CeilNode::getInputName(unsigned idx) const {
  if (idx == 0) { return "Input"; }
  idx -= 1;
  llvm_unreachable("Invalid index");
}

NodeValue CeilNode::getNthInput(unsigned idx) {
  if (idx == 0) { return Input_; }
  idx -= 1;
  llvm_unreachable("Invalid index");
}

void CeilNode::setNthInput(unsigned idx, NodeValue val) {
  if (idx == 0) { Input_ = val; return; }
  idx -= 1;
  llvm_unreachable("Invalid index");
}

llvm::StringRef CeilNode::getOutputName(unsigned idx) const {
  if (idx == 0) { return "Result"; }
  llvm_unreachable("Invalid index");
}

std::string CeilNode::getDebugDesc() const {
  DescriptionBuilder db(getKindName());
  db.addParam("name", getName());
  if (hasPredicate()) db.addParam("Predicate", "Yes");
  db
    .addParam("Input", *(getInput().getType()))
    .addParam("users", getNumUsers());
  db.addParam("Result", *(getResult().getType()));
  return db;
}

void CeilNode::visit(Node *parent, NodeWalker *visitor) {
  if (!visitor->shouldVisit(parent, this)) { return; }
  visitor->pre(parent, this);
if (hasPredicate())
 getPredicate().getNode()->visit(this, visitor);
  getInput().getNode()->visit(this, visitor);
  visitor->post(parent, this);
}

bool CeilNode::isEqual(const CeilNode &other) const {
  return true &&
      Input_ == other.Input_ &&
      predicate_ == other.predicate_ &&
      getType(0) == other.getType(0);
}

Node* CeilNode::clone() const {
  return new CeilNode(getName(), getResult().getType(), getInput());
}

llvm::hash_code CeilNode::getHash() const {
  return llvm::hash_combine(
      Input_);
}

unsigned RoundNode::getNumInputs() const {
  return 1;
}

std::string RoundNode::getInputName(unsigned idx) const {
  if (idx == 0) { return "Input"; }
  idx -= 1;
  llvm_unreachable("Invalid index");
}

NodeValue RoundNode::getNthInput(unsigned idx) {
  if (idx == 0) { return Input_; }
  idx -= 1;
  llvm_unreachable("Invalid index");
}

void RoundNode::setNthInput(unsigned idx, NodeValue val) {
  if (idx == 0) { Input_ = val; return; }
  idx -= 1;
  llvm_unreachable("Invalid index");
}

llvm::StringRef RoundNode::getOutputName(unsigned idx) const {
  if (idx == 0) { return "Result"; }
  llvm_unreachable("Invalid index");
}

std::string RoundNode::getDebugDesc() const {
  DescriptionBuilder db(getKindName());
  db.addParam("name", getName());
  if (hasPredicate()) db.addParam("Predicate", "Yes");
  db
    .addParam("Input", *(getInput().getType()))
    .addParam("users", getNumUsers());
  db.addParam("Result", *(getResult().getType()));
  return db;
}

void RoundNode::visit(Node *parent, NodeWalker *visitor) {
  if (!visitor->shouldVisit(parent, this)) { return; }
  visitor->pre(parent, this);
if (hasPredicate())
 getPredicate().getNode()->visit(this, visitor);
  getInput().getNode()->visit(this, visitor);
  visitor->post(parent, this);
}

bool RoundNode::isEqual(const RoundNode &other) const {
  return true &&
      Input_ == other.Input_ &&
      predicate_ == other.predicate_ &&
      getType(0) == other.getType(0);
}

Node* RoundNode::clone() const {
  return new RoundNode(getName(), getResult().getType(), getInput());
}

llvm::hash_code RoundNode::getHash() const {
  return llvm::hash_combine(
      Input_);
}

unsigned SqrtNode::getNumInputs() const {
  return 1;
}

std::string SqrtNode::getInputName(unsigned idx) const {
  if (idx == 0) { return "Input"; }
  idx -= 1;
  llvm_unreachable("Invalid index");
}

NodeValue SqrtNode::getNthInput(unsigned idx) {
  if (idx == 0) { return Input_; }
  idx -= 1;
  llvm_unreachable("Invalid index");
}

void SqrtNode::setNthInput(unsigned idx, NodeValue val) {
  if (idx == 0) { Input_ = val; return; }
  idx -= 1;
  llvm_unreachable("Invalid index");
}

llvm::StringRef SqrtNode::getOutputName(unsigned idx) const {
  if (idx == 0) { return "Result"; }
  llvm_unreachable("Invalid index");
}

std::string SqrtNode::getDebugDesc() const {
  DescriptionBuilder db(getKindName());
  db.addParam("name", getName());
  if (hasPredicate()) db.addParam("Predicate", "Yes");
  db
    .addParam("Input", *(getInput().getType()))
    .addParam("users", getNumUsers());
  db.addParam("Result", *(getResult().getType()));
  return db;
}

void SqrtNode::visit(Node *parent, NodeWalker *visitor) {
  if (!visitor->shouldVisit(parent, this)) { return; }
  visitor->pre(parent, this);
if (hasPredicate())
 getPredicate().getNode()->visit(this, visitor);
  getInput().getNode()->visit(this, visitor);
  visitor->post(parent, this);
}

bool SqrtNode::isEqual(const SqrtNode &other) const {
  return true &&
      Input_ == other.Input_ &&
      predicate_ == other.predicate_ &&
      getType(0) == other.getType(0);
}

Node* SqrtNode::clone() const {
  return new SqrtNode(getName(), getResult().getType(), getInput());
}

llvm::hash_code SqrtNode::getHash() const {
  return llvm::hash_combine(
      Input_);
}

unsigned RsqrtNode::getNumInputs() const {
  return 1;
}

std::string RsqrtNode::getInputName(unsigned idx) const {
  if (idx == 0) { return "Input"; }
  idx -= 1;
  llvm_unreachable("Invalid index");
}

NodeValue RsqrtNode::getNthInput(unsigned idx) {
  if (idx == 0) { return Input_; }
  idx -= 1;
  llvm_unreachable("Invalid index");
}

void RsqrtNode::setNthInput(unsigned idx, NodeValue val) {
  if (idx == 0) { Input_ = val; return; }
  idx -= 1;
  llvm_unreachable("Invalid index");
}

llvm::StringRef RsqrtNode::getOutputName(unsigned idx) const {
  if (idx == 0) { return "Result"; }
  llvm_unreachable("Invalid index");
}

std::string RsqrtNode::getDebugDesc() const {
  DescriptionBuilder db(getKindName());
  db.addParam("name", getName());
  if (hasPredicate()) db.addParam("Predicate", "Yes");
  db
    .addParam("Input", *(getInput().getType()))
    .addParam("users", getNumUsers());
  db.addParam("Result", *(getResult().getType()));
  return db;
}

void RsqrtNode::visit(Node *parent, NodeWalker *visitor) {
  if (!visitor->shouldVisit(parent, this)) { return; }
  visitor->pre(parent, this);
if (hasPredicate())
 getPredicate().getNode()->visit(this, visitor);
  getInput().getNode()->visit(this, visitor);
  visitor->post(parent, this);
}

bool RsqrtNode::isEqual(const RsqrtNode &other) const {
  return true &&
      Input_ == other.Input_ &&
      predicate_ == other.predicate_ &&
      getType(0) == other.getType(0);
}

Node* RsqrtNode::clone() const {
  return new RsqrtNode(getName(), getResult().getType(), getInput());
}

llvm::hash_code RsqrtNode::getHash() const {
  return llvm::hash_combine(
      Input_);
}

unsigned ReciprocalNode::getNumInputs() const {
  return 1;
}

std::string ReciprocalNode::getInputName(unsigned idx) const {
  if (idx == 0) { return "Input"; }
  idx -= 1;
  llvm_unreachable("Invalid index");
}

NodeValue ReciprocalNode::getNthInput(unsigned idx) {
  if (idx == 0) { return Input_; }
  idx -= 1;
  llvm_unreachable("Invalid index");
}

void ReciprocalNode::setNthInput(unsigned idx, NodeValue val) {
  if (idx == 0) { Input_ = val; return; }
  idx -= 1;
  llvm_unreachable("Invalid index");
}

llvm::StringRef ReciprocalNode::getOutputName(unsigned idx) const {
  if (idx == 0) { return "Result"; }
  llvm_unreachable("Invalid index");
}

std::string ReciprocalNode::getDebugDesc() const {
  DescriptionBuilder db(getKindName());
  db.addParam("name", getName());
  if (hasPredicate()) db.addParam("Predicate", "Yes");
  db
    .addParam("Input", *(getInput().getType()))
    .addParam("users", getNumUsers());
  db.addParam("Result", *(getResult().getType()));
  return db;
}

void ReciprocalNode::visit(Node *parent, NodeWalker *visitor) {
  if (!visitor->shouldVisit(parent, this)) { return; }
  visitor->pre(parent, this);
if (hasPredicate())
 getPredicate().getNode()->visit(this, visitor);
  getInput().getNode()->visit(this, visitor);
  visitor->post(parent, this);
}

bool ReciprocalNode::isEqual(const ReciprocalNode &other) const {
  return true &&
      Input_ == other.Input_ &&
      predicate_ == other.predicate_ &&
      getType(0) == other.getType(0);
}

Node* ReciprocalNode::clone() const {
  return new ReciprocalNode(getName(), getResult().getType(), getInput());
}

llvm::hash_code ReciprocalNode::getHash() const {
  return llvm::hash_combine(
      Input_);
}

unsigned SinNode::getNumInputs() const {
  return 1;
}

std::string SinNode::getInputName(unsigned idx) const {
  if (idx == 0) { return "Input"; }
  idx -= 1;
  llvm_unreachable("Invalid index");
}

NodeValue SinNode::getNthInput(unsigned idx) {
  if (idx == 0) { return Input_; }
  idx -= 1;
  llvm_unreachable("Invalid index");
}

void SinNode::setNthInput(unsigned idx, NodeValue val) {
  if (idx == 0) { Input_ = val; return; }
  idx -= 1;
  llvm_unreachable("Invalid index");
}

llvm::StringRef SinNode::getOutputName(unsigned idx) const {
  if (idx == 0) { return "Result"; }
  llvm_unreachable("Invalid index");
}

std::string SinNode::getDebugDesc() const {
  DescriptionBuilder db(getKindName());
  db.addParam("name", getName());
  if (hasPredicate()) db.addParam("Predicate", "Yes");
  db
    .addParam("Input", *(getInput().getType()))
    .addParam("users", getNumUsers());
  db.addParam("Result", *(getResult().getType()));
  return db;
}

void SinNode::visit(Node *parent, NodeWalker *visitor) {
  if (!visitor->shouldVisit(parent, this)) { return; }
  visitor->pre(parent, this);
if (hasPredicate())
 getPredicate().getNode()->visit(this, visitor);
  getInput().getNode()->visit(this, visitor);
  visitor->post(parent, this);
}

bool SinNode::isEqual(const SinNode &other) const {
  return true &&
      Input_ == other.Input_ &&
      predicate_ == other.predicate_ &&
      getType(0) == other.getType(0);
}

Node* SinNode::clone() const {
  return new SinNode(getName(), getResult().getType(), getInput());
}

llvm::hash_code SinNode::getHash() const {
  return llvm::hash_combine(
      Input_);
}

unsigned CosNode::getNumInputs() const {
  return 1;
}

std::string CosNode::getInputName(unsigned idx) const {
  if (idx == 0) { return "Input"; }
  idx -= 1;
  llvm_unreachable("Invalid index");
}

NodeValue CosNode::getNthInput(unsigned idx) {
  if (idx == 0) { return Input_; }
  idx -= 1;
  llvm_unreachable("Invalid index");
}

void CosNode::setNthInput(unsigned idx, NodeValue val) {
  if (idx == 0) { Input_ = val; return; }
  idx -= 1;
  llvm_unreachable("Invalid index");
}

llvm::StringRef CosNode::getOutputName(unsigned idx) const {
  if (idx == 0) { return "Result"; }
  llvm_unreachable("Invalid index");
}

std::string CosNode::getDebugDesc() const {
  DescriptionBuilder db(getKindName());
  db.addParam("name", getName());
  if (hasPredicate()) db.addParam("Predicate", "Yes");
  db
    .addParam("Input", *(getInput().getType()))
    .addParam("users", getNumUsers());
  db.addParam("Result", *(getResult().getType()));
  return db;
}

void CosNode::visit(Node *parent, NodeWalker *visitor) {
  if (!visitor->shouldVisit(parent, this)) { return; }
  visitor->pre(parent, this);
if (hasPredicate())
 getPredicate().getNode()->visit(this, visitor);
  getInput().getNode()->visit(this, visitor);
  visitor->post(parent, this);
}

bool CosNode::isEqual(const CosNode &other) const {
  return true &&
      Input_ == other.Input_ &&
      predicate_ == other.predicate_ &&
      getType(0) == other.getType(0);
}

Node* CosNode::clone() const {
  return new CosNode(getName(), getResult().getType(), getInput());
}

llvm::hash_code CosNode::getHash() const {
  return llvm::hash_combine(
      Input_);
}

unsigned LogNode::getNumInputs() const {
  return 1;
}

std::string LogNode::getInputName(unsigned idx) const {
  if (idx == 0) { return "Input"; }
  idx -= 1;
  llvm_unreachable("Invalid index");
}

NodeValue LogNode::getNthInput(unsigned idx) {
  if (idx == 0) { return Input_; }
  idx -= 1;
  llvm_unreachable("Invalid index");
}

void LogNode::setNthInput(unsigned idx, NodeValue val) {
  if (idx == 0) { Input_ = val; return; }
  idx -= 1;
  llvm_unreachable("Invalid index");
}

llvm::StringRef LogNode::getOutputName(unsigned idx) const {
  if (idx == 0) { return "Result"; }
  llvm_unreachable("Invalid index");
}

std::string LogNode::getDebugDesc() const {
  DescriptionBuilder db(getKindName());
  db.addParam("name", getName());
  if (hasPredicate()) db.addParam("Predicate", "Yes");
  db
    .addParam("Input", *(getInput().getType()))
    .addParam("users", getNumUsers());
  db.addParam("Result", *(getResult().getType()));
  return db;
}

void LogNode::visit(Node *parent, NodeWalker *visitor) {
  if (!visitor->shouldVisit(parent, this)) { return; }
  visitor->pre(parent, this);
if (hasPredicate())
 getPredicate().getNode()->visit(this, visitor);
  getInput().getNode()->visit(this, visitor);
  visitor->post(parent, this);
}

bool LogNode::isEqual(const LogNode &other) const {
  return true &&
      Input_ == other.Input_ &&
      predicate_ == other.predicate_ &&
      getType(0) == other.getType(0);
}

Node* LogNode::clone() const {
  return new LogNode(getName(), getResult().getType(), getInput());
}

llvm::hash_code LogNode::getHash() const {
  return llvm::hash_combine(
      Input_);
}

unsigned ExpNode::getNumInputs() const {
  return 1;
}

std::string ExpNode::getInputName(unsigned idx) const {
  if (idx == 0) { return "Input"; }
  idx -= 1;
  llvm_unreachable("Invalid index");
}

NodeValue ExpNode::getNthInput(unsigned idx) {
  if (idx == 0) { return Input_; }
  idx -= 1;
  llvm_unreachable("Invalid index");
}

void ExpNode::setNthInput(unsigned idx, NodeValue val) {
  if (idx == 0) { Input_ = val; return; }
  idx -= 1;
  llvm_unreachable("Invalid index");
}

llvm::StringRef ExpNode::getOutputName(unsigned idx) const {
  if (idx == 0) { return "Result"; }
  llvm_unreachable("Invalid index");
}

std::string ExpNode::getDebugDesc() const {
  DescriptionBuilder db(getKindName());
  db.addParam("name", getName());
  if (hasPredicate()) db.addParam("Predicate", "Yes");
  db
    .addParam("Input", *(getInput().getType()))
    .addParam("users", getNumUsers());
  db.addParam("Result", *(getResult().getType()));
  return db;
}

void ExpNode::visit(Node *parent, NodeWalker *visitor) {
  if (!visitor->shouldVisit(parent, this)) { return; }
  visitor->pre(parent, this);
if (hasPredicate())
 getPredicate().getNode()->visit(this, visitor);
  getInput().getNode()->visit(this, visitor);
  visitor->post(parent, this);
}

bool ExpNode::isEqual(const ExpNode &other) const {
  return true &&
      Input_ == other.Input_ &&
      predicate_ == other.predicate_ &&
      getType(0) == other.getType(0);
}

Node* ExpNode::clone() const {
  return new ExpNode(getName(), getResult().getType(), getInput());
}

llvm::hash_code ExpNode::getHash() const {
  return llvm::hash_combine(
      Input_);
}

unsigned LogitNode::getNumInputs() const {
  return 1;
}

std::string LogitNode::getInputName(unsigned idx) const {
  if (idx == 0) { return "Input"; }
  idx -= 1;
  llvm_unreachable("Invalid index");
}

NodeValue LogitNode::getNthInput(unsigned idx) {
  if (idx == 0) { return Input_; }
  idx -= 1;
  llvm_unreachable("Invalid index");
}

void LogitNode::setNthInput(unsigned idx, NodeValue val) {
  if (idx == 0) { Input_ = val; return; }
  idx -= 1;
  llvm_unreachable("Invalid index");
}

llvm::StringRef LogitNode::getOutputName(unsigned idx) const {
  if (idx == 0) { return "Result"; }
  llvm_unreachable("Invalid index");
}

std::string LogitNode::getDebugDesc() const {
  DescriptionBuilder db(getKindName());
  db.addParam("name", getName());
  if (hasPredicate()) db.addParam("Predicate", "Yes");
  db
    .addParam("Input", *(getInput().getType()))
    .addParam("Epsilon", getEpsilon())
    .addParam("users", getNumUsers());
  db.addParam("Result", *(getResult().getType()));
  return db;
}

void LogitNode::visit(Node *parent, NodeWalker *visitor) {
  if (!visitor->shouldVisit(parent, this)) { return; }
  visitor->pre(parent, this);
if (hasPredicate())
 getPredicate().getNode()->visit(this, visitor);
  getInput().getNode()->visit(this, visitor);
  visitor->post(parent, this);
}

bool LogitNode::isEqual(const LogitNode &other) const {
  return true &&
      Input_ == other.Input_ &&
      predicate_ == other.predicate_ &&
      Epsilon_ == other.Epsilon_ &&
      getType(0) == other.getType(0);
}

Node* LogitNode::clone() const {
  return new LogitNode(getName(), getResult().getType(), getInput(), getEpsilon());
}

llvm::hash_code LogitNode::getHash() const {
  return llvm::hash_combine(
      toBinary(Epsilon_),
      Input_);
}

unsigned SelectNode::getNumInputs() const {
  return 3;
}

std::string SelectNode::getInputName(unsigned idx) const {
  if (idx == 0) { return "Cond"; }
  if (idx == 1) { return "LHS"; }
  if (idx == 2) { return "RHS"; }
  idx -= 3;
  llvm_unreachable("Invalid index");
}

NodeValue SelectNode::getNthInput(unsigned idx) {
  if (idx == 0) { return Cond_; }
  if (idx == 1) { return LHS_; }
  if (idx == 2) { return RHS_; }
  idx -= 3;
  llvm_unreachable("Invalid index");
}

void SelectNode::setNthInput(unsigned idx, NodeValue val) {
  if (idx == 0) { Cond_ = val; return; }
  if (idx == 1) { LHS_ = val; return; }
  if (idx == 2) { RHS_ = val; return; }
  idx -= 3;
  llvm_unreachable("Invalid index");
}

llvm::StringRef SelectNode::getOutputName(unsigned idx) const {
  if (idx == 0) { return "Result"; }
  llvm_unreachable("Invalid index");
}

std::string SelectNode::getDebugDesc() const {
  DescriptionBuilder db(getKindName());
  db.addParam("name", getName());
  if (hasPredicate()) db.addParam("Predicate", "Yes");
  db
    .addParam("Cond", *(getCond().getType()))
    .addParam("LHS", *(getLHS().getType()))
    .addParam("RHS", *(getRHS().getType()))
    .addParam("users", getNumUsers());
  db.addParam("Result", *(getResult().getType()));
  return db;
}

void SelectNode::visit(Node *parent, NodeWalker *visitor) {
  if (!visitor->shouldVisit(parent, this)) { return; }
  visitor->pre(parent, this);
if (hasPredicate())
 getPredicate().getNode()->visit(this, visitor);
  getCond().getNode()->visit(this, visitor);
  getLHS().getNode()->visit(this, visitor);
  getRHS().getNode()->visit(this, visitor);
  visitor->post(parent, this);
}

bool SelectNode::isEqual(const SelectNode &other) const {
  return true &&
      Cond_ == other.Cond_ &&
      LHS_ == other.LHS_ &&
      RHS_ == other.RHS_ &&
      predicate_ == other.predicate_ &&
      getType(0) == other.getType(0);
}

Node* SelectNode::clone() const {
  return new SelectNode(getName(), getResult().getType(), getCond(), getLHS(), getRHS());
}

llvm::hash_code SelectNode::getHash() const {
  return llvm::hash_combine(
      Cond_,
      LHS_,
      RHS_);
}

unsigned BatchedAddNode::getNumInputs() const {
  return 2;
}

std::string BatchedAddNode::getInputName(unsigned idx) const {
  if (idx == 0) { return "Batch"; }
  if (idx == 1) { return "Slice"; }
  idx -= 2;
  llvm_unreachable("Invalid index");
}

NodeValue BatchedAddNode::getNthInput(unsigned idx) {
  if (idx == 0) { return Batch_; }
  if (idx == 1) { return Slice_; }
  idx -= 2;
  llvm_unreachable("Invalid index");
}

void BatchedAddNode::setNthInput(unsigned idx, NodeValue val) {
  if (idx == 0) { Batch_ = val; return; }
  if (idx == 1) { Slice_ = val; return; }
  idx -= 2;
  llvm_unreachable("Invalid index");
}

llvm::StringRef BatchedAddNode::getOutputName(unsigned idx) const {
  if (idx == 0) { return "Result"; }
  llvm_unreachable("Invalid index");
}

std::string BatchedAddNode::getDebugDesc() const {
  DescriptionBuilder db(getKindName());
  db.addParam("name", getName());
  if (hasPredicate()) db.addParam("Predicate", "Yes");
  db
    .addParam("Batch", *(getBatch().getType()))
    .addParam("Slice", *(getSlice().getType()))
    .addParam("users", getNumUsers());
  db.addParam("Result", *(getResult().getType()));
  return db;
}

void BatchedAddNode::visit(Node *parent, NodeWalker *visitor) {
  if (!visitor->shouldVisit(parent, this)) { return; }
  visitor->pre(parent, this);
if (hasPredicate())
 getPredicate().getNode()->visit(this, visitor);
  getBatch().getNode()->visit(this, visitor);
  getSlice().getNode()->visit(this, visitor);
  visitor->post(parent, this);
}

bool BatchedAddNode::isEqual(const BatchedAddNode &other) const {
  return true &&
      Batch_ == other.Batch_ &&
      Slice_ == other.Slice_ &&
      predicate_ == other.predicate_ &&
      getType(0) == other.getType(0);
}

Node* BatchedAddNode::clone() const {
  return new BatchedAddNode(getName(), getResult().getType(), getBatch(), getSlice());
}

llvm::hash_code BatchedAddNode::getHash() const {
  return llvm::hash_combine(
      Batch_,
      Slice_);
}

unsigned MatMulNode::getNumInputs() const {
  return 2;
}

std::string MatMulNode::getInputName(unsigned idx) const {
  if (idx == 0) { return "LHS"; }
  if (idx == 1) { return "RHS"; }
  idx -= 2;
  llvm_unreachable("Invalid index");
}

NodeValue MatMulNode::getNthInput(unsigned idx) {
  if (idx == 0) { return LHS_; }
  if (idx == 1) { return RHS_; }
  idx -= 2;
  llvm_unreachable("Invalid index");
}

void MatMulNode::setNthInput(unsigned idx, NodeValue val) {
  if (idx == 0) { LHS_ = val; return; }
  if (idx == 1) { RHS_ = val; return; }
  idx -= 2;
  llvm_unreachable("Invalid index");
}

llvm::StringRef MatMulNode::getOutputName(unsigned idx) const {
  if (idx == 0) { return "Result"; }
  llvm_unreachable("Invalid index");
}

std::string MatMulNode::getDebugDesc() const {
  DescriptionBuilder db(getKindName());
  db.addParam("name", getName());
  if (hasPredicate()) db.addParam("Predicate", "Yes");
  db
    .addParam("LHS", *(getLHS().getType()))
    .addParam("RHS", *(getRHS().getType()))
    .addParam("users", getNumUsers());
  db.addParam("Result", *(getResult().getType()));
  return db;
}

void MatMulNode::visit(Node *parent, NodeWalker *visitor) {
  if (!visitor->shouldVisit(parent, this)) { return; }
  visitor->pre(parent, this);
if (hasPredicate())
 getPredicate().getNode()->visit(this, visitor);
  getLHS().getNode()->visit(this, visitor);
  getRHS().getNode()->visit(this, visitor);
  visitor->post(parent, this);
}

bool MatMulNode::isEqual(const MatMulNode &other) const {
  return true &&
      LHS_ == other.LHS_ &&
      RHS_ == other.RHS_ &&
      predicate_ == other.predicate_ &&
      getType(0) == other.getType(0);
}

Node* MatMulNode::clone() const {
  return new MatMulNode(getName(), getResult().getType(), getLHS(), getRHS());
}

llvm::hash_code MatMulNode::getHash() const {
  return llvm::hash_combine(
      LHS_,
      RHS_);
}

unsigned BatchMatMulNode::getNumInputs() const {
  return 2;
}

std::string BatchMatMulNode::getInputName(unsigned idx) const {
  if (idx == 0) { return "LHS"; }
  if (idx == 1) { return "RHS"; }
  idx -= 2;
  llvm_unreachable("Invalid index");
}

NodeValue BatchMatMulNode::getNthInput(unsigned idx) {
  if (idx == 0) { return LHS_; }
  if (idx == 1) { return RHS_; }
  idx -= 2;
  llvm_unreachable("Invalid index");
}

void BatchMatMulNode::setNthInput(unsigned idx, NodeValue val) {
  if (idx == 0) { LHS_ = val; return; }
  if (idx == 1) { RHS_ = val; return; }
  idx -= 2;
  llvm_unreachable("Invalid index");
}

llvm::StringRef BatchMatMulNode::getOutputName(unsigned idx) const {
  if (idx == 0) { return "Result"; }
  llvm_unreachable("Invalid index");
}

std::string BatchMatMulNode::getDebugDesc() const {
  DescriptionBuilder db(getKindName());
  db.addParam("name", getName());
  if (hasPredicate()) db.addParam("Predicate", "Yes");
  db
    .addParam("LHS", *(getLHS().getType()))
    .addParam("RHS", *(getRHS().getType()))
    .addParam("users", getNumUsers());
  db.addParam("Result", *(getResult().getType()));
  return db;
}

void BatchMatMulNode::visit(Node *parent, NodeWalker *visitor) {
  if (!visitor->shouldVisit(parent, this)) { return; }
  visitor->pre(parent, this);
if (hasPredicate())
 getPredicate().getNode()->visit(this, visitor);
  getLHS().getNode()->visit(this, visitor);
  getRHS().getNode()->visit(this, visitor);
  visitor->post(parent, this);
}

bool BatchMatMulNode::isEqual(const BatchMatMulNode &other) const {
  return true &&
      LHS_ == other.LHS_ &&
      RHS_ == other.RHS_ &&
      predicate_ == other.predicate_ &&
      getType(0) == other.getType(0);
}

Node* BatchMatMulNode::clone() const {
  return new BatchMatMulNode(getName(), getResult().getType(), getLHS(), getRHS());
}

llvm::hash_code BatchMatMulNode::getHash() const {
  return llvm::hash_combine(
      LHS_,
      RHS_);
}

unsigned BatchedReduceAddNode::getNumInputs() const {
  return 1;
}

std::string BatchedReduceAddNode::getInputName(unsigned idx) const {
  if (idx == 0) { return "Batch"; }
  idx -= 1;
  llvm_unreachable("Invalid index");
}

NodeValue BatchedReduceAddNode::getNthInput(unsigned idx) {
  if (idx == 0) { return Batch_; }
  idx -= 1;
  llvm_unreachable("Invalid index");
}

void BatchedReduceAddNode::setNthInput(unsigned idx, NodeValue val) {
  if (idx == 0) { Batch_ = val; return; }
  idx -= 1;
  llvm_unreachable("Invalid index");
}

llvm::StringRef BatchedReduceAddNode::getOutputName(unsigned idx) const {
  if (idx == 0) { return "Result"; }
  llvm_unreachable("Invalid index");
}

std::string BatchedReduceAddNode::getDebugDesc() const {
  DescriptionBuilder db(getKindName());
  db.addParam("name", getName());
  if (hasPredicate()) db.addParam("Predicate", "Yes");
  db
    .addParam("Batch", *(getBatch().getType()))
    .addParam("Axis", getAxis())
    .addParam("users", getNumUsers());
  db.addParam("Result", *(getResult().getType()));
  return db;
}

void BatchedReduceAddNode::visit(Node *parent, NodeWalker *visitor) {
  if (!visitor->shouldVisit(parent, this)) { return; }
  visitor->pre(parent, this);
if (hasPredicate())
 getPredicate().getNode()->visit(this, visitor);
  getBatch().getNode()->visit(this, visitor);
  visitor->post(parent, this);
}

bool BatchedReduceAddNode::isEqual(const BatchedReduceAddNode &other) const {
  return true &&
      Batch_ == other.Batch_ &&
      predicate_ == other.predicate_ &&
      Axis_ == other.Axis_ &&
      getType(0) == other.getType(0);
}

Node* BatchedReduceAddNode::clone() const {
  return new BatchedReduceAddNode(getName(), getResult().getType(), getBatch(), getAxis());
}

llvm::hash_code BatchedReduceAddNode::getHash() const {
  return llvm::hash_combine(
      Axis_,
      Batch_);
}

unsigned BatchedReduceMeanNode::getNumInputs() const {
  return 1;
}

std::string BatchedReduceMeanNode::getInputName(unsigned idx) const {
  if (idx == 0) { return "Batch"; }
  idx -= 1;
  llvm_unreachable("Invalid index");
}

NodeValue BatchedReduceMeanNode::getNthInput(unsigned idx) {
  if (idx == 0) { return Batch_; }
  idx -= 1;
  llvm_unreachable("Invalid index");
}

void BatchedReduceMeanNode::setNthInput(unsigned idx, NodeValue val) {
  if (idx == 0) { Batch_ = val; return; }
  idx -= 1;
  llvm_unreachable("Invalid index");
}

llvm::StringRef BatchedReduceMeanNode::getOutputName(unsigned idx) const {
  if (idx == 0) { return "Result"; }
  llvm_unreachable("Invalid index");
}

std::string BatchedReduceMeanNode::getDebugDesc() const {
  DescriptionBuilder db(getKindName());
  db.addParam("name", getName());
  if (hasPredicate()) db.addParam("Predicate", "Yes");
  db
    .addParam("Batch", *(getBatch().getType()))
    .addParam("Axes", getAxes())
    .addParam("users", getNumUsers());
  db.addParam("Result", *(getResult().getType()));
  return db;
}

void BatchedReduceMeanNode::visit(Node *parent, NodeWalker *visitor) {
  if (!visitor->shouldVisit(parent, this)) { return; }
  visitor->pre(parent, this);
if (hasPredicate())
 getPredicate().getNode()->visit(this, visitor);
  getBatch().getNode()->visit(this, visitor);
  visitor->post(parent, this);
}

bool BatchedReduceMeanNode::isEqual(const BatchedReduceMeanNode &other) const {
  return true &&
      Batch_ == other.Batch_ &&
      predicate_ == other.predicate_ &&
      Axes_ == other.Axes_ &&
      getType(0) == other.getType(0);
}

Node* BatchedReduceMeanNode::clone() const {
  return new BatchedReduceMeanNode(getName(), getResult().getType(), getBatch(), getAxes());
}

llvm::hash_code BatchedReduceMeanNode::getHash() const {
  return llvm::hash_combine(
      llvm::hash_combine_range(Axes_.begin(), Axes_.end()),
      Batch_);
}

unsigned BatchedReduceMinNode::getNumInputs() const {
  return 1;
}

std::string BatchedReduceMinNode::getInputName(unsigned idx) const {
  if (idx == 0) { return "Batch"; }
  idx -= 1;
  llvm_unreachable("Invalid index");
}

NodeValue BatchedReduceMinNode::getNthInput(unsigned idx) {
  if (idx == 0) { return Batch_; }
  idx -= 1;
  llvm_unreachable("Invalid index");
}

void BatchedReduceMinNode::setNthInput(unsigned idx, NodeValue val) {
  if (idx == 0) { Batch_ = val; return; }
  idx -= 1;
  llvm_unreachable("Invalid index");
}

llvm::StringRef BatchedReduceMinNode::getOutputName(unsigned idx) const {
  if (idx == 0) { return "Result"; }
  llvm_unreachable("Invalid index");
}

std::string BatchedReduceMinNode::getDebugDesc() const {
  DescriptionBuilder db(getKindName());
  db.addParam("name", getName());
  if (hasPredicate()) db.addParam("Predicate", "Yes");
  db
    .addParam("Batch", *(getBatch().getType()))
    .addParam("Axes", getAxes())
    .addParam("users", getNumUsers());
  db.addParam("Result", *(getResult().getType()));
  return db;
}

void BatchedReduceMinNode::visit(Node *parent, NodeWalker *visitor) {
  if (!visitor->shouldVisit(parent, this)) { return; }
  visitor->pre(parent, this);
if (hasPredicate())
 getPredicate().getNode()->visit(this, visitor);
  getBatch().getNode()->visit(this, visitor);
  visitor->post(parent, this);
}

bool BatchedReduceMinNode::isEqual(const BatchedReduceMinNode &other) const {
  return true &&
      Batch_ == other.Batch_ &&
      predicate_ == other.predicate_ &&
      Axes_ == other.Axes_ &&
      getType(0) == other.getType(0);
}

Node* BatchedReduceMinNode::clone() const {
  return new BatchedReduceMinNode(getName(), getResult().getType(), getBatch(), getAxes());
}

llvm::hash_code BatchedReduceMinNode::getHash() const {
  return llvm::hash_combine(
      llvm::hash_combine_range(Axes_.begin(), Axes_.end()),
      Batch_);
}

unsigned ChannelShuffleNode::getNumInputs() const {
  return 1;
}

std::string ChannelShuffleNode::getInputName(unsigned idx) const {
  if (idx == 0) { return "Input"; }
  idx -= 1;
  llvm_unreachable("Invalid index");
}

NodeValue ChannelShuffleNode::getNthInput(unsigned idx) {
  if (idx == 0) { return Input_; }
  idx -= 1;
  llvm_unreachable("Invalid index");
}

void ChannelShuffleNode::setNthInput(unsigned idx, NodeValue val) {
  if (idx == 0) { Input_ = val; return; }
  idx -= 1;
  llvm_unreachable("Invalid index");
}

llvm::StringRef ChannelShuffleNode::getOutputName(unsigned idx) const {
  if (idx == 0) { return "Result"; }
  llvm_unreachable("Invalid index");
}

std::string ChannelShuffleNode::getDebugDesc() const {
  DescriptionBuilder db(getKindName());
  db.addParam("name", getName());
  if (hasPredicate()) db.addParam("Predicate", "Yes");
  db
    .addParam("Input", *(getInput().getType()))
    .addParam("Group", getGroup())
    .addParam("Kernel", getKernel())
    .addParam("users", getNumUsers());
  db.addParam("Result", *(getResult().getType()));
  return db;
}

void ChannelShuffleNode::visit(Node *parent, NodeWalker *visitor) {
  if (!visitor->shouldVisit(parent, this)) { return; }
  visitor->pre(parent, this);
if (hasPredicate())
 getPredicate().getNode()->visit(this, visitor);
  getInput().getNode()->visit(this, visitor);
  visitor->post(parent, this);
}

bool ChannelShuffleNode::isEqual(const ChannelShuffleNode &other) const {
  return true &&
      Input_ == other.Input_ &&
      predicate_ == other.predicate_ &&
      Group_ == other.Group_ &&
      Kernel_ == other.Kernel_ &&
      getType(0) == other.getType(0);
}

Node* ChannelShuffleNode::clone() const {
  return new ChannelShuffleNode(getName(), getResult().getType(), getInput(), getGroup(), getKernel());
}

llvm::hash_code ChannelShuffleNode::getHash() const {
  return llvm::hash_combine(
      Group_,
      Kernel_,
      Input_);
}

unsigned CumSumNode::getNumInputs() const {
  return 1;
}

std::string CumSumNode::getInputName(unsigned idx) const {
  if (idx == 0) { return "Input"; }
  idx -= 1;
  llvm_unreachable("Invalid index");
}

NodeValue CumSumNode::getNthInput(unsigned idx) {
  if (idx == 0) { return Input_; }
  idx -= 1;
  llvm_unreachable("Invalid index");
}

void CumSumNode::setNthInput(unsigned idx, NodeValue val) {
  if (idx == 0) { Input_ = val; return; }
  idx -= 1;
  llvm_unreachable("Invalid index");
}

llvm::StringRef CumSumNode::getOutputName(unsigned idx) const {
  if (idx == 0) { return "Result"; }
  llvm_unreachable("Invalid index");
}

std::string CumSumNode::getDebugDesc() const {
  DescriptionBuilder db(getKindName());
  db.addParam("name", getName());
  if (hasPredicate()) db.addParam("Predicate", "Yes");
  db
    .addParam("Input", *(getInput().getType()))
    .addParam("Exclusive", getExclusive())
    .addParam("Reverse", getReverse())
    .addParam("users", getNumUsers());
  db.addParam("Result", *(getResult().getType()));
  return db;
}

void CumSumNode::visit(Node *parent, NodeWalker *visitor) {
  if (!visitor->shouldVisit(parent, this)) { return; }
  visitor->pre(parent, this);
if (hasPredicate())
 getPredicate().getNode()->visit(this, visitor);
  getInput().getNode()->visit(this, visitor);
  visitor->post(parent, this);
}

bool CumSumNode::isEqual(const CumSumNode &other) const {
  return true &&
      Input_ == other.Input_ &&
      predicate_ == other.predicate_ &&
      Exclusive_ == other.Exclusive_ &&
      Reverse_ == other.Reverse_ &&
      getType(0) == other.getType(0);
}

Node* CumSumNode::clone() const {
  return new CumSumNode(getName(), getResult().getType(), getInput(), getExclusive(), getReverse());
}

llvm::hash_code CumSumNode::getHash() const {
  return llvm::hash_combine(
      Exclusive_,
      Reverse_,
      Input_);
}

unsigned LengthsSumNode::getNumInputs() const {
  return 2;
}

std::string LengthsSumNode::getInputName(unsigned idx) const {
  if (idx == 0) { return "Data"; }
  if (idx == 1) { return "Lengths"; }
  idx -= 2;
  llvm_unreachable("Invalid index");
}

NodeValue LengthsSumNode::getNthInput(unsigned idx) {
  if (idx == 0) { return Data_; }
  if (idx == 1) { return Lengths_; }
  idx -= 2;
  llvm_unreachable("Invalid index");
}

void LengthsSumNode::setNthInput(unsigned idx, NodeValue val) {
  if (idx == 0) { Data_ = val; return; }
  if (idx == 1) { Lengths_ = val; return; }
  idx -= 2;
  llvm_unreachable("Invalid index");
}

llvm::StringRef LengthsSumNode::getOutputName(unsigned idx) const {
  if (idx == 0) { return "Result"; }
  llvm_unreachable("Invalid index");
}

std::string LengthsSumNode::getDebugDesc() const {
  DescriptionBuilder db(getKindName());
  db.addParam("name", getName());
  if (hasPredicate()) db.addParam("Predicate", "Yes");
  db
    .addParam("Data", *(getData().getType()))
    .addParam("Lengths", *(getLengths().getType()))
    .addParam("users", getNumUsers());
  db.addParam("Result", *(getResult().getType()));
  return db;
}

void LengthsSumNode::visit(Node *parent, NodeWalker *visitor) {
  if (!visitor->shouldVisit(parent, this)) { return; }
  visitor->pre(parent, this);
if (hasPredicate())
 getPredicate().getNode()->visit(this, visitor);
  getData().getNode()->visit(this, visitor);
  getLengths().getNode()->visit(this, visitor);
  visitor->post(parent, this);
}

bool LengthsSumNode::isEqual(const LengthsSumNode &other) const {
  return true &&
      Data_ == other.Data_ &&
      Lengths_ == other.Lengths_ &&
      predicate_ == other.predicate_ &&
      getType(0) == other.getType(0);
}

Node* LengthsSumNode::clone() const {
  return new LengthsSumNode(getName(), getResult().getType(), getData(), getLengths());
}

llvm::hash_code LengthsSumNode::getHash() const {
  return llvm::hash_combine(
      Data_,
      Lengths_);
}

unsigned SparseLengthsSumGradNode::getNumInputs() const {
  return 5;
}

std::string SparseLengthsSumGradNode::getInputName(unsigned idx) const {
  if (idx == 0) { return "Data"; }
  if (idx == 1) { return "Indices"; }
  if (idx == 2) { return "Lengths"; }
  if (idx == 3) { return "OriginalOutputForResult"; }
  if (idx == 4) { return "GradOfOriginalOutputNamedResult"; }
  idx -= 5;
  llvm_unreachable("Invalid index");
}

NodeValue SparseLengthsSumGradNode::getNthInput(unsigned idx) {
  if (idx == 0) { return Data_; }
  if (idx == 1) { return Indices_; }
  if (idx == 2) { return Lengths_; }
  if (idx == 3) { return OriginalOutputForResult_; }
  if (idx == 4) { return GradOfOriginalOutputNamedResult_; }
  idx -= 5;
  llvm_unreachable("Invalid index");
}

void SparseLengthsSumGradNode::setNthInput(unsigned idx, NodeValue val) {
  if (idx == 0) { Data_ = val; return; }
  if (idx == 1) { Indices_ = val; return; }
  if (idx == 2) { Lengths_ = val; return; }
  if (idx == 3) { OriginalOutputForResult_ = val; return; }
  if (idx == 4) { GradOfOriginalOutputNamedResult_ = val; return; }
  idx -= 5;
  llvm_unreachable("Invalid index");
}

llvm::StringRef SparseLengthsSumGradNode::getOutputName(unsigned idx) const {
  if (idx == 0) { return "GradOfInputNamedData"; }
  if (idx == 1) { return "GradOfInputNamedIndices"; }
  if (idx == 2) { return "GradOfInputNamedLengths"; }
  llvm_unreachable("Invalid index");
}

std::string SparseLengthsSumGradNode::getDebugDesc() const {
  DescriptionBuilder db(getKindName());
  db.addParam("name", getName());
  if (hasPredicate()) db.addParam("Predicate", "Yes");
  db
    .addParam("Data", *(getData().getType()))
    .addParam("Indices", *(getIndices().getType()))
    .addParam("Lengths", *(getLengths().getType()))
    .addParam("OriginalOutputForResult", *(getOriginalOutputForResult().getType()))
    .addParam("GradOfOriginalOutputNamedResult", *(getGradOfOriginalOutputNamedResult().getType()))
    .addParam("LengthsMode", getLengthsMode())
    .addParam("AvgLength", getAvgLength())
    .addParam("users", getNumUsers());
  db.addParam("GradOfInputNamedData", *(getGradOfInputNamedData().getType()));
  db.addParam("GradOfInputNamedIndices", *(getGradOfInputNamedIndices().getType()));
  db.addParam("GradOfInputNamedLengths", *(getGradOfInputNamedLengths().getType()));
  return db;
}

void SparseLengthsSumGradNode::visit(Node *parent, NodeWalker *visitor) {
  if (!visitor->shouldVisit(parent, this)) { return; }
  visitor->pre(parent, this);
if (hasPredicate())
 getPredicate().getNode()->visit(this, visitor);
  getData().getNode()->visit(this, visitor);
  getIndices().getNode()->visit(this, visitor);
  getLengths().getNode()->visit(this, visitor);
  getOriginalOutputForResult().getNode()->visit(this, visitor);
  getGradOfOriginalOutputNamedResult().getNode()->visit(this, visitor);
  visitor->post(parent, this);
}

bool SparseLengthsSumGradNode::isEqual(const SparseLengthsSumGradNode &other) const {
  return true &&
      Data_ == other.Data_ &&
      Indices_ == other.Indices_ &&
      Lengths_ == other.Lengths_ &&
      OriginalOutputForResult_ == other.OriginalOutputForResult_ &&
      GradOfOriginalOutputNamedResult_ == other.GradOfOriginalOutputNamedResult_ &&
      predicate_ == other.predicate_ &&
      LengthsMode_ == other.LengthsMode_ &&
      AvgLength_ == other.AvgLength_ &&
      getType(0) == other.getType(0) &&
      getType(1) == other.getType(1) &&
      getType(2) == other.getType(2);
}

Node* SparseLengthsSumGradNode::clone() const {
  return new SparseLengthsSumGradNode(getName(), getData(), getIndices(), getLengths(), getOriginalOutputForResult(), getGradOfOriginalOutputNamedResult(), getLengthsMode(), getAvgLength());
}

llvm::hash_code SparseLengthsSumGradNode::getHash() const {
  return llvm::hash_combine(
      LengthsMode_,
      toBinary(AvgLength_),
      Data_,
      Indices_,
      Lengths_,
      OriginalOutputForResult_,
      GradOfOriginalOutputNamedResult_);
}

unsigned SparseLengthsSumNode::getNumInputs() const {
  return 3;
}

std::string SparseLengthsSumNode::getInputName(unsigned idx) const {
  if (idx == 0) { return "Data"; }
  if (idx == 1) { return "Indices"; }
  if (idx == 2) { return "Lengths"; }
  idx -= 3;
  llvm_unreachable("Invalid index");
}

NodeValue SparseLengthsSumNode::getNthInput(unsigned idx) {
  if (idx == 0) { return Data_; }
  if (idx == 1) { return Indices_; }
  if (idx == 2) { return Lengths_; }
  idx -= 3;
  llvm_unreachable("Invalid index");
}

void SparseLengthsSumNode::setNthInput(unsigned idx, NodeValue val) {
  if (idx == 0) { Data_ = val; return; }
  if (idx == 1) { Indices_ = val; return; }
  if (idx == 2) { Lengths_ = val; return; }
  idx -= 3;
  llvm_unreachable("Invalid index");
}

llvm::StringRef SparseLengthsSumNode::getOutputName(unsigned idx) const {
  if (idx == 0) { return "Result"; }
  llvm_unreachable("Invalid index");
}

std::string SparseLengthsSumNode::getDebugDesc() const {
  DescriptionBuilder db(getKindName());
  db.addParam("name", getName());
  if (hasPredicate()) db.addParam("Predicate", "Yes");
  db
    .addParam("Data", *(getData().getType()))
    .addParam("Indices", *(getIndices().getType()))
    .addParam("Lengths", *(getLengths().getType()))
    .addParam("LengthsMode", getLengthsMode())
    .addParam("AvgLength", getAvgLength())
    .addParam("users", getNumUsers());
  db.addParam("Result", *(getResult().getType()));
  return db;
}

void SparseLengthsSumNode::visit(Node *parent, NodeWalker *visitor) {
  if (!visitor->shouldVisit(parent, this)) { return; }
  visitor->pre(parent, this);
if (hasPredicate())
 getPredicate().getNode()->visit(this, visitor);
  getData().getNode()->visit(this, visitor);
  getIndices().getNode()->visit(this, visitor);
  getLengths().getNode()->visit(this, visitor);
  visitor->post(parent, this);
}

bool SparseLengthsSumNode::isEqual(const SparseLengthsSumNode &other) const {
  return true &&
      Data_ == other.Data_ &&
      Indices_ == other.Indices_ &&
      Lengths_ == other.Lengths_ &&
      predicate_ == other.predicate_ &&
      LengthsMode_ == other.LengthsMode_ &&
      AvgLength_ == other.AvgLength_ &&
      getType(0) == other.getType(0);
}

Node* SparseLengthsSumNode::clone() const {
  return new SparseLengthsSumNode(getName(), getResult().getType(), getData(), getIndices(), getLengths(), getLengthsMode(), getAvgLength());
}

llvm::hash_code SparseLengthsSumNode::getHash() const {
  return llvm::hash_combine(
      LengthsMode_,
      toBinary(AvgLength_),
      Data_,
      Indices_,
      Lengths_);
}

SparseLengthsSumGradNode *SparseLengthsSumNode::getGrad(GraphGradMapper &builder) {
  auto *x = new SparseLengthsSumGradNode(getName().str() + "_grad", getData(), getIndices(), getLengths(), getResult(), builder.getGradient(getResult()), getLengthsMode(), getAvgLength());
  builder.addGradient(getData(), x->getGradOfInputNamedData());
  builder.addGradient(getIndices(), x->getGradOfInputNamedIndices());
  builder.addGradient(getLengths(), x->getGradOfInputNamedLengths());
  return x;
}

unsigned SparseLengthsWeightedSumGradNode::getNumInputs() const {
  return 6;
}

std::string SparseLengthsWeightedSumGradNode::getInputName(unsigned idx) const {
  if (idx == 0) { return "Data"; }
  if (idx == 1) { return "Weights"; }
  if (idx == 2) { return "Indices"; }
  if (idx == 3) { return "Lengths"; }
  if (idx == 4) { return "OriginalOutputForResult"; }
  if (idx == 5) { return "GradOfOriginalOutputNamedResult"; }
  idx -= 6;
  llvm_unreachable("Invalid index");
}

NodeValue SparseLengthsWeightedSumGradNode::getNthInput(unsigned idx) {
  if (idx == 0) { return Data_; }
  if (idx == 1) { return Weights_; }
  if (idx == 2) { return Indices_; }
  if (idx == 3) { return Lengths_; }
  if (idx == 4) { return OriginalOutputForResult_; }
  if (idx == 5) { return GradOfOriginalOutputNamedResult_; }
  idx -= 6;
  llvm_unreachable("Invalid index");
}

void SparseLengthsWeightedSumGradNode::setNthInput(unsigned idx, NodeValue val) {
  if (idx == 0) { Data_ = val; return; }
  if (idx == 1) { Weights_ = val; return; }
  if (idx == 2) { Indices_ = val; return; }
  if (idx == 3) { Lengths_ = val; return; }
  if (idx == 4) { OriginalOutputForResult_ = val; return; }
  if (idx == 5) { GradOfOriginalOutputNamedResult_ = val; return; }
  idx -= 6;
  llvm_unreachable("Invalid index");
}

llvm::StringRef SparseLengthsWeightedSumGradNode::getOutputName(unsigned idx) const {
  if (idx == 0) { return "GradOfInputNamedData"; }
  if (idx == 1) { return "GradOfInputNamedWeights"; }
  if (idx == 2) { return "GradOfInputNamedIndices"; }
  if (idx == 3) { return "GradOfInputNamedLengths"; }
  llvm_unreachable("Invalid index");
}

std::string SparseLengthsWeightedSumGradNode::getDebugDesc() const {
  DescriptionBuilder db(getKindName());
  db.addParam("name", getName());
  if (hasPredicate()) db.addParam("Predicate", "Yes");
  db
    .addParam("Data", *(getData().getType()))
    .addParam("Weights", *(getWeights().getType()))
    .addParam("Indices", *(getIndices().getType()))
    .addParam("Lengths", *(getLengths().getType()))
    .addParam("OriginalOutputForResult", *(getOriginalOutputForResult().getType()))
    .addParam("GradOfOriginalOutputNamedResult", *(getGradOfOriginalOutputNamedResult().getType()))
    .addParam("LengthsMode", getLengthsMode())
    .addParam("AvgLength", getAvgLength())
    .addParam("users", getNumUsers());
  db.addParam("GradOfInputNamedData", *(getGradOfInputNamedData().getType()));
  db.addParam("GradOfInputNamedWeights", *(getGradOfInputNamedWeights().getType()));
  db.addParam("GradOfInputNamedIndices", *(getGradOfInputNamedIndices().getType()));
  db.addParam("GradOfInputNamedLengths", *(getGradOfInputNamedLengths().getType()));
  return db;
}

void SparseLengthsWeightedSumGradNode::visit(Node *parent, NodeWalker *visitor) {
  if (!visitor->shouldVisit(parent, this)) { return; }
  visitor->pre(parent, this);
if (hasPredicate())
 getPredicate().getNode()->visit(this, visitor);
  getData().getNode()->visit(this, visitor);
  getWeights().getNode()->visit(this, visitor);
  getIndices().getNode()->visit(this, visitor);
  getLengths().getNode()->visit(this, visitor);
  getOriginalOutputForResult().getNode()->visit(this, visitor);
  getGradOfOriginalOutputNamedResult().getNode()->visit(this, visitor);
  visitor->post(parent, this);
}

bool SparseLengthsWeightedSumGradNode::isEqual(const SparseLengthsWeightedSumGradNode &other) const {
  return true &&
      Data_ == other.Data_ &&
      Weights_ == other.Weights_ &&
      Indices_ == other.Indices_ &&
      Lengths_ == other.Lengths_ &&
      OriginalOutputForResult_ == other.OriginalOutputForResult_ &&
      GradOfOriginalOutputNamedResult_ == other.GradOfOriginalOutputNamedResult_ &&
      predicate_ == other.predicate_ &&
      LengthsMode_ == other.LengthsMode_ &&
      AvgLength_ == other.AvgLength_ &&
      getType(0) == other.getType(0) &&
      getType(1) == other.getType(1) &&
      getType(2) == other.getType(2) &&
      getType(3) == other.getType(3);
}

Node* SparseLengthsWeightedSumGradNode::clone() const {
  return new SparseLengthsWeightedSumGradNode(getName(), getData(), getWeights(), getIndices(), getLengths(), getOriginalOutputForResult(), getGradOfOriginalOutputNamedResult(), getLengthsMode(), getAvgLength());
}

llvm::hash_code SparseLengthsWeightedSumGradNode::getHash() const {
  return llvm::hash_combine(
      LengthsMode_,
      toBinary(AvgLength_),
      Data_,
      Weights_,
      Indices_,
      Lengths_,
      OriginalOutputForResult_,
      GradOfOriginalOutputNamedResult_);
}

unsigned SparseLengthsWeightedSumNode::getNumInputs() const {
  return 4;
}

std::string SparseLengthsWeightedSumNode::getInputName(unsigned idx) const {
  if (idx == 0) { return "Data"; }
  if (idx == 1) { return "Weights"; }
  if (idx == 2) { return "Indices"; }
  if (idx == 3) { return "Lengths"; }
  idx -= 4;
  llvm_unreachable("Invalid index");
}

NodeValue SparseLengthsWeightedSumNode::getNthInput(unsigned idx) {
  if (idx == 0) { return Data_; }
  if (idx == 1) { return Weights_; }
  if (idx == 2) { return Indices_; }
  if (idx == 3) { return Lengths_; }
  idx -= 4;
  llvm_unreachable("Invalid index");
}

void SparseLengthsWeightedSumNode::setNthInput(unsigned idx, NodeValue val) {
  if (idx == 0) { Data_ = val; return; }
  if (idx == 1) { Weights_ = val; return; }
  if (idx == 2) { Indices_ = val; return; }
  if (idx == 3) { Lengths_ = val; return; }
  idx -= 4;
  llvm_unreachable("Invalid index");
}

llvm::StringRef SparseLengthsWeightedSumNode::getOutputName(unsigned idx) const {
  if (idx == 0) { return "Result"; }
  llvm_unreachable("Invalid index");
}

std::string SparseLengthsWeightedSumNode::getDebugDesc() const {
  DescriptionBuilder db(getKindName());
  db.addParam("name", getName());
  if (hasPredicate()) db.addParam("Predicate", "Yes");
  db
    .addParam("Data", *(getData().getType()))
    .addParam("Weights", *(getWeights().getType()))
    .addParam("Indices", *(getIndices().getType()))
    .addParam("Lengths", *(getLengths().getType()))
    .addParam("LengthsMode", getLengthsMode())
    .addParam("AvgLength", getAvgLength())
    .addParam("users", getNumUsers());
  db.addParam("Result", *(getResult().getType()));
  return db;
}

void SparseLengthsWeightedSumNode::visit(Node *parent, NodeWalker *visitor) {
  if (!visitor->shouldVisit(parent, this)) { return; }
  visitor->pre(parent, this);
if (hasPredicate())
 getPredicate().getNode()->visit(this, visitor);
  getData().getNode()->visit(this, visitor);
  getWeights().getNode()->visit(this, visitor);
  getIndices().getNode()->visit(this, visitor);
  getLengths().getNode()->visit(this, visitor);
  visitor->post(parent, this);
}

bool SparseLengthsWeightedSumNode::isEqual(const SparseLengthsWeightedSumNode &other) const {
  return true &&
      Data_ == other.Data_ &&
      Weights_ == other.Weights_ &&
      Indices_ == other.Indices_ &&
      Lengths_ == other.Lengths_ &&
      predicate_ == other.predicate_ &&
      LengthsMode_ == other.LengthsMode_ &&
      AvgLength_ == other.AvgLength_ &&
      getType(0) == other.getType(0);
}

Node* SparseLengthsWeightedSumNode::clone() const {
  return new SparseLengthsWeightedSumNode(getName(), getResult().getType(), getData(), getWeights(), getIndices(), getLengths(), getLengthsMode(), getAvgLength());
}

llvm::hash_code SparseLengthsWeightedSumNode::getHash() const {
  return llvm::hash_combine(
      LengthsMode_,
      toBinary(AvgLength_),
      Data_,
      Weights_,
      Indices_,
      Lengths_);
}

SparseLengthsWeightedSumGradNode *SparseLengthsWeightedSumNode::getGrad(GraphGradMapper &builder) {
  auto *x = new SparseLengthsWeightedSumGradNode(getName().str() + "_grad", getData(), getWeights(), getIndices(), getLengths(), getResult(), builder.getGradient(getResult()), getLengthsMode(), getAvgLength());
  builder.addGradient(getData(), x->getGradOfInputNamedData());
  builder.addGradient(getWeights(), x->getGradOfInputNamedWeights());
  builder.addGradient(getIndices(), x->getGradOfInputNamedIndices());
  builder.addGradient(getLengths(), x->getGradOfInputNamedLengths());
  return x;
}

unsigned EmbeddingBagNode::getNumInputs() const {
  return 4;
}

std::string EmbeddingBagNode::getInputName(unsigned idx) const {
  if (idx == 0) { return "Data"; }
  if (idx == 1) { return "Weights"; }
  if (idx == 2) { return "Indices"; }
  if (idx == 3) { return "Offsets"; }
  idx -= 4;
  llvm_unreachable("Invalid index");
}

NodeValue EmbeddingBagNode::getNthInput(unsigned idx) {
  if (idx == 0) { return Data_; }
  if (idx == 1) { return Weights_; }
  if (idx == 2) { return Indices_; }
  if (idx == 3) { return Offsets_; }
  idx -= 4;
  llvm_unreachable("Invalid index");
}

void EmbeddingBagNode::setNthInput(unsigned idx, NodeValue val) {
  if (idx == 0) { Data_ = val; return; }
  if (idx == 1) { Weights_ = val; return; }
  if (idx == 2) { Indices_ = val; return; }
  if (idx == 3) { Offsets_ = val; return; }
  idx -= 4;
  llvm_unreachable("Invalid index");
}

llvm::StringRef EmbeddingBagNode::getOutputName(unsigned idx) const {
  if (idx == 0) { return "Result"; }
  llvm_unreachable("Invalid index");
}

std::string EmbeddingBagNode::getDebugDesc() const {
  DescriptionBuilder db(getKindName());
  db.addParam("name", getName());
  if (hasPredicate()) db.addParam("Predicate", "Yes");
  db
    .addParam("Data", *(getData().getType()))
    .addParam("Weights", *(getWeights().getType()))
    .addParam("Indices", *(getIndices().getType()))
    .addParam("Offsets", *(getOffsets().getType()))
    .addParam("HasEndOffset", getHasEndOffset())
    .addParam("LengthsMode", getLengthsMode())
    .addParam("AvgLength", getAvgLength())
    .addParam("users", getNumUsers());
  db.addParam("Result", *(getResult().getType()));
  return db;
}

void EmbeddingBagNode::visit(Node *parent, NodeWalker *visitor) {
  if (!visitor->shouldVisit(parent, this)) { return; }
  visitor->pre(parent, this);
if (hasPredicate())
 getPredicate().getNode()->visit(this, visitor);
  getData().getNode()->visit(this, visitor);
  getWeights().getNode()->visit(this, visitor);
  getIndices().getNode()->visit(this, visitor);
  getOffsets().getNode()->visit(this, visitor);
  visitor->post(parent, this);
}

bool EmbeddingBagNode::isEqual(const EmbeddingBagNode &other) const {
  return true &&
      Data_ == other.Data_ &&
      Weights_ == other.Weights_ &&
      Indices_ == other.Indices_ &&
      Offsets_ == other.Offsets_ &&
      predicate_ == other.predicate_ &&
      HasEndOffset_ == other.HasEndOffset_ &&
      LengthsMode_ == other.LengthsMode_ &&
      AvgLength_ == other.AvgLength_ &&
      getType(0) == other.getType(0);
}

Node* EmbeddingBagNode::clone() const {
  return new EmbeddingBagNode(getName(), getResult().getType(), getData(), getWeights(), getIndices(), getOffsets(), getHasEndOffset(), getLengthsMode(), getAvgLength());
}

llvm::hash_code EmbeddingBagNode::getHash() const {
  return llvm::hash_combine(
      HasEndOffset_,
      LengthsMode_,
      toBinary(AvgLength_),
      Data_,
      Weights_,
      Indices_,
      Offsets_);
}

unsigned EmbeddingBagByteRowwiseOffsetsNode::getNumInputs() const {
  return 4;
}

std::string EmbeddingBagByteRowwiseOffsetsNode::getInputName(unsigned idx) const {
  if (idx == 0) { return "Data"; }
  if (idx == 1) { return "Weights"; }
  if (idx == 2) { return "Indices"; }
  if (idx == 3) { return "Offsets"; }
  idx -= 4;
  llvm_unreachable("Invalid index");
}

NodeValue EmbeddingBagByteRowwiseOffsetsNode::getNthInput(unsigned idx) {
  if (idx == 0) { return Data_; }
  if (idx == 1) { return Weights_; }
  if (idx == 2) { return Indices_; }
  if (idx == 3) { return Offsets_; }
  idx -= 4;
  llvm_unreachable("Invalid index");
}

void EmbeddingBagByteRowwiseOffsetsNode::setNthInput(unsigned idx, NodeValue val) {
  if (idx == 0) { Data_ = val; return; }
  if (idx == 1) { Weights_ = val; return; }
  if (idx == 2) { Indices_ = val; return; }
  if (idx == 3) { Offsets_ = val; return; }
  idx -= 4;
  llvm_unreachable("Invalid index");
}

llvm::StringRef EmbeddingBagByteRowwiseOffsetsNode::getOutputName(unsigned idx) const {
  if (idx == 0) { return "Result"; }
  llvm_unreachable("Invalid index");
}

std::string EmbeddingBagByteRowwiseOffsetsNode::getDebugDesc() const {
  DescriptionBuilder db(getKindName());
  db.addParam("name", getName());
  if (hasPredicate()) db.addParam("Predicate", "Yes");
  db
    .addParam("Data", *(getData().getType()))
    .addParam("Weights", *(getWeights().getType()))
    .addParam("Indices", *(getIndices().getType()))
    .addParam("Offsets", *(getOffsets().getType()))
    .addParam("UseFP16Accumulation", getUseFP16Accumulation())
    .addParam("HasEndOffset", getHasEndOffset())
    .addParam("LengthsMode", getLengthsMode())
    .addParam("AvgLength", getAvgLength())
    .addParam("users", getNumUsers());
  db.addParam("Result", *(getResult().getType()));
  return db;
}

void EmbeddingBagByteRowwiseOffsetsNode::visit(Node *parent, NodeWalker *visitor) {
  if (!visitor->shouldVisit(parent, this)) { return; }
  visitor->pre(parent, this);
if (hasPredicate())
 getPredicate().getNode()->visit(this, visitor);
  getData().getNode()->visit(this, visitor);
  getWeights().getNode()->visit(this, visitor);
  getIndices().getNode()->visit(this, visitor);
  getOffsets().getNode()->visit(this, visitor);
  visitor->post(parent, this);
}

bool EmbeddingBagByteRowwiseOffsetsNode::isEqual(const EmbeddingBagByteRowwiseOffsetsNode &other) const {
  return true &&
      Data_ == other.Data_ &&
      Weights_ == other.Weights_ &&
      Indices_ == other.Indices_ &&
      Offsets_ == other.Offsets_ &&
      predicate_ == other.predicate_ &&
      UseFP16Accumulation_ == other.UseFP16Accumulation_ &&
      HasEndOffset_ == other.HasEndOffset_ &&
      LengthsMode_ == other.LengthsMode_ &&
      AvgLength_ == other.AvgLength_ &&
      getType(0) == other.getType(0);
}

Node* EmbeddingBagByteRowwiseOffsetsNode::clone() const {
  return new EmbeddingBagByteRowwiseOffsetsNode(getName(), getResult().getType(), getData(), getWeights(), getIndices(), getOffsets(), getUseFP16Accumulation(), getHasEndOffset(), getLengthsMode(), getAvgLength());
}

llvm::hash_code EmbeddingBagByteRowwiseOffsetsNode::getHash() const {
  return llvm::hash_combine(
      UseFP16Accumulation_,
      HasEndOffset_,
      LengthsMode_,
      toBinary(AvgLength_),
      Data_,
      Weights_,
      Indices_,
      Offsets_);
}

unsigned RowwiseQuantizedSparseLengthsWeightedSumNode::getNumInputs() const {
  return 6;
}

std::string RowwiseQuantizedSparseLengthsWeightedSumNode::getInputName(unsigned idx) const {
  if (idx == 0) { return "Data"; }
  if (idx == 1) { return "Scales"; }
  if (idx == 2) { return "Offsets"; }
  if (idx == 3) { return "Weights"; }
  if (idx == 4) { return "Indices"; }
  if (idx == 5) { return "Lengths"; }
  idx -= 6;
  llvm_unreachable("Invalid index");
}

NodeValue RowwiseQuantizedSparseLengthsWeightedSumNode::getNthInput(unsigned idx) {
  if (idx == 0) { return Data_; }
  if (idx == 1) { return Scales_; }
  if (idx == 2) { return Offsets_; }
  if (idx == 3) { return Weights_; }
  if (idx == 4) { return Indices_; }
  if (idx == 5) { return Lengths_; }
  idx -= 6;
  llvm_unreachable("Invalid index");
}

void RowwiseQuantizedSparseLengthsWeightedSumNode::setNthInput(unsigned idx, NodeValue val) {
  if (idx == 0) { Data_ = val; return; }
  if (idx == 1) { Scales_ = val; return; }
  if (idx == 2) { Offsets_ = val; return; }
  if (idx == 3) { Weights_ = val; return; }
  if (idx == 4) { Indices_ = val; return; }
  if (idx == 5) { Lengths_ = val; return; }
  idx -= 6;
  llvm_unreachable("Invalid index");
}

llvm::StringRef RowwiseQuantizedSparseLengthsWeightedSumNode::getOutputName(unsigned idx) const {
  if (idx == 0) { return "Result"; }
  llvm_unreachable("Invalid index");
}

std::string RowwiseQuantizedSparseLengthsWeightedSumNode::getDebugDesc() const {
  DescriptionBuilder db(getKindName());
  db.addParam("name", getName());
  if (hasPredicate()) db.addParam("Predicate", "Yes");
  db
    .addParam("Data", *(getData().getType()))
    .addParam("Scales", *(getScales().getType()))
    .addParam("Offsets", *(getOffsets().getType()))
    .addParam("Weights", *(getWeights().getType()))
    .addParam("Indices", *(getIndices().getType()))
    .addParam("Lengths", *(getLengths().getType()))
    .addParam("UseFP16Accumulation", getUseFP16Accumulation())
    .addParam("LengthsMode", getLengthsMode())
    .addParam("AvgLength", getAvgLength())
    .addParam("users", getNumUsers());
  db.addParam("Result", *(getResult().getType()));
  return db;
}

void RowwiseQuantizedSparseLengthsWeightedSumNode::visit(Node *parent, NodeWalker *visitor) {
  if (!visitor->shouldVisit(parent, this)) { return; }
  visitor->pre(parent, this);
if (hasPredicate())
 getPredicate().getNode()->visit(this, visitor);
  getData().getNode()->visit(this, visitor);
  getScales().getNode()->visit(this, visitor);
  getOffsets().getNode()->visit(this, visitor);
  getWeights().getNode()->visit(this, visitor);
  getIndices().getNode()->visit(this, visitor);
  getLengths().getNode()->visit(this, visitor);
  visitor->post(parent, this);
}

bool RowwiseQuantizedSparseLengthsWeightedSumNode::isEqual(const RowwiseQuantizedSparseLengthsWeightedSumNode &other) const {
  return true &&
      Data_ == other.Data_ &&
      Scales_ == other.Scales_ &&
      Offsets_ == other.Offsets_ &&
      Weights_ == other.Weights_ &&
      Indices_ == other.Indices_ &&
      Lengths_ == other.Lengths_ &&
      predicate_ == other.predicate_ &&
      UseFP16Accumulation_ == other.UseFP16Accumulation_ &&
      LengthsMode_ == other.LengthsMode_ &&
      AvgLength_ == other.AvgLength_ &&
      getType(0) == other.getType(0);
}

Node* RowwiseQuantizedSparseLengthsWeightedSumNode::clone() const {
  return new RowwiseQuantizedSparseLengthsWeightedSumNode(getName(), getResult().getType(), getData(), getScales(), getOffsets(), getWeights(), getIndices(), getLengths(), getUseFP16Accumulation(), getLengthsMode(), getAvgLength());
}

llvm::hash_code RowwiseQuantizedSparseLengthsWeightedSumNode::getHash() const {
  return llvm::hash_combine(
      UseFP16Accumulation_,
      LengthsMode_,
      toBinary(AvgLength_),
      Data_,
      Scales_,
      Offsets_,
      Weights_,
      Indices_,
      Lengths_);
}

unsigned FusedRowwiseQuantizedSparseLengthsWeightedSumNode::getNumInputs() const {
  return 4;
}

std::string FusedRowwiseQuantizedSparseLengthsWeightedSumNode::getInputName(unsigned idx) const {
  if (idx == 0) { return "Data"; }
  if (idx == 1) { return "Weights"; }
  if (idx == 2) { return "Indices"; }
  if (idx == 3) { return "Lengths"; }
  idx -= 4;
  llvm_unreachable("Invalid index");
}

NodeValue FusedRowwiseQuantizedSparseLengthsWeightedSumNode::getNthInput(unsigned idx) {
  if (idx == 0) { return Data_; }
  if (idx == 1) { return Weights_; }
  if (idx == 2) { return Indices_; }
  if (idx == 3) { return Lengths_; }
  idx -= 4;
  llvm_unreachable("Invalid index");
}

void FusedRowwiseQuantizedSparseLengthsWeightedSumNode::setNthInput(unsigned idx, NodeValue val) {
  if (idx == 0) { Data_ = val; return; }
  if (idx == 1) { Weights_ = val; return; }
  if (idx == 2) { Indices_ = val; return; }
  if (idx == 3) { Lengths_ = val; return; }
  idx -= 4;
  llvm_unreachable("Invalid index");
}

llvm::StringRef FusedRowwiseQuantizedSparseLengthsWeightedSumNode::getOutputName(unsigned idx) const {
  if (idx == 0) { return "Result"; }
  llvm_unreachable("Invalid index");
}

std::string FusedRowwiseQuantizedSparseLengthsWeightedSumNode::getDebugDesc() const {
  DescriptionBuilder db(getKindName());
  db.addParam("name", getName());
  if (hasPredicate()) db.addParam("Predicate", "Yes");
  db
    .addParam("Data", *(getData().getType()))
    .addParam("Weights", *(getWeights().getType()))
    .addParam("Indices", *(getIndices().getType()))
    .addParam("Lengths", *(getLengths().getType()))
    .addParam("UseFP16Accumulation", getUseFP16Accumulation())
    .addParam("LengthsMode", getLengthsMode())
    .addParam("AvgLength", getAvgLength())
    .addParam("users", getNumUsers());
  db.addParam("Result", *(getResult().getType()));
  return db;
}

void FusedRowwiseQuantizedSparseLengthsWeightedSumNode::visit(Node *parent, NodeWalker *visitor) {
  if (!visitor->shouldVisit(parent, this)) { return; }
  visitor->pre(parent, this);
if (hasPredicate())
 getPredicate().getNode()->visit(this, visitor);
  getData().getNode()->visit(this, visitor);
  getWeights().getNode()->visit(this, visitor);
  getIndices().getNode()->visit(this, visitor);
  getLengths().getNode()->visit(this, visitor);
  visitor->post(parent, this);
}

bool FusedRowwiseQuantizedSparseLengthsWeightedSumNode::isEqual(const FusedRowwiseQuantizedSparseLengthsWeightedSumNode &other) const {
  return true &&
      Data_ == other.Data_ &&
      Weights_ == other.Weights_ &&
      Indices_ == other.Indices_ &&
      Lengths_ == other.Lengths_ &&
      predicate_ == other.predicate_ &&
      UseFP16Accumulation_ == other.UseFP16Accumulation_ &&
      LengthsMode_ == other.LengthsMode_ &&
      AvgLength_ == other.AvgLength_ &&
      getType(0) == other.getType(0);
}

Node* FusedRowwiseQuantizedSparseLengthsWeightedSumNode::clone() const {
  return new FusedRowwiseQuantizedSparseLengthsWeightedSumNode(getName(), getResult().getType(), getData(), getWeights(), getIndices(), getLengths(), getUseFP16Accumulation(), getLengthsMode(), getAvgLength());
}

llvm::hash_code FusedRowwiseQuantizedSparseLengthsWeightedSumNode::getHash() const {
  return llvm::hash_combine(
      UseFP16Accumulation_,
      LengthsMode_,
      toBinary(AvgLength_),
      Data_,
      Weights_,
      Indices_,
      Lengths_);
}

unsigned FusedRowwiseQuantizedSparseLengthsSumNode::getNumInputs() const {
  return 3;
}

std::string FusedRowwiseQuantizedSparseLengthsSumNode::getInputName(unsigned idx) const {
  if (idx == 0) { return "Data"; }
  if (idx == 1) { return "Indices"; }
  if (idx == 2) { return "Lengths"; }
  idx -= 3;
  llvm_unreachable("Invalid index");
}

NodeValue FusedRowwiseQuantizedSparseLengthsSumNode::getNthInput(unsigned idx) {
  if (idx == 0) { return Data_; }
  if (idx == 1) { return Indices_; }
  if (idx == 2) { return Lengths_; }
  idx -= 3;
  llvm_unreachable("Invalid index");
}

void FusedRowwiseQuantizedSparseLengthsSumNode::setNthInput(unsigned idx, NodeValue val) {
  if (idx == 0) { Data_ = val; return; }
  if (idx == 1) { Indices_ = val; return; }
  if (idx == 2) { Lengths_ = val; return; }
  idx -= 3;
  llvm_unreachable("Invalid index");
}

llvm::StringRef FusedRowwiseQuantizedSparseLengthsSumNode::getOutputName(unsigned idx) const {
  if (idx == 0) { return "Result"; }
  llvm_unreachable("Invalid index");
}

std::string FusedRowwiseQuantizedSparseLengthsSumNode::getDebugDesc() const {
  DescriptionBuilder db(getKindName());
  db.addParam("name", getName());
  if (hasPredicate()) db.addParam("Predicate", "Yes");
  db
    .addParam("Data", *(getData().getType()))
    .addParam("Indices", *(getIndices().getType()))
    .addParam("Lengths", *(getLengths().getType()))
    .addParam("UseFP16Accumulation", getUseFP16Accumulation())
    .addParam("LengthsMode", getLengthsMode())
    .addParam("AvgLength", getAvgLength())
    .addParam("users", getNumUsers());
  db.addParam("Result", *(getResult().getType()));
  return db;
}

void FusedRowwiseQuantizedSparseLengthsSumNode::visit(Node *parent, NodeWalker *visitor) {
  if (!visitor->shouldVisit(parent, this)) { return; }
  visitor->pre(parent, this);
if (hasPredicate())
 getPredicate().getNode()->visit(this, visitor);
  getData().getNode()->visit(this, visitor);
  getIndices().getNode()->visit(this, visitor);
  getLengths().getNode()->visit(this, visitor);
  visitor->post(parent, this);
}

bool FusedRowwiseQuantizedSparseLengthsSumNode::isEqual(const FusedRowwiseQuantizedSparseLengthsSumNode &other) const {
  return true &&
      Data_ == other.Data_ &&
      Indices_ == other.Indices_ &&
      Lengths_ == other.Lengths_ &&
      predicate_ == other.predicate_ &&
      UseFP16Accumulation_ == other.UseFP16Accumulation_ &&
      LengthsMode_ == other.LengthsMode_ &&
      AvgLength_ == other.AvgLength_ &&
      getType(0) == other.getType(0);
}

Node* FusedRowwiseQuantizedSparseLengthsSumNode::clone() const {
  return new FusedRowwiseQuantizedSparseLengthsSumNode(getName(), getResult().getType(), getData(), getIndices(), getLengths(), getUseFP16Accumulation(), getLengthsMode(), getAvgLength());
}

llvm::hash_code FusedRowwiseQuantizedSparseLengthsSumNode::getHash() const {
  return llvm::hash_combine(
      UseFP16Accumulation_,
      LengthsMode_,
      toBinary(AvgLength_),
      Data_,
      Indices_,
      Lengths_);
}

unsigned LengthsToRangesNode::getNumInputs() const {
  return 1;
}

std::string LengthsToRangesNode::getInputName(unsigned idx) const {
  if (idx == 0) { return "Lengths"; }
  idx -= 1;
  llvm_unreachable("Invalid index");
}

NodeValue LengthsToRangesNode::getNthInput(unsigned idx) {
  if (idx == 0) { return Lengths_; }
  idx -= 1;
  llvm_unreachable("Invalid index");
}

void LengthsToRangesNode::setNthInput(unsigned idx, NodeValue val) {
  if (idx == 0) { Lengths_ = val; return; }
  idx -= 1;
  llvm_unreachable("Invalid index");
}

llvm::StringRef LengthsToRangesNode::getOutputName(unsigned idx) const {
  if (idx == 0) { return "Result"; }
  llvm_unreachable("Invalid index");
}

std::string LengthsToRangesNode::getDebugDesc() const {
  DescriptionBuilder db(getKindName());
  db.addParam("name", getName());
  if (hasPredicate()) db.addParam("Predicate", "Yes");
  db
    .addParam("Lengths", *(getLengths().getType()))
    .addParam("users", getNumUsers());
  db.addParam("Result", *(getResult().getType()));
  return db;
}

void LengthsToRangesNode::visit(Node *parent, NodeWalker *visitor) {
  if (!visitor->shouldVisit(parent, this)) { return; }
  visitor->pre(parent, this);
if (hasPredicate())
 getPredicate().getNode()->visit(this, visitor);
  getLengths().getNode()->visit(this, visitor);
  visitor->post(parent, this);
}

bool LengthsToRangesNode::isEqual(const LengthsToRangesNode &other) const {
  return true &&
      Lengths_ == other.Lengths_ &&
      predicate_ == other.predicate_ &&
      getType(0) == other.getType(0);
}

Node* LengthsToRangesNode::clone() const {
  return new LengthsToRangesNode(getName(), getResult().getType(), getLengths());
}

llvm::hash_code LengthsToRangesNode::getHash() const {
  return llvm::hash_combine(
      Lengths_);
}

unsigned LengthsRangeFillNode::getNumInputs() const {
  return 1;
}

std::string LengthsRangeFillNode::getInputName(unsigned idx) const {
  if (idx == 0) { return "Lengths"; }
  idx -= 1;
  llvm_unreachable("Invalid index");
}

NodeValue LengthsRangeFillNode::getNthInput(unsigned idx) {
  if (idx == 0) { return Lengths_; }
  idx -= 1;
  llvm_unreachable("Invalid index");
}

void LengthsRangeFillNode::setNthInput(unsigned idx, NodeValue val) {
  if (idx == 0) { Lengths_ = val; return; }
  idx -= 1;
  llvm_unreachable("Invalid index");
}

llvm::StringRef LengthsRangeFillNode::getOutputName(unsigned idx) const {
  if (idx == 0) { return "Result"; }
  llvm_unreachable("Invalid index");
}

std::string LengthsRangeFillNode::getDebugDesc() const {
  DescriptionBuilder db(getKindName());
  db.addParam("name", getName());
  if (hasPredicate()) db.addParam("Predicate", "Yes");
  db
    .addParam("Lengths", *(getLengths().getType()))
    .addParam("users", getNumUsers());
  db.addParam("Result", *(getResult().getType()));
  return db;
}

void LengthsRangeFillNode::visit(Node *parent, NodeWalker *visitor) {
  if (!visitor->shouldVisit(parent, this)) { return; }
  visitor->pre(parent, this);
if (hasPredicate())
 getPredicate().getNode()->visit(this, visitor);
  getLengths().getNode()->visit(this, visitor);
  visitor->post(parent, this);
}

bool LengthsRangeFillNode::isEqual(const LengthsRangeFillNode &other) const {
  return true &&
      Lengths_ == other.Lengths_ &&
      predicate_ == other.predicate_ &&
      getType(0) == other.getType(0);
}

Node* LengthsRangeFillNode::clone() const {
  return new LengthsRangeFillNode(getName(), getResult().getType(), getLengths());
}

llvm::hash_code LengthsRangeFillNode::getHash() const {
  return llvm::hash_combine(
      Lengths_);
}

unsigned SparseToDenseNode::getNumInputs() const {
  return 2;
}

std::string SparseToDenseNode::getInputName(unsigned idx) const {
  if (idx == 0) { return "Indices"; }
  if (idx == 1) { return "Values"; }
  idx -= 2;
  llvm_unreachable("Invalid index");
}

NodeValue SparseToDenseNode::getNthInput(unsigned idx) {
  if (idx == 0) { return Indices_; }
  if (idx == 1) { return Values_; }
  idx -= 2;
  llvm_unreachable("Invalid index");
}

void SparseToDenseNode::setNthInput(unsigned idx, NodeValue val) {
  if (idx == 0) { Indices_ = val; return; }
  if (idx == 1) { Values_ = val; return; }
  idx -= 2;
  llvm_unreachable("Invalid index");
}

llvm::StringRef SparseToDenseNode::getOutputName(unsigned idx) const {
  if (idx == 0) { return "Result"; }
  llvm_unreachable("Invalid index");
}

std::string SparseToDenseNode::getDebugDesc() const {
  DescriptionBuilder db(getKindName());
  db.addParam("name", getName());
  if (hasPredicate()) db.addParam("Predicate", "Yes");
  db
    .addParam("Indices", *(getIndices().getType()))
    .addParam("Values", *(getValues().getType()))
    .addParam("users", getNumUsers());
  db.addParam("Result", *(getResult().getType()));
  return db;
}

void SparseToDenseNode::visit(Node *parent, NodeWalker *visitor) {
  if (!visitor->shouldVisit(parent, this)) { return; }
  visitor->pre(parent, this);
if (hasPredicate())
 getPredicate().getNode()->visit(this, visitor);
  getIndices().getNode()->visit(this, visitor);
  getValues().getNode()->visit(this, visitor);
  visitor->post(parent, this);
}

bool SparseToDenseNode::isEqual(const SparseToDenseNode &other) const {
  return true &&
      Indices_ == other.Indices_ &&
      Values_ == other.Values_ &&
      predicate_ == other.predicate_ &&
      getType(0) == other.getType(0);
}

Node* SparseToDenseNode::clone() const {
  return new SparseToDenseNode(getName(), getResult().getType(), getIndices(), getValues());
}

llvm::hash_code SparseToDenseNode::getHash() const {
  return llvm::hash_combine(
      Indices_,
      Values_);
}

unsigned SparseToDenseMaskNode::getNumInputs() const {
  return 4;
}

std::string SparseToDenseMaskNode::getInputName(unsigned idx) const {
  if (idx == 0) { return "Indices"; }
  if (idx == 1) { return "Values"; }
  if (idx == 2) { return "DefaultValue"; }
  if (idx == 3) { return "Lengths"; }
  idx -= 4;
  llvm_unreachable("Invalid index");
}

NodeValue SparseToDenseMaskNode::getNthInput(unsigned idx) {
  if (idx == 0) { return Indices_; }
  if (idx == 1) { return Values_; }
  if (idx == 2) { return DefaultValue_; }
  if (idx == 3) { return Lengths_; }
  idx -= 4;
  llvm_unreachable("Invalid index");
}

void SparseToDenseMaskNode::setNthInput(unsigned idx, NodeValue val) {
  if (idx == 0) { Indices_ = val; return; }
  if (idx == 1) { Values_ = val; return; }
  if (idx == 2) { DefaultValue_ = val; return; }
  if (idx == 3) { Lengths_ = val; return; }
  idx -= 4;
  llvm_unreachable("Invalid index");
}

llvm::StringRef SparseToDenseMaskNode::getOutputName(unsigned idx) const {
  if (idx == 0) { return "Result"; }
  llvm_unreachable("Invalid index");
}

std::string SparseToDenseMaskNode::getDebugDesc() const {
  DescriptionBuilder db(getKindName());
  db.addParam("name", getName());
  if (hasPredicate()) db.addParam("Predicate", "Yes");
  db
    .addParam("Indices", *(getIndices().getType()))
    .addParam("Values", *(getValues().getType()))
    .addParam("DefaultValue", *(getDefaultValue().getType()))
    .addParam("Lengths", *(getLengths().getType()))
    .addParam("Mask", getMask())
    .addParam("users", getNumUsers());
  db.addParam("Result", *(getResult().getType()));
  return db;
}

void SparseToDenseMaskNode::visit(Node *parent, NodeWalker *visitor) {
  if (!visitor->shouldVisit(parent, this)) { return; }
  visitor->pre(parent, this);
if (hasPredicate())
 getPredicate().getNode()->visit(this, visitor);
  getIndices().getNode()->visit(this, visitor);
  getValues().getNode()->visit(this, visitor);
  getDefaultValue().getNode()->visit(this, visitor);
  getLengths().getNode()->visit(this, visitor);
  visitor->post(parent, this);
}

bool SparseToDenseMaskNode::isEqual(const SparseToDenseMaskNode &other) const {
  return true &&
      Indices_ == other.Indices_ &&
      Values_ == other.Values_ &&
      DefaultValue_ == other.DefaultValue_ &&
      Lengths_ == other.Lengths_ &&
      predicate_ == other.predicate_ &&
      Mask_ == other.Mask_ &&
      getType(0) == other.getType(0);
}

Node* SparseToDenseMaskNode::clone() const {
  return new SparseToDenseMaskNode(getName(), getResult().getType(), getIndices(), getValues(), getDefaultValue(), getLengths(), getMask());
}

llvm::hash_code SparseToDenseMaskNode::getHash() const {
  return llvm::hash_combine(
      llvm::hash_combine_range(Mask_.begin(), Mask_.end()),
      Indices_,
      Values_,
      DefaultValue_,
      Lengths_);
}

unsigned IsNaNNode::getNumInputs() const {
  return 1;
}

std::string IsNaNNode::getInputName(unsigned idx) const {
  if (idx == 0) { return "Input"; }
  idx -= 1;
  llvm_unreachable("Invalid index");
}

NodeValue IsNaNNode::getNthInput(unsigned idx) {
  if (idx == 0) { return Input_; }
  idx -= 1;
  llvm_unreachable("Invalid index");
}

void IsNaNNode::setNthInput(unsigned idx, NodeValue val) {
  if (idx == 0) { Input_ = val; return; }
  idx -= 1;
  llvm_unreachable("Invalid index");
}

llvm::StringRef IsNaNNode::getOutputName(unsigned idx) const {
  if (idx == 0) { return "Result"; }
  llvm_unreachable("Invalid index");
}

std::string IsNaNNode::getDebugDesc() const {
  DescriptionBuilder db(getKindName());
  db.addParam("name", getName());
  if (hasPredicate()) db.addParam("Predicate", "Yes");
  db
    .addParam("Input", *(getInput().getType()))
    .addParam("users", getNumUsers());
  db.addParam("Result", *(getResult().getType()));
  return db;
}

void IsNaNNode::visit(Node *parent, NodeWalker *visitor) {
  if (!visitor->shouldVisit(parent, this)) { return; }
  visitor->pre(parent, this);
if (hasPredicate())
 getPredicate().getNode()->visit(this, visitor);
  getInput().getNode()->visit(this, visitor);
  visitor->post(parent, this);
}

bool IsNaNNode::isEqual(const IsNaNNode &other) const {
  return true &&
      Input_ == other.Input_ &&
      predicate_ == other.predicate_ &&
      getType(0) == other.getType(0);
}

Node* IsNaNNode::clone() const {
  return new IsNaNNode(getName(), getResult().getType(), getInput());
}

llvm::hash_code IsNaNNode::getHash() const {
  return llvm::hash_combine(
      Input_);
}

unsigned ReplaceNaNNode::getNumInputs() const {
  return 1;
}

std::string ReplaceNaNNode::getInputName(unsigned idx) const {
  if (idx == 0) { return "Input"; }
  idx -= 1;
  llvm_unreachable("Invalid index");
}

NodeValue ReplaceNaNNode::getNthInput(unsigned idx) {
  if (idx == 0) { return Input_; }
  idx -= 1;
  llvm_unreachable("Invalid index");
}

void ReplaceNaNNode::setNthInput(unsigned idx, NodeValue val) {
  if (idx == 0) { Input_ = val; return; }
  idx -= 1;
  llvm_unreachable("Invalid index");
}

llvm::StringRef ReplaceNaNNode::getOutputName(unsigned idx) const {
  if (idx == 0) { return "Result"; }
  llvm_unreachable("Invalid index");
}

std::string ReplaceNaNNode::getDebugDesc() const {
  DescriptionBuilder db(getKindName());
  db.addParam("name", getName());
  if (hasPredicate()) db.addParam("Predicate", "Yes");
  db
    .addParam("Input", *(getInput().getType()))
    .addParam("Value", getValue())
    .addParam("users", getNumUsers());
  db.addParam("Result", *(getResult().getType()));
  return db;
}

void ReplaceNaNNode::visit(Node *parent, NodeWalker *visitor) {
  if (!visitor->shouldVisit(parent, this)) { return; }
  visitor->pre(parent, this);
if (hasPredicate())
 getPredicate().getNode()->visit(this, visitor);
  getInput().getNode()->visit(this, visitor);
  visitor->post(parent, this);
}

bool ReplaceNaNNode::isEqual(const ReplaceNaNNode &other) const {
  return true &&
      Input_ == other.Input_ &&
      predicate_ == other.predicate_ &&
      Value_ == other.Value_ &&
      getType(0) == other.getType(0);
}

Node* ReplaceNaNNode::clone() const {
  return new ReplaceNaNNode(getName(), getResult().getType(), getInput(), getValue());
}

llvm::hash_code ReplaceNaNNode::getHash() const {
  return llvm::hash_combine(
      toBinary(Value_),
      Input_);
}

unsigned ModuloNode::getNumInputs() const {
  return 1;
}

std::string ModuloNode::getInputName(unsigned idx) const {
  if (idx == 0) { return "Input"; }
  idx -= 1;
  llvm_unreachable("Invalid index");
}

NodeValue ModuloNode::getNthInput(unsigned idx) {
  if (idx == 0) { return Input_; }
  idx -= 1;
  llvm_unreachable("Invalid index");
}

void ModuloNode::setNthInput(unsigned idx, NodeValue val) {
  if (idx == 0) { Input_ = val; return; }
  idx -= 1;
  llvm_unreachable("Invalid index");
}

llvm::StringRef ModuloNode::getOutputName(unsigned idx) const {
  if (idx == 0) { return "Result"; }
  llvm_unreachable("Invalid index");
}

std::string ModuloNode::getDebugDesc() const {
  DescriptionBuilder db(getKindName());
  db.addParam("name", getName());
  if (hasPredicate()) db.addParam("Predicate", "Yes");
  db
    .addParam("Input", *(getInput().getType()))
    .addParam("Divisor", getDivisor())
    .addParam("SignFollowDivisor", getSignFollowDivisor())
    .addParam("users", getNumUsers());
  db.addParam("Result", *(getResult().getType()));
  return db;
}

void ModuloNode::visit(Node *parent, NodeWalker *visitor) {
  if (!visitor->shouldVisit(parent, this)) { return; }
  visitor->pre(parent, this);
if (hasPredicate())
 getPredicate().getNode()->visit(this, visitor);
  getInput().getNode()->visit(this, visitor);
  visitor->post(parent, this);
}

bool ModuloNode::isEqual(const ModuloNode &other) const {
  return true &&
      Input_ == other.Input_ &&
      predicate_ == other.predicate_ &&
      Divisor_ == other.Divisor_ &&
      SignFollowDivisor_ == other.SignFollowDivisor_ &&
      getType(0) == other.getType(0);
}

Node* ModuloNode::clone() const {
  return new ModuloNode(getName(), getResult().getType(), getInput(), getDivisor(), getSignFollowDivisor());
}

llvm::hash_code ModuloNode::getHash() const {
  return llvm::hash_combine(
      Divisor_,
      SignFollowDivisor_,
      Input_);
}

unsigned BatchedPairwiseDotProductNode::getNumInputs() const {
  return 0 + Inputs_.size();
}

std::string BatchedPairwiseDotProductNode::getInputName(unsigned idx) const {
  idx -= 0;
  if (idx < Inputs_.size()) { return "Inputs" + std::to_string(idx); }
  idx -= Inputs_.size();
  llvm_unreachable("Invalid index");
}

NodeValue BatchedPairwiseDotProductNode::getNthInput(unsigned idx) {
  idx -= 0;
  if (idx < Inputs_.size()) { return Inputs_[idx]; }
  idx -= Inputs_.size();
  llvm_unreachable("Invalid index");
}

void BatchedPairwiseDotProductNode::setNthInput(unsigned idx, NodeValue val) {
  idx -= 0;
  if (idx < Inputs_.size()) { Inputs_[idx] = val; return; }
  idx -= Inputs_.size();
  llvm_unreachable("Invalid index");
}

llvm::StringRef BatchedPairwiseDotProductNode::getOutputName(unsigned idx) const {
  if (idx == 0) { return "Result"; }
  llvm_unreachable("Invalid index");
}

std::string BatchedPairwiseDotProductNode::getDebugDesc() const {
  DescriptionBuilder db(getKindName());
  db.addParam("name", getName());
  if (hasPredicate()) db.addParam("Predicate", "Yes");
  db
    .addParam("users", getNumUsers());
  {
  unsigned mIndex = 0;
  for (const auto &II : getInputs()) {
    db.addParam("Inputs"+std::to_string(mIndex++), *II.getType());
  }
  }
  db.addParam("Result", *(getResult().getType()));
  return db;
}

void BatchedPairwiseDotProductNode::visit(Node *parent, NodeWalker *visitor) {
  if (!visitor->shouldVisit(parent, this)) { return; }
  visitor->pre(parent, this);
if (hasPredicate())
 getPredicate().getNode()->visit(this, visitor);
  for (auto &I : Inputs_) { I.getNode()->visit(this, visitor); }
  visitor->post(parent, this);
}

bool BatchedPairwiseDotProductNode::isEqual(const BatchedPairwiseDotProductNode &other) const {
  return true &&
      predicate_ == other.predicate_ &&
      Inputs_ == other.Inputs_ &&
      getType(0) == other.getType(0);
}

Node* BatchedPairwiseDotProductNode::clone() const {
  return new BatchedPairwiseDotProductNode(getName(), getResult().getType(), getInputs());
}

llvm::hash_code BatchedPairwiseDotProductNode::getHash() const {
  return llvm::hash_combine(
      llvm::hash_combine_range(Inputs_.begin(), Inputs_.end()));
}

unsigned BatchedPairwiseDotProductGradNode::getNumInputs() const {
  return 1 + OriginalInputs_.size();
}

std::string BatchedPairwiseDotProductGradNode::getInputName(unsigned idx) const {
  if (idx == 0) { return "OutputGrad"; }
  idx -= 1;
  if (idx < OriginalInputs_.size()) { return "OriginalInputs" + std::to_string(idx); }
  idx -= OriginalInputs_.size();
  llvm_unreachable("Invalid index");
}

NodeValue BatchedPairwiseDotProductGradNode::getNthInput(unsigned idx) {
  if (idx == 0) { return OutputGrad_; }
  idx -= 1;
  if (idx < OriginalInputs_.size()) { return OriginalInputs_[idx]; }
  idx -= OriginalInputs_.size();
  llvm_unreachable("Invalid index");
}

void BatchedPairwiseDotProductGradNode::setNthInput(unsigned idx, NodeValue val) {
  if (idx == 0) { OutputGrad_ = val; return; }
  idx -= 1;
  if (idx < OriginalInputs_.size()) { OriginalInputs_[idx] = val; return; }
  idx -= OriginalInputs_.size();
  llvm_unreachable("Invalid index");
}

llvm::StringRef BatchedPairwiseDotProductGradNode::getOutputName(unsigned idx) const {
  llvm_unreachable("Invalid index");
}

std::string BatchedPairwiseDotProductGradNode::getDebugDesc() const {
  DescriptionBuilder db(getKindName());
  db.addParam("name", getName());
  if (hasPredicate()) db.addParam("Predicate", "Yes");
  db
    .addParam("OutputGrad", *(getOutputGrad().getType()))
    .addParam("users", getNumUsers());
  {
  unsigned mIndex = 0;
  for (const auto &II : getOriginalInputs()) {
    db.addParam("OriginalInputs"+std::to_string(mIndex++), *II.getType());
  }
  }
  return db;
}

void BatchedPairwiseDotProductGradNode::visit(Node *parent, NodeWalker *visitor) {
  if (!visitor->shouldVisit(parent, this)) { return; }
  visitor->pre(parent, this);
if (hasPredicate())
 getPredicate().getNode()->visit(this, visitor);
  getOutputGrad().getNode()->visit(this, visitor);
  for (auto &I : OriginalInputs_) { I.getNode()->visit(this, visitor); }
  visitor->post(parent, this);
}

bool BatchedPairwiseDotProductGradNode::isEqual(const BatchedPairwiseDotProductGradNode &other) const {
  return true &&
      OutputGrad_ == other.OutputGrad_ &&
      predicate_ == other.predicate_ &&
      OriginalInputs_ == other.OriginalInputs_;
}

Node* BatchedPairwiseDotProductGradNode::clone() const {
  return new BatchedPairwiseDotProductGradNode(getName(), getOutputGrad(), getOriginalInputs());
}

llvm::hash_code BatchedPairwiseDotProductGradNode::getHash() const {
  return llvm::hash_combine(
      llvm::hash_combine_range(OriginalInputs_.begin(), OriginalInputs_.end()),
      OutputGrad_);
}

unsigned ReluGradNode::getNumInputs() const {
  return 3;
}

std::string ReluGradNode::getInputName(unsigned idx) const {
  if (idx == 0) { return "Input"; }
  if (idx == 1) { return "OriginalOutputForResult"; }
  if (idx == 2) { return "GradOfOriginalOutputNamedResult"; }
  idx -= 3;
  llvm_unreachable("Invalid index");
}

NodeValue ReluGradNode::getNthInput(unsigned idx) {
  if (idx == 0) { return Input_; }
  if (idx == 1) { return OriginalOutputForResult_; }
  if (idx == 2) { return GradOfOriginalOutputNamedResult_; }
  idx -= 3;
  llvm_unreachable("Invalid index");
}

void ReluGradNode::setNthInput(unsigned idx, NodeValue val) {
  if (idx == 0) { Input_ = val; return; }
  if (idx == 1) { OriginalOutputForResult_ = val; return; }
  if (idx == 2) { GradOfOriginalOutputNamedResult_ = val; return; }
  idx -= 3;
  llvm_unreachable("Invalid index");
}

llvm::StringRef ReluGradNode::getOutputName(unsigned idx) const {
  if (idx == 0) { return "GradOfInputNamedInput"; }
  llvm_unreachable("Invalid index");
}

std::string ReluGradNode::getDebugDesc() const {
  DescriptionBuilder db(getKindName());
  db.addParam("name", getName());
  if (hasPredicate()) db.addParam("Predicate", "Yes");
  db
    .addParam("Input", *(getInput().getType()))
    .addParam("OriginalOutputForResult", *(getOriginalOutputForResult().getType()))
    .addParam("GradOfOriginalOutputNamedResult", *(getGradOfOriginalOutputNamedResult().getType()))
    .addParam("users", getNumUsers());
  db.addParam("GradOfInputNamedInput", *(getGradOfInputNamedInput().getType()));
  return db;
}

void ReluGradNode::visit(Node *parent, NodeWalker *visitor) {
  if (!visitor->shouldVisit(parent, this)) { return; }
  visitor->pre(parent, this);
if (hasPredicate())
 getPredicate().getNode()->visit(this, visitor);
  getInput().getNode()->visit(this, visitor);
  getOriginalOutputForResult().getNode()->visit(this, visitor);
  getGradOfOriginalOutputNamedResult().getNode()->visit(this, visitor);
  visitor->post(parent, this);
}

bool ReluGradNode::isEqual(const ReluGradNode &other) const {
  return true &&
      Input_ == other.Input_ &&
      OriginalOutputForResult_ == other.OriginalOutputForResult_ &&
      GradOfOriginalOutputNamedResult_ == other.GradOfOriginalOutputNamedResult_ &&
      predicate_ == other.predicate_ &&
      getType(0) == other.getType(0);
}

Node* ReluGradNode::clone() const {
  return new ReluGradNode(getName(), getInput(), getOriginalOutputForResult(), getGradOfOriginalOutputNamedResult());
}

llvm::hash_code ReluGradNode::getHash() const {
  return llvm::hash_combine(
      Input_,
      OriginalOutputForResult_,
      GradOfOriginalOutputNamedResult_);
}

unsigned ReluNode::getNumInputs() const {
  return 1;
}

std::string ReluNode::getInputName(unsigned idx) const {
  if (idx == 0) { return "Input"; }
  idx -= 1;
  llvm_unreachable("Invalid index");
}

NodeValue ReluNode::getNthInput(unsigned idx) {
  if (idx == 0) { return Input_; }
  idx -= 1;
  llvm_unreachable("Invalid index");
}

void ReluNode::setNthInput(unsigned idx, NodeValue val) {
  if (idx == 0) { Input_ = val; return; }
  idx -= 1;
  llvm_unreachable("Invalid index");
}

llvm::StringRef ReluNode::getOutputName(unsigned idx) const {
  if (idx == 0) { return "Result"; }
  llvm_unreachable("Invalid index");
}

std::string ReluNode::getDebugDesc() const {
  DescriptionBuilder db(getKindName());
  db.addParam("name", getName());
  if (hasPredicate()) db.addParam("Predicate", "Yes");
  db
    .addParam("Input", *(getInput().getType()))
    .addParam("users", getNumUsers());
  db.addParam("Result", *(getResult().getType()));
  return db;
}

void ReluNode::visit(Node *parent, NodeWalker *visitor) {
  if (!visitor->shouldVisit(parent, this)) { return; }
  visitor->pre(parent, this);
if (hasPredicate())
 getPredicate().getNode()->visit(this, visitor);
  getInput().getNode()->visit(this, visitor);
  visitor->post(parent, this);
}

bool ReluNode::isEqual(const ReluNode &other) const {
  return true &&
      Input_ == other.Input_ &&
      predicate_ == other.predicate_ &&
      getType(0) == other.getType(0);
}

Node* ReluNode::clone() const {
  return new ReluNode(getName(), getResult().getType(), getInput());
}

llvm::hash_code ReluNode::getHash() const {
  return llvm::hash_combine(
      Input_);
}

ReluGradNode *ReluNode::getGrad(GraphGradMapper &builder) {
  auto *x = new ReluGradNode(getName().str() + "_grad", getInput(), getResult(), builder.getGradient(getResult()));
  builder.addGradient(getInput(), x->getGradOfInputNamedInput());
  return x;
}

unsigned ClipNode::getNumInputs() const {
  return 1;
}

std::string ClipNode::getInputName(unsigned idx) const {
  if (idx == 0) { return "Input"; }
  idx -= 1;
  llvm_unreachable("Invalid index");
}

NodeValue ClipNode::getNthInput(unsigned idx) {
  if (idx == 0) { return Input_; }
  idx -= 1;
  llvm_unreachable("Invalid index");
}

void ClipNode::setNthInput(unsigned idx, NodeValue val) {
  if (idx == 0) { Input_ = val; return; }
  idx -= 1;
  llvm_unreachable("Invalid index");
}

llvm::StringRef ClipNode::getOutputName(unsigned idx) const {
  if (idx == 0) { return "Result"; }
  llvm_unreachable("Invalid index");
}

std::string ClipNode::getDebugDesc() const {
  DescriptionBuilder db(getKindName());
  db.addParam("name", getName());
  if (hasPredicate()) db.addParam("Predicate", "Yes");
  db
    .addParam("Input", *(getInput().getType()))
    .addParam("Min", getMin())
    .addParam("Max", getMax())
    .addParam("users", getNumUsers());
  db.addParam("Result", *(getResult().getType()));
  return db;
}

void ClipNode::visit(Node *parent, NodeWalker *visitor) {
  if (!visitor->shouldVisit(parent, this)) { return; }
  visitor->pre(parent, this);
if (hasPredicate())
 getPredicate().getNode()->visit(this, visitor);
  getInput().getNode()->visit(this, visitor);
  visitor->post(parent, this);
}

bool ClipNode::isEqual(const ClipNode &other) const {
  return true &&
      Input_ == other.Input_ &&
      predicate_ == other.predicate_ &&
      Min_ == other.Min_ &&
      Max_ == other.Max_ &&
      getType(0) == other.getType(0);
}

Node* ClipNode::clone() const {
  return new ClipNode(getName(), getResult().getType(), getInput(), getMin(), getMax());
}

llvm::hash_code ClipNode::getHash() const {
  return llvm::hash_combine(
      toBinary(Min_),
      toBinary(Max_),
      Input_);
}

unsigned PReluNode::getNumInputs() const {
  return 2;
}

std::string PReluNode::getInputName(unsigned idx) const {
  if (idx == 0) { return "Input"; }
  if (idx == 1) { return "Slope"; }
  idx -= 2;
  llvm_unreachable("Invalid index");
}

NodeValue PReluNode::getNthInput(unsigned idx) {
  if (idx == 0) { return Input_; }
  if (idx == 1) { return Slope_; }
  idx -= 2;
  llvm_unreachable("Invalid index");
}

void PReluNode::setNthInput(unsigned idx, NodeValue val) {
  if (idx == 0) { Input_ = val; return; }
  if (idx == 1) { Slope_ = val; return; }
  idx -= 2;
  llvm_unreachable("Invalid index");
}

llvm::StringRef PReluNode::getOutputName(unsigned idx) const {
  if (idx == 0) { return "Result"; }
  llvm_unreachable("Invalid index");
}

std::string PReluNode::getDebugDesc() const {
  DescriptionBuilder db(getKindName());
  db.addParam("name", getName());
  if (hasPredicate()) db.addParam("Predicate", "Yes");
  db
    .addParam("Input", *(getInput().getType()))
    .addParam("Slope", *(getSlope().getType()))
    .addParam("users", getNumUsers());
  db.addParam("Result", *(getResult().getType()));
  return db;
}

void PReluNode::visit(Node *parent, NodeWalker *visitor) {
  if (!visitor->shouldVisit(parent, this)) { return; }
  visitor->pre(parent, this);
if (hasPredicate())
 getPredicate().getNode()->visit(this, visitor);
  getInput().getNode()->visit(this, visitor);
  getSlope().getNode()->visit(this, visitor);
  visitor->post(parent, this);
}

bool PReluNode::isEqual(const PReluNode &other) const {
  return true &&
      Input_ == other.Input_ &&
      Slope_ == other.Slope_ &&
      predicate_ == other.predicate_ &&
      getType(0) == other.getType(0);
}

Node* PReluNode::clone() const {
  return new PReluNode(getName(), getResult().getType(), getInput(), getSlope());
}

llvm::hash_code PReluNode::getHash() const {
  return llvm::hash_combine(
      Input_,
      Slope_);
}

unsigned SigmoidGradNode::getNumInputs() const {
  return 3;
}

std::string SigmoidGradNode::getInputName(unsigned idx) const {
  if (idx == 0) { return "Input"; }
  if (idx == 1) { return "OriginalOutputForResult"; }
  if (idx == 2) { return "GradOfOriginalOutputNamedResult"; }
  idx -= 3;
  llvm_unreachable("Invalid index");
}

NodeValue SigmoidGradNode::getNthInput(unsigned idx) {
  if (idx == 0) { return Input_; }
  if (idx == 1) { return OriginalOutputForResult_; }
  if (idx == 2) { return GradOfOriginalOutputNamedResult_; }
  idx -= 3;
  llvm_unreachable("Invalid index");
}

void SigmoidGradNode::setNthInput(unsigned idx, NodeValue val) {
  if (idx == 0) { Input_ = val; return; }
  if (idx == 1) { OriginalOutputForResult_ = val; return; }
  if (idx == 2) { GradOfOriginalOutputNamedResult_ = val; return; }
  idx -= 3;
  llvm_unreachable("Invalid index");
}

llvm::StringRef SigmoidGradNode::getOutputName(unsigned idx) const {
  if (idx == 0) { return "GradOfInputNamedInput"; }
  llvm_unreachable("Invalid index");
}

std::string SigmoidGradNode::getDebugDesc() const {
  DescriptionBuilder db(getKindName());
  db.addParam("name", getName());
  if (hasPredicate()) db.addParam("Predicate", "Yes");
  db
    .addParam("Input", *(getInput().getType()))
    .addParam("OriginalOutputForResult", *(getOriginalOutputForResult().getType()))
    .addParam("GradOfOriginalOutputNamedResult", *(getGradOfOriginalOutputNamedResult().getType()))
    .addParam("users", getNumUsers());
  db.addParam("GradOfInputNamedInput", *(getGradOfInputNamedInput().getType()));
  return db;
}

void SigmoidGradNode::visit(Node *parent, NodeWalker *visitor) {
  if (!visitor->shouldVisit(parent, this)) { return; }
  visitor->pre(parent, this);
if (hasPredicate())
 getPredicate().getNode()->visit(this, visitor);
  getInput().getNode()->visit(this, visitor);
  getOriginalOutputForResult().getNode()->visit(this, visitor);
  getGradOfOriginalOutputNamedResult().getNode()->visit(this, visitor);
  visitor->post(parent, this);
}

bool SigmoidGradNode::isEqual(const SigmoidGradNode &other) const {
  return true &&
      Input_ == other.Input_ &&
      OriginalOutputForResult_ == other.OriginalOutputForResult_ &&
      GradOfOriginalOutputNamedResult_ == other.GradOfOriginalOutputNamedResult_ &&
      predicate_ == other.predicate_ &&
      getType(0) == other.getType(0);
}

Node* SigmoidGradNode::clone() const {
  return new SigmoidGradNode(getName(), getInput(), getOriginalOutputForResult(), getGradOfOriginalOutputNamedResult());
}

llvm::hash_code SigmoidGradNode::getHash() const {
  return llvm::hash_combine(
      Input_,
      OriginalOutputForResult_,
      GradOfOriginalOutputNamedResult_);
}

unsigned SigmoidNode::getNumInputs() const {
  return 1;
}

std::string SigmoidNode::getInputName(unsigned idx) const {
  if (idx == 0) { return "Input"; }
  idx -= 1;
  llvm_unreachable("Invalid index");
}

NodeValue SigmoidNode::getNthInput(unsigned idx) {
  if (idx == 0) { return Input_; }
  idx -= 1;
  llvm_unreachable("Invalid index");
}

void SigmoidNode::setNthInput(unsigned idx, NodeValue val) {
  if (idx == 0) { Input_ = val; return; }
  idx -= 1;
  llvm_unreachable("Invalid index");
}

llvm::StringRef SigmoidNode::getOutputName(unsigned idx) const {
  if (idx == 0) { return "Result"; }
  llvm_unreachable("Invalid index");
}

std::string SigmoidNode::getDebugDesc() const {
  DescriptionBuilder db(getKindName());
  db.addParam("name", getName());
  if (hasPredicate()) db.addParam("Predicate", "Yes");
  db
    .addParam("Input", *(getInput().getType()))
    .addParam("users", getNumUsers());
  db.addParam("Result", *(getResult().getType()));
  return db;
}

void SigmoidNode::visit(Node *parent, NodeWalker *visitor) {
  if (!visitor->shouldVisit(parent, this)) { return; }
  visitor->pre(parent, this);
if (hasPredicate())
 getPredicate().getNode()->visit(this, visitor);
  getInput().getNode()->visit(this, visitor);
  visitor->post(parent, this);
}

bool SigmoidNode::isEqual(const SigmoidNode &other) const {
  return true &&
      Input_ == other.Input_ &&
      predicate_ == other.predicate_ &&
      getType(0) == other.getType(0);
}

Node* SigmoidNode::clone() const {
  return new SigmoidNode(getName(), getResult().getType(), getInput());
}

llvm::hash_code SigmoidNode::getHash() const {
  return llvm::hash_combine(
      Input_);
}

SigmoidGradNode *SigmoidNode::getGrad(GraphGradMapper &builder) {
  auto *x = new SigmoidGradNode(getName().str() + "_grad", getInput(), getResult(), builder.getGradient(getResult()));
  builder.addGradient(getInput(), x->getGradOfInputNamedInput());
  return x;
}

unsigned SwishNode::getNumInputs() const {
  return 1;
}

std::string SwishNode::getInputName(unsigned idx) const {
  if (idx == 0) { return "Input"; }
  idx -= 1;
  llvm_unreachable("Invalid index");
}

NodeValue SwishNode::getNthInput(unsigned idx) {
  if (idx == 0) { return Input_; }
  idx -= 1;
  llvm_unreachable("Invalid index");
}

void SwishNode::setNthInput(unsigned idx, NodeValue val) {
  if (idx == 0) { Input_ = val; return; }
  idx -= 1;
  llvm_unreachable("Invalid index");
}

llvm::StringRef SwishNode::getOutputName(unsigned idx) const {
  if (idx == 0) { return "Result"; }
  llvm_unreachable("Invalid index");
}

std::string SwishNode::getDebugDesc() const {
  DescriptionBuilder db(getKindName());
  db.addParam("name", getName());
  if (hasPredicate()) db.addParam("Predicate", "Yes");
  db
    .addParam("Input", *(getInput().getType()))
    .addParam("users", getNumUsers());
  db.addParam("Result", *(getResult().getType()));
  return db;
}

void SwishNode::visit(Node *parent, NodeWalker *visitor) {
  if (!visitor->shouldVisit(parent, this)) { return; }
  visitor->pre(parent, this);
if (hasPredicate())
 getPredicate().getNode()->visit(this, visitor);
  getInput().getNode()->visit(this, visitor);
  visitor->post(parent, this);
}

bool SwishNode::isEqual(const SwishNode &other) const {
  return true &&
      Input_ == other.Input_ &&
      predicate_ == other.predicate_ &&
      getType(0) == other.getType(0);
}

Node* SwishNode::clone() const {
  return new SwishNode(getName(), getResult().getType(), getInput());
}

llvm::hash_code SwishNode::getHash() const {
  return llvm::hash_combine(
      Input_);
}

unsigned TanhGradNode::getNumInputs() const {
  return 3;
}

std::string TanhGradNode::getInputName(unsigned idx) const {
  if (idx == 0) { return "Input"; }
  if (idx == 1) { return "OriginalOutputForResult"; }
  if (idx == 2) { return "GradOfOriginalOutputNamedResult"; }
  idx -= 3;
  llvm_unreachable("Invalid index");
}

NodeValue TanhGradNode::getNthInput(unsigned idx) {
  if (idx == 0) { return Input_; }
  if (idx == 1) { return OriginalOutputForResult_; }
  if (idx == 2) { return GradOfOriginalOutputNamedResult_; }
  idx -= 3;
  llvm_unreachable("Invalid index");
}

void TanhGradNode::setNthInput(unsigned idx, NodeValue val) {
  if (idx == 0) { Input_ = val; return; }
  if (idx == 1) { OriginalOutputForResult_ = val; return; }
  if (idx == 2) { GradOfOriginalOutputNamedResult_ = val; return; }
  idx -= 3;
  llvm_unreachable("Invalid index");
}

llvm::StringRef TanhGradNode::getOutputName(unsigned idx) const {
  if (idx == 0) { return "GradOfInputNamedInput"; }
  llvm_unreachable("Invalid index");
}

std::string TanhGradNode::getDebugDesc() const {
  DescriptionBuilder db(getKindName());
  db.addParam("name", getName());
  if (hasPredicate()) db.addParam("Predicate", "Yes");
  db
    .addParam("Input", *(getInput().getType()))
    .addParam("OriginalOutputForResult", *(getOriginalOutputForResult().getType()))
    .addParam("GradOfOriginalOutputNamedResult", *(getGradOfOriginalOutputNamedResult().getType()))
    .addParam("users", getNumUsers());
  db.addParam("GradOfInputNamedInput", *(getGradOfInputNamedInput().getType()));
  return db;
}

void TanhGradNode::visit(Node *parent, NodeWalker *visitor) {
  if (!visitor->shouldVisit(parent, this)) { return; }
  visitor->pre(parent, this);
if (hasPredicate())
 getPredicate().getNode()->visit(this, visitor);
  getInput().getNode()->visit(this, visitor);
  getOriginalOutputForResult().getNode()->visit(this, visitor);
  getGradOfOriginalOutputNamedResult().getNode()->visit(this, visitor);
  visitor->post(parent, this);
}

bool TanhGradNode::isEqual(const TanhGradNode &other) const {
  return true &&
      Input_ == other.Input_ &&
      OriginalOutputForResult_ == other.OriginalOutputForResult_ &&
      GradOfOriginalOutputNamedResult_ == other.GradOfOriginalOutputNamedResult_ &&
      predicate_ == other.predicate_ &&
      getType(0) == other.getType(0);
}

Node* TanhGradNode::clone() const {
  return new TanhGradNode(getName(), getInput(), getOriginalOutputForResult(), getGradOfOriginalOutputNamedResult());
}

llvm::hash_code TanhGradNode::getHash() const {
  return llvm::hash_combine(
      Input_,
      OriginalOutputForResult_,
      GradOfOriginalOutputNamedResult_);
}

unsigned TanhNode::getNumInputs() const {
  return 1;
}

std::string TanhNode::getInputName(unsigned idx) const {
  if (idx == 0) { return "Input"; }
  idx -= 1;
  llvm_unreachable("Invalid index");
}

NodeValue TanhNode::getNthInput(unsigned idx) {
  if (idx == 0) { return Input_; }
  idx -= 1;
  llvm_unreachable("Invalid index");
}

void TanhNode::setNthInput(unsigned idx, NodeValue val) {
  if (idx == 0) { Input_ = val; return; }
  idx -= 1;
  llvm_unreachable("Invalid index");
}

llvm::StringRef TanhNode::getOutputName(unsigned idx) const {
  if (idx == 0) { return "Result"; }
  llvm_unreachable("Invalid index");
}

std::string TanhNode::getDebugDesc() const {
  DescriptionBuilder db(getKindName());
  db.addParam("name", getName());
  if (hasPredicate()) db.addParam("Predicate", "Yes");
  db
    .addParam("Input", *(getInput().getType()))
    .addParam("users", getNumUsers());
  db.addParam("Result", *(getResult().getType()));
  return db;
}

void TanhNode::visit(Node *parent, NodeWalker *visitor) {
  if (!visitor->shouldVisit(parent, this)) { return; }
  visitor->pre(parent, this);
if (hasPredicate())
 getPredicate().getNode()->visit(this, visitor);
  getInput().getNode()->visit(this, visitor);
  visitor->post(parent, this);
}

bool TanhNode::isEqual(const TanhNode &other) const {
  return true &&
      Input_ == other.Input_ &&
      predicate_ == other.predicate_ &&
      getType(0) == other.getType(0);
}

Node* TanhNode::clone() const {
  return new TanhNode(getName(), getResult().getType(), getInput());
}

llvm::hash_code TanhNode::getHash() const {
  return llvm::hash_combine(
      Input_);
}

TanhGradNode *TanhNode::getGrad(GraphGradMapper &builder) {
  auto *x = new TanhGradNode(getName().str() + "_grad", getInput(), getResult(), builder.getGradient(getResult()));
  builder.addGradient(getInput(), x->getGradOfInputNamedInput());
  return x;
}

unsigned ReshapeNode::getNumInputs() const {
  return 1;
}

std::string ReshapeNode::getInputName(unsigned idx) const {
  if (idx == 0) { return "Input"; }
  idx -= 1;
  llvm_unreachable("Invalid index");
}

NodeValue ReshapeNode::getNthInput(unsigned idx) {
  if (idx == 0) { return Input_; }
  idx -= 1;
  llvm_unreachable("Invalid index");
}

void ReshapeNode::setNthInput(unsigned idx, NodeValue val) {
  if (idx == 0) { Input_ = val; return; }
  idx -= 1;
  llvm_unreachable("Invalid index");
}

llvm::StringRef ReshapeNode::getOutputName(unsigned idx) const {
  if (idx == 0) { return "Result"; }
  llvm_unreachable("Invalid index");
}

std::string ReshapeNode::getDebugDesc() const {
  DescriptionBuilder db(getKindName());
  db.addParam("name", getName());
  if (hasPredicate()) db.addParam("Predicate", "Yes");
  db
    .addParam("Input", *(getInput().getType()))
    .addParam("Dims", getDims())
    .addParam("Layout", getLayout())
    .addParam("users", getNumUsers());
  db.addParam("Result", *(getResult().getType()));
  return db;
}

void ReshapeNode::visit(Node *parent, NodeWalker *visitor) {
  if (!visitor->shouldVisit(parent, this)) { return; }
  visitor->pre(parent, this);
if (hasPredicate())
 getPredicate().getNode()->visit(this, visitor);
  getInput().getNode()->visit(this, visitor);
  visitor->post(parent, this);
}

bool ReshapeNode::isEqual(const ReshapeNode &other) const {
  return true &&
      Input_ == other.Input_ &&
      predicate_ == other.predicate_ &&
      Dims_ == other.Dims_ &&
      Layout_ == other.Layout_ &&
      getType(0) == other.getType(0);
}

Node* ReshapeNode::clone() const {
  return new ReshapeNode(getName(), getResult().getType(), getInput(), getDims(), getLayout());
}

llvm::hash_code ReshapeNode::getHash() const {
  return llvm::hash_combine(
      llvm::hash_combine_range(Dims_.begin(), Dims_.end()),
      Layout_,
      Input_);
}

unsigned TransposeNode::getNumInputs() const {
  return 1;
}

std::string TransposeNode::getInputName(unsigned idx) const {
  if (idx == 0) { return "Input"; }
  idx -= 1;
  llvm_unreachable("Invalid index");
}

NodeValue TransposeNode::getNthInput(unsigned idx) {
  if (idx == 0) { return Input_; }
  idx -= 1;
  llvm_unreachable("Invalid index");
}

void TransposeNode::setNthInput(unsigned idx, NodeValue val) {
  if (idx == 0) { Input_ = val; return; }
  idx -= 1;
  llvm_unreachable("Invalid index");
}

llvm::StringRef TransposeNode::getOutputName(unsigned idx) const {
  if (idx == 0) { return "Result"; }
  llvm_unreachable("Invalid index");
}

std::string TransposeNode::getDebugDesc() const {
  DescriptionBuilder db(getKindName());
  db.addParam("name", getName());
  if (hasPredicate()) db.addParam("Predicate", "Yes");
  db
    .addParam("Input", *(getInput().getType()))
    .addParam("Shuffle", getShuffle())
    .addParam("Layout", getLayout())
    .addParam("users", getNumUsers());
  db.addParam("Result", *(getResult().getType()));
  return db;
}

void TransposeNode::visit(Node *parent, NodeWalker *visitor) {
  if (!visitor->shouldVisit(parent, this)) { return; }
  visitor->pre(parent, this);
if (hasPredicate())
 getPredicate().getNode()->visit(this, visitor);
  getInput().getNode()->visit(this, visitor);
  visitor->post(parent, this);
}

bool TransposeNode::isEqual(const TransposeNode &other) const {
  return true &&
      Input_ == other.Input_ &&
      predicate_ == other.predicate_ &&
      Shuffle_ == other.Shuffle_ &&
      Layout_ == other.Layout_ &&
      getType(0) == other.getType(0);
}

Node* TransposeNode::clone() const {
  return new TransposeNode(getName(), getResult().getType(), getInput(), getShuffle(), getLayout());
}

llvm::hash_code TransposeNode::getHash() const {
  return llvm::hash_combine(
      llvm::hash_combine_range(Shuffle_.begin(), Shuffle_.end()),
      Layout_,
      Input_);
}

unsigned ConcatNode::getNumInputs() const {
  return 0 + Inputs_.size();
}

std::string ConcatNode::getInputName(unsigned idx) const {
  idx -= 0;
  if (idx < Inputs_.size()) { return "Inputs" + std::to_string(idx); }
  idx -= Inputs_.size();
  llvm_unreachable("Invalid index");
}

NodeValue ConcatNode::getNthInput(unsigned idx) {
  idx -= 0;
  if (idx < Inputs_.size()) { return Inputs_[idx]; }
  idx -= Inputs_.size();
  llvm_unreachable("Invalid index");
}

void ConcatNode::setNthInput(unsigned idx, NodeValue val) {
  idx -= 0;
  if (idx < Inputs_.size()) { Inputs_[idx] = val; return; }
  idx -= Inputs_.size();
  llvm_unreachable("Invalid index");
}

llvm::StringRef ConcatNode::getOutputName(unsigned idx) const {
  if (idx == 0) { return "Result"; }
  llvm_unreachable("Invalid index");
}

std::string ConcatNode::getDebugDesc() const {
  DescriptionBuilder db(getKindName());
  db.addParam("name", getName());
  if (hasPredicate()) db.addParam("Predicate", "Yes");
  db
    .addParam("Dim", getDim())
    .addParam("users", getNumUsers());
  {
  unsigned mIndex = 0;
  for (const auto &II : getInputs()) {
    db.addParam("Inputs"+std::to_string(mIndex++), *II.getType());
  }
  }
  db.addParam("Result", *(getResult().getType()));
  return db;
}

void ConcatNode::visit(Node *parent, NodeWalker *visitor) {
  if (!visitor->shouldVisit(parent, this)) { return; }
  visitor->pre(parent, this);
if (hasPredicate())
 getPredicate().getNode()->visit(this, visitor);
  for (auto &I : Inputs_) { I.getNode()->visit(this, visitor); }
  visitor->post(parent, this);
}

bool ConcatNode::isEqual(const ConcatNode &other) const {
  return true &&
      predicate_ == other.predicate_ &&
      Inputs_ == other.Inputs_ &&
      Dim_ == other.Dim_ &&
      getType(0) == other.getType(0);
}

Node* ConcatNode::clone() const {
  return new ConcatNode(getName(), getResult().getType(), getInputs(), getDim());
}

llvm::hash_code ConcatNode::getHash() const {
  return llvm::hash_combine(
      llvm::hash_combine_range(Inputs_.begin(), Inputs_.end()),
      Dim_);
}

unsigned SliceNode::getNumInputs() const {
  return 1;
}

std::string SliceNode::getInputName(unsigned idx) const {
  if (idx == 0) { return "Input"; }
  idx -= 1;
  llvm_unreachable("Invalid index");
}

NodeValue SliceNode::getNthInput(unsigned idx) {
  if (idx == 0) { return Input_; }
  idx -= 1;
  llvm_unreachable("Invalid index");
}

void SliceNode::setNthInput(unsigned idx, NodeValue val) {
  if (idx == 0) { Input_ = val; return; }
  idx -= 1;
  llvm_unreachable("Invalid index");
}

llvm::StringRef SliceNode::getOutputName(unsigned idx) const {
  if (idx == 0) { return "Result"; }
  llvm_unreachable("Invalid index");
}

std::string SliceNode::getDebugDesc() const {
  DescriptionBuilder db(getKindName());
  db.addParam("name", getName());
  if (hasPredicate()) db.addParam("Predicate", "Yes");
  db
    .addParam("Input", *(getInput().getType()))
    .addParam("Start", getStart())
    .addParam("users", getNumUsers());
  db.addParam("Result", *(getResult().getType()));
  return db;
}

void SliceNode::visit(Node *parent, NodeWalker *visitor) {
  if (!visitor->shouldVisit(parent, this)) { return; }
  visitor->pre(parent, this);
if (hasPredicate())
 getPredicate().getNode()->visit(this, visitor);
  getInput().getNode()->visit(this, visitor);
  visitor->post(parent, this);
}

bool SliceNode::isEqual(const SliceNode &other) const {
  return true &&
      Input_ == other.Input_ &&
      predicate_ == other.predicate_ &&
      Start_ == other.Start_ &&
      getType(0) == other.getType(0);
}

Node* SliceNode::clone() const {
  return new SliceNode(getName(), getResult().getType(), getInput(), getStart());
}

llvm::hash_code SliceNode::getHash() const {
  return llvm::hash_combine(
      llvm::hash_combine_range(Start_.begin(), Start_.end()),
      Input_);
}

unsigned InsertTensorNode::getNumInputs() const {
  return 2;
}

std::string InsertTensorNode::getInputName(unsigned idx) const {
  if (idx == 0) { return "Big"; }
  if (idx == 1) { return "Small"; }
  idx -= 2;
  llvm_unreachable("Invalid index");
}

NodeValue InsertTensorNode::getNthInput(unsigned idx) {
  if (idx == 0) { return Big_; }
  if (idx == 1) { return Small_; }
  idx -= 2;
  llvm_unreachable("Invalid index");
}

void InsertTensorNode::setNthInput(unsigned idx, NodeValue val) {
  if (idx == 0) { Big_ = val; return; }
  if (idx == 1) { Small_ = val; return; }
  idx -= 2;
  llvm_unreachable("Invalid index");
}

llvm::StringRef InsertTensorNode::getOutputName(unsigned idx) const {
  if (idx == 0) { return "Result"; }
  llvm_unreachable("Invalid index");
}

std::string InsertTensorNode::getDebugDesc() const {
  DescriptionBuilder db(getKindName());
  db.addParam("name", getName());
  if (hasPredicate()) db.addParam("Predicate", "Yes");
  db
    .addParam("Big", *(getBig().getType()))
    .addParam("Small", *(getSmall().getType()))
    .addParam("Start", getStart())
    .addParam("Count", getCount())
    .addParam("Axis", getAxis())
    .addParam("users", getNumUsers());
  db.addParam("Result", *(getResult().getType()));
  return db;
}

void InsertTensorNode::visit(Node *parent, NodeWalker *visitor) {
  if (!visitor->shouldVisit(parent, this)) { return; }
  visitor->pre(parent, this);
if (hasPredicate())
 getPredicate().getNode()->visit(this, visitor);
  getBig().getNode()->visit(this, visitor);
  getSmall().getNode()->visit(this, visitor);
  visitor->post(parent, this);
}

bool InsertTensorNode::isEqual(const InsertTensorNode &other) const {
  return true &&
      Big_ == other.Big_ &&
      Small_ == other.Small_ &&
      predicate_ == other.predicate_ &&
      Start_ == other.Start_ &&
      Count_ == other.Count_ &&
      Axis_ == other.Axis_ &&
      getType(0) == other.getType(0);
}

Node* InsertTensorNode::clone() const {
  return new InsertTensorNode(getName(), getBig(), getSmall(), getStart(), getCount(), getAxis());
}

llvm::hash_code InsertTensorNode::getHash() const {
  return llvm::hash_combine(
      llvm::hash_combine_range(Start_.begin(), Start_.end()),
      Count_,
      Axis_,
      Big_,
      Small_);
}

unsigned GatherNode::getNumInputs() const {
  return 2;
}

std::string GatherNode::getInputName(unsigned idx) const {
  if (idx == 0) { return "Data"; }
  if (idx == 1) { return "Indices"; }
  idx -= 2;
  llvm_unreachable("Invalid index");
}

NodeValue GatherNode::getNthInput(unsigned idx) {
  if (idx == 0) { return Data_; }
  if (idx == 1) { return Indices_; }
  idx -= 2;
  llvm_unreachable("Invalid index");
}

void GatherNode::setNthInput(unsigned idx, NodeValue val) {
  if (idx == 0) { Data_ = val; return; }
  if (idx == 1) { Indices_ = val; return; }
  idx -= 2;
  llvm_unreachable("Invalid index");
}

llvm::StringRef GatherNode::getOutputName(unsigned idx) const {
  if (idx == 0) { return "Result"; }
  llvm_unreachable("Invalid index");
}

std::string GatherNode::getDebugDesc() const {
  DescriptionBuilder db(getKindName());
  db.addParam("name", getName());
  if (hasPredicate()) db.addParam("Predicate", "Yes");
  db
    .addParam("Data", *(getData().getType()))
    .addParam("Indices", *(getIndices().getType()))
    .addParam("BatchDims", getBatchDims())
    .addParam("users", getNumUsers());
  db.addParam("Result", *(getResult().getType()));
  return db;
}

void GatherNode::visit(Node *parent, NodeWalker *visitor) {
  if (!visitor->shouldVisit(parent, this)) { return; }
  visitor->pre(parent, this);
if (hasPredicate())
 getPredicate().getNode()->visit(this, visitor);
  getData().getNode()->visit(this, visitor);
  getIndices().getNode()->visit(this, visitor);
  visitor->post(parent, this);
}

bool GatherNode::isEqual(const GatherNode &other) const {
  return true &&
      Data_ == other.Data_ &&
      Indices_ == other.Indices_ &&
      predicate_ == other.predicate_ &&
      BatchDims_ == other.BatchDims_ &&
      getType(0) == other.getType(0);
}

Node* GatherNode::clone() const {
  return new GatherNode(getName(), getResult().getType(), getData(), getIndices(), getBatchDims());
}

llvm::hash_code GatherNode::getHash() const {
  return llvm::hash_combine(
      BatchDims_,
      Data_,
      Indices_);
}

unsigned GatherRangesNode::getNumInputs() const {
  return 2;
}

std::string GatherRangesNode::getInputName(unsigned idx) const {
  if (idx == 0) { return "Data"; }
  if (idx == 1) { return "Ranges"; }
  idx -= 2;
  llvm_unreachable("Invalid index");
}

NodeValue GatherRangesNode::getNthInput(unsigned idx) {
  if (idx == 0) { return Data_; }
  if (idx == 1) { return Ranges_; }
  idx -= 2;
  llvm_unreachable("Invalid index");
}

void GatherRangesNode::setNthInput(unsigned idx, NodeValue val) {
  if (idx == 0) { Data_ = val; return; }
  if (idx == 1) { Ranges_ = val; return; }
  idx -= 2;
  llvm_unreachable("Invalid index");
}

llvm::StringRef GatherRangesNode::getOutputName(unsigned idx) const {
  if (idx == 0) { return "Output"; }
  if (idx == 1) { return "Lengths"; }
  llvm_unreachable("Invalid index");
}

std::string GatherRangesNode::getDebugDesc() const {
  DescriptionBuilder db(getKindName());
  db.addParam("name", getName());
  if (hasPredicate()) db.addParam("Predicate", "Yes");
  db
    .addParam("Data", *(getData().getType()))
    .addParam("Ranges", *(getRanges().getType()))
    .addParam("users", getNumUsers());
  db.addParam("Output", *(getOutput().getType()));
  db.addParam("Lengths", *(getLengths().getType()));
  return db;
}

void GatherRangesNode::visit(Node *parent, NodeWalker *visitor) {
  if (!visitor->shouldVisit(parent, this)) { return; }
  visitor->pre(parent, this);
if (hasPredicate())
 getPredicate().getNode()->visit(this, visitor);
  getData().getNode()->visit(this, visitor);
  getRanges().getNode()->visit(this, visitor);
  visitor->post(parent, this);
}

bool GatherRangesNode::isEqual(const GatherRangesNode &other) const {
  return true &&
      Data_ == other.Data_ &&
      Ranges_ == other.Ranges_ &&
      predicate_ == other.predicate_ &&
      getType(0) == other.getType(0) &&
      getType(1) == other.getType(1);
}

Node* GatherRangesNode::clone() const {
  return new GatherRangesNode(getName(), getOutput().getType(), getLengths().getType(), getData(), getRanges());
}

llvm::hash_code GatherRangesNode::getHash() const {
  return llvm::hash_combine(
      Data_,
      Ranges_);
}

unsigned ScatterDataNode::getNumInputs() const {
  return 3;
}

std::string ScatterDataNode::getInputName(unsigned idx) const {
  if (idx == 0) { return "Data"; }
  if (idx == 1) { return "Indices"; }
  if (idx == 2) { return "Slices"; }
  idx -= 3;
  llvm_unreachable("Invalid index");
}

NodeValue ScatterDataNode::getNthInput(unsigned idx) {
  if (idx == 0) { return Data_; }
  if (idx == 1) { return Indices_; }
  if (idx == 2) { return Slices_; }
  idx -= 3;
  llvm_unreachable("Invalid index");
}

void ScatterDataNode::setNthInput(unsigned idx, NodeValue val) {
  if (idx == 0) { Data_ = val; return; }
  if (idx == 1) { Indices_ = val; return; }
  if (idx == 2) { Slices_ = val; return; }
  idx -= 3;
  llvm_unreachable("Invalid index");
}

llvm::StringRef ScatterDataNode::getOutputName(unsigned idx) const {
  if (idx == 0) { return "Result"; }
  llvm_unreachable("Invalid index");
}

std::string ScatterDataNode::getDebugDesc() const {
  DescriptionBuilder db(getKindName());
  db.addParam("name", getName());
  if (hasPredicate()) db.addParam("Predicate", "Yes");
  db
    .addParam("Data", *(getData().getType()))
    .addParam("Indices", *(getIndices().getType()))
    .addParam("Slices", *(getSlices().getType()))
    .addParam("Cumulative", getCumulative())
    .addParam("users", getNumUsers());
  db.addParam("Result", *(getResult().getType()));
  return db;
}

void ScatterDataNode::visit(Node *parent, NodeWalker *visitor) {
  if (!visitor->shouldVisit(parent, this)) { return; }
  visitor->pre(parent, this);
if (hasPredicate())
 getPredicate().getNode()->visit(this, visitor);
  getData().getNode()->visit(this, visitor);
  getIndices().getNode()->visit(this, visitor);
  getSlices().getNode()->visit(this, visitor);
  visitor->post(parent, this);
}

bool ScatterDataNode::isEqual(const ScatterDataNode &other) const {
  return true &&
      Data_ == other.Data_ &&
      Indices_ == other.Indices_ &&
      Slices_ == other.Slices_ &&
      predicate_ == other.predicate_ &&
      Cumulative_ == other.Cumulative_ &&
      getType(0) == other.getType(0);
}

Node* ScatterDataNode::clone() const {
  return new ScatterDataNode(getName(), getData(), getIndices(), getSlices(), getCumulative());
}

llvm::hash_code ScatterDataNode::getHash() const {
  return llvm::hash_combine(
      Cumulative_,
      Data_,
      Indices_,
      Slices_);
}

unsigned TileNode::getNumInputs() const {
  return 1;
}

std::string TileNode::getInputName(unsigned idx) const {
  if (idx == 0) { return "Input"; }
  idx -= 1;
  llvm_unreachable("Invalid index");
}

NodeValue TileNode::getNthInput(unsigned idx) {
  if (idx == 0) { return Input_; }
  idx -= 1;
  llvm_unreachable("Invalid index");
}

void TileNode::setNthInput(unsigned idx, NodeValue val) {
  if (idx == 0) { Input_ = val; return; }
  idx -= 1;
  llvm_unreachable("Invalid index");
}

llvm::StringRef TileNode::getOutputName(unsigned idx) const {
  if (idx == 0) { return "Result"; }
  llvm_unreachable("Invalid index");
}

std::string TileNode::getDebugDesc() const {
  DescriptionBuilder db(getKindName());
  db.addParam("name", getName());
  if (hasPredicate()) db.addParam("Predicate", "Yes");
  db
    .addParam("Input", *(getInput().getType()))
    .addParam("Count", getCount())
    .addParam("Axis", getAxis())
    .addParam("users", getNumUsers());
  db.addParam("Result", *(getResult().getType()));
  return db;
}

void TileNode::visit(Node *parent, NodeWalker *visitor) {
  if (!visitor->shouldVisit(parent, this)) { return; }
  visitor->pre(parent, this);
if (hasPredicate())
 getPredicate().getNode()->visit(this, visitor);
  getInput().getNode()->visit(this, visitor);
  visitor->post(parent, this);
}

bool TileNode::isEqual(const TileNode &other) const {
  return true &&
      Input_ == other.Input_ &&
      predicate_ == other.predicate_ &&
      Count_ == other.Count_ &&
      Axis_ == other.Axis_ &&
      getType(0) == other.getType(0);
}

Node* TileNode::clone() const {
  return new TileNode(getName(), getResult().getType(), getInput(), getCount(), getAxis());
}

llvm::hash_code TileNode::getHash() const {
  return llvm::hash_combine(
      Count_,
      Axis_,
      Input_);
}

unsigned BatchOneHotNode::getNumInputs() const {
  return 3;
}

std::string BatchOneHotNode::getInputName(unsigned idx) const {
  if (idx == 0) { return "Data"; }
  if (idx == 1) { return "Lengths"; }
  if (idx == 2) { return "Values"; }
  idx -= 3;
  llvm_unreachable("Invalid index");
}

NodeValue BatchOneHotNode::getNthInput(unsigned idx) {
  if (idx == 0) { return Data_; }
  if (idx == 1) { return Lengths_; }
  if (idx == 2) { return Values_; }
  idx -= 3;
  llvm_unreachable("Invalid index");
}

void BatchOneHotNode::setNthInput(unsigned idx, NodeValue val) {
  if (idx == 0) { Data_ = val; return; }
  if (idx == 1) { Lengths_ = val; return; }
  if (idx == 2) { Values_ = val; return; }
  idx -= 3;
  llvm_unreachable("Invalid index");
}

llvm::StringRef BatchOneHotNode::getOutputName(unsigned idx) const {
  if (idx == 0) { return "Result"; }
  llvm_unreachable("Invalid index");
}

std::string BatchOneHotNode::getDebugDesc() const {
  DescriptionBuilder db(getKindName());
  db.addParam("name", getName());
  if (hasPredicate()) db.addParam("Predicate", "Yes");
  db
    .addParam("Data", *(getData().getType()))
    .addParam("Lengths", *(getLengths().getType()))
    .addParam("Values", *(getValues().getType()))
    .addParam("users", getNumUsers());
  db.addParam("Result", *(getResult().getType()));
  return db;
}

void BatchOneHotNode::visit(Node *parent, NodeWalker *visitor) {
  if (!visitor->shouldVisit(parent, this)) { return; }
  visitor->pre(parent, this);
if (hasPredicate())
 getPredicate().getNode()->visit(this, visitor);
  getData().getNode()->visit(this, visitor);
  getLengths().getNode()->visit(this, visitor);
  getValues().getNode()->visit(this, visitor);
  visitor->post(parent, this);
}

bool BatchOneHotNode::isEqual(const BatchOneHotNode &other) const {
  return true &&
      Data_ == other.Data_ &&
      Lengths_ == other.Lengths_ &&
      Values_ == other.Values_ &&
      predicate_ == other.predicate_ &&
      getType(0) == other.getType(0);
}

Node* BatchOneHotNode::clone() const {
  return new BatchOneHotNode(getName(), getResult().getType(), getData(), getLengths(), getValues());
}

llvm::hash_code BatchOneHotNode::getHash() const {
  return llvm::hash_combine(
      Data_,
      Lengths_,
      Values_);
}

unsigned SpaceToDepthNode::getNumInputs() const {
  return 1;
}

std::string SpaceToDepthNode::getInputName(unsigned idx) const {
  if (idx == 0) { return "Input"; }
  idx -= 1;
  llvm_unreachable("Invalid index");
}

NodeValue SpaceToDepthNode::getNthInput(unsigned idx) {
  if (idx == 0) { return Input_; }
  idx -= 1;
  llvm_unreachable("Invalid index");
}

void SpaceToDepthNode::setNthInput(unsigned idx, NodeValue val) {
  if (idx == 0) { Input_ = val; return; }
  idx -= 1;
  llvm_unreachable("Invalid index");
}

llvm::StringRef SpaceToDepthNode::getOutputName(unsigned idx) const {
  if (idx == 0) { return "Result"; }
  llvm_unreachable("Invalid index");
}

std::string SpaceToDepthNode::getDebugDesc() const {
  DescriptionBuilder db(getKindName());
  db.addParam("name", getName());
  if (hasPredicate()) db.addParam("Predicate", "Yes");
  db
    .addParam("Input", *(getInput().getType()))
    .addParam("BlockSize", getBlockSize())
    .addParam("users", getNumUsers());
  db.addParam("Result", *(getResult().getType()));
  return db;
}

void SpaceToDepthNode::visit(Node *parent, NodeWalker *visitor) {
  if (!visitor->shouldVisit(parent, this)) { return; }
  visitor->pre(parent, this);
if (hasPredicate())
 getPredicate().getNode()->visit(this, visitor);
  getInput().getNode()->visit(this, visitor);
  visitor->post(parent, this);
}

bool SpaceToDepthNode::isEqual(const SpaceToDepthNode &other) const {
  return true &&
      Input_ == other.Input_ &&
      predicate_ == other.predicate_ &&
      BlockSize_ == other.BlockSize_ &&
      getType(0) == other.getType(0);
}

Node* SpaceToDepthNode::clone() const {
  return new SpaceToDepthNode(getName(), getResult().getType(), getInput(), getBlockSize());
}

llvm::hash_code SpaceToDepthNode::getHash() const {
  return llvm::hash_combine(
      BlockSize_,
      Input_);
}

unsigned ResizeNearestNode::getNumInputs() const {
  return 1;
}

std::string ResizeNearestNode::getInputName(unsigned idx) const {
  if (idx == 0) { return "Input"; }
  idx -= 1;
  llvm_unreachable("Invalid index");
}

NodeValue ResizeNearestNode::getNthInput(unsigned idx) {
  if (idx == 0) { return Input_; }
  idx -= 1;
  llvm_unreachable("Invalid index");
}

void ResizeNearestNode::setNthInput(unsigned idx, NodeValue val) {
  if (idx == 0) { Input_ = val; return; }
  idx -= 1;
  llvm_unreachable("Invalid index");
}

llvm::StringRef ResizeNearestNode::getOutputName(unsigned idx) const {
  if (idx == 0) { return "Result"; }
  llvm_unreachable("Invalid index");
}

std::string ResizeNearestNode::getDebugDesc() const {
  DescriptionBuilder db(getKindName());
  db.addParam("name", getName());
  if (hasPredicate()) db.addParam("Predicate", "Yes");
  db
    .addParam("Input", *(getInput().getType()))
    .addParam("Scale", getScale())
    .addParam("users", getNumUsers());
  db.addParam("Result", *(getResult().getType()));
  return db;
}

void ResizeNearestNode::visit(Node *parent, NodeWalker *visitor) {
  if (!visitor->shouldVisit(parent, this)) { return; }
  visitor->pre(parent, this);
if (hasPredicate())
 getPredicate().getNode()->visit(this, visitor);
  getInput().getNode()->visit(this, visitor);
  visitor->post(parent, this);
}

bool ResizeNearestNode::isEqual(const ResizeNearestNode &other) const {
  return true &&
      Input_ == other.Input_ &&
      predicate_ == other.predicate_ &&
      Scale_ == other.Scale_ &&
      getType(0) == other.getType(0);
}

Node* ResizeNearestNode::clone() const {
  return new ResizeNearestNode(getName(), getResult().getType(), getInput(), getScale());
}

llvm::hash_code ResizeNearestNode::getHash() const {
  return llvm::hash_combine(
      [](const std::vector<float>& floatVec) -> llvm::hash_code {
        std::vector<size_t> sizeVec = toBinary(floatVec);
        return llvm::hash_combine_range(sizeVec.begin(), sizeVec.end());
      }(Scale_),
      Input_);
}

unsigned ResizeBilinearNode::getNumInputs() const {
  return 1;
}

std::string ResizeBilinearNode::getInputName(unsigned idx) const {
  if (idx == 0) { return "Input"; }
  idx -= 1;
  llvm_unreachable("Invalid index");
}

NodeValue ResizeBilinearNode::getNthInput(unsigned idx) {
  if (idx == 0) { return Input_; }
  idx -= 1;
  llvm_unreachable("Invalid index");
}

void ResizeBilinearNode::setNthInput(unsigned idx, NodeValue val) {
  if (idx == 0) { Input_ = val; return; }
  idx -= 1;
  llvm_unreachable("Invalid index");
}

llvm::StringRef ResizeBilinearNode::getOutputName(unsigned idx) const {
  if (idx == 0) { return "Result"; }
  llvm_unreachable("Invalid index");
}

std::string ResizeBilinearNode::getDebugDesc() const {
  DescriptionBuilder db(getKindName());
  db.addParam("name", getName());
  if (hasPredicate()) db.addParam("Predicate", "Yes");
  db
    .addParam("Input", *(getInput().getType()))
    .addParam("Scale", getScale())
    .addParam("users", getNumUsers());
  db.addParam("Result", *(getResult().getType()));
  return db;
}

void ResizeBilinearNode::visit(Node *parent, NodeWalker *visitor) {
  if (!visitor->shouldVisit(parent, this)) { return; }
  visitor->pre(parent, this);
if (hasPredicate())
 getPredicate().getNode()->visit(this, visitor);
  getInput().getNode()->visit(this, visitor);
  visitor->post(parent, this);
}

bool ResizeBilinearNode::isEqual(const ResizeBilinearNode &other) const {
  return true &&
      Input_ == other.Input_ &&
      predicate_ == other.predicate_ &&
      Scale_ == other.Scale_ &&
      getType(0) == other.getType(0);
}

Node* ResizeBilinearNode::clone() const {
  return new ResizeBilinearNode(getName(), getResult().getType(), getInput(), getScale());
}

llvm::hash_code ResizeBilinearNode::getHash() const {
  return llvm::hash_combine(
      [](const std::vector<float>& floatVec) -> llvm::hash_code {
        std::vector<size_t> sizeVec = toBinary(floatVec);
        return llvm::hash_combine_range(sizeVec.begin(), sizeVec.end());
      }(Scale_),
      Input_);
}

unsigned FlipNode::getNumInputs() const {
  return 1;
}

std::string FlipNode::getInputName(unsigned idx) const {
  if (idx == 0) { return "Input"; }
  idx -= 1;
  llvm_unreachable("Invalid index");
}

NodeValue FlipNode::getNthInput(unsigned idx) {
  if (idx == 0) { return Input_; }
  idx -= 1;
  llvm_unreachable("Invalid index");
}

void FlipNode::setNthInput(unsigned idx, NodeValue val) {
  if (idx == 0) { Input_ = val; return; }
  idx -= 1;
  llvm_unreachable("Invalid index");
}

llvm::StringRef FlipNode::getOutputName(unsigned idx) const {
  if (idx == 0) { return "Result"; }
  llvm_unreachable("Invalid index");
}

std::string FlipNode::getDebugDesc() const {
  DescriptionBuilder db(getKindName());
  db.addParam("name", getName());
  if (hasPredicate()) db.addParam("Predicate", "Yes");
  db
    .addParam("Input", *(getInput().getType()))
    .addParam("Axis", getAxis())
    .addParam("users", getNumUsers());
  db.addParam("Result", *(getResult().getType()));
  return db;
}

void FlipNode::visit(Node *parent, NodeWalker *visitor) {
  if (!visitor->shouldVisit(parent, this)) { return; }
  visitor->pre(parent, this);
if (hasPredicate())
 getPredicate().getNode()->visit(this, visitor);
  getInput().getNode()->visit(this, visitor);
  visitor->post(parent, this);
}

bool FlipNode::isEqual(const FlipNode &other) const {
  return true &&
      Input_ == other.Input_ &&
      predicate_ == other.predicate_ &&
      Axis_ == other.Axis_ &&
      getType(0) == other.getType(0);
}

Node* FlipNode::clone() const {
  return new FlipNode(getName(), getResult().getType(), getInput(), getAxis());
}

llvm::hash_code FlipNode::getHash() const {
  return llvm::hash_combine(
      Axis_,
      Input_);
}

unsigned SplatNode::getNumInputs() const {
  return 0;
}

std::string SplatNode::getInputName(unsigned idx) const {
  idx -= 0;
  llvm_unreachable("Invalid index");
}

NodeValue SplatNode::getNthInput(unsigned idx) {
  idx -= 0;
  llvm_unreachable("Invalid index");
}

void SplatNode::setNthInput(unsigned idx, NodeValue val) {
  idx -= 0;
  llvm_unreachable("Invalid index");
}

llvm::StringRef SplatNode::getOutputName(unsigned idx) const {
  if (idx == 0) { return "Result"; }
  llvm_unreachable("Invalid index");
}

std::string SplatNode::getDebugDesc() const {
  DescriptionBuilder db(getKindName());
  db.addParam("name", getName());
  if (hasPredicate()) db.addParam("Predicate", "Yes");
  db
    .addParam("Value", getValue())
    .addParam("users", getNumUsers());
  db.addParam("Result", *(getResult().getType()));
  return db;
}

void SplatNode::visit(Node *parent, NodeWalker *visitor) {
  if (!visitor->shouldVisit(parent, this)) { return; }
  visitor->pre(parent, this);
if (hasPredicate())
 getPredicate().getNode()->visit(this, visitor);
  visitor->post(parent, this);
}

bool SplatNode::isEqual(const SplatNode &other) const {
  return true &&
      predicate_ == other.predicate_ &&
      Value_ == other.Value_ &&
      getType(0) == other.getType(0);
}

Node* SplatNode::clone() const {
  return new SplatNode(getName(), getResult().getType(), getValue());
}

llvm::hash_code SplatNode::getHash() const {
  return llvm::hash_combine(
      toBinary(Value_));
}

unsigned TouchNode::getNumInputs() const {
  return 0;
}

std::string TouchNode::getInputName(unsigned idx) const {
  idx -= 0;
  llvm_unreachable("Invalid index");
}

NodeValue TouchNode::getNthInput(unsigned idx) {
  idx -= 0;
  llvm_unreachable("Invalid index");
}

void TouchNode::setNthInput(unsigned idx, NodeValue val) {
  idx -= 0;
  llvm_unreachable("Invalid index");
}

llvm::StringRef TouchNode::getOutputName(unsigned idx) const {
  if (idx == 0) { return "Result"; }
  llvm_unreachable("Invalid index");
}

std::string TouchNode::getDebugDesc() const {
  DescriptionBuilder db(getKindName());
  db.addParam("name", getName());
  if (hasPredicate()) db.addParam("Predicate", "Yes");
  db
    .addParam("users", getNumUsers());
  db.addParam("Result", *(getResult().getType()));
  return db;
}

void TouchNode::visit(Node *parent, NodeWalker *visitor) {
  if (!visitor->shouldVisit(parent, this)) { return; }
  visitor->pre(parent, this);
if (hasPredicate())
 getPredicate().getNode()->visit(this, visitor);
  visitor->post(parent, this);
}

bool TouchNode::isEqual(const TouchNode &other) const {
  return true &&
      predicate_ == other.predicate_ &&
      getType(0) == other.getType(0);
}

Node* TouchNode::clone() const {
  return new TouchNode(getName(), getResult().getType());
}

llvm::hash_code TouchNode::getHash() const {
  return llvm::hash_combine(0);
 }

unsigned SGDNode::getNumInputs() const {
  return 2;
}

std::string SGDNode::getInputName(unsigned idx) const {
  if (idx == 0) { return "Gradient"; }
  if (idx == 1) { return "Weight"; }
  idx -= 2;
  llvm_unreachable("Invalid index");
}

NodeValue SGDNode::getNthInput(unsigned idx) {
  if (idx == 0) { return Gradient_; }
  if (idx == 1) { return Weight_; }
  idx -= 2;
  llvm_unreachable("Invalid index");
}

void SGDNode::setNthInput(unsigned idx, NodeValue val) {
  if (idx == 0) { Gradient_ = val; return; }
  if (idx == 1) { Weight_ = val; return; }
  idx -= 2;
  llvm_unreachable("Invalid index");
}

llvm::StringRef SGDNode::getOutputName(unsigned idx) const {
  if (idx == 0) { return "UpdatedWeight"; }
  llvm_unreachable("Invalid index");
}

std::string SGDNode::getDebugDesc() const {
  DescriptionBuilder db(getKindName());
  db.addParam("name", getName());
  if (hasPredicate()) db.addParam("Predicate", "Yes");
  db
    .addParam("Gradient", *(getGradient().getType()))
    .addParam("Weight", *(getWeight().getType()))
    .addParam("L1Decay", getL1Decay())
    .addParam("L2Decay", getL2Decay())
    .addParam("LearningRate", getLearningRate())
    .addParam("Momentum", getMomentum())
    .addParam("BatchSize", getBatchSize())
    .addParam("users", getNumUsers());
  db.addParam("UpdatedWeight", *(getUpdatedWeight().getType()));
  return db;
}

void SGDNode::visit(Node *parent, NodeWalker *visitor) {
  if (!visitor->shouldVisit(parent, this)) { return; }
  visitor->pre(parent, this);
if (hasPredicate())
 getPredicate().getNode()->visit(this, visitor);
  getGradient().getNode()->visit(this, visitor);
  getWeight().getNode()->visit(this, visitor);
  visitor->post(parent, this);
}

bool SGDNode::isEqual(const SGDNode &other) const {
  return true &&
      Gradient_ == other.Gradient_ &&
      Weight_ == other.Weight_ &&
      predicate_ == other.predicate_ &&
      L1Decay_ == other.L1Decay_ &&
      L2Decay_ == other.L2Decay_ &&
      LearningRate_ == other.LearningRate_ &&
      Momentum_ == other.Momentum_ &&
      BatchSize_ == other.BatchSize_ &&
      getType(0) == other.getType(0);
}

Node* SGDNode::clone() const {
  return new SGDNode(getName(), getGradient(), getWeight(), getL1Decay(), getL2Decay(), getLearningRate(), getMomentum(), getBatchSize());
}

llvm::hash_code SGDNode::getHash() const {
  return llvm::hash_combine(
      toBinary(L1Decay_),
      toBinary(L2Decay_),
      toBinary(LearningRate_),
      toBinary(Momentum_),
      BatchSize_,
      Gradient_,
      Weight_);
}

unsigned TraceEventNode::getNumInputs() const {
  return 1;
}

std::string TraceEventNode::getInputName(unsigned idx) const {
  if (idx == 0) { return "Data"; }
  idx -= 1;
  llvm_unreachable("Invalid index");
}

NodeValue TraceEventNode::getNthInput(unsigned idx) {
  if (idx == 0) { return Data_; }
  idx -= 1;
  llvm_unreachable("Invalid index");
}

void TraceEventNode::setNthInput(unsigned idx, NodeValue val) {
  if (idx == 0) { Data_ = val; return; }
  idx -= 1;
  llvm_unreachable("Invalid index");
}

llvm::StringRef TraceEventNode::getOutputName(unsigned idx) const {
  llvm_unreachable("Invalid index");
}

std::string TraceEventNode::getDebugDesc() const {
  DescriptionBuilder db(getKindName());
  db.addParam("name", getName());
  if (hasPredicate()) db.addParam("Predicate", "Yes");
  db
    .addParam("Data", *(getData().getType()))
    .addParam("EventName", getEventName())
    .addParam("EventType", getEventType())
    .addParam("Index", getIndex())
    .addParam("users", getNumUsers());
  return db;
}

void TraceEventNode::visit(Node *parent, NodeWalker *visitor) {
  if (!visitor->shouldVisit(parent, this)) { return; }
  visitor->pre(parent, this);
if (hasPredicate())
 getPredicate().getNode()->visit(this, visitor);
  getData().getNode()->visit(this, visitor);
  visitor->post(parent, this);
}

bool TraceEventNode::isEqual(const TraceEventNode &other) const {
  return true &&
      Data_ == other.Data_ &&
      predicate_ == other.predicate_ &&
      EventName_ == other.EventName_ &&
      EventType_ == other.EventType_ &&
      Index_ == other.Index_;
}

Node* TraceEventNode::clone() const {
  return new TraceEventNode(getName(), getData(), getEventName(), getEventType(), getIndex());
}

llvm::hash_code TraceEventNode::getHash() const {
  return llvm::hash_combine(
      EventName_,
      EventType_,
      Index_,
      Data_);
}

unsigned QuantizationProfileNode::getNumInputs() const {
  return 3;
}

std::string QuantizationProfileNode::getInputName(unsigned idx) const {
  if (idx == 0) { return "Input"; }
  if (idx == 1) { return "Histogram"; }
  if (idx == 2) { return "ComputationInfo"; }
  idx -= 3;
  llvm_unreachable("Invalid index");
}

NodeValue QuantizationProfileNode::getNthInput(unsigned idx) {
  if (idx == 0) { return Input_; }
  if (idx == 1) { return Histogram_; }
  if (idx == 2) { return ComputationInfo_; }
  idx -= 3;
  llvm_unreachable("Invalid index");
}

void QuantizationProfileNode::setNthInput(unsigned idx, NodeValue val) {
  if (idx == 0) { Input_ = val; return; }
  if (idx == 1) { Histogram_ = val; return; }
  if (idx == 2) { ComputationInfo_ = val; return; }
  idx -= 3;
  llvm_unreachable("Invalid index");
}

llvm::StringRef QuantizationProfileNode::getOutputName(unsigned idx) const {
  llvm_unreachable("Invalid index");
}

std::string QuantizationProfileNode::getDebugDesc() const {
  DescriptionBuilder db(getKindName());
  db.addParam("name", getName());
  if (hasPredicate()) db.addParam("Predicate", "Yes");
  db
    .addParam("Input", *(getInput().getType()))
    .addParam("Histogram", *(getHistogram().getType()))
    .addParam("ComputationInfo", *(getComputationInfo().getType()))
    .addParam("ProfiledNodeName", getProfiledNodeName())
    .addParam("ProfiledOutputNumber", getProfiledOutputNumber())
    .addParam("users", getNumUsers());
  return db;
}

void QuantizationProfileNode::visit(Node *parent, NodeWalker *visitor) {
  if (!visitor->shouldVisit(parent, this)) { return; }
  visitor->pre(parent, this);
if (hasPredicate())
 getPredicate().getNode()->visit(this, visitor);
  getInput().getNode()->visit(this, visitor);
  getHistogram().getNode()->visit(this, visitor);
  getComputationInfo().getNode()->visit(this, visitor);
  visitor->post(parent, this);
}

bool QuantizationProfileNode::isEqual(const QuantizationProfileNode &other) const {
  return true &&
      Input_ == other.Input_ &&
      Histogram_ == other.Histogram_ &&
      ComputationInfo_ == other.ComputationInfo_ &&
      predicate_ == other.predicate_ &&
      ProfiledNodeName_ == other.ProfiledNodeName_ &&
      ProfiledOutputNumber_ == other.ProfiledOutputNumber_;
}

Node* QuantizationProfileNode::clone() const {
  return new QuantizationProfileNode(getName(), getInput(), getHistogram(), getComputationInfo(), getProfiledNodeName(), getProfiledOutputNumber());
}

llvm::hash_code QuantizationProfileNode::getHash() const {
  return llvm::hash_combine(
      ProfiledNodeName_,
      ProfiledOutputNumber_,
      Input_,
      Histogram_,
      ComputationInfo_);
}
Placeholder *QuantizationProfileNode::getHistogramPlaceholder() const { return llvm::cast<Placeholder>(Histogram_.getNode()); };
Placeholder *QuantizationProfileNode::getComputationInfoPlaceholder() const { return llvm::cast<Placeholder>(ComputationInfo_.getNode()); };

unsigned IntLookupTableNode::getNumInputs() const {
  return 2;
}

std::string IntLookupTableNode::getInputName(unsigned idx) const {
  if (idx == 0) { return "Input"; }
  if (idx == 1) { return "Mapping"; }
  idx -= 2;
  llvm_unreachable("Invalid index");
}

NodeValue IntLookupTableNode::getNthInput(unsigned idx) {
  if (idx == 0) { return Input_; }
  if (idx == 1) { return Mapping_; }
  idx -= 2;
  llvm_unreachable("Invalid index");
}

void IntLookupTableNode::setNthInput(unsigned idx, NodeValue val) {
  if (idx == 0) { Input_ = val; return; }
  if (idx == 1) { Mapping_ = val; return; }
  idx -= 2;
  llvm_unreachable("Invalid index");
}

llvm::StringRef IntLookupTableNode::getOutputName(unsigned idx) const {
  if (idx == 0) { return "Result"; }
  llvm_unreachable("Invalid index");
}

std::string IntLookupTableNode::getDebugDesc() const {
  DescriptionBuilder db(getKindName());
  db.addParam("name", getName());
  if (hasPredicate()) db.addParam("Predicate", "Yes");
  db
    .addParam("Input", *(getInput().getType()))
    .addParam("Mapping", *(getMapping().getType()))
    .addParam("users", getNumUsers());
  db.addParam("Result", *(getResult().getType()));
  return db;
}

void IntLookupTableNode::visit(Node *parent, NodeWalker *visitor) {
  if (!visitor->shouldVisit(parent, this)) { return; }
  visitor->pre(parent, this);
if (hasPredicate())
 getPredicate().getNode()->visit(this, visitor);
  getInput().getNode()->visit(this, visitor);
  getMapping().getNode()->visit(this, visitor);
  visitor->post(parent, this);
}

bool IntLookupTableNode::isEqual(const IntLookupTableNode &other) const {
  return true &&
      Input_ == other.Input_ &&
      Mapping_ == other.Mapping_ &&
      predicate_ == other.predicate_ &&
      getType(0) == other.getType(0);
}

Node* IntLookupTableNode::clone() const {
  return new IntLookupTableNode(getName(), getResult().getType(), getInput(), getMapping());
}

llvm::hash_code IntLookupTableNode::getHash() const {
  return llvm::hash_combine(
      Input_,
      Mapping_);
}

unsigned QuantizeNode::getNumInputs() const {
  return 1;
}

std::string QuantizeNode::getInputName(unsigned idx) const {
  if (idx == 0) { return "Input"; }
  idx -= 1;
  llvm_unreachable("Invalid index");
}

NodeValue QuantizeNode::getNthInput(unsigned idx) {
  if (idx == 0) { return Input_; }
  idx -= 1;
  llvm_unreachable("Invalid index");
}

void QuantizeNode::setNthInput(unsigned idx, NodeValue val) {
  if (idx == 0) { Input_ = val; return; }
  idx -= 1;
  llvm_unreachable("Invalid index");
}

llvm::StringRef QuantizeNode::getOutputName(unsigned idx) const {
  if (idx == 0) { return "Result"; }
  llvm_unreachable("Invalid index");
}

std::string QuantizeNode::getDebugDesc() const {
  DescriptionBuilder db(getKindName());
  db.addParam("name", getName());
  if (hasPredicate()) db.addParam("Predicate", "Yes");
  db
    .addParam("Input", *(getInput().getType()))
    .addParam("users", getNumUsers());
  db.addParam("Result", *(getResult().getType()));
  return db;
}

void QuantizeNode::visit(Node *parent, NodeWalker *visitor) {
  if (!visitor->shouldVisit(parent, this)) { return; }
  visitor->pre(parent, this);
if (hasPredicate())
 getPredicate().getNode()->visit(this, visitor);
  getInput().getNode()->visit(this, visitor);
  visitor->post(parent, this);
}

bool QuantizeNode::isEqual(const QuantizeNode &other) const {
  return true &&
      Input_ == other.Input_ &&
      predicate_ == other.predicate_ &&
      getType(0) == other.getType(0);
}

Node* QuantizeNode::clone() const {
  return new QuantizeNode(getName(), getResult().getType(), getInput());
}

llvm::hash_code QuantizeNode::getHash() const {
  return llvm::hash_combine(
      Input_);
}

unsigned DequantizeNode::getNumInputs() const {
  return 1;
}

std::string DequantizeNode::getInputName(unsigned idx) const {
  if (idx == 0) { return "Input"; }
  idx -= 1;
  llvm_unreachable("Invalid index");
}

NodeValue DequantizeNode::getNthInput(unsigned idx) {
  if (idx == 0) { return Input_; }
  idx -= 1;
  llvm_unreachable("Invalid index");
}

void DequantizeNode::setNthInput(unsigned idx, NodeValue val) {
  if (idx == 0) { Input_ = val; return; }
  idx -= 1;
  llvm_unreachable("Invalid index");
}

llvm::StringRef DequantizeNode::getOutputName(unsigned idx) const {
  if (idx == 0) { return "Result"; }
  llvm_unreachable("Invalid index");
}

std::string DequantizeNode::getDebugDesc() const {
  DescriptionBuilder db(getKindName());
  db.addParam("name", getName());
  if (hasPredicate()) db.addParam("Predicate", "Yes");
  db
    .addParam("Input", *(getInput().getType()))
    .addParam("users", getNumUsers());
  db.addParam("Result", *(getResult().getType()));
  return db;
}

void DequantizeNode::visit(Node *parent, NodeWalker *visitor) {
  if (!visitor->shouldVisit(parent, this)) { return; }
  visitor->pre(parent, this);
if (hasPredicate())
 getPredicate().getNode()->visit(this, visitor);
  getInput().getNode()->visit(this, visitor);
  visitor->post(parent, this);
}

bool DequantizeNode::isEqual(const DequantizeNode &other) const {
  return true &&
      Input_ == other.Input_ &&
      predicate_ == other.predicate_ &&
      getType(0) == other.getType(0);
}

Node* DequantizeNode::clone() const {
  return new DequantizeNode(getName(), getResult().getType(), getInput());
}

llvm::hash_code DequantizeNode::getHash() const {
  return llvm::hash_combine(
      Input_);
}

unsigned RescaleQuantizedNode::getNumInputs() const {
  return 1;
}

std::string RescaleQuantizedNode::getInputName(unsigned idx) const {
  if (idx == 0) { return "Input"; }
  idx -= 1;
  llvm_unreachable("Invalid index");
}

NodeValue RescaleQuantizedNode::getNthInput(unsigned idx) {
  if (idx == 0) { return Input_; }
  idx -= 1;
  llvm_unreachable("Invalid index");
}

void RescaleQuantizedNode::setNthInput(unsigned idx, NodeValue val) {
  if (idx == 0) { Input_ = val; return; }
  idx -= 1;
  llvm_unreachable("Invalid index");
}

llvm::StringRef RescaleQuantizedNode::getOutputName(unsigned idx) const {
  if (idx == 0) { return "Result"; }
  llvm_unreachable("Invalid index");
}

std::string RescaleQuantizedNode::getDebugDesc() const {
  DescriptionBuilder db(getKindName());
  db.addParam("name", getName());
  if (hasPredicate()) db.addParam("Predicate", "Yes");
  db
    .addParam("Input", *(getInput().getType()))
    .addParam("users", getNumUsers());
  db.addParam("Result", *(getResult().getType()));
  return db;
}

void RescaleQuantizedNode::visit(Node *parent, NodeWalker *visitor) {
  if (!visitor->shouldVisit(parent, this)) { return; }
  visitor->pre(parent, this);
if (hasPredicate())
 getPredicate().getNode()->visit(this, visitor);
  getInput().getNode()->visit(this, visitor);
  visitor->post(parent, this);
}

bool RescaleQuantizedNode::isEqual(const RescaleQuantizedNode &other) const {
  return true &&
      Input_ == other.Input_ &&
      predicate_ == other.predicate_ &&
      getType(0) == other.getType(0);
}

Node* RescaleQuantizedNode::clone() const {
  return new RescaleQuantizedNode(getName(), getResult().getType(), getInput());
}

llvm::hash_code RescaleQuantizedNode::getHash() const {
  return llvm::hash_combine(
      Input_);
}

unsigned TopKNode::getNumInputs() const {
  return 1;
}

std::string TopKNode::getInputName(unsigned idx) const {
  if (idx == 0) { return "Input"; }
  idx -= 1;
  llvm_unreachable("Invalid index");
}

NodeValue TopKNode::getNthInput(unsigned idx) {
  if (idx == 0) { return Input_; }
  idx -= 1;
  llvm_unreachable("Invalid index");
}

void TopKNode::setNthInput(unsigned idx, NodeValue val) {
  if (idx == 0) { Input_ = val; return; }
  idx -= 1;
  llvm_unreachable("Invalid index");
}

llvm::StringRef TopKNode::getOutputName(unsigned idx) const {
  if (idx == 0) { return "Values"; }
  if (idx == 1) { return "Indices"; }
  llvm_unreachable("Invalid index");
}

std::string TopKNode::getDebugDesc() const {
  DescriptionBuilder db(getKindName());
  db.addParam("name", getName());
  if (hasPredicate()) db.addParam("Predicate", "Yes");
  db
    .addParam("Input", *(getInput().getType()))
    .addParam("K", getK())
    .addParam("users", getNumUsers());
  db.addParam("Values", *(getValues().getType()));
  db.addParam("Indices", *(getIndices().getType()));
  return db;
}

void TopKNode::visit(Node *parent, NodeWalker *visitor) {
  if (!visitor->shouldVisit(parent, this)) { return; }
  visitor->pre(parent, this);
if (hasPredicate())
 getPredicate().getNode()->visit(this, visitor);
  getInput().getNode()->visit(this, visitor);
  visitor->post(parent, this);
}

bool TopKNode::isEqual(const TopKNode &other) const {
  return true &&
      Input_ == other.Input_ &&
      predicate_ == other.predicate_ &&
      K_ == other.K_ &&
      getType(0) == other.getType(0) &&
      getType(1) == other.getType(1);
}

Node* TopKNode::clone() const {
  return new TopKNode(getName(), getValues().getType(), getIndices().getType(), getInput(), getK());
}

llvm::hash_code TopKNode::getHash() const {
  return llvm::hash_combine(
      K_,
      Input_);
}

unsigned ConvertToNode::getNumInputs() const {
  return 1;
}

std::string ConvertToNode::getInputName(unsigned idx) const {
  if (idx == 0) { return "Input"; }
  idx -= 1;
  llvm_unreachable("Invalid index");
}

NodeValue ConvertToNode::getNthInput(unsigned idx) {
  if (idx == 0) { return Input_; }
  idx -= 1;
  llvm_unreachable("Invalid index");
}

void ConvertToNode::setNthInput(unsigned idx, NodeValue val) {
  if (idx == 0) { Input_ = val; return; }
  idx -= 1;
  llvm_unreachable("Invalid index");
}

llvm::StringRef ConvertToNode::getOutputName(unsigned idx) const {
  if (idx == 0) { return "Result"; }
  llvm_unreachable("Invalid index");
}

std::string ConvertToNode::getDebugDesc() const {
  DescriptionBuilder db(getKindName());
  db.addParam("name", getName());
  if (hasPredicate()) db.addParam("Predicate", "Yes");
  db
    .addParam("Input", *(getInput().getType()))
    .addParam("users", getNumUsers());
  db.addParam("Result", *(getResult().getType()));
  return db;
}

void ConvertToNode::visit(Node *parent, NodeWalker *visitor) {
  if (!visitor->shouldVisit(parent, this)) { return; }
  visitor->pre(parent, this);
if (hasPredicate())
 getPredicate().getNode()->visit(this, visitor);
  getInput().getNode()->visit(this, visitor);
  visitor->post(parent, this);
}

bool ConvertToNode::isEqual(const ConvertToNode &other) const {
  return true &&
      Input_ == other.Input_ &&
      predicate_ == other.predicate_ &&
      getType(0) == other.getType(0);
}

Node* ConvertToNode::clone() const {
  return new ConvertToNode(getName(), getResult().getType(), getInput());
}

llvm::hash_code ConvertToNode::getHash() const {
  return llvm::hash_combine(
      Input_);
}

unsigned AudioSpectrogramNode::getNumInputs() const {
  return 5;
}

std::string AudioSpectrogramNode::getInputName(unsigned idx) const {
  if (idx == 0) { return "Input"; }
  if (idx == 1) { return "Window"; }
  if (idx == 2) { return "TwiddleFactors"; }
  if (idx == 3) { return "BitReverseIndices"; }
  if (idx == 4) { return "ComplexToRealWeights"; }
  idx -= 5;
  llvm_unreachable("Invalid index");
}

NodeValue AudioSpectrogramNode::getNthInput(unsigned idx) {
  if (idx == 0) { return Input_; }
  if (idx == 1) { return Window_; }
  if (idx == 2) { return TwiddleFactors_; }
  if (idx == 3) { return BitReverseIndices_; }
  if (idx == 4) { return ComplexToRealWeights_; }
  idx -= 5;
  llvm_unreachable("Invalid index");
}

void AudioSpectrogramNode::setNthInput(unsigned idx, NodeValue val) {
  if (idx == 0) { Input_ = val; return; }
  if (idx == 1) { Window_ = val; return; }
  if (idx == 2) { TwiddleFactors_ = val; return; }
  if (idx == 3) { BitReverseIndices_ = val; return; }
  if (idx == 4) { ComplexToRealWeights_ = val; return; }
  idx -= 5;
  llvm_unreachable("Invalid index");
}

llvm::StringRef AudioSpectrogramNode::getOutputName(unsigned idx) const {
  if (idx == 0) { return "Spectrogram"; }
  llvm_unreachable("Invalid index");
}

std::string AudioSpectrogramNode::getDebugDesc() const {
  DescriptionBuilder db(getKindName());
  db.addParam("name", getName());
  if (hasPredicate()) db.addParam("Predicate", "Yes");
  db
    .addParam("Input", *(getInput().getType()))
    .addParam("Window", *(getWindow().getType()))
    .addParam("TwiddleFactors", *(getTwiddleFactors().getType()))
    .addParam("BitReverseIndices", *(getBitReverseIndices().getType()))
    .addParam("ComplexToRealWeights", *(getComplexToRealWeights().getType()))
    .addParam("WindowSize", getWindowSize())
    .addParam("WindowStride", getWindowStride())
    .addParam("MagnitudeSquared", getMagnitudeSquared())
    .addParam("users", getNumUsers());
  db.addParam("Spectrogram", *(getSpectrogram().getType()));
  return db;
}

void AudioSpectrogramNode::visit(Node *parent, NodeWalker *visitor) {
  if (!visitor->shouldVisit(parent, this)) { return; }
  visitor->pre(parent, this);
if (hasPredicate())
 getPredicate().getNode()->visit(this, visitor);
  getInput().getNode()->visit(this, visitor);
  getWindow().getNode()->visit(this, visitor);
  getTwiddleFactors().getNode()->visit(this, visitor);
  getBitReverseIndices().getNode()->visit(this, visitor);
  getComplexToRealWeights().getNode()->visit(this, visitor);
  visitor->post(parent, this);
}

bool AudioSpectrogramNode::isEqual(const AudioSpectrogramNode &other) const {
  return true &&
      Input_ == other.Input_ &&
      Window_ == other.Window_ &&
      TwiddleFactors_ == other.TwiddleFactors_ &&
      BitReverseIndices_ == other.BitReverseIndices_ &&
      ComplexToRealWeights_ == other.ComplexToRealWeights_ &&
      predicate_ == other.predicate_ &&
      WindowSize_ == other.WindowSize_ &&
      WindowStride_ == other.WindowStride_ &&
      MagnitudeSquared_ == other.MagnitudeSquared_ &&
      getType(0) == other.getType(0);
}

Node* AudioSpectrogramNode::clone() const {
  return new AudioSpectrogramNode(getName(), getSpectrogram().getType(), getInput(), getWindow(), getTwiddleFactors(), getBitReverseIndices(), getComplexToRealWeights(), getWindowSize(), getWindowStride(), getMagnitudeSquared());
}

llvm::hash_code AudioSpectrogramNode::getHash() const {
  return llvm::hash_combine(
      WindowSize_,
      WindowStride_,
      MagnitudeSquared_,
      Input_,
      Window_,
      TwiddleFactors_,
      BitReverseIndices_,
      ComplexToRealWeights_);
}

unsigned MFCCNode::getNumInputs() const {
  return 4;
}

std::string MFCCNode::getInputName(unsigned idx) const {
  if (idx == 0) { return "Spectrogram"; }
  if (idx == 1) { return "MelWeights"; }
  if (idx == 2) { return "MelRanges"; }
  if (idx == 3) { return "DctMat"; }
  idx -= 4;
  llvm_unreachable("Invalid index");
}

NodeValue MFCCNode::getNthInput(unsigned idx) {
  if (idx == 0) { return Spectrogram_; }
  if (idx == 1) { return MelWeights_; }
  if (idx == 2) { return MelRanges_; }
  if (idx == 3) { return DctMat_; }
  idx -= 4;
  llvm_unreachable("Invalid index");
}

void MFCCNode::setNthInput(unsigned idx, NodeValue val) {
  if (idx == 0) { Spectrogram_ = val; return; }
  if (idx == 1) { MelWeights_ = val; return; }
  if (idx == 2) { MelRanges_ = val; return; }
  if (idx == 3) { DctMat_ = val; return; }
  idx -= 4;
  llvm_unreachable("Invalid index");
}

llvm::StringRef MFCCNode::getOutputName(unsigned idx) const {
  if (idx == 0) { return "Coefficients"; }
  llvm_unreachable("Invalid index");
}

std::string MFCCNode::getDebugDesc() const {
  DescriptionBuilder db(getKindName());
  db.addParam("name", getName());
  if (hasPredicate()) db.addParam("Predicate", "Yes");
  db
    .addParam("Spectrogram", *(getSpectrogram().getType()))
    .addParam("MelWeights", *(getMelWeights().getType()))
    .addParam("MelRanges", *(getMelRanges().getType()))
    .addParam("DctMat", *(getDctMat().getType()))
    .addParam("SampleRate", getSampleRate())
    .addParam("LowerFrequency", getLowerFrequency())
    .addParam("UpperFrequency", getUpperFrequency())
    .addParam("FilterBankCount", getFilterBankCount())
    .addParam("NumCoefficients", getNumCoefficients())
    .addParam("users", getNumUsers());
  db.addParam("Coefficients", *(getCoefficients().getType()));
  return db;
}

void MFCCNode::visit(Node *parent, NodeWalker *visitor) {
  if (!visitor->shouldVisit(parent, this)) { return; }
  visitor->pre(parent, this);
if (hasPredicate())
 getPredicate().getNode()->visit(this, visitor);
  getSpectrogram().getNode()->visit(this, visitor);
  getMelWeights().getNode()->visit(this, visitor);
  getMelRanges().getNode()->visit(this, visitor);
  getDctMat().getNode()->visit(this, visitor);
  visitor->post(parent, this);
}

bool MFCCNode::isEqual(const MFCCNode &other) const {
  return true &&
      Spectrogram_ == other.Spectrogram_ &&
      MelWeights_ == other.MelWeights_ &&
      MelRanges_ == other.MelRanges_ &&
      DctMat_ == other.DctMat_ &&
      predicate_ == other.predicate_ &&
      SampleRate_ == other.SampleRate_ &&
      LowerFrequency_ == other.LowerFrequency_ &&
      UpperFrequency_ == other.UpperFrequency_ &&
      FilterBankCount_ == other.FilterBankCount_ &&
      NumCoefficients_ == other.NumCoefficients_ &&
      getType(0) == other.getType(0);
}

Node* MFCCNode::clone() const {
  return new MFCCNode(getName(), getCoefficients().getType(), getSpectrogram(), getMelWeights(), getMelRanges(), getDctMat(), getSampleRate(), getLowerFrequency(), getUpperFrequency(), getFilterBankCount(), getNumCoefficients());
}

llvm::hash_code MFCCNode::getHash() const {
  return llvm::hash_combine(
      toBinary(SampleRate_),
      toBinary(LowerFrequency_),
      toBinary(UpperFrequency_),
      FilterBankCount_,
      NumCoefficients_,
      Spectrogram_,
      MelWeights_,
      MelRanges_,
      DctMat_);
}

unsigned NonMaxSuppressionNode::getNumInputs() const {
  return 2;
}

std::string NonMaxSuppressionNode::getInputName(unsigned idx) const {
  if (idx == 0) { return "Boxes"; }
  if (idx == 1) { return "Scores"; }
  idx -= 2;
  llvm_unreachable("Invalid index");
}

NodeValue NonMaxSuppressionNode::getNthInput(unsigned idx) {
  if (idx == 0) { return Boxes_; }
  if (idx == 1) { return Scores_; }
  idx -= 2;
  llvm_unreachable("Invalid index");
}

void NonMaxSuppressionNode::setNthInput(unsigned idx, NodeValue val) {
  if (idx == 0) { Boxes_ = val; return; }
  if (idx == 1) { Scores_ = val; return; }
  idx -= 2;
  llvm_unreachable("Invalid index");
}

llvm::StringRef NonMaxSuppressionNode::getOutputName(unsigned idx) const {
  if (idx == 0) { return "Indices"; }
  if (idx == 1) { return "NumberOfSelectedIndices"; }
  llvm_unreachable("Invalid index");
}

std::string NonMaxSuppressionNode::getDebugDesc() const {
  DescriptionBuilder db(getKindName());
  db.addParam("name", getName());
  if (hasPredicate()) db.addParam("Predicate", "Yes");
  db
    .addParam("Boxes", *(getBoxes().getType()))
    .addParam("Scores", *(getScores().getType()))
    .addParam("CenterPointBox", getCenterPointBox())
    .addParam("MaxOutputBoxesPerClass", getMaxOutputBoxesPerClass())
    .addParam("IouThreshold", getIouThreshold())
    .addParam("ScoreThreshold", getScoreThreshold())
    .addParam("IsTFVersion4", getIsTFVersion4())
    .addParam("users", getNumUsers());
  db.addParam("Indices", *(getIndices().getType()));
  db.addParam("NumberOfSelectedIndices", *(getNumberOfSelectedIndices().getType()));
  return db;
}

void NonMaxSuppressionNode::visit(Node *parent, NodeWalker *visitor) {
  if (!visitor->shouldVisit(parent, this)) { return; }
  visitor->pre(parent, this);
if (hasPredicate())
 getPredicate().getNode()->visit(this, visitor);
  getBoxes().getNode()->visit(this, visitor);
  getScores().getNode()->visit(this, visitor);
  visitor->post(parent, this);
}

bool NonMaxSuppressionNode::isEqual(const NonMaxSuppressionNode &other) const {
  return true &&
      Boxes_ == other.Boxes_ &&
      Scores_ == other.Scores_ &&
      predicate_ == other.predicate_ &&
      CenterPointBox_ == other.CenterPointBox_ &&
      MaxOutputBoxesPerClass_ == other.MaxOutputBoxesPerClass_ &&
      IouThreshold_ == other.IouThreshold_ &&
      ScoreThreshold_ == other.ScoreThreshold_ &&
      IsTFVersion4_ == other.IsTFVersion4_ &&
      getType(0) == other.getType(0) &&
      getType(1) == other.getType(1);
}

Node* NonMaxSuppressionNode::clone() const {
  return new NonMaxSuppressionNode(getName(), getIndices().getType(), getNumberOfSelectedIndices().getType(), getBoxes(), getScores(), getCenterPointBox(), getMaxOutputBoxesPerClass(), getIouThreshold(), getScoreThreshold(), getIsTFVersion4());
}

llvm::hash_code NonMaxSuppressionNode::getHash() const {
  return llvm::hash_combine(
      CenterPointBox_,
      MaxOutputBoxesPerClass_,
      toBinary(IouThreshold_),
      toBinary(ScoreThreshold_),
      IsTFVersion4_,
      Boxes_,
      Scores_);
}

unsigned CPUMaxSplatNode::getNumInputs() const {
  return 1;
}

std::string CPUMaxSplatNode::getInputName(unsigned idx) const {
  if (idx == 0) { return "Input"; }
  idx -= 1;
  llvm_unreachable("Invalid index");
}

NodeValue CPUMaxSplatNode::getNthInput(unsigned idx) {
  if (idx == 0) { return Input_; }
  idx -= 1;
  llvm_unreachable("Invalid index");
}

void CPUMaxSplatNode::setNthInput(unsigned idx, NodeValue val) {
  if (idx == 0) { Input_ = val; return; }
  idx -= 1;
  llvm_unreachable("Invalid index");
}

llvm::StringRef CPUMaxSplatNode::getOutputName(unsigned idx) const {
  if (idx == 0) { return "Result"; }
  llvm_unreachable("Invalid index");
}

std::string CPUMaxSplatNode::getDebugDesc() const {
  DescriptionBuilder db(getKindName());
  db.addParam("name", getName());
  if (hasPredicate()) db.addParam("Predicate", "Yes");
  db
    .addParam("Input", *(getInput().getType()))
    .addParam("SplatValue", getSplatValue())
    .addParam("users", getNumUsers());
  db.addParam("Result", *(getResult().getType()));
  return db;
}

void CPUMaxSplatNode::visit(Node *parent, NodeWalker *visitor) {
  if (!visitor->shouldVisit(parent, this)) { return; }
  visitor->pre(parent, this);
if (hasPredicate())
 getPredicate().getNode()->visit(this, visitor);
  getInput().getNode()->visit(this, visitor);
  visitor->post(parent, this);
}

bool CPUMaxSplatNode::isEqual(const CPUMaxSplatNode &other) const {
  return true &&
      Input_ == other.Input_ &&
      predicate_ == other.predicate_ &&
      SplatValue_ == other.SplatValue_ &&
      getType(0) == other.getType(0);
}

Node* CPUMaxSplatNode::clone() const {
  return new CPUMaxSplatNode(getName(), getInput(), getSplatValue());
}

llvm::hash_code CPUMaxSplatNode::getHash() const {
  return llvm::hash_combine(
      toBinary(SplatValue_),
      Input_);
}

unsigned CPUConvDKKC8Node::getNumInputs() const {
  return 3;
}

std::string CPUConvDKKC8Node::getInputName(unsigned idx) const {
  if (idx == 0) { return "Input"; }
  if (idx == 1) { return "Filter"; }
  if (idx == 2) { return "Bias"; }
  idx -= 3;
  llvm_unreachable("Invalid index");
}

NodeValue CPUConvDKKC8Node::getNthInput(unsigned idx) {
  if (idx == 0) { return Input_; }
  if (idx == 1) { return Filter_; }
  if (idx == 2) { return Bias_; }
  idx -= 3;
  llvm_unreachable("Invalid index");
}

void CPUConvDKKC8Node::setNthInput(unsigned idx, NodeValue val) {
  if (idx == 0) { Input_ = val; return; }
  if (idx == 1) { Filter_ = val; return; }
  if (idx == 2) { Bias_ = val; return; }
  idx -= 3;
  llvm_unreachable("Invalid index");
}

llvm::StringRef CPUConvDKKC8Node::getOutputName(unsigned idx) const {
  if (idx == 0) { return "Result"; }
  llvm_unreachable("Invalid index");
}

std::string CPUConvDKKC8Node::getDebugDesc() const {
  DescriptionBuilder db(getKindName());
  db.addParam("name", getName());
  if (hasPredicate()) db.addParam("Predicate", "Yes");
  db
    .addParam("Input", *(getInput().getType()))
    .addParam("Filter", *(getFilter().getType()))
    .addParam("Bias", *(getBias().getType()))
    .addParam("Kernels", getKernels())
    .addParam("Strides", getStrides())
    .addParam("Pads", getPads())
    .addParam("Group", getGroup())
    .addParam("users", getNumUsers());
  db.addParam("Result", *(getResult().getType()));
  return db;
}

void CPUConvDKKC8Node::visit(Node *parent, NodeWalker *visitor) {
  if (!visitor->shouldVisit(parent, this)) { return; }
  visitor->pre(parent, this);
if (hasPredicate())
 getPredicate().getNode()->visit(this, visitor);
  getInput().getNode()->visit(this, visitor);
  getFilter().getNode()->visit(this, visitor);
  getBias().getNode()->visit(this, visitor);
  visitor->post(parent, this);
}

bool CPUConvDKKC8Node::isEqual(const CPUConvDKKC8Node &other) const {
  return true &&
      Input_ == other.Input_ &&
      Filter_ == other.Filter_ &&
      Bias_ == other.Bias_ &&
      predicate_ == other.predicate_ &&
      Kernels_ == other.Kernels_ &&
      Strides_ == other.Strides_ &&
      Pads_ == other.Pads_ &&
      Group_ == other.Group_ &&
      getType(0) == other.getType(0);
}

Node* CPUConvDKKC8Node::clone() const {
  return new CPUConvDKKC8Node(getName(), getResult().getType(), getInput(), getFilter(), getBias(), getKernels(), getStrides(), getPads(), getGroup());
}

llvm::hash_code CPUConvDKKC8Node::getHash() const {
  return llvm::hash_combine(
      llvm::hash_combine_range(Kernels_.begin(), Kernels_.end()),
      llvm::hash_combine_range(Strides_.begin(), Strides_.end()),
      llvm::hash_combine_range(Pads_.begin(), Pads_.end()),
      Group_,
      Input_,
      Filter_,
      Bias_);
}

#include "glow/CPUSpecificNodesVerification.h"
